/**
\file           
\brief      Main bridging file. Contains startup routines
*/
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <inttypes.h>
#include <init.h>
#include <barrier.h>
#include <rcu_lock.h>
#include <fm_types.h>
#include <ll_struct.h>
#include <stack.h>
#include <swi.h>
#include <timer_wheel.h>
#include <dpa_alloc.h>
#include <shell.h>
#include <wdog.h>
#include <sg_api.h>
#include <dpa/dpa_eth.h>
#include <br.h>

static volatile enum application_state bridging_state = INVALID;
volatile uint32_t g_num_bridge_lwe;
uint32_t recv_channel_map;

__PERCPU uint32_t ageing_timer;
__PERCPU struct dpa_buff_allocator *buff_allocator;
__PERCPU struct bman_pool *bpool[MAX_NUM_BMAN_POOLS];
int num_fq_close_pending;


static void br_shutdown_complete(uint32_t timer_id, void *p_data)
{
	wdog_poke();

	if (0 == num_fq_close_pending) {
		APP_INFO("Moving application state to SHUTDOWN_COMPLETE");
		bridging_state = SHUTDOWN_COMPLETE;
		return;
	}

	APP_WARN("Num Tx FQ close pending: %d", num_fq_close_pending);
	APP_WARN("Waiting for some more time for Tx FQs to retire");

	timer_id = reset_timer(timer_id, 0);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Resetting shutdown complete timer failed");
}

static void br_shutdown_rx_complete(uint32_t timer_id, void *p_data)
{
	wdog_poke();

	if (num_fq_close_pending) {
		APP_WARN("Num Rx FQ close pending: %d", num_fq_close_pending);
		APP_WARN("Waiting for some more time for FQs to retire");

		timer_id = reset_timer(timer_id, 0);
		if (INV_TIMER_ID == timer_id)
			APP_ERROR("Reset shutdown Rx complete timer failed");
		return;
	}

	APP_INFO("Moving application state to SHUTDOWN_RX_COMPLETE");
	bridging_state = SHUTDOWN_RX_COMPLETE;

	/* At this stage, all the traffic which was received has been
	 * processed and we can now move forward to retire Tx FQs
	 */
	num_fq_close_pending = dpa_eth_tx_fq_retire();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring Tx FQs failed");
		return;
	}

	APP_INFO("Number of Tx FQs close pending: %d", num_fq_close_pending);

	/*
	 * If there are no Tx FQs which are pending to be closed, directly
	 * shutdown the application
	 */
	if (!num_fq_close_pending) {
		br_shutdown_complete(INV_TIMER_ID, NULL);
		return;
	}

	/*Start a single shot timer to wait for all Tx FQs to get retired*/
	timer_id = start_timer(BR_SHUTDOWN_TIMER_MS, false,
				NULL, SWI_PRI_HIGH,
				br_shutdown_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown complete timer failed");
	else
		APP_INFO("Started shutdown complete timer ID = %u", timer_id);
}


/* This function processes bridging shutdown event*/
static void br_process_shutdown(void)
{
	uint32_t timer_id;

	APP_INFO("Processing shutdown command ..");
	/*Bootcore is responsible for partition cleanup*/
	bridging_state = SHUTDOWN_PENDING;

	/*Disable both Rx mac ports*/
	if (-1 == dpa_eth_disable_ports(true, false))
		APP_ERROR("Disabling MAC-Rx port failed");
	else
		APP_INFO("Disabled MAC-Rx ports");

	/*Put all Rx frame queues in retired state*/
	/*This would effectively disable the MAC-less Rx port*/
	num_fq_close_pending = dpa_eth_rx_fq_retire();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring Rx FQs failed");
		return;
	}

	APP_INFO("Number of Rx FQs close pending: %d", num_fq_close_pending);

	/*If there is no FQ close pending, directly move to the next state*/
	if (!num_fq_close_pending) {
		br_shutdown_rx_complete(INV_TIMER_ID, 0);
		return;
	}

	/*Start a timer to complete Rx shutdown*/
	timer_id = start_timer(BR_SHUTDOWN_TIMER_MS,
				false, NULL, SWI_PRI_HIGH,
				br_shutdown_rx_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown RX complete timer failed");
	else
		APP_INFO("Started shutdown Rx complete timer = %u", timer_id);
}

/*This function is called by all cores running bridging*/
static int br_common_init(void)
{
	int ecode;

	/* Inform RCU that one more LWE is running */
	rcu_join(lwe_instance);

	/* Start the timer framework*/
	if (0 != timer_framework_start()) {
		APP_ERROR("timer_framework_start failed");
		return -1;
	}

	/* Start a periodic timer for running ageing algo */
	ageing_timer = start_timer(BR_AGEING_INTERVAL_MS, true, NULL,
					SWI_PRI_HIGH, br_aging_run, NULL);
	if (INV_TIMER_ID == ageing_timer) {
		APP_ERROR("Ageing timer start failed");
		return -1;
	}

	APP_INFO("Ageing timer start SUCCESS");

	buff_allocator = dpa_allocator_create(&ecode);
	if (NULL == buff_allocator) {
		APP_ERROR("dpa_buff_allocator_create failed");
		return -1;
	}

	APP_INFO("Buffer allocator created...");
	return 0;
}

/*
 * This function is used to move a FQ to OOS state, destroy it and free
 * the associated memory with FQ structure.
 */
static void close_fq(uint32_t swi_id, void *data)
{
	struct qman_fq *fq = data;
	uint32_t fqid;

	fqid = fq->fqid;
	if (0 > qman_oos_fq(fq)) {
		APP_ERROR("Moving FQID: %u to OOS failed", fq->fqid);
		free(fq);
		return;
	}

	qman_destroy_fq(fq, 0);
	free(fq);
	APP_INFO("Closed FQID: %u", fqid);

	num_fq_close_pending--;
}

/* This function waits till the FQ becomes empty and then closes it */
static void wait_fq_empty(uint32_t swi_id, void *data)
{
	struct qman_fq *fq = data;
	uint32_t vdqcr;
	uint32_t flags;

	APP_INFO("Waiting for FQID: %u to become empty", fq->fqid);

	vdqcr = QM_VDQCR_PRECEDENCE_VDQCR | QM_VDQCR_NUMFRAMES_TILLEMPTY;

	/* NB: the use of QMAN_VOLATILE_FLAG_WAIT is prohibited in LWE
	 * applications, and indeed was unsafe when called from a callback (as
	 * is the case here, in at least one of the code-paths). To avoid
	 * thrashing, we implement a throttled poll for completion. */
	if (qman_volatile_dequeue(fq, 0, vdqcr)) {
		APP_ERROR("qman_volatile_dequeue failure for FQID: %u",
			fq->fqid);
		free(fq);
		return;
	}
loop:
	qman_fq_state(fq, NULL, &flags);
	if (flags & QMAN_FQ_STATE_VDQCR) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	/*Now the FQ is empty, close it*/
	close_fq(SWI_ID_INV, data);
}

/* This function processes FQ state change notifications. This function is
 * called from ISR context. Thus this function must not make any blocking
 * and heavy weight APIs. When the FQ get retired and also gets empty, a
 * SWI is posted towards application's main loop for taking FQs OOS and
 * destroy them.
 */
static void br_proc_fq_state_chg(struct qman_portal *qm,
				struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	uint8_t verb;
	uint8_t fqs;

	/*Extract the response verb*/
	verb = msg->verb & ~QM_MR_VERB_VBIT;

	if (QM_MR_VERB_FQRL == verb) {
		if (0 > swi_add(SWI_ID_INV, fq, close_fq, SWI_PRI_HIGH))
			APP_ERROR("Cannot add SWI for closing FQ");

		return;
	}

	fqs = msg->fq.fqs;

	if ((QM_MR_VERB_FQRN == verb) || (QM_MR_VERB_FQRNI == verb)) {
		if (fqs & QM_MR_FQS_NOTEMPTY) {
			APP_INFO("Retired FQID: %u not yet empty", fq->fqid);

			if (0 > swi_add(SWI_ID_INV, fq, wait_fq_empty,
					SWI_PRI_HIGH))
				APP_ERROR("Cannot add SWI for emptying FQ");
			return;
		}

		if (fqs & QM_MR_FQS_ORLPRESENT) {
			/*We need to wait till we get FQRL*/
			APP_INFO("Retired FQID: %u has ORL list", fq->fqid);
			return;
		}

		/*Else FQ is empty and can be destroyed*/
		if (0 > swi_add(SWI_ID_INV, fq,	close_fq, SWI_PRI_HIGH)) {
			APP_ERROR("Cannot add SWI for closing FQ");
			return;
		}

		return;
	}

	APP_ERROR("Unexpected FQ state change verb: %d", verb);
}

/*This function is called by only control core in bridging*/
void br_proc_ern(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	APP_WARN("ERN on FQID: %u", msg->ern.fqid);
	lwe_free_fd(buff_allocator, &(msg->ern.fd));
}


static int br_ctrl_init(void)
{
	struct td_param pcd_td_param;
	struct qman_fq_cb rx_pcd_cb = { br_proc_pcd_rx_fq, NULL,
						NULL, br_proc_fq_state_chg };
	struct qman_fq_cb rx_err_cb = { br_proc_rx_err_fq, NULL,
						NULL, br_proc_fq_state_chg };
	struct qman_fq_cb tx_err_cb = { br_proc_tx_err_fq, NULL,
						NULL, br_proc_fq_state_chg };
	struct qman_fq_cb tx_cb = { NULL, br_proc_ern, NULL, br_proc_fq_state_chg };
	struct qman_fq_cb tx_confirm_cb = { br_proc_tx_confirm_fq, NULL,
						NULL, br_proc_fq_state_chg };
	struct qman_fq_cb rx_default_cb = { br_proc_default_rx_fq, NULL,
						NULL, br_proc_fq_state_chg };

	if (0 >= dpa_eth_read(&recv_channel_map)) {
		APP_ERROR("dpa_eth_read failed");
		return -1;
	}

	if (0 != dpa_eth_fq_create(&rx_default_cb,
				&rx_pcd_cb,
				&rx_err_cb,
				&tx_cb,
				&tx_confirm_cb,
				&tx_err_cb, 0)) {
		APP_ERROR("dpa_eth_fq_create failed");
		return -1;
	}

	APP_INFO("Num ethernet ports discovered: %u", g_num_dpa_eth_ports);

	/*Assuming we do not buffer more than 128 frames of 256 bytes per FQ*/
	pcd_td_param.mant = 128;
	pcd_td_param.exp = 8;
	if (0 != dpa_eth_fq_init(64, 64, 64, NULL, &pcd_td_param)) {
		APP_ERROR("dpa_eth_fq_init failed");
		return -1;
	}

	if (0 > br_swdb_init()) {
		APP_ERROR("br_swdb_init failed");
		return -1;
	}

	br_aging_init();

	if (0 != br_stats_init()) {
		APP_ERROR("br_stats_init failed");
		return -1;
	}

	return 0;
}

/*BMAN pool initialisation function*/
static int br_bpools_init(void)
{
	struct bman_pool_params params;
	uint32_t bpid;

	/*Create BMAN pools */
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		/*Else create the pool object */
		params.flags = BMAN_POOL_FLAG_STOCKPILE;
		params.cb = NULL;
		params.cb_ctx = NULL;
		params.bpid = bpid;

		bpool[bpid] = bman_new_pool(&params);
		if (NULL == bpool[bpid]) {
			APP_ERROR("bman_new_pool failed for BPID: %u", bpid);
			return -1;
		}

		if (0 != dpa_allocator_add_bpool(buff_allocator,
						dpa_bpools[bpid].size,
						bpid, bpool[bpid])) {
			APP_ERROR("dpa_allocator_add_bpool failed");
			return -1;
		}

		APP_INFO("Added BPID: %u, Size: %u", bpid,
			 dpa_bpools[bpid].size);
	}

	return 0;
}

/*BMAN pool destroy function*/
static void br_bpools_destroy(void)
{
	uint32_t bpid;

	/*Free the BMAN buffer pools*/
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		bman_free_pool(bpool[bpid]);
		APP_INFO("BPID: %u destroyed", bpid);
	}

	/*Free the DPA allocator*/
	dpa_allocator_free(buff_allocator);
	APP_INFO("DPA allocator freed");
}

/*This function waits for shutdown process to complete and also cleans up
per core resources*/
static void wait_for_shutdown_complete(void)
{
	void *p_data;
	uint32_t swi_id;

	/*Stop the ageing timer*/
	if (0 != stop_timer(ageing_timer))
		APP_ERROR("Stopping ageing timer failed");
	else
		APP_INFO("Ageing timer stopped");

	while (1) {
		if (SHUTDOWN_COMPLETE == bridging_state)
			break;

		/*This would dequeue any remanining notifications from QMAN*/
		qman_poll();

		while (SWI_ID_INV != (swi_id = swi_activate(&p_data))) {
			switch (swi_id) {
			case SWI_ID_TIMER_EXPIRY:
				proc_timer_expiry((uint32_t) p_data);
				break;

			default:
				break;
			}
		}

		rcu_synchronize_cpu(lwe_instance);
	}

	rcu_leave(lwe_instance);
	APP_INFO("Left RCU protection");

	br_bpools_destroy();
	APP_INFO("Destroyed BMAN pools and buffer allocator");

	if (0 != timer_framework_stop()) {
		APP_ERROR("timer_framework_stop failed");
		return;
	}

	APP_INFO("Timer framework stopped");

	/*Dequeue any pending SWI if any due to decrementer
	already expired in between*/
	while (SWI_ID_INV != (swi_id = swi_activate(&p_data))) {
		switch (swi_id) {
		case SWI_ID_TIMER_EXPIRY:
			APP_INFO("Consuming pending timer tick");
			break;

		default:
			break;
		}
	}

}

void br_wdog_handler(void)
{
	APP_WARN("Watchdog expired !!");
}

/**
\brief   Main function of bridge code.

\details Bridge startup routine. Initialises the DPA components, switching
	database etc. The bootcore assumes the role of control core. Only
	boot core does the one time global initialisation needed by all
	secondary cores. The per core resources like portals are initialised
	by the individual cores themselves. After the initialisation is done,
	the cores enter into an infinite loop dequeueing frames from a pool
	channel.
\note    Should never return.
*/
int main(int argc, char *argv[])
{
	uint32_t swi_id;
	void *p_data;

	APP_INFO("Bridge started on LWE ID : %u", lwe_instance);
	atomic_inc((unsigned long *)&g_num_bridge_lwe);
	APP_INFO("Now %u LWE instances run bridging", g_num_bridge_lwe);

	/*Start the watchdog with double the ageing interval*/
	if (0 != wdog_init(br_wdog_handler, TCR_WRC_RESET,
			BR_AGEING_INTERVAL_MS * 2)) {
		APP_ERROR("Watchdog initialisation failed");
		goto err_out;
	}

	/*Boot core initialisation*/
	if (0 == lwe_instance) {
		APP_INFO("Welcome to FSL Bridge ... \r\n");

		/*Initialise the RCU library so that the cores can join*/
		if (NULL == rcu_initialize()) {
			APP_ERROR("RCU library initialisation failed");
			goto err_out;
		}

		bridging_state = INITIALIZING;
	} else {
		while (INITIALIZING != bridging_state);
	}

	/*Common initialisation routine for all cores*/
	if (0 > br_common_init()) {
		APP_ERROR("br_common_init failed");
		goto err_out;
	}

	/*Control core initialises global database*/
	if (0 == lwe_instance) {
		if (0 > br_ctrl_init()) {
			APP_ERROR("br_ctrl_init failed");
			goto err_out;
		}
	}

	/*Wait till DPA lib init finishes */
	do {
		rcu_synchronize_cpu(lwe_instance);
	} while (0 == g_num_dpa_eth_ports);


	APP_INFO("Frames to be recv on channel map: 0x%x", recv_channel_map);
	qman_static_dequeue_add(recv_channel_map);

	/*Create BMAN pools */
	if (0 > br_bpools_init()) {
		APP_ERROR("br_bpool_init failed");
		goto err_out;
	}

	if (0 == lwe_instance) {
		if (0 != dpa_eth_enable_ports(true)) {
			APP_ERROR("dpa_eth_enable_ports failed");
			goto err_out;
		}

		APP_INFO("Bridging initialisation completed ..");

		/*All init done now. Give a goahead to other partitions */
		bridging_state = RUNNING;
	}

	do {
		rcu_synchronize_cpu(lwe_instance);
	} while (RUNNING != bridging_state);

	/*The following magic breakpoint halts the simulation. 
	   This is needed to take a checkpoint to be used in hybrid 
	   simics model */
	MAGIC_BREAKPOINT;

	APP_INFO("Starting frame dequeue...");

	while (1) {
		if (unlikely(SHUTDOWN_PENDING == bridging_state))
			break;

		qman_poll();

		while (unlikely(SWI_ID_INV !=
				(swi_id = swi_activate(&p_data)))) {

			switch (swi_id) {
			case SWI_ID_TIMER_EXPIRY:
				proc_timer_expiry((uint32_t) p_data);
				break;

			/*Shutdown SWI is received only on bootcore*/
			case SWI_ID_SHUTDOWN:
				br_process_shutdown();
				break;

			default:
				break;
			}
		}

		rcu_synchronize_cpu(lwe_instance);
	}

	/*The following code is executed only in SHUTDOWN_PENDING state*/
	/*Here, all cores wait for shutdown to complete*/
	wait_for_shutdown_complete();
	wdog_stop();
	APP_INFO("Shutdown completed on core: %u", lwe_instance);
	return 0;

err_out:
	APP_ERROR("FATAL error, exiting\r\n");
	return -1;
}
