/**
 \file           br.h
 \brief          Common datatypes and hash-defines
 */
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _BR_H
#define _BR_H

#include <string.h>
#include <compat.h>
#include <fsl_qman.h>

#define LOG_LEVEL  4
#define PRINT_INFO

#include <init.h>
#include <statistics.h>
#include <rcu_lock.h>
#include "fm_types.h"
#include <app_common.h>
#include <ether.h>

#define BR_HASH_BITS 8
#define BR_HASH_SIZE (1 << BR_HASH_BITS)
#define INV_LWE_ID        255

#define BR_NUM_SWDB_BUCKETS	1024
#define BR_BUCKETS_PER_INTERVAL (BR_NUM_SWDB_BUCKETS / 8)
#define BR_EXPIRE_TIME_MS	100

/*
 * Because simics virtual time progresses slow, 
 * hence these timeout value are kept low for demo purposes
 */
#define BR_AGEING_INTERVAL_MS		2000
#define BR_SHUTDOWN_TIMER_MS		50

/*State of bridging application*/
enum application_state {
	INVALID,
	INITIALIZING,
	RUNNING,
	SHUTDOWN_PENDING,
	SHUTDOWN_RX_COMPLETE,
	SHUTDOWN_COMPLETE
};

/**
\brief          Switching Database entry

\details        Lookup entry per dest mac address 
                Contains per entry data as well as link to next node 
*/
typedef struct br_swdb_entry {
	struct br_swdb_entry *p_next;/**< Next node */
	uint32_t port;		     /**< Egress Port(s) */
	uint32_t port_index;	     /**< Index in the port structure*/
	uint64_t ageing_stamp;	    /**< Last ageing time-stamp */
	uint8_t mac_addr[ETH_ALEN]; /**< Dst Mac Address (key) */
	uint8_t is_local;	    /**< Is Local MAC? */
	uint8_t is_static;	    /**< Not to be aged? */
} br_swdb_entry_t;

/**
\brief          Switching DB per-bucket head
*/
typedef struct br_swdb_bucket {
	br_swdb_entry_t *p_head;  /**< Bucket head */
	uint32_t wlock;	  /**< Read/Write lock (per bucket) */
} br_swdb_bucket_t;

/**
\brief          Switching hash-table (max buckets)
*/
typedef struct {
	struct br_swdb_bucket buckets[BR_NUM_SWDB_BUCKETS];
} br_swdb_t;

/**
\brief          Bridging Statistics counter identifiers
*/
enum br_stats_type_e {
	BR_STAT_RX_TOTAL = 0,
	BR_STAT_RX_UNICAST,
	BR_STAT_RX_MULTICAST,
	BR_STAT_RX_BROADCAST,
	BR_STAT_RX_ERROR,
	BR_STAT_RX_UNSUPPORTED,
	BR_STAT_RX_LKUPFAIL,
	BR_STAT_RX_LOCAL,
	BR_STAT_TX_UNICAST,
	BR_STAT_TX_MULTICAST,
	BR_STAT_TX_BROADCAST,
	BR_STAT_TX_ERROR,
	BR_NUM_STATS_TYPE
};

/**
\brief          Bridging staistics block.

\details        Bridge statistics (frame/octet counter pairs). 
	`	Either of following types:
		i) Named Members
		ii) Unnamed list array
*/
union br_stats_s {
	struct {
		struct stat64_pair_t rx_total;	/**< total pkts recvd */
		struct stat64_pair_t rx_unicast;/**< unicast pkts recvd */
		struct stat64_pair_t rx_multicast;/**< mcast pkts recvd */
		struct stat64_pair_t rx_broadcast;/**< bcast pkts recvd */
		struct stat64_pair_t rx_error;	/**< error pkts */
		struct stat64_pair_t rx_unsupported;/**< unsupported frames */
		struct stat64_pair_t rx_lookup_fail;/**< lookup failures */
		struct stat64_pair_t rx_local;	/**< locally destined packets */
		struct stat64_pair_t tx_unicast;/**< total unicast pkts txed */
		struct stat64_pair_t tx_multicast;/**< total multicast pkts txed */
		struct stat64_pair_t tx_broadcast;/**< total broadcast pkts txed */
		struct stat64_pair_t tx_error;	/**< packet transmit errors */
	} stats_type;

	struct stat64_pair_t stats_array[BR_NUM_STATS_TYPE];
} __attribute__ ((aligned(16)));	/*Stats area must be 16 byte aligned */

/* Extern Data */
extern struct dpa_buff_allocator *buff_allocator;
extern uint32_t my_lwe_id;
extern br_swdb_t *br_swdb;
extern uint32_t g_idle_count;

extern union br_stats_s *br_stats;

extern struct qman_portal *g_qm_portal;
extern struct dpa_buff_allocator *buff_allocator;

#define PRINT_STAT(type)  \
	printf(#type "-> Pkts: %llu	Bytes: %llu\r\n", \
		 br_stats->stats_array[type].inc.stat, \
		 br_stats->stats_array[type].acc.stat)

void br_synchronize_cpus(void);
int br_swdb_init(void);
void br_swdb_cleanup(void);
int br_swdb_insert(struct br_swdb_bucket *, struct br_swdb_entry *);
br_swdb_entry_t *br_swdb_learn_mac(uint8_t * p_s_mac, uint32_t sport);
int br_swdb_delete(struct br_swdb_entry *p_entry,
		   struct br_swdb_bucket *p_bucket);
struct br_swdb_entry *br_swdb_entry_search(uint8_t * p_macaddr,
					   struct br_swdb_bucket **p_p_bucket);
void br_swdb_lock(uint32_t * lock);
void br_aging_init(void);
void br_aging_run(uint32_t timer_id, void *p_data);
int br_stats_init(void);
void br_stats_update(enum br_stats_type_e stats_type, uint32_t frame_size);

enum qman_cb_dqrr_result br_proc_default_rx_fq(struct qman_portal *p,
					       struct qman_fq *fq,
					       const struct qm_dqrr_entry
					       *entry);

enum qman_cb_dqrr_result br_proc_pcd_rx_fq(struct qman_portal *p,
					   struct qman_fq *fq,
					   const struct qm_dqrr_entry *entry);

enum qman_cb_dqrr_result br_proc_rx_err_fq(struct qman_portal *p,
					   struct qman_fq *fq,
					   const struct qm_dqrr_entry *entry);

enum qman_cb_dqrr_result br_proc_tx_confirm_fq(struct qman_portal *p,
					       struct qman_fq *fq,
					       const struct qm_dqrr_entry
					       *entry);

enum qman_cb_dqrr_result br_proc_tx_err_fq(struct qman_portal *p,
					   struct qman_fq *fq,
					   const struct qm_dqrr_entry *entry);

#endif
