/**
 \file       br_aging.c
 \brief      Implements ageing support in bridging.
*/
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <rcu_lock.h>
#include <timer_wheel.h>
#include <wdog.h>
#include <br.h>


static volatile uint32_t aging_bookmark;
static uint32_t aging_lock;

/**
 \brief          Initialises the ageing functionality
 \return         None
*/
void br_aging_init(void)
{
	aging_bookmark = 0;
	aging_lock = 0;
}

/**
 \brief		Runs the ageing routine.
 \details	This routines checks whether a learnt MAC entry has not been
		used for long time. If it has not been used for long, it is
		deleted.
 \return	None
*/
void br_aging_run(uint32_t timer_id, void *p_data)
{
	uint32_t i;
	struct br_swdb_entry *p_entry;
	struct br_swdb_entry **p_p_entry;
	struct br_swdb_bucket *p_bucket;
	uint32_t bookmark;

	/*Watchdog is poked from the ageing timer*/
	wdog_poke();

	if (0 == spin_trylock(&aging_lock)) {
		/*Lock already held. Some other core already running ageing.*/
		return;
	}

	APP_INFO("Running ageing ...");

	bookmark = aging_bookmark;
	aging_bookmark = aging_bookmark + BR_BUCKETS_PER_INTERVAL;

	if (aging_bookmark >= BR_NUM_SWDB_BUCKETS)
		aging_bookmark = 0;

	rcu_read_lock();

	for (i = 0; i < BR_BUCKETS_PER_INTERVAL; i++) {
		p_bucket = &(br_swdb->buckets[bookmark]);

		br_swdb_lock(&(p_bucket->wlock));

		p_p_entry = &(p_bucket->p_head);
		p_entry = rcu_dereference(*p_p_entry);

		while (NULL != p_entry) {
			if ((get_current_time() - p_entry->ageing_stamp) >
			    BR_EXPIRE_TIME_MS) {
				APP_INFO("Ageing out MAC entry: " MAC_FMT
					 " from port: %d",
					 NMAC_STR(p_entry->mac_addr),
					 p_entry->port);
				br_swdb_delete(p_entry, p_bucket);
			}

			p_p_entry = &(p_entry->p_next);
			p_entry = rcu_dereference(*p_p_entry);
		}

		spin_unlock(&(p_bucket->wlock));
		bookmark++;
	}

	rcu_read_unlock();

	spin_unlock(&aging_lock);

	APP_INFO("Exiting ageing ...");
}
