/**
\file           
\brief          Bridge packet processing and forwarding code
*/
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <misc.h>
#include <sg_api.h>
#include <dpa_alloc.h>
#include <dpa/dpa_eth.h>
#include <net/frame_desc.h>
#include <br.h>

/**
\brief          Parses ethernet header and copies in a structure

\details        Parses ethernet header and copies in ethernet header structure

\param[in]      p_pkt - Ethernet packet.
\param[out]      eth  - Ethernet header

\return         void
 
*/
static void br_parse_pkt(struct ethhdr *p_eth, const struct qm_fd *p_fd)
{
	if (unlikely((0xFF == p_eth->h_dest[0]) &&
		     (0xFF == p_eth->h_dest[1]) &&
		     (0xFF == p_eth->h_dest[2]) &&
		     (0xFF == p_eth->h_dest[3]) &&
		     (0xFF == p_eth->h_dest[4]) &&
		     (0xFF == p_eth->h_dest[5]))) {
		br_stats_update(BR_STAT_RX_BROADCAST, p_fd->length20);
	} else {
		/*All other packets we assume to be unicast */
		br_stats_update(BR_STAT_RX_UNICAST, p_fd->length20);
	}
}

/**
\brief          Check for allowed ethernet frame types

\details        Checks whether the ethernet frame is allowed to be processed.

\param[in]      p_pkt - Ethernet packet.
\param[in]      eth  - Ethernet header

\return         1 - Can be processed
  		0 - Cannot be processed
*/
static int br_is_allowed(uint8_t * p_pkt, struct ethhdr *p_eth)
{
	/*Currently we do not support VLAN */
	if (unlikely((p_eth->h_proto == ETH_P_8021Q) ||
		     (p_eth->h_proto == ETH_P_PAUSE))) {
		return 0;
	}

	/*Source address cannot be broadcast address */
	if (unlikely((0xFF == p_eth->h_source[0]) &&
		     (0xFF == p_eth->h_source[1]) &&
		     (0xFF == p_eth->h_source[2]) &&
		     (0xFF == p_eth->h_source[3]) &&
		     (0xFF == p_eth->h_source[4]) &&
		     (0xFF == p_eth->h_source[5]))) {
		APP_ERROR("SMAC can't be broadcast address");
		return 0;
	}

	return 1;
}

/**
\brief          Frame sending function

\details        Sends a frame through a given port

\param[in]      p_fd - Frame descriptor of frame to be sent
\param[in]      port - Port through which to send the frame

\return         None
  
*/
static void br_send_out(const struct qm_fd *p_fd, uint32_t port_index)
{
	uint32_t fq_id;
	uint8_t *p_dmac;
	struct qman_fq *fq;
	phys_addr_t addr;

	addr = p_fd->addr_hi;
	addr = (addr << 32) | (p_fd->addr_lo);
	addr = addr + p_fd->offset;

	p_dmac = (uint8_t *) ptov(addr);

	/*we use only one egress queue */
	fq_id = dpa_eth_ports[port_index].egress.fq_start;

	APP_DEBUG("Sending DMAC: " MAC_FMT " -> Port: %u, FQID: %u",
		  NMAC_STR(p_dmac), port_index, fq_id);

	fq = dpa_eth_ports[port_index].egress.fq[0];

	/* NB: the use of QMAN_ENQUEUE_FLAG_WAIT is prohibited in LWE
	 * applications. To avoid thrashing, we implement a throttled poll for
	 * retrying the enqueue. */
loop:
	if (qman_enqueue(fq, p_fd, 0)) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}
}

/**
\brief          Copy frame

\details        Makes a copy of the frame

\param[in]      p_fd - Frame descriptor of frame to be copied.
\param[out]     p_new_fd - New frame descriptor allocated

\return         New frame descriptor
  
*/
static struct qm_fd *br_copy_frame(const struct qm_fd *p_fd,
				   struct qm_fd *p_new_fd)
{
	uint8_t *p_pkt;
	struct bm_buffer buf;
	uint32_t size;
	phys_addr_t addr_new;
	phys_addr_t addr_orig;

	size = p_fd->length20;

	if (unlikely(0 > dpa_allocator_get_buff(buff_allocator, size, &buf))) {
		APP_ERROR("dpa_allocator_get_buff fail for size = %u", size);
		return NULL;
	}

	addr_new = buf.hi;
	addr_new = (addr_new << 32) | buf.lo;

	p_pkt = (uint8_t *) ptov(addr_new);

	/* Copy the actual frame data
	 */

	addr_orig = p_fd->addr_hi;
	addr_orig = (addr_orig << 32) | p_fd->addr_lo;
	addr_orig = addr_orig + p_fd->offset;

	memcpy(p_pkt, ptov(addr_orig), p_fd->length20);

	p_new_fd->bpid = buf.bpid;
	p_new_fd->format = qm_fd_contig;
	p_new_fd->offset = 0;
	p_new_fd->length20 = p_fd->length20;

	p_new_fd->addr_hi = (uint8_t) (buf.hi);
	p_new_fd->addr_lo = (uint32_t) (buf.lo);
	return p_new_fd;
}

/**
\brief          Frame broadcast function

\details        Broadcasts frame on all port except the source port on 
		which the frame arrived. For this first we send the frame to
		all FMAN ports with requiring a sent confirmation and send a 
		copy of frame to all mac-less ports

\param[in]      p_fd - Frame descriptor of frame to be broadcast
\parama[in]	src_port - Source port on which frame arrived

\return         None
*/
static void br_broadcast(const struct qm_fd *p_fd, uint32_t src_port_index)
{
	uint32_t port;
	struct qm_fd new_fd;
	uint32_t *p_tx_confirm;
	phys_addr_t addr;

	/*Send the frame on all ports except the source port. */
	APP_INFO("Broadcasting frame except sport: %u", src_port_index);

	/*Make a list of FMAN ports on which to broadcast the frame at 
	   the start of frame. Each frame has an start offset of 32 bytes 
	   before parse result. We can store uin32_t flag for upto maximum 
	   of 32 ports in this area */
	addr = p_fd->addr_hi;
	addr = (addr << 32) | p_fd->addr_lo;
	p_tx_confirm = (uint32_t *) ptov(addr);

	*p_tx_confirm = 0;

	for (port = 0; port < g_num_dpa_eth_ports; port++) {
		if (unlikely(port == src_port_index)) {
			continue;
		}

		/*For all MAC-less ports, send them an individual copy */
		if (likely(MAC_ETH == dpa_eth_ports[port].type)) {
			/*For FMAN ports, set the TX confirmation flag ON. */
			*p_tx_confirm = *p_tx_confirm | (1 << port);
			continue;
		}

		/*Else for mac-less, make a copy of the frame and send it */
		if (unlikely(NULL == br_copy_frame(p_fd, &new_fd))) {
			APP_ERROR("Error copying the frame");
			continue;
		}

		new_fd.status = 0;
		br_send_out(&new_fd, port);
	}

	/*Now send the original frame to all FMAN ports with 
	   requiring TX  confirmation */

	new_fd = *p_fd;		/*We copy FD here since original FD is in DQRR memory 
				   and writes to this memory are ignored */

	for (port = 0; port < g_num_dpa_eth_ports; port++) {
		if (unlikely(port == src_port_index)) {
			continue;
		}

		/*Here skip all MAC-less ports */
		if (unlikely(MAC_LESS_ETH == dpa_eth_ports[port].type)) {
			continue;
		}

		new_fd.cmd = FM_FD_CMD_FCO;
		new_fd.cmd =
		    new_fd.cmd | dpa_eth_ports[port].egress_confirm.fq_start;

		br_send_out(&new_fd, port);
	}
}

/**
\brief          Determines port index from FQID

\details        Determines port index using FQID on which the frame arrived.

\param[in]      fq_id - Frame queue number

\return         port index
*/
static uint32_t fqid_to_port(uint32_t fq_id)
{
	uint32_t fq_start;
	uint32_t fq_end;
	uint32_t port_index;

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		/*Check in PCD frame queues */
		fq_start = dpa_eth_ports[port_index].pcd_ingress.fq_start;
		fq_end =
		    fq_start + dpa_eth_ports[port_index].pcd_ingress.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end) {
			return port_index;
		}

		/*Check in default FQs */
		fq_start = dpa_eth_ports[port_index].ingress.fq_start;
		fq_end = fq_start + dpa_eth_ports[port_index].ingress.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end) {
			return port_index;
		}
	}

	return INV_PORT;
}

/**
\brief          Determines port index from TX confirm FQID

\details        Determines port index using FQID on which the frame arrived.

\param[in]      fq_id - Frame queue number

\return         port index
*/

static uint32_t tx_confirm_fqid_to_port(uint32_t fq_id)
{
	uint32_t fq_start;
	uint32_t fq_end;
	uint32_t port_index;

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		/*Check in PCD frame queues */
		fq_start = dpa_eth_ports[port_index].egress_confirm.fq_start;
		fq_end =
		    fq_start +
		    dpa_eth_ports[port_index].egress_confirm.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end) {
			return port_index;
		}
	}

	return INV_PORT;
}

/**
\brief          Determines port index from TX err FQID

\details        Determines port index using FQID on which the frame arrived.

\param[in]      fq_id - Frame queue number

\return         port index
*/
uint32_t tx_err_fqid_to_port(uint32_t fq_id)
{
	uint32_t fq_start;
	uint32_t fq_end;
	uint32_t port_index;

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		/*Check in Tx err frame queues */
		fq_start = dpa_eth_ports[port_index].egress_err.fq_start;
		fq_end =
		    fq_start + dpa_eth_ports[port_index].egress_err.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end) {
			return port_index;
		}
	}

	return INV_PORT;
}

/**
\brief          Ethernet frame processing function

\details        Parses ethernet packet, learns source MAC address and forwards to packet to 
		the required port as indicated by switching database.

\param[in]      p_fd - Frame descriptor of frame to be forwarded.

\return         None

\note           Called for every incoming frame
*/
void br_process_ethsw(const struct qm_fd *p_fd, uint32_t fqid)
{
	br_swdb_entry_t *p_entry;
	struct ethhdr *p_eth;
	uint8_t *p_pkt;
	struct br_swdb_bucket *p_bucket;
	uint32_t src_port;
	phys_addr_t addr;

	/*
	 * Currently we are supporting only simple frame format. 
	 * Also we assume that only 32 bit addresses are being used for frames.
	 */
	if (unlikely(qm_fd_contig != p_fd->format)) {
		APP_ERROR("Unsupported frame format: %d, FQID: %u, len: %u",
		     p_fd->format, fqid, p_fd->length20);
		lwe_free_fd(buff_allocator, p_fd);
		return;
	}

	src_port = fqid_to_port(fqid);
	if (unlikely(INV_PORT == src_port)) {
		APP_ERROR("Frame from unknown port. Ingress FQID: %u", fqid);
		lwe_free_fd(buff_allocator, p_fd);
		return;
	}

	br_stats_update(BR_STAT_RX_TOTAL, p_fd->length20);

	addr = p_fd->addr_hi;
	addr = (addr << 32) | p_fd->addr_lo;

	p_pkt = (uint8_t *) ptov(addr);
	p_pkt = p_pkt + p_fd->offset;

	p_eth = (struct ethhdr *)(p_pkt);

	APP_DEBUG("Recv " MAC_FMT " -> " MAC_FMT
		  " FQ:%u sport:%u BP:%d len:%u Off:%u",
		  NMAC_STR(p_eth->h_source), NMAC_STR(p_eth->h_dest), fqid,
		  src_port, p_fd->bpid, p_fd->length20, p_fd->offset);

	if (unlikely(0 == br_is_allowed(p_pkt, p_eth))) {
		APP_ERROR("Frame not supported. Dropping it.");
		br_stats_update(BR_STAT_RX_UNSUPPORTED, p_fd->length20);
		lwe_free_fd(buff_allocator, p_fd);
		return;
	}

	br_parse_pkt(p_eth, p_fd);

	rcu_read_lock();
	br_swdb_learn_mac(&(p_eth->h_source[0]), src_port);

	/*Get the entry for destination MAC */
	p_entry = br_swdb_entry_search(&(p_eth->h_dest[0]), &p_bucket);

	if (unlikely(NULL == p_entry)) {
		/*
		 * We would never get entry for multicast and broadcast DMAC
		 */
		br_stats_update(BR_STAT_RX_LKUPFAIL, p_fd->length20);
		br_broadcast(p_fd, src_port);
	} else {
		br_send_out(p_fd, p_entry->port);
	}

	rcu_read_unlock();
}

/**
\brief          Ethernet frame processing function

\details        Processes frames received on default rx FQ. 

\param[in]      p_fd - Frame descriptor of frame to be forwarded
\param[in]      fqid - Frame queue on which frame was received

\return         None
*/
enum qman_cb_dqrr_result br_proc_default_rx_fq(struct qman_portal *p,
					       struct qman_fq *fq,
					       const struct qm_dqrr_entry
					       *entry)
{
	const struct qm_fd *fd;

	APP_WARN("Frame on default RX FQID: %u", entry->fqid);

	fd = &(entry->fd);
	br_process_ethsw(fd, entry->fqid);
	return qman_cb_dqrr_consume;
}

/**
\brief          Ethernet frame processing function

\details        Processes frames received on PCD rx FQ. 

\param[in]      p_fd - Frame descriptor of frame to be forwarded
\param[in]      fqid - Frame queue on which frame was received

\return         None
*/
enum qman_cb_dqrr_result br_proc_pcd_rx_fq(struct qman_portal *p,
					   struct qman_fq *fq,
					   const struct qm_dqrr_entry *entry)
{
	const struct qm_fd *fd;

	fd = &(entry->fd);

	br_process_ethsw(fd, entry->fqid);
	return qman_cb_dqrr_consume;
}

/**
\brief          Ethernet frame processing function

\details        Processes frames received on rx error FQ. 

\param[in]      p_fd - Frame descriptor of frame to be forwarded
\param[in]      fqid - Frame queue on which frame was received

\return         None
*/
enum qman_cb_dqrr_result br_proc_rx_err_fq(struct qman_portal *p,
					   struct qman_fq *fq,
					   const struct qm_dqrr_entry *entry)
{
	APP_WARN("Frame on RX error FQID: %u", entry->fqid);
	lwe_free_fd(buff_allocator, &(entry->fd));
	return qman_cb_dqrr_consume;
}

/**
\brief          Ethernet frame processing function

\details        Processes frames received on TX confirm FQ. 

\param[in]      p_fd - Frame descriptor of frame to be forwarded
\param[in]      fqid - Frame queue on which frame was received

\return         None
*/
enum qman_cb_dqrr_result br_proc_tx_confirm_fq(struct qman_portal *p,
					       struct qman_fq *fq,
					       const struct qm_dqrr_entry
					       *entry)
{
	volatile uint32_t *p_tx_confirm;
	uint32_t port;
	uint32_t new_flag;
	uint32_t old_flag;
	const struct qm_fd *p_fd;
	phys_addr_t addr;

	APP_DEBUG("Frame on TX confirm FQID: %u. ", entry->fqid);

	port = tx_confirm_fqid_to_port(entry->fqid);
	if (unlikely(INV_PORT == port)) {
		APP_ERROR("tx_confirm_fqid_to_port failed");
		return qman_cb_dqrr_consume;
	}

	p_fd = &(entry->fd);

	addr = p_fd->addr_hi;
	addr = (addr << 32) | p_fd->addr_lo;

	/*Mark the TX confirm flag at start of frame to '0' for this port */
	p_tx_confirm = (uint32_t *) ptov(addr);
reset_tx_confirm:
	old_flag = *p_tx_confirm;
	new_flag = old_flag & (~(1 << port));

	/*If confirmation is not remaining from any of the port, then free FD */
	if (unlikely(0 == new_flag)) {
		lwe_free_fd(buff_allocator, p_fd);
		APP_DEBUG("Freed...");
		return qman_cb_dqrr_consume;
	}

	if (unlikely(0 == compare_and_swap32((uint32_t *) p_tx_confirm,
					     old_flag, new_flag))) {
		APP_DEBUG("Retrying...");
		goto reset_tx_confirm;
	}

	APP_DEBUG("Skipped...");
	return qman_cb_dqrr_consume;
}

/**
\brief          Ethernet frame processing function

\details        Processes frames received on TX error FQ. 

\param[in]      p_fd - Frame descriptor of frame to be forwarded
\param[in]      fqid - Frame queue on which frame was received

\return         None
*/
enum qman_cb_dqrr_result br_proc_tx_err_fq(struct qman_portal *p,
					   struct qman_fq *fq,
					   const struct qm_dqrr_entry *entry)
{
	APP_WARN("Frame on TX error FQID: %u", entry->fqid);

	lwe_free_fd(buff_allocator, &(entry->fd));
	return qman_cb_dqrr_consume;
}
