/**
 \file           br_swdb.c
 \brief          Switching database handling routines
*/
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include "br.h"
#include "rcu_lock.h"
#include "hash.h"
#include "timer_wheel.h"

br_swdb_t *br_swdb = NULL;

/**
\brief          Hash calculation function
\details        This function calculates a hash value using the MAC address
\param[in]      p_mac - MAC address
\return         Hash value
*/
uint32_t mac_hash(uint8_t * p_mac)
{
	uint32_t a, b;

	a = (p_mac[0] << 24) | (p_mac[1] << 16) | (p_mac[2] << 8) | (p_mac[3]);
	b = (p_mac[4] << 8) | (p_mac[5]);

	return (hash_2words(a, b, 0) % BR_NUM_SWDB_BUCKETS);
}

/**
\brief          Initialises the switching database

\details        This function will allocate memory and perform preliminary 
		initialization of switching database
		All the buckets are empty and locks are cleared initially

\param[in]      none

\return         Value indicating success(0) or failure

      \note           To be called only once, from single partition
*/
int br_swdb_init(void)
{
	uint32_t i;
	br_swdb = malloc(sizeof(br_swdb_t));

	if (unlikely(NULL == br_swdb)) {
		APP_ERROR("Error allocing swdb");
		return -1;
	}

	for (i = 0; i < BR_NUM_SWDB_BUCKETS; i++) {
		(br_swdb->buckets[i]).p_head = NULL;
		(br_swdb->buckets[i]).wlock = 0;	/*Unlocked */
	}

	return 0;
}

/**
\brief          Acquire a write lock

\details        Called while updating and existing entry (for example 
		port shifts) or creating a new entry in DB.

\param[in]      lock    - Lock to be acquired

      \return         none
*/
void br_swdb_lock(uint32_t * lock)
{
	do {
		rcu_synchronize_cpu(lwe_instance);
	} while (unlikely(spin_trylock(lock) == 0));
}

/**
\brief          Search a bucket for the dest address(key)

\details        This function will traverse the bucket and return the first
		matching entry

\param[in]      p_macaddr - mac address in byte array (search key)
\param[in]      p_p_bucket - Bucket head to traverse

\return         Matched Entry or NULL (if lookup failure)

      \note           This function is called with rcu_read_locked
*/
struct br_swdb_entry *br_swdb_entry_search(uint8_t * p_macaddr,
					   struct br_swdb_bucket **p_p_bucket)
{
	struct br_swdb_entry *p_node;
	struct br_swdb_entry **p_p_head;
	struct br_swdb_bucket *p_bucket;
	uint32_t swdb_index;

	swdb_index = mac_hash(p_macaddr);
	p_bucket = &(br_swdb->buckets[swdb_index]);
	p_p_head = &p_bucket->p_head;

	p_node = rcu_dereference(*p_p_head);

	while (NULL != p_node) {
		if (0 == memcmp(&(p_node->mac_addr[0]), p_macaddr, ETH_ALEN)) {
			break;
		}
		p_p_head = &(p_node->p_next);
		p_node = rcu_dereference(*p_p_head);
	}

	*p_p_bucket = p_bucket;
	return p_node;
}

/**
\brief          Create new entry in switching database

\details        This function will create a new entry structure for the 
		new mac address learnt.
		The created entry needs to be added into the switching db list

\param[in]      src_port    - Source port  on which the MAC was learnt
\param[in]      p_mac_Addr  - Learnt MAC Address (byte array)

\return         Pointer to the created entry structure

\note           Entry may be created outside a lock, however, insertion has 
      		to be within a write lock.
*/
br_swdb_entry_t *br_swdb_entry_create(uint32_t src_port, uint8_t * p_mac_addr)
{

	br_swdb_entry_t *p_entry;

	p_entry = (br_swdb_entry_t *) malloc(sizeof(br_swdb_entry_t));

	if (likely(p_entry)) {
		memcpy(&(p_entry->mac_addr[0]), p_mac_addr, ETH_ALEN);
		p_entry->port = src_port;
		p_entry->port_index = src_port;
		p_entry->is_static = 0;
		p_entry->ageing_stamp = get_current_time();
	}

	return p_entry;
}

/**
\brief          Inserts a new entry into the head of a bucket.

\param[in]      p_entry   - Entry to be inserted, all the fields are assumed to
			    be initialised 
\param[in]      p_bucket  - Head of the bucket in which to insert the entry.

\return         Pointer to the created entry

\note           Uses RCU assignment. To be called after acquiring the write 
      		lock on bucket
*/
struct br_swdb_entry *br_swdb_entry_insert(struct br_swdb_entry *p_entry,
					   br_swdb_bucket_t * p_bucket)
{
	p_entry->p_next = p_bucket->p_head;
	rcu_assign_pointer(p_bucket->p_head, p_entry);

	return p_entry;
}

/**
\brief          Learn new MAC address on a new port

\details        New MAC addresses are learnt as and when packets arrive from
	        ethernet ports and contain SMAC which are non-existent.
		First checks if the entry with the same MAC address exists.
		If yes, then check if the source port update is required
		(Host has moved to a different port since last update)

\param[in]      p_smac  - Src MAC Address to be learnt
\param[in]      sport   - Ingress port on which to 'learn' the MAC address

\return         Returns pointer to the created or matching entry
*/
br_swdb_entry_t *br_swdb_learn_mac(uint8_t * p_smac, uint32_t sport)
{
	br_swdb_entry_t *p_entry;
	struct br_swdb_bucket *p_bucket;

	p_entry = br_swdb_entry_search(p_smac, &p_bucket);

	if (likely(NULL != p_entry)) {
		if (likely(sport == p_entry->port)) {
			p_entry->ageing_stamp = get_current_time();
			return p_entry;
		}

		APP_INFO("Relearning MAC : " MAC_FMT " on port id: %u",
		  NMAC_STR(p_smac), sport);

		/* Need to remove old entry and insert new entry
		 * This is case of MAC entry update
		 */
		br_swdb_lock(&(p_bucket->wlock));
		p_entry = br_swdb_entry_search(p_smac, &p_bucket);

		/*Already deleted by ageing */
		if (unlikely(NULL == p_entry)) {
			APP_INFO("MAC : " MAC_FMT " already deleted by ageing",
				NMAC_STR(p_smac));
			goto create_new_entry_locked;
		}

		/*Already updated by other LWE */
		if (unlikely(sport == p_entry->port)) {
			spin_unlock(&(p_bucket->wlock));
			APP_INFO("MAC : " MAC_FMT " already updated",
				NMAC_STR(p_smac));
			return p_entry;
		}

		br_swdb_delete(p_entry, p_bucket);
		APP_INFO("Deleted old MAC : " MAC_FMT " from port: %u",
				NMAC_STR(p_smac), p_entry->port);

		goto create_new_entry_locked;
	}

	/* This is a new entry.  
	 * Create and insert this to the head of bucket if other have not
	 * done so yet
	 */
	br_swdb_lock(&(p_bucket->wlock));

	p_entry = br_swdb_entry_search(p_smac, &p_bucket);
	/* Already learnt by other LWE */
	if (unlikely(NULL != p_entry)) {
		spin_unlock(&(p_bucket->wlock));
		APP_INFO("MAC : " MAC_FMT " already learnt",
				NMAC_STR(p_smac));
		return p_entry;
	}

create_new_entry_locked:

	p_entry = br_swdb_entry_create(sport, p_smac);
	if (unlikely(NULL == p_entry)) {
		spin_unlock(&(p_bucket->wlock));
		APP_ERROR("Error creating SMAC " MAC_FMT, NMAC_STR(p_smac));
		return NULL;
	}

	br_swdb_entry_insert(p_entry, p_bucket);

	spin_unlock(&(p_bucket->wlock));

	APP_INFO("Learnt MAC : " MAC_FMT " on port id: %u",
		  NMAC_STR(p_smac), sport);


	return p_entry;
}

/**
\brief          Free an entry

\details        Assumes that the entry has already been removed from the 
		switching db list, frees the associated pointer.

\param[in]      p_entry   - Pointer to the entry to be freed
\param[in]      p_context - Unused as of now

\return         none

*/
void __br_swdb_free_entry(void *p_entry, void *p_context)
{
	free(p_entry);
}

/**
\brief          Remove an entry from switching database using RCU

\details        This routine removes an entry from a bucket and hands over
		to the RCU reclaimer.

\param[in]      p_entry    - Entry pointer to remove
\param[in]      p_bucket   - Bucket head from where to remove the entry

\return         Value indicating success(0).

\note           Typically called after an entry has been aged out
*/
int
br_swdb_delete(struct br_swdb_entry *p_entry, struct br_swdb_bucket *p_bucket)
{
	struct br_swdb_entry *p_node;
	struct br_swdb_entry **p_p_node;

	p_p_node = &(p_bucket->p_head);
	p_node = rcu_dereference(*p_p_node);

	if (p_node == p_entry) {
		rcu_assign_pointer(p_bucket->p_head, p_entry->p_next);
	} else {
		while (p_node->p_next != p_entry) {
			p_p_node = &(p_node->p_next);
			p_node = rcu_dereference(*p_p_node);
		}

		rcu_assign_pointer(p_node->p_next, p_entry->p_next);
	}

	rcu_free(__br_swdb_free_entry, p_entry, NULL);	/*Hand it to reclaimer */

	return 0;
}
