/**
 \file dpa_torture.c
 */
/*
 * Copyright (C) 2010-2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <io.h>
#include <errno.h>
#include <bitops.h>
#include <init.h>
#include <mmap.h>
#include <sg_api.h>
#include <dpa_alloc.h>
#include <fsl_qman.h>
#include <ip/ip.h>
#include <fm_types.h>
#include <barrier.h>
#include <app_common.h>
#include <dpa/dpa_eth.h>
#include <net/frame_desc.h>
#include <timebase.h>
#include <timer_wheel.h>
#include <shell.h>
#include <limits.h>
#include <ethernet/eth.h>
#include <dpa_torture.h>

#define BRG_QUEUE_COUNT	8
#define BRG_QUEUE_ID	0x4044
#define Max_D_P		1500
#define NUM_OF_PORTS	4
#define MAX_BUF_LEN	1578

__PERCPU struct dpa_buff_allocator *buff_allocator;
__PERCPU struct bman_pool *bpool[MAX_NUM_BMAN_POOLS];

uint32_t ingress_lock;
uint32_t egress_lock;
struct lwe_barrier ingress_barrier;
struct lwe_barrier egress_barrier;
struct qman_fq **ingress_brg_fqs;
struct qman_fq **egress_brg_fqs, *pfq;
struct qman_fq *egress_fq;
uint32_t recv_channel_map_ingress;
uint32_t recv_channel_map_egress;
struct pkt_annot *pkt_annotations;
struct qm_fd fd[MAX_BUF_LEN];
struct dpa_port_t *dp;
uint8_t eth_port;
int cnt_traffic_ports;
uint64_t start_time = 0;

__STATS unsigned long pkt_cnt_1;
__STATS unsigned long pkt_cnt_2;
__STATS unsigned long pkt_cnt_3;
__STATS unsigned long pkt_cnt_4;

/* per-port array of Tx FQs */
static struct qman_fq **egress_tx_fqs[NUM_OF_PORTS];

static enum qman_cb_dqrr_result rx_brg_fq(struct qman_portal *p,
						struct qman_fq *fq,
						const struct
						qm_dqrr_entry *dqrr);

struct qman_fq_cb qman_rx_cb = {rx_brg_fq, NULL, NULL, NULL};

/* creates the bridge frame queues for ingress */
int create_brg_queues(void)
{
	int i;
	uint32_t flags;
	struct qman_fq *pfq;

	int count = BRG_QUEUE_COUNT;
	int fqid = BRG_QUEUE_ID;

	struct qman_fq_cb qman_tx_cb = \
		{NULL, NULL, NULL, NULL};

	flags = QMAN_FQ_FLAG_LOCKED | QMAN_FQ_FLAG_NO_MODIFY;

	ingress_brg_fqs = (struct qman_fq **) memalign(CACHE_LINE_SIZE, \
				sizeof(struct qman_fq *) * count);
	if (ingress_brg_fqs == NULL) {
		APP_ERROR("error allocating ingress_brg_fqs");
		return -ENOMEM;
	}

	memset((char *)ingress_brg_fqs, 0, sizeof(struct qman_fq *) * count);

	for (i = 0; i < count; i++, fqid++) {
		pfq = (struct qman_fq *) memalign(CACHE_LINE_SIZE, \
			sizeof(struct qman_fq));
		if (pfq == NULL) {
			APP_ERROR("error allocating FQ with ID = 0x%x", fqid);
			goto err_ret;
		}

		ingress_brg_fqs[i] = pfq;
		pfq->cb = qman_tx_cb;
		if (qman_create_fq(fqid, flags , pfq) != 0) {
			APP_ERROR("qman_create_fq failed for FQID 0x%x", fqid);
			goto err_ret;
		}
		APP_INFO("qman_create_fq successful for FQID %d", fqid);
	}

	APP_INFO("Created bridge queues");
	return 0;

err_ret:
	for (i = 0; i < count; i++) {
		if (ingress_brg_fqs[i] == NULL)
			continue;
		qman_destroy_fq(ingress_brg_fqs[i], 0);
		free(ingress_brg_fqs[i]);
	}
	free(ingress_brg_fqs);

	return -1;
}

/**
\brief Creates the frame queues associated with DPA ethernet ports for ingress

\details Memory is allocated for FQ structures and FQs are created.

\param[in]	rx_default_cb - Callback function for default Rx FQs
\param[in]	priv_data_size - Private data size appended to each FQ
				structure.

\retval 0 - SUCCESS
\retval -1 - Failure

\note   Called only once from control core of partition
*/
int dpatort_fq_create_ingress(struct qman_fq_cb *rx_default_cb,
			      uint32_t priv_data_size)
{
	uint32_t port_id;
	uint32_t flags;
	int ret;

	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
			APP_INFO("Skip for MAC-less port: %u", port_id);
			continue;
		}

		APP_INFO("Creating FQs for port id: %u", port_id);

		flags = QMAN_FQ_FLAG_NO_ENQUEUE | QMAN_FQ_FLAG_LOCKED;
		ret = create_fqs(&dpa_eth_ports[port_id].ingress, flags,
				 rx_default_cb, "Rx default", priv_data_size);

		if (unlikely(0 != ret)) {
			APP_ERROR("create_fqs failed for Rx default");
			return -1;
		}
	}

	return 0;
}

/**
\brief Initialises the frame queues associated with DPA ethernet ports
	for ingress

\details RX queues are initialised with stashing ON.
	 TX queues are not initialised.

\param[in]	data_stash_size - Stashing size for frame data
\param[in]	ann_stash_size - Stashing size for frame annotation
\param[in]	ctx_stash_size - Stashing size for FQ context

\retval 0 - SUCCESS
\retval -1 - Failure

\note   Called only once from control core of the partition
*/
int dpatort_fq_init_ingress(uint32_t data_stash_size,
			    uint32_t ann_stash_size, uint32_t ctx_stash_size)
{
	uint32_t port_id;
	uint32_t flags;

	uint32_t ctx_lines;
	uint32_t ann_lines;
	uint32_t data_lines;
	uint32_t ctx_a_excl;

	struct qm_mcc_initfq opts;

	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		/*For MAC-less ports, no other FQ needs to be initialized */
		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
			APP_INFO("Skip FQs-init for MAC-less port: %u",
				 port_id);
			continue;
		}

		APP_INFO("Initializing FQs for port id: %u", port_id);

		flags = QMAN_INITFQ_FLAG_SCHED;
		opts.we_mask = QM_INITFQ_WE_DESTWQ |
		    QM_INITFQ_WE_CONTEXTA | QM_INITFQ_WE_FQCTRL;
		opts.fqd.fq_ctrl =
			 QM_FQCTRL_CTXASTASHING | QM_FQCTRL_LOCKINCACHE |
				QM_FQCTRL_CPCSTASH | QM_FQCTRL_AVOIDBLOCK;

		ctx_a_excl =
		    QM_STASHING_EXCL_ANNOTATION | QM_STASHING_EXCL_DATA;

		ctx_lines = GET_STASH_LINES(ctx_stash_size);
		data_lines = GET_STASH_LINES(data_stash_size);
		ann_lines = GET_STASH_LINES(ann_stash_size);

		opts.fqd.context_a.stashing.exclusive = ctx_a_excl;
		opts.fqd.context_a.stashing.annotation_cl = ann_lines;
		opts.fqd.context_a.stashing.data_cl = data_lines;
		opts.fqd.context_a.stashing.context_cl = ctx_lines;

		/*No need to set the address in context_a.
				This is set by driver itself */
		if (0 != init_fqs(&dpa_eth_ports[port_id].ingress, flags,
				  &opts, "Rx default")) {
			APP_ERROR("init_fqs failed for Rx default FQs");
			return -1;
		}
	}

	return 0;
}

/**
\brief		Calculates Port Information from Frame Queue ID

\param[in]	fq_id - Frame Queue Identifier

\return		Port ID of incoming frame
*/
static uint32_t fqid_to_port(uint32_t fq_id)
{
	uint32_t fq_start;
	uint32_t fq_end;
	uint32_t port_index;

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		fq_start = dpa_eth_ports[port_index].pcd_ingress.fq_start;
		fq_end = fq_start +
			dpa_eth_ports[port_index].pcd_ingress.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end)
			return port_index;
	}

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		fq_start = dpa_eth_ports[port_index].ingress.fq_start;
		fq_end = fq_start + dpa_eth_ports[port_index].ingress.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end)
			return port_index;
	}

	return INV_PORT;
}

/**
\brief		Send packet to bridge

\param[in]	p_fd - Frame descriptor to be send out

\return		0 on success, -ve value for failure
*/
static int dpatort_brg_send_out(struct qm_fd *p_fd,
				const struct qm_dqrr_entry *dqrr)
{
	struct qman_fq *p_fq;
	int32_t ret;

	p_fq = ingress_brg_fqs[0];

loop:
	ret = qman_enqueue(p_fq, p_fd, 0);

	if (unlikely(ret)) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	APP_DEBUG("Sent packet to bridge\n");
	return 0;
}

/* handler called when frame is received on RX default
			queues on ingress cores */
enum qman_cb_dqrr_result rx_default_fq(struct qman_portal *p,
					       struct qman_fq *fq,
					       const struct qm_dqrr_entry
					       *dqrr)
{
	uint8_t *p_pkt;

	uint8_t src_port;
	struct sg_entry_t *sgentry;
	phys_addr_t addr = 0;
	int32_t ret;
	uint32_t offset;

	APP_DEBUG("Rx packet on default FQID: %d", dqrr->fqid);

	if (likely(dqrr->fd.format == 0)) {
		APP_DEBUG("%s: received packet is of contiguous format",
			 __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo;
		offset = dqrr->fd.offset;
	} else if (unlikely(dqrr->fd.format == QM_FD_FORMAT_SG)) {
		APP_DEBUG("%s: received packet is of sg format", __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo + dqrr->fd.offset;
		sgentry = (struct sg_entry_t *)ptov(addr);

		addr = sgentry->addr_hi;
		addr = (addr << 32) + sgentry->addr_lo;
		offset = sgentry->offset;
	} else {
		APP_ERROR("Unsupported Frame Format");
		lwe_free_fd(buff_allocator, &dqrr->fd);
		return qman_cb_dqrr_consume;
	}

	pkt_annotations = ptov(addr);
	p_pkt = (uint8_t *)pkt_annotations + offset;

	src_port = fqid_to_port(dqrr->fqid);
	if (unlikely(INV_PORT == src_port)) {
		APP_ERROR("Frame from unknown port, Dropping it.");
		lwe_free_fd(buff_allocator, &dqrr->fd);
		return qman_cb_dqrr_consume;
	}

	pkt_annotations->port_id = (src_port + 1) % cnt_traffic_ports;

	ret = dpatort_brg_send_out((struct qm_fd *)&dqrr->fd, dqrr);

	if (unlikely(0 > ret)) {
		lwe_free_fd(buff_allocator, &dqrr->fd);
		return qman_cb_dqrr_consume;
	}

	return 0;
}

/*This function is called by all cores*/
static int dpatort_common_init(void)
{
	int ecode;

	/* Initialise the timer framework */
	if (0 != timer_framework_start()) {
		APP_ERROR("timer_framework_start Failed");
		return -1;
	}

	buff_allocator = dpa_allocator_create(&ecode);
	if (NULL == buff_allocator) {
		APP_ERROR("dpa_buff_allocator_create failed");
		return -1;
	}

	APP_INFO("Buffer allocator created...");
	return 0;
}

/*BMAN pool initialisation function*/
static int dpatort_bpools_init(void)
{
	struct bman_pool_params params;
	uint32_t bpid;

	/*Create BMAN pools */
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		/*Else create the pool object */
		params.flags = BMAN_POOL_FLAG_STOCKPILE;
		params.cb = NULL;
		params.cb_ctx = NULL;
		params.bpid = bpid;

		bpool[bpid] = bman_new_pool(&params);
		if (NULL == bpool[bpid]) {
			APP_ERROR("bman_new_pool failed for BPID: %u", bpid);
			return -1;
		}

		if (0 != dpa_allocator_add_bpool(buff_allocator,
						dpa_bpools[bpid].size,
						bpid, bpool[bpid])) {
			APP_ERROR("dpa_allocator_add_bpool failed");
			return -1;
		}

		APP_INFO("Added BPID: %u, Size: %u", bpid,
			 dpa_bpools[bpid].size);
	}

	return 0;
}

/**
\brief		Send packet to ethernet port

\param[in]	p_fd - Frame descriptor to be send out

\return		0 on success, -ve value for failure
*/
static int dpatort_eth_send_out(struct qm_fd *p_fd,
				const struct qm_dqrr_entry *dqrr)
{
	struct qman_fq *p_fq;
	int32_t ret;

	p_fq = egress_tx_fqs[pkt_annotations->port_id][0];

loop:
	ret = qman_enqueue(p_fq, p_fd, 0);

	if (unlikely(ret)) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	if (pkt_annotations->port_id == 0)
		decorated_notify_inc_32(&pkt_cnt_1);
	else if (pkt_annotations->port_id == 1)
		decorated_notify_inc_32(&pkt_cnt_2);
	else if (pkt_annotations->port_id == 2)
		decorated_notify_inc_32(&pkt_cnt_3);
	else if (pkt_annotations->port_id == 3)
		decorated_notify_inc_32(&pkt_cnt_4);

	APP_DEBUG("Sent packet to Fman");
	return 0;
}

/* handler called when frame is received on bridge queues on
				egress cores */
static enum qman_cb_dqrr_result rx_brg_fq(struct qman_portal *p,
						struct qman_fq *fq,
						const struct
						qm_dqrr_entry *dqrr)
{
	uint8_t *p_pkt;

	struct sg_entry_t *sgentry;
	phys_addr_t addr = 0;
	int32_t ret;
	uint32_t offset;

	if (likely(dqrr->fd.format == 0)) {
		APP_DEBUG("%s: received packet is of contiguous format",
			 __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo;
		offset = dqrr->fd.offset;
	} else if (unlikely(dqrr->fd.format == QM_FD_FORMAT_SG)) {
		APP_DEBUG("%s: received packet is of sg format", __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo + dqrr->fd.offset;
		sgentry = (struct sg_entry_t *)ptov(addr);

		addr = sgentry->addr_hi;
		addr = (addr << 32) + sgentry->addr_lo;
		offset = sgentry->offset;
	} else {
		APP_ERROR("Unsupported Frame Format");
		lwe_free_fd(buff_allocator, &dqrr->fd);
		return qman_cb_dqrr_consume;
	}

	pkt_annotations = ptov(addr);
	p_pkt = (uint8_t *)pkt_annotations + offset;

	ret = dpatort_eth_send_out((struct qm_fd *)&dqrr->fd, dqrr);

	if (unlikely(0 > ret)) {
		lwe_free_fd(buff_allocator, &dqrr->fd);
		return qman_cb_dqrr_consume;
	}

	return 0;
}

/* create and init the bridge FQs */
int egress_brg_queues_create_init(void)
{
	int i;
	struct qm_mcc_initfq opts;

	int count = BRG_QUEUE_COUNT;
	int fqid = BRG_QUEUE_ID;

	egress_brg_fqs = (struct qman_fq **) memalign(CACHE_LINE_SIZE, \
					sizeof(struct qman_fq *) * count);
	if (egress_brg_fqs == NULL) {
		APP_ERROR("error allocating egress_brg_fqs");
		return -ENOMEM;
	}

	memset((char *)egress_brg_fqs, 0, sizeof(struct qman_fq *) * count);

	for (i = 0; i < count; i++, fqid++) {
		pfq = (struct qman_fq *) memalign(CACHE_LINE_SIZE, \
						sizeof(struct qman_fq));
		if (pfq == NULL) {
			APP_ERROR("error allocating FQ with ID = 0x%x", fqid);
			goto err_ret;
		}

		egress_brg_fqs[i] = pfq;
		pfq->cb = qman_rx_cb;
		if (qman_create_fq(fqid, QMAN_FQ_FLAG_NO_ENQUEUE, pfq) != 0) {
			APP_ERROR("qman_create_fq failed for FQID 0x%x", fqid);
			goto err_ret;
		}
		APP_INFO("created fqid:%d", fqid);

		memset((char *)&opts, 0, sizeof(opts));

		opts.fqd.dest.channel =
			0x2d;
		opts.fqd.dest.wq = 1;

		opts.we_mask = QM_INITFQ_WE_DESTWQ;
		opts.fqd.fq_ctrl = QM_FQCTRL_CTXASTASHING |
					QM_FQCTRL_HOLDACTIVE;
		opts.fqd.context_a.stashing.data_cl = 0;
		if (unlikely(qman_init_fq(pfq, \
				QMAN_INITFQ_FLAG_SCHED, &opts) != 0)) {
			APP_ERROR("qman_init_fq failed for FQID 0x%x", fqid);
			goto err_ret;
		}
	}

	APP_INFO("Created and Inited bridge queues");

	return 0;

err_ret:
	for (i = 0; i < count; i++) {
		if (egress_brg_fqs[i] == NULL)
			continue;
		qman_destroy_fq(egress_brg_fqs[i], 0);
		free(egress_brg_fqs[i]);
	}
	free(egress_brg_fqs);

	return -1;
}

void qman_tx_err_cb(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg) {
	APP_INFO("error callback called");
}

/* create Tx queues */
int egress_transmit_queues_create(void)
{
	int i, j, count, fqid;
	struct qman_fq *pfq;
	struct qman_fq_cb qman_tx_cb = {NULL, qman_tx_err_cb, NULL, NULL};

	for (i = 0; i < 4; i++) {
		count = dpa_eth_ports[i].egress.fq_count;
		if (count == 0)
			continue;
		egress_tx_fqs[i] = \
			(struct qman_fq **) memalign(CACHE_LINE_SIZE,
			sizeof(struct qman_fq *) * count);
		if (egress_tx_fqs[i] == NULL) {
			APP_ERROR("error allocating " \
				"egress_tx_fqs for port %d", i);
			goto err_ret;
		}
		memset((char *)egress_tx_fqs[i], 0,
			sizeof(struct qman_fq *) * count);

		fqid = dpa_eth_ports[i].egress.fq_start;
		for (j = 0; j < count; j++, fqid++) {
			pfq = (struct qman_fq *) memalign(CACHE_LINE_SIZE,
				sizeof(struct qman_fq));
			if (pfq == NULL) {
				APP_ERROR("error allocating Tx FQ " \
					"with ID = 0x%x", fqid);
				goto err_ret;
			}

			egress_tx_fqs[i][j] = pfq;
			pfq->cb = qman_tx_cb;
			if (qman_create_fq(fqid, QMAN_FQ_FLAG_LOCKED | \
				QMAN_FQ_FLAG_TO_DCPORTAL, pfq) != 0) {
				APP_ERROR("qman_create_fq failed for FQID " \
					"0x%x", fqid);
				goto err_ret;
			}
			APP_INFO("qman_create_fq successful for FQID " \
					"0x%x %d %d", fqid, count, i);
		}
	}

	APP_INFO("Created Tx default queues");

	return 0;

err_ret:
	for (i = 0; i < NUM_OF_PORTS; i++) {
		count = dpa_eth_ports[i].egress.fq_count;
		if (count == 0)
			continue;
		for (j = 0; j < count; j++) {
			qman_destroy_fq(egress_tx_fqs[i][j], 0);
			free(egress_tx_fqs[i][j]);
		}
		free(egress_tx_fqs[i]);
	}

	return -1;
}

/* init Tx queues */
int egress_transmit_queues_init(void)
{
	int i, j, count;
	struct qm_mcc_initfq opts;

	opts.fqd.fq_ctrl = QM_FQCTRL_AVOIDBLOCK;
	opts.we_mask = QM_INITFQ_WE_DESTWQ | QM_INITFQ_WE_CONTEXTA |
		    QM_INITFQ_WE_CONTEXTB;
	opts.fqd.context_a.hi = 0x80000000;
	opts.fqd.context_a.lo = 0x0;
	opts.fqd.context_b = 0;	/*By default no confirmation needed */

	for (i = 0; i < 4; i++) {
		opts.fqd.dest.channel =
			dpa_eth_ports[i].egress.channel;
		opts.fqd.dest.wq = dpa_eth_ports[i].egress.work_queue;
		count = dpa_eth_ports[i].egress.fq_count;
		if (count == 0)
			continue;
		for (j = 0; j < count; j++) {
			if (qman_init_fq(egress_tx_fqs[i][j],
				QMAN_INITFQ_FLAG_SCHED, &opts) != 0) {
				APP_ERROR("qm_init_fq failed for Tx" \
					" fq_id: 0x%x", egress_tx_fqs[i][j]->\
						fqid);
				return -1;
			}
		}
	}
	APP_INFO("Inited Tx default queues");
	return 0;
}

/* create and initialize both the bridge and Tx FQs */
int create_and_init_queues(void)
{
	APP_INFO("Creating and initializing bridge FQs");
	if (egress_brg_queues_create_init() == -1)
		return -1;

	APP_INFO("Creating Tx FQs");
	if (egress_transmit_queues_create() == -1)
		return -1;

	APP_INFO("Initializing Tx FQs");
	if (egress_transmit_queues_init() == -1)
		return -1;

	return 0;
}

/* generate frame */
int generate_frame(uint32_t port_num, int num_of_frame, char *frame_data)
{
	int i;
	struct bm_buffer bman_buf[num_of_frame];
	struct pkt_annot *p_annot;
	struct ethernet_header_t *eth_hdr;
	uint16_t eth_payload_len;
	uint8_t *eth_payload;
	phys_addr_t addr;
	int len;

	eth_payload_len = strlen(frame_data);
	APP_DEBUG("payload length is: %d", eth_payload_len);

	len = 64 + ETH_HDR_LEN + eth_payload_len;

	for (i = 0; i < num_of_frame; i++) {
		if (0 >= dpa_allocator_get_buff(buff_allocator,
			len, &bman_buf[i])) {
			APP_ERROR("%s: couldn't allocate buf of size %d",
				__func__, len);
			return -ENOMEM;
		}
	}

	for (i = 0; i < num_of_frame; i++) {
		addr = 0;
		fd[i].addr_hi = (uint32_t)bman_buf[i].hi;
		fd[i].addr_lo = (uint32_t)bman_buf[i].lo;
		fd[i].bpid = bman_buf[i].bpid;
		fd[i].format = 0;
		fd[i].offset = 64;
		fd[i].length20 = fd[i].offset + ETH_HDR_LEN +
				eth_payload_len + 4;

		addr = (uint32_t) bman_buf[i].hi;
		addr = (addr << 32) | (uint32_t) bman_buf[i].lo;
		p_annot = ptov(addr);
		eth_hdr = (struct ethernet_header_t *)p_annot + fd[i].offset;

		memcpy(eth_hdr->destination.bytes,
			&dpa_eth_ports[(port_num + 1) % cnt_traffic_ports].mac_addr,
			ETH_ADDR_LEN);
		memcpy(eth_hdr->source.bytes, &dpa_eth_ports[port_num].mac_addr,
			ETH_ADDR_LEN);

		eth_hdr->proto = ETH_P_IP;
		APP_DEBUG("eth prototype: %x", eth_hdr->proto);

		eth_payload = ((uint8_t *)eth_hdr) + ETH_HDR_LEN;
		memcpy(eth_payload, (uint8_t *)frame_data, eth_payload_len);

		APP_DEBUG("packet%d formed\n", i + 1);
	}

	return 0;
}

/* pump traffic to specified port */
int pump_traffic(uint32_t port_num, int num_of_frame)
{
	int i;
	int32_t ret;

	egress_fq = egress_tx_fqs[port_num][0];

	for (i = 0; i < num_of_frame; i++) {
loop:
		ret = qman_enqueue(egress_fq, (struct qm_fd *)&fd[i], 0);

		if (unlikely(ret)) {
			uint64_t now, then = get_atb();
			do {
				now = get_atb();
			} while (now < (then + 1000));
			goto loop;
		}
	}

	return 0;
}

/* generate and pump frame to specified port */
int send_frame(uint32_t port_num, int num_of_frame, char *frame_data)
{
	generate_frame(port_num, num_of_frame, frame_data);
	pump_traffic(port_num, num_of_frame);

	return 0;
}

/* handler called when user gives send_frame cmd */
int shell_cmd_send_frame(int argc, char *argv[])
{
	int i, r;
	uint32_t port_num;
	int num_of_frame;
	char *frame_data;
	char *p;
	char *frame_data_pattern;
	int frame_data_len, frame_data_pattern_len;

	if (argc != 5) {
		printf("cmd usage : send_frame <port_num> "
			"<num_of_frame> <frame_data_len> "
			"<frame_data_pattern>\n");
		printf("valid port_num :\n");
		printf("0 - %d\n", cnt_traffic_ports - 1);
		printf("valid num_of_frame :\n");
		printf("1 - 10\n");
		printf("valid frame_data length :\n");
		printf("46 - 1499 (bytes)\n");
		printf("valid frame_data_pattern length :\n");
		printf("1 - 10 (bytes)\n");
		return -1;
	}

	port_num = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == port_num)
			|| (0 != errno && 0 == port_num)) {
		printf("strtoul failed converting command's argument1\n");
		return -1;
	}

	if (port_num < 0 || port_num > 3) {
		printf("wrong port_num entered\n");
		return -1;
	}

	num_of_frame = strtoul(argv[2], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == num_of_frame)
			|| (0 != errno && 0 == num_of_frame)) {
		printf("strtoul failed converting command's argument2\n");
		return -1;
	}

	if (num_of_frame < 1 || num_of_frame > 10) {
		printf("wrong num_of_frame entered\n");
		return -1;
	}

	frame_data_len  = strtoul(argv[3], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == frame_data_len)
			|| (0 != errno && 0 == frame_data_len)) {
		printf("strtoul failed converting command's argument3\n");
		return -1;
	}

	if (frame_data_len < 46 || frame_data_len > 1499) {
		printf("wrong frame_data length entered\n");
		return -1;
	}

	frame_data_pattern = argv[4];
	frame_data_pattern_len = strlen(frame_data_pattern);

	if (frame_data_pattern_len < 1 || frame_data_pattern_len > 10) {
		printf("frame_data_pattern of wrong length entered\n");
		return -1;
	}

	frame_data = malloc(1500 * sizeof(char));

	memset(frame_data, 0, 1500);

	p = frame_data;

	for (i = 0; i < frame_data_len / frame_data_pattern_len; i++) {
		memcpy(p, frame_data_pattern, frame_data_pattern_len);
		p += frame_data_pattern_len;
	}

	r = frame_data_len % frame_data_pattern_len;

	if (r != 0)
		memcpy(p, frame_data_pattern, r);

	send_frame(port_num, num_of_frame, frame_data);

	/* once initial seeding is done, note down the start time */
	start_time = get_timebase();

	free(frame_data);

	return 0;
}

/* handler called when user gives display_traffic_rate cmd */
int shell_cmd_display_traffic_rate(int argc, char *argv[])
{
	uint64_t curr_time = 0;
	uint64_t diff = 0;
	int total_num_pkts = 0;

	if (argc != 1) {
		printf("cmd usage: display_traffic_rate\n");
		return -1;
	}

	curr_time = get_timebase();

	if (start_time == 0)
		curr_time = 0;

	diff = curr_time - start_time;

	APP_INFO("traffic rate on port0: "
		"%d packets / %lld time units\n",
		atomic_read(&pkt_cnt_1), diff);
	APP_INFO("traffic rate on port1: "
		"%d packets / %lld time units\n",
		atomic_read(&pkt_cnt_2), diff);
	APP_INFO("traffic rate on port2: "
		"%d packets / %lld time units\n",
		atomic_read(&pkt_cnt_3), diff);
	APP_INFO("traffic rate on port3: "
		"%d packets / %lld time units\n",
		atomic_read(&pkt_cnt_4), diff);

	total_num_pkts = atomic_read(&pkt_cnt_1) + atomic_read(&pkt_cnt_2) +
			atomic_read(&pkt_cnt_3) + atomic_read(&pkt_cnt_4);

	APP_INFO("total traffic rate on ports 0-3: "
		"%d packets / %lld time units\n",
		total_num_pkts, diff);

	return 0;
}

/**
\brief   Main function of dpa_torture application.

\details Cores have been divided into two sets - ingress and egress.
	 Even numbered cores act as ingress cores and odd numbered
	 cores act as egress cores. All 1G ethernet ports are put in
	 loopback mode. Frames arriving at any 1G ethernet port are
	 dequeued by ingress cores and then enqueued to bridge queues.
	 Frames present in bridge queues are then dequeued by egress
	 cores and enqueued back to one of 1G ethernet ports. Thus,
	 traffic once generated and pumped into the ports keeps on
	 looping indefinitely between the ports, cores and bridge.

\note    Should never return.
*/
int main(int argc, char *argv[])
{
	struct qman_fq_cb rx_def_cb = {
			rx_default_fq, NULL, NULL, NULL };

	void *p_data = NULL;
	int port_id;
	int cnt_macless_ports = 0;

	APP_INFO("Welcome to dpa torture application!");

	spin_lock(&ingress_lock);

	if (lwe_instance == 0) {
		atomic_set(&pkt_cnt_1, 0);
		atomic_set(&pkt_cnt_2, 0);
		atomic_set(&pkt_cnt_3, 0);
		atomic_set(&pkt_cnt_4, 0);

		if (barrier_init(&ingress_barrier,
			get_online_core_mask() & 0x55)) {
			APP_INFO("Barrier initialization failed for ingress");
			return -EINVAL;
		}
		APP_INFO("Barrier initialized for ingress");
	}

	spin_unlock(&ingress_lock);

	spin_lock(&egress_lock);

	if (lwe_instance == 1) {
		if (barrier_init(&egress_barrier,
			get_online_core_mask() & 0xaa)) {
			APP_INFO("Barrier initialization failed for egress");
			return -EINVAL;
		}
		APP_INFO("Barrier initialized for egress");
	}

	spin_unlock(&egress_lock);

	/*Creating and initializing QMAN queues for ingress
						cores only */
	if (lwe_instance == 0) {
		if (dpa_eth_read(&recv_channel_map_ingress) < 0) {
			APP_ERROR("dpa_eth_read failed");
			goto err_out;
		}

		if (dpatort_fq_create_ingress(&rx_def_cb, 0) < 0) {
			APP_ERROR("dpatort_fq_create_ingress failed");
			goto err_out;
		}
		APP_INFO("queues created");

		APP_INFO("Num Ethernet ports discovered: %d",
			 g_num_dpa_eth_ports);

		for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
			if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
				cnt_macless_ports++;
				continue;
			}
		};

		cnt_traffic_ports = g_num_dpa_eth_ports - cnt_macless_ports;

		if (create_brg_queues() < 0) {
			APP_ERROR("create_brg_queues failed.");
			goto err_out;
		}
		APP_INFO("Bridge queues created successfully");

		if (unlikely(dpatort_fq_init_ingress(64, 64, 64) != 0)) {
			APP_ERROR("dpatort_fq_init_ingress failed");
			goto err_out;
		}
		APP_INFO("queues initialized");
	}

	/* waiting till DPA lib init finishes */
	while (g_num_dpa_eth_ports == 0)
		;

	if (lwe_instance % 2 == 0)
		barrier_sync(&ingress_barrier);

	if (lwe_instance % 2 != 0)
		barrier_sync(&egress_barrier);

	/*Common initialisation routine for all cores*/
	if (0 > dpatort_common_init()) {
		APP_ERROR("dpatort_common_init failed");
		goto err_out;
	}

	/*Create BMAN pools */
	if (0 > dpatort_bpools_init()) {
		APP_ERROR("dpatort_bpool_init failed");
		goto err_out;
	}

	if (lwe_instance % 2 == 0)
		APP_INFO("Frames to be recv on channel map: 0x%x",
				recv_channel_map_ingress);

	if (lwe_instance == 1) {
		recv_channel_map_egress = recv_channel_map_egress |
		    QM_SDQCR_CHANNELS_POOL(13);
	}

	if (lwe_instance % 2 != 0)
		APP_INFO("Frames to be recv on channel map: 0x%x",
				recv_channel_map_egress);

	if (lwe_instance == 1)
		create_and_init_queues();

	if (lwe_instance % 2 == 0) {
		if (barrier_sync(&ingress_barrier) < 0)
			panic("barrier sync for ingress failed. Exiting...");
	}

	if (lwe_instance % 2 != 0) {
		if (barrier_sync(&egress_barrier) < 0)
			panic("barrier sync for egress failed. Exiting...");
	}

	if (lwe_instance == 0) {
		if (unlikely(dpa_eth_enable_ports(false) != 0)) {
			APP_ERROR("dpa_eth_enable_ports failed");
			goto err_out;
		}
	}

	if (lwe_instance % 2 == 0) {
		APP_INFO("Waiting for recv channel info...");
		while (0 == recv_channel_map_ingress)
			;
	}

	if (lwe_instance % 2 != 0) {
		APP_INFO("Waiting for recv channel info...");
		while (0 == recv_channel_map_egress)
			;
	}

	/*Schedule the channel_map */
	if (lwe_instance % 2 != 0) {
		/* subscribe the portal to the channel for static dequeues */
		APP_INFO("recv_channel_map on the odd numbered "
			"(egress) cores: 0x%x\n", recv_channel_map_egress);
		qman_static_dequeue_add(recv_channel_map_egress);
	}

	/* better for all the egress cores start jobs at the same time */
	if (lwe_instance % 2 != 0) {
		if (barrier_sync(&egress_barrier) < 0)
			panic("barrier sync failed for egress. Exiting...");

		APP_INFO("Starting frame dequeue...");
	}

	/*Schedule the channel_map */
	if (lwe_instance % 2 == 0) {
		/* subscribe the portal to the channel for static dequeues */
		APP_INFO("recv_channel_map on the even numbered "
			"(ingress) cores: 0x%x\n", recv_channel_map_ingress);
		qman_static_dequeue_add(recv_channel_map_ingress);
	}

	/* better for all the ingress cores start jobs at the same time */
	if (lwe_instance % 2 == 0) {
		if (barrier_sync(&ingress_barrier) < 0)
			panic("barrier sync failed for ingress. Exiting...");

		APP_INFO("Starting frame dequeue...");
	}

	if (lwe_instance == 0) {
		/* Add shell command for mac loopback */
		if (0 != add_command("loopback", shell_cmd_loopback,
				"mac loopback")) {
			APP_ERROR("add_command failed for mac loopback");
			return -1;
		}

		/* Add command for generate and pump frames */
		if (0 != add_command("send_frame", shell_cmd_send_frame,
				"generate and pump frames")) {
			APP_ERROR("add_command failed for send_frame");
			return -1;
		}

		/* Add command for displaying traffic rate flowing */
		/* through DPA ethernet ports */
		if (0 != add_command("display_traffic_rate",
				shell_cmd_display_traffic_rate,
				"display traffic rate flowing through DPA "
				"ethernet ports")) {
			APP_ERROR("add_command failed for "
					"display_traffic_rate");
			return -1;
		}
	}

	while (1) {
		qman_poll();
		swi_activate(&p_data);
	}

	return 0;

err_out:
	APP_ERROR("FATAL error, exiting");
	return -1;
}
