/** @file
 * Decorated Storage Performance Measurement
 */
/*
 * Copyright (c) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <app_common.h>
#include <timebase.h>
#include <statistics.h>
#include <compat.h>
#include <dspm.h>
#include <errno.h>
#include <limits.h>
#include <swi.h>
#include <barrier.h>
#include <init.h>
#include <shell.h>
#include <misc.h>
#include <msl_math_api_gcc.h>
#include <lwe_malloc.h>

#define TEST_1 1
#define TEST_2 2
#define TEST_3 3
#define TEST_4 4
#define TEST_5 5
#define TEST_6 6
#define TEST_7 7
#define TEST_8 8

struct lwe_barrier bar;

__PERCPU uint32_t curr_lwe_id = 0;	/**< LWE ID for each core */

__STATS unsigned long var;

unsigned long var1;

__STATS struct stat32_pair_t var2;

struct stat32_pair_t var3;

spinlock_t dseq_lock = SPIN_LOCK_UNLOCKED;

static inline unsigned long read(unsigned long *a)
{
	return *a;
}

static inline unsigned long set(unsigned long *a, uint32_t val)
{
	*a = val;
	return *a;
}

static inline unsigned long dec(unsigned long *a)
{
	*a -= 1;
	return *a;
}

static inline unsigned long inc(unsigned long *a)
{
	*a += 1;
	return *a;
}

static inline unsigned long add(unsigned long *a, uint32_t val)
{
	*a = *a + val;
	return *a;
}

int test_decorated_load_clear_32(int argc, char *argv[])
{
	int i, core_mask, num_cores_run_test, pass;
	uint64_t t1 = 0, t2 = 0, t3 = 0, t4 = 0;
	uint64_t avg_time_decor_func = 0, avg_time_non_decor_func = 0;
	uint32_t *p_stat, *p_nonstat, *res;
	unsigned long int num_exec;

	/* Necessary checks are carried out on these
	 * arguments in shell cmd handler (for DSPM) from where
	 * they are passed
	 */
	num_exec = strtoul(argv[0], NULL, 0);
	if ((ULONG_MAX == num_exec && ERANGE == errno)
		 || (0 != errno && 0 == num_exec)) {
		APP_ERROR("failed to convert argument 1 of"
				"test_decorated_load_clear_32");
		return -1;
	}
	core_mask = strtoul(argv[1], NULL, 0);

	num_cores_run_test = __msl_count_trailing_one32(core_mask);

	pass = 0;

	p_stat = stats_malloc(num_exec * sizeof(uint32_t));

	if (p_stat == NULL) {
		APP_ERROR("stats_malloc failed\n");
		return -1;
	}

	memset(p_stat, 0, num_exec * sizeof(uint32_t));

	for (i = 0; i < num_exec; i++)
		*(p_stat + i) = 0x12345678;

	res = malloc(num_exec * sizeof(uint32_t));

	if (res == NULL) {
		APP_ERROR("malloc failed\n");
		stats_free(p_stat);
		return -1;
	}

	memset(res, 0, num_exec * sizeof(uint32_t));

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(res);
			return -1;
		}
	}

	t1 = get_timebase();
	APP_DEBUG("t1 :: %llx\n", t1);

	for (i = 0; i < num_exec; i++)
		*(res + i) = decorated_load_clear_32(p_stat + i);

	t2 = get_timebase();
	APP_DEBUG("t2 :: %llx\n", t2);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(res);
			return -1;
		}
	}

	for (i = 0; i < num_exec; i++) {
		if (*(res + i) == 0x12345678 && *(p_stat + i) == 0) {
			pass = 1;
		} else {
			pass = 0;
			break;
		}
	}

	if (pass) {
		avg_time_decor_func = (t2 - t1)/num_exec;
	} else {
		APP_ERROR("decorated_load_clear_32 didn't function "
			"correctly\n");
		stats_free(p_stat);
		free(res);
		return -1;
	}

	p_nonstat = malloc(num_exec * sizeof(uint32_t));

	if (p_nonstat == NULL) {
		APP_ERROR("malloc failed \n");
		stats_free(p_stat);
		free(res);
		return -1;
	}

	for (i = 0; i < num_exec; i++)
		*(p_nonstat + i) = 0x12345678;

	memset(res, 0, num_exec * sizeof(uint32_t));

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(res);
			free(p_nonstat);
			return -1;
		}
	}

	t3 = get_timebase();
	APP_DEBUG("t3 :: %llx\n", t3);

	for (i = 0; i < num_exec; i++) {
		spin_lock(&dseq_lock);

		*(res + i) = read((unsigned long *) p_nonstat + i);
		set((unsigned long *) p_nonstat + i, 0x0); /* clear */

		spin_unlock(&dseq_lock);
	}

	t4 = get_timebase();
	APP_DEBUG("t4 :: %llx\n", t4);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(res);
			free(p_nonstat);
			return -1;
		}
	}

	for (i = 0; i < num_exec; i++) {
		if (*(res + i) == 0x12345678 && *(p_nonstat + i) == 0) {
			pass = 1;
		} else {
			pass = 0;
			break;
		}
	}

	if (pass) {
		avg_time_non_decor_func = (t4 - t3)/num_exec;
	} else {
		APP_ERROR("equivalent functions (wrt decorated_load_clear_32) "
			"didn't function correctly\n");
		stats_free(p_stat);
		free(res);
		free(p_nonstat);
		return -1;
	}

	stats_free(p_stat);
	free(res);
	free(p_nonstat);

	APP_INFO("Average time taken by %ld iterations of "
		"decorated_load_clear_32 is %lld units\n", num_exec,
		avg_time_decor_func);

	APP_INFO("Average time taken by %ld iterations of equivalent functions "
		"(wrt decorated_load_clear_32) is %lld units\n", num_exec,
		avg_time_non_decor_func);

	if (avg_time_non_decor_func > avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_clear_32 when compared to same number "
			"of iterations of equivalent functions is lesser by "
			"%lld units\n", num_exec,
			avg_time_non_decor_func - avg_time_decor_func);
	} else if (avg_time_non_decor_func < avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_clear_32 when compared to same number "
			"of iterations of equivalent functions is greater by "
			"%lld units\n", num_exec,
			avg_time_decor_func - avg_time_non_decor_func);
	} else {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_clear_32 is equal to average time "
			"taken by same number of iterations "
			"of equivalent functions ", num_exec);
	}

	return 0;
}

int test_decorated_load_set_32(int argc, char *argv[])
{
	int i, core_mask, num_cores_run_test, pass;
	unsigned long int num_exec;
	uint64_t t1 = 0, t2 = 0, t3 = 0, t4 = 0;
	uint64_t avg_time_decor_func = 0, avg_time_non_decor_func = 0;
	uint32_t *p_stat, *p_nonstat, *res;

	/* Necessary checks are carried out on these
	 * arguments in shell cmd handler (for DSPM) from where
	 * they are passed
	 */
	num_exec = strtoul(argv[0], NULL, 0);
	if ((ULONG_MAX == num_exec && ERANGE == errno)
		|| (0 != errno && 0 == num_exec)) {
		APP_ERROR("failed to convert argument 1 of"
				"test_decorated_load_set_32");
		return -1;
	}
	core_mask = strtoul(argv[1], NULL, 0);

	num_cores_run_test = __msl_count_trailing_one32(core_mask);

	pass = 0;

	p_stat = stats_malloc(num_exec * sizeof(uint32_t));

	if (p_stat == NULL) {
		APP_ERROR("stats_malloc failed \n");
		return -1;
	}

	memset(p_stat, 0, num_exec * sizeof(uint32_t));

	for (i = 0; i < num_exec; i++)
		*(p_stat + i) = 0x12345678;

	res = malloc(num_exec * sizeof(uint32_t));

	if (res == NULL) {
		APP_ERROR("malloc failed\n");
		stats_free(p_stat);
		return -1;
	}

	memset(res, 0, num_exec * sizeof(uint32_t));

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(res);
			return -1;
		}
	}

	t1 = get_timebase();
	APP_DEBUG("t1 :: %llx\n", t1);

	for (i = 0; i < num_exec; i++)
		*(res + i) = decorated_load_set_32(p_stat + i);

	t2 = get_timebase();
	APP_DEBUG("t2 :: %llx\n", t2);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(res);
			return -1;
		}
	}

	for (i = 0; i < num_exec; i++) {
		if (*(res + i) == 0x12345678 && *(p_stat + i) == 0xffffffff) {
			pass = 1;
		} else {
			pass = 0;
			break;
		}
	}

	if (pass) {
		avg_time_decor_func = (t2 - t1)/num_exec;
	} else {
		APP_ERROR("decorated_load_set_32 didn't function "
			"correctly\n");
		stats_free(p_stat);
		free(res);
		return -1;
	}

	p_nonstat = malloc(num_exec * sizeof(uint32_t));

	if (p_nonstat == NULL) {
		APP_ERROR("malloc failed \n");
		stats_free(p_stat);
		free(res);
		return -1;
	}

	for (i = 0; i < num_exec; i++)
		*(p_nonstat + i) = 0x12345678;

	memset(res, 0, num_exec * sizeof(uint32_t));

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(res);
			free(p_nonstat);
			return -1;
		}
	}

	t3 = get_timebase();
	APP_DEBUG("t3 :: %llx\n", t3);

	for (i = 0; i < num_exec; i++) {
		spin_lock(&dseq_lock);

		*(res + i) = read((unsigned long *) p_nonstat + i);
		set((unsigned long *) p_nonstat + i, 0xffffffff); /* set */

		spin_unlock(&dseq_lock);
	}

	t4 = get_timebase();
	APP_DEBUG("t4 :: %llx\n", t4);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(res);
			free(p_nonstat);
			return -1;
		}
	}

	for (i = 0; i < num_exec; i++) {
		if (*(res + i) == 0x12345678 &&
			*(p_nonstat + i) == 0xffffffff) {
			pass = 1;
		} else {
			pass = 0;
			break;
		}
	}

	if (pass) {
		avg_time_non_decor_func = (t4 - t3)/num_exec;
	} else {
		APP_ERROR("equivalent functions (wrt decorated_load_set_32) "
			"didn't function correctly\n");
		stats_free(p_stat);
		free(res);
		free(p_nonstat);
		return -1;
	}

	stats_free(p_stat);
	free(res);
	free(p_nonstat);

	APP_INFO("Average time taken by %ld iterations of"
			"decorated_load_set_32 is %lld units\n",
					num_exec, avg_time_decor_func);

	APP_INFO("Average time taken by %ld iterations of equivalent functions "
		"(wrt decorated_load_set_32) is %lld units\n", num_exec,
		avg_time_non_decor_func);

	if (avg_time_non_decor_func > avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_set_32 when compared to same number "
			"of iterations of equivalent functions is lesser by "
			"%lld units\n", num_exec,
			avg_time_non_decor_func - avg_time_decor_func);
	} else if (avg_time_non_decor_func < avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_set_32 when compared to same number "
			"of iterations of equivalent functions is greater by "
			"%lld units\n", num_exec,
			avg_time_decor_func - avg_time_non_decor_func);
	} else {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_set_32 is equal to average time "
			"taken by same number of iterations of equivalent "
			"functions ", num_exec);
	}

	return 0;
}

int test_decorated_load_dec_32(int argc, char *argv[])
{
	int i, core_mask, num_cores_run_test;
	unsigned long int num_exec;
	uint64_t t1 = 0, t2 = 0, t3 = 0, t4 = 0;
	uint64_t avg_time_decor_func = 0, avg_time_non_decor_func = 0;

	/* Necessary checks are carried out on these
	 * arguments in shell cmd handler (for DSPM) from where
	 * they are passed
	 */
	num_exec = strtoul(argv[0], NULL, 0);
	if ((ULONG_MAX == num_exec && ERANGE == errno)
		|| (0 != errno && 0 == num_exec)) {
		APP_ERROR("Failed to convert argument 1 of"
				"test_decorated_load_dec_32");
		return -1;
	}
	core_mask = strtoul(argv[1], NULL, 0);

	num_cores_run_test = __msl_count_trailing_one32(core_mask);

	if (curr_lwe_id == 0)
		atomic_set(&var, num_cores_run_test * num_exec);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t1 = get_timebase();
	APP_DEBUG("t1 :: %llx\n", t1);

	for (i = 0; i < num_exec; i++)
		decorated_load_dec_32(&var);

	t2 = get_timebase();
	APP_DEBUG("t2 :: %llx\n", t2);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if (!(atomic_read(&var))) {
			avg_time_decor_func = (t2 - t1)/num_exec;
		} else {
			APP_ERROR("decorated_load_dec_32 didn't function "
				"correctly\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0)
		atomic_set(&var1, num_cores_run_test * num_exec);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t3 = get_timebase();
	APP_DEBUG("t3 :: %llx\n", t3);

	for (i = 0; i < num_exec; i++) {
		spin_lock(&dseq_lock);

		read(&var1);
		dec(&var1); /* dec */

		spin_unlock(&dseq_lock);
	}

	t4 = get_timebase();
	APP_DEBUG("t4 :: %llx\n", t4);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if (!(atomic_read(&var1))) {
			avg_time_non_decor_func = (t4 - t3)/num_exec;
		} else {
			APP_ERROR("equivalent fucntions "
				"(wrt decorated_load_dec_32) "
				"didn't function correctly\n");
			return -1;
		}
	}

	APP_INFO("Average time taken by %ld iterations of"
			"decorated_load_dec_32 is %lld units\n",
					 num_exec, avg_time_decor_func);

	APP_INFO("Average time taken by %ld iterations of equivalent functions "
		"(wrt decorated_load_dec_32) is %lld units\n", num_exec,
		avg_time_non_decor_func);

	if (avg_time_non_decor_func > avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_dec_32 when compared to same number "
			"of iterations of equivalent functions is lesser by "
			"%lld units\n", num_exec,
			avg_time_non_decor_func - avg_time_decor_func);
	} else if (avg_time_non_decor_func < avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_dec_32 when compared to same number "
			"of iterations of equivalent functions is greater by "
			"%lld units\n", num_exec,
			avg_time_decor_func - avg_time_non_decor_func);
	} else {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_dec_32 is equal to average time "
			"taken by same number of iterations of equivalent "
			"functions ", num_exec);
	}

	return 0;
}

int test_decorated_load_inc_32(int argc, char *argv[])
{
	int i, core_mask, num_cores_run_test;
	unsigned long int num_exec;
	uint64_t t1 = 0, t2 = 0, t3 = 0, t4 = 0;
	uint64_t avg_time_decor_func = 0, avg_time_non_decor_func = 0;

	/* Necessary checks are carried out on these
	 * arguments in shell cmd handler (for DSPM) from where
	 * they are passed
	 */
	num_exec = strtoul(argv[0], NULL, 0);
	if ((ULONG_MAX == num_exec && ERANGE == errno)
		|| (0 != errno && 0 == num_exec)) {
		APP_ERROR("Failed to convert argument 1 of"
				"test_decorated_load_inc_32");
		return -1;
	}
	core_mask = strtoul(argv[1], NULL, 0);

	num_cores_run_test = __msl_count_trailing_one32(core_mask);

	if (curr_lwe_id == 0)
		atomic_set(&var, 0x0);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t1 = get_timebase();
	APP_DEBUG("t1 :: %llx\n", t1);

	for (i = 0; i < num_exec; i++)
		decorated_load_inc_32(&var);

	t2 = get_timebase();
	APP_DEBUG("t2 :: %llx\n", t2);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if ((atomic_read(&var)) == num_cores_run_test * num_exec) {
			avg_time_decor_func = (t2 - t1)/num_exec;
		} else {
			APP_ERROR("decorated_load_inc_32 didn't function "
				"correctly\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0)
		atomic_set(&var1, 0x0);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t3 = get_timebase();
	APP_DEBUG("t3 :: %llx\n", t3);

	for (i = 0; i < num_exec; i++) {
		spin_lock(&dseq_lock);

		read(&var1);
		inc(&var1); /* inc */

		spin_unlock(&dseq_lock);
	}

	t4 = get_timebase();
	APP_DEBUG("t4 :: %llx\n", t4);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if ((atomic_read(&var1)) == num_cores_run_test * num_exec) {
			avg_time_non_decor_func = (t4 - t3)/num_exec;
		} else {
			APP_ERROR("equivalent functions "
				"(wrt decorated_load_inc_32)"
				"didn't function correctly\n");
			return -1;
		}
	}

	APP_INFO("Average time taken by %ld iterations of "
		"decorated_load_inc_32 is %lld units\n", num_exec,
		avg_time_decor_func);

	APP_INFO("Average time taken by %ld iterations of equivalent functions "
		"(wrt decorated_load_inc_32) is %lld units\n", num_exec,
		avg_time_non_decor_func);

	if (avg_time_non_decor_func > avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_inc_32 when compared to same number "
			"of iterations of equivalent functions is lesser by "
			"%lld units\n", num_exec,
			avg_time_non_decor_func - avg_time_decor_func);
	} else if (avg_time_non_decor_func < avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_inc_32 when compared to same number "
			"of iterations of equivalent functions is greater by "
			"%lld units\n", num_exec,
			avg_time_decor_func - avg_time_non_decor_func);
	} else {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_load_inc_32 is equal to average time "
			"taken by same number of iterations of equivalent"
			"functions ", num_exec);
	}

	return 0;
}

int test_decorated_store_32_acc_32(int argc, char *argv[])
{
	int i, core_mask, num_cores_run_test;
	unsigned long int num_exec;
	uint64_t t1 = 0, t2 = 0, t3 = 0, t4 = 0;
	uint64_t avg_time_decor_func = 0, avg_time_non_decor_func = 0;

	/* Necessary checks are carried out on these
	 * arguments in shell cmd handler (for DSPM) from where
	 * they are passed
	 */
	num_exec = strtoul(argv[0], NULL, 0);
	if ((ULONG_MAX == num_exec && ERANGE == errno)
		|| (0 != errno && 0 == num_exec)) {
		APP_ERROR("Failed to convert argument 1 of"
				"test_decorated_store_32_acc_32");
		return -1;
	}
	core_mask = strtoul(argv[1], NULL, 0);

	num_cores_run_test = __msl_count_trailing_one32(core_mask);

	if (curr_lwe_id == 0)
		atomic_set(&var, 0x0);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t1 = get_timebase();
	APP_DEBUG("t1 :: %llx\n", t1);

	for (i = 0; i < num_exec; i++)
		decorated_store_32_acc_32(&var, 0x12345678);

	t2 = get_timebase();
	APP_DEBUG("t2 :: %llx\n", t2);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if (atomic_read(&var) == num_cores_run_test *
			num_exec * 0x12345678) {
			avg_time_decor_func = (t2 - t1)/num_exec;
		} else {
			APP_ERROR("decorated_store_32_acc_32 didn't function "
				"correctly\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0)
		atomic_set(&var1, 0x0);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t3 = get_timebase();
	APP_DEBUG("t3 :: %llx\n", t3);

	for (i = 0; i < num_exec; i++)
		atomic_add(&var1, 0x12345678);

	t4 = get_timebase();
	APP_DEBUG("t4 :: %llx\n", t4);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if (atomic_read(&var1) == num_cores_run_test *
			num_exec * 0x12345678) {
			avg_time_non_decor_func = (t4 - t3)/num_exec;
		} else {
			APP_ERROR("equivalent functions "
				"(wrt decorated_store_32_acc_32)"
				"didn't function correctly\n");
			return -1;
		}
	}

	APP_INFO("Average time taken by %ld iterations of "
		"decorated_store_32_acc_32 is %lld units\n", num_exec,
		avg_time_decor_func);

	APP_INFO("Average time taken by %ld iterations of equivalent functions "
		"(wrt decorated_store_32_acc_32) is %lld units\n", num_exec,
		avg_time_non_decor_func);

	if (avg_time_non_decor_func > avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_store_32_acc_32 when compared to same "
			"number of iterations of equivalent functions is "
			"lesser by %lld units\n", num_exec,
			avg_time_non_decor_func - avg_time_decor_func);
	} else if (avg_time_non_decor_func < avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_store_32_acc_32 when compared to "
			"same number of iterations of equivalent functions "
			"is greater by %lld units\n", num_exec,
			avg_time_decor_func - avg_time_non_decor_func);
	} else {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_store_32_acc_32 is equal to average time "
			"taken by same number of iterations of equivalent "
			"functions ", num_exec);
	}

	return 0;
}

int test_decorated_store_32_inc_acc_32(int argc, char *argv[])
{
	int i, core_mask, num_cores_run_test;
	unsigned long int num_exec;
	uint64_t t1 = 0, t2 = 0, t3 = 0, t4 = 0;
	uint64_t avg_time_decor_func = 0, avg_time_non_decor_func = 0;

	/* Necessary checks are carried out on these
	 * arguments in shell cmd handler (for DSPM) from where
	 * they are passed
	 */
	num_exec = strtoul(argv[0], NULL, 0);
	if ((ULONG_MAX == num_exec && ERANGE == errno)
		|| (0 != errno && 0 == num_exec)) {
		APP_ERROR("Failed to convert argument 1 of"
				 "test_decorated_store_32_inc_acc_32");
		return -1;
	}
	core_mask = strtoul(argv[1], NULL, 0);

	num_cores_run_test = __msl_count_trailing_one32(core_mask);

	if (curr_lwe_id == 0) {
		atomic_set((unsigned long *) &var2.inc, 0x0);
		atomic_set((unsigned long *) &var2.acc, 0x0);
	}

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t1 = get_timebase();
	APP_DEBUG("t1 :: %llx\n", t1);

	for (i = 0; i < num_exec; i++)
		decorated_store_32_inc_acc_32(&var2, 0x12345678);

	t2 = get_timebase();
	APP_DEBUG("t2 :: %llx\n", t2);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if (atomic_read((unsigned long *) &var2.inc) ==
			num_cores_run_test * num_exec &&
			atomic_read((unsigned long *) &var2.acc) ==
			num_cores_run_test * num_exec * 0x12345678) {
			avg_time_decor_func = (t2 - t1)/num_exec;
		} else {
			APP_ERROR("decorated_store_32_inc_acc_32 didn't "
				"function correctly\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		atomic_set((unsigned long *) &var3.inc, 0x0);
		atomic_set((unsigned long *) &var3.acc, 0x0);
	}

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t3 = get_timebase();
	APP_DEBUG("t3 :: %llx\n", t3);

	for (i = 0; i < num_exec; i++) {
		spin_lock(&dseq_lock);

		inc((unsigned long *) &var3.inc); /* inc */
		add((unsigned long *) &var3.acc, 0x12345678); /* accumulate */

		spin_unlock(&dseq_lock);
	}

	t4 = get_timebase();
	APP_DEBUG("t4 :: %llx\n", t4);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if (atomic_read((unsigned long *) &var3.inc) ==
			num_cores_run_test * num_exec &&
			atomic_read((unsigned long *) &var3.acc) ==
			num_cores_run_test * num_exec * 0x12345678) {
			avg_time_non_decor_func = (t4 - t3)/num_exec;
		} else {
			APP_ERROR("equivalent functions "
				"(wrt decorated_store_32_inc_acc_32 "
				"didn't function correctly\n");
			return -1;
		}
	}

	APP_INFO("Average time taken by %ld iterations of "
		"decorated_store_32_inc_acc_32 is "
		"%lld units\n", num_exec, avg_time_decor_func);

	APP_INFO("Average time taken by %ld iterations of equivalent functions "
		"(wrt decorated_store_32_inc_acc_32) is %lld units\n", num_exec,
		avg_time_non_decor_func);

	if (avg_time_non_decor_func > avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_store_32_inc_acc_32 when compared to same "
			"number of iterations of equivalent functions "
			"is lesser by %lld units\n", num_exec,
			avg_time_non_decor_func - avg_time_decor_func);
	} else if (avg_time_non_decor_func < avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_store_32_inc_acc_32 when compared to same "
			"number of iterations of equivalent functions "
			"is greater by %lld units\n", num_exec,
			avg_time_decor_func - avg_time_non_decor_func);
	} else {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_store_32_inc_acc_32 is equal to average "
			"time taken by same number of iterations of "
			"equivalent functions ", num_exec);
	}

	return 0;
}

int test_decorated_notify_inc_32(int argc, char *argv[])
{
	int i, core_mask, num_cores_run_test;
	unsigned long int num_exec;
	uint64_t t1 = 0, t2 = 0, t3 = 0, t4 = 0;
	uint64_t avg_time_decor_func = 0, avg_time_non_decor_func = 0;

	/* Necessary checks are carried out on these
	 * arguments in shell cmd handler (for DSPM) from where
	 * they are passed
	 */
	num_exec = strtoul(argv[0], NULL, 0);
	if ((ULONG_MAX == num_exec && ERANGE == errno)
		|| (0 != errno && 0 == num_exec)) {
		APP_ERROR("Failed to convert argument 1 of"
				"test_decorated_notify_inc_32");
		return -1;
	}
	core_mask = strtoul(argv[1], NULL, 0);

	num_cores_run_test = __msl_count_trailing_one32(core_mask);

	if (curr_lwe_id == 0)
		atomic_set(&var, 0x0);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t1 = get_timebase();
	APP_DEBUG("t1 :: %llx\n", t1);

	for (i = 0; i < num_exec; i++)
		decorated_notify_inc_32(&var);

	t2 = get_timebase();
	APP_DEBUG("t2 :: %llx\n", t2);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if (atomic_read(&var) == num_cores_run_test * num_exec) {
			avg_time_decor_func = (t2 - t1)/num_exec;
		} else {
			APP_ERROR("decorated_notify_inc_32 didn't "
				"function correctly\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0)
		atomic_set(&var1, 0x0);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	t3 = get_timebase();
	APP_DEBUG("t3 :: %llx\n", t3);

	for (i = 0; i < num_exec; i++)
		atomic_inc(&var1);

	t4 = get_timebase();
	APP_DEBUG("t4 :: %llx\n", t4);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		if (atomic_read(&var1) == num_cores_run_test * num_exec) {
			avg_time_non_decor_func = (t4 - t3)/num_exec;
		} else {
			APP_ERROR("equivalent functions "
				"(wrt decorated_notify_inc_32) "
				"didn't function correctly\n");
			return -1;
		}
	}

	APP_INFO("Average time taken by %ld iterations of "
		"decorated_notify_inc_32 is "
		"%lld units\n", num_exec, avg_time_decor_func);

	APP_INFO("Average time taken by %ld iterations of equivalent functions "
		"(wrt decorated_notify_inc_32) is %lld units\n", num_exec,
		avg_time_non_decor_func);

	if (avg_time_non_decor_func > avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_notify_inc_32 when compared to same number "
			"of iterations of equivalent functions "
			"is lesser by %lld units\n", num_exec,
			avg_time_non_decor_func - avg_time_decor_func);
	} else if (avg_time_non_decor_func < avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_notify_inc_32 when "
			"compared to same number of iterations of equivalent "
			"functions is greater by %lld units\n", num_exec,
			avg_time_decor_func - avg_time_non_decor_func);
	} else {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_notify_inc_32 is equal to average time "
			"taken by same number of iterations of equivalent "
			"functions ", num_exec);
	}

	return 0;
}

int test_decorated_notify_clear_32(int argc, char *argv[])
{
	int i, core_mask, num_cores_run_test, pass;
	unsigned long int num_exec;
	uint64_t t1 = 0, t2 = 0, t3 = 0, t4 = 0;
	uint64_t avg_time_decor_func = 0, avg_time_non_decor_func = 0;
	uint32_t *p_stat, *p_nonstat;

	/* Necessary checks are carried out on these
	 * arguments in shell cmd handler (for DSPM) from where
	 * they are passed
	 */
	num_exec = strtoul(argv[0], NULL, 0);
	if ((ULONG_MAX == num_exec && ERANGE == errno)
		|| (0 != errno && 0 == num_exec)) {
		APP_ERROR("Failed to convert argument 1 of"
				" test_decorated_notify_clear_32");
		return -1;
	}
	core_mask = strtoul(argv[1], NULL, 0);

	num_cores_run_test = __msl_count_trailing_one32(core_mask);

	pass = 0;

	p_stat = stats_malloc(num_exec * sizeof(uint32_t));

	if (p_stat == NULL) {
		APP_ERROR("stats_malloc failed \n");
		return -1;
	}

	memset(p_stat, 0, num_exec * sizeof(uint32_t));

	for (i = 0; i < num_exec; i++)
		*(p_stat + i) = 0x12345678;

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			return -1;
		}
	}

	t1 = get_timebase();
	APP_DEBUG("t1 :: %llx\n", t1);

	for (i = 0; i < num_exec; i++)
		decorated_notify_clear_32(p_stat + i);

	t2 = get_timebase();
	APP_DEBUG("t2 :: %llx\n", t2);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			return -1;
		}
	}

	for (i = 0; i < num_exec; i++) {
		if (*(p_stat + i) == 0) {
			pass = 1;
		} else {
			pass = 0;
			break;
		}
	}

	if (pass) {
		avg_time_decor_func = (t2 - t1)/num_exec;
	} else {
		APP_ERROR("decorated_notify_clear_32 didn't function "
			"correctly\n");
		stats_free(p_stat);
		return -1;
	}

	p_nonstat = malloc(num_exec * sizeof(uint32_t));

	if (p_nonstat == NULL) {
		APP_ERROR("malloc failed \n");
		stats_free(p_stat);
		return -1;
	}

	memset(p_nonstat, 0, num_exec * sizeof(uint32_t));

	for (i = 0; i < num_exec; i++)
		*(p_nonstat + i) = 0x12345678;

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(p_nonstat);
			return -1;
		}
	}

	t3 = get_timebase();
	APP_DEBUG("t3 :: %llx\n", t3);

	for (i = 0; i < num_exec; i++)
		atomic_set((unsigned long *) p_nonstat + i, 0x0);

	t4 = get_timebase();
	APP_DEBUG("t4 :: %llx\n", t4);

	/* Don't run barrier_sync if application is
	 * running on a single core
	 */
	if (num_cores_run_test > 1) {
		if (0 > barrier_sync(&bar)) {
			APP_ERROR("barrier sync failed\n");
			stats_free(p_stat);
			free(p_nonstat);
			return -1;
		}
	}

	for (i = 0; i < num_exec; i++) {
		if (*(p_nonstat + i) == 0) {
			pass = 1;
		} else {
			pass = 0;
			break;
		}
	}

	if (pass) {
		avg_time_non_decor_func = (t4 - t3)/num_exec;
	} else {
		APP_ERROR("equivalent functions "
			"(wrt decorated_notify_clear_32 didn't function "
			"correctly\n");
		stats_free(p_stat);
		free(p_nonstat);
		return -1;
	}

	stats_free(p_stat);
	free(p_nonstat);

	APP_INFO("Average time taken by %ld iterations of "
		"decorated_notify_clear_32 is "
		"%lld units\n", num_exec, avg_time_decor_func);

	APP_INFO("Average time taken by %ld iterations of equivalent functions "
		"(wrt decorated_notify_clear_32) is %lld units\n", num_exec,
		avg_time_non_decor_func);

	if (avg_time_non_decor_func > avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_notify_clear_32 when "
			"compared to same number of iterations of equivalent "
			"functions is lesser by %lld units\n", num_exec,
			avg_time_non_decor_func - avg_time_decor_func);
	} else if (avg_time_non_decor_func < avg_time_decor_func) {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_notify_clear_32 when "
			"compared to same number of iterations of equivalent "
			"functions is greater by %lld units\n", num_exec,
			avg_time_decor_func - avg_time_non_decor_func);
	} else {
		APP_INFO("Average time taken by %ld iterations of "
			"decorated_notify_clear_32 is equal to average time "
			"taken by same number of iterations of equivalent "
			"functions ", num_exec);
	}

	return 0;
}

int shell_cmd_dspm(int argc, char *argv[])
{
	int test_id, core_mask;
	uint8_t core_num;
	unsigned long int num_exec;
	char *argp[2] = {argv[2], argv[3]};
	errno = 0;

	if (argc != 4) {
		printf("usage: dspm <test_id> <num_times_execution> "
			"<core_mask>\n");
		printf("Valid test_id values:\n");
		printf("-----------------------\n");
		printf("       1 - test case 1 - decorated_load_clear_32 vs "
			"spin_lock\n"
			"                                                    "
			"read\n"
			"                                                    "
			"set\n"
			"                                                    "
			"spin_unlock\n");
		printf("       2 - test case 2 - decorated_load_set_32 vs "
			"spin_lock\n"
			"						  "
			"read\n"
			"						  "
			"set\n"
			"						  "
			"spin_unlock\n");
		printf("       3 - test case 3 - decorated_load_dec_32 vs "
			"spin_lock\n"
			"						  "
			"read\n"
			"						  "
			"dec\n"
			"						  "
			"spin_unlock\n");
		printf("       4 - test case 4 - decorated_load_inc_32 vs "
			"spin_lock\n"
			"						  "
			"read\n"
			"						  "
			"inc\n"
			"						  "
			"spin_unlock\n");
		printf("       5 - test case 5 - decorated_store_32_acc_32 "
			"vs atomic_add \n");
		printf("       6 - test case 6 - "
			"decorated_store_32_inc_acc_32 vs spin_lock\n"
			"						"
			"	  inc\n"
			"						"
			"	  add\n"
			"						"
			"	  spin_unlock\n");
		printf("       7 - test case 7 - decorated_notify_inc_32 vs "
			"atomic_inc\n");
		printf("       8 - test case 8 - decorated_notify_clear_32 "
			"vs atomic_set\n");
		printf("Valid num_times_execution values:\n");
		printf("-----------------------\n");
		printf("       1 - 100000\n");
		printf("Valid core_mask values:\n");
		printf("-----------------------\n");
		printf("       0x1 - 0x%x\n", ((1 << get_num_cores()) - 1));
		printf("       Note : You must include boot core in "
			"core_mask\n");
		return -1;
	}

	test_id = strtoul(argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == test_id)
			|| (0 != errno && 0 == test_id)) {
		printf("strtoul failed converting command's argument1\n");
		return -1;
	}

	if (test_id < 1 || test_id > 8) {
		printf("wrong test_id entered\n");
		return -1;
	}

	num_exec = strtoul(argv[2], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == num_exec)
			|| (0 != errno && 0 == num_exec)) {
		printf("strtoul failed converting command argument2\n");
		return -1;
	}

	if (num_exec < 1 || num_exec > 100000) {
		printf("wrong number of times of execution entered\n");
		return -1;
	}

	core_mask = strtoul(argv[3], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == core_mask)
			|| (0 != errno && 0 == core_mask)) {
		printf("strtoul failed converting command's argument3\n");
		return -1;
	}

	if (core_mask < 1 || core_mask > ((1 << get_num_cores()) - 1)) {
		printf("wrong core_mask entered\n");
		return -1;
	}

	if (curr_lwe_id == 0) {
		if (0 > barrier_init(&bar, core_mask)) {
			APP_ERROR("barrier init failed\n");
			return -1;
		}
	}

	if (curr_lwe_id == 0) {
		switch (test_id) {
		case TEST_1:
			for (core_num = 1; core_num < 7; core_num++) {
				if ((core_mask & (1 << core_num))) {
					if (0 > start_cpu(core_num,
						test_decorated_load_clear_32,
						2, argp)) {
						APP_ERROR("start_cpu failed "
							"for core %d\n",
							core_num);
						return -1;
					}
				}
			}

			if (core_mask & (1 << 0))
				test_decorated_load_clear_32(2, argp);
			break;
		case TEST_2:
			for (core_num = 1; core_num < 7; core_num++) {
				if ((core_mask & (1 << core_num))) {
					if (0 > start_cpu(core_num,
						test_decorated_load_set_32,
						2, argp)) {
						APP_ERROR("start_cpu failed "
							"for core %d\n",
							core_num);
						return -1;
					}
				}
			}

			if (core_mask & (1 << 0))
				test_decorated_load_set_32(2, argp);
			break;
		case TEST_3:
			for (core_num = 1; core_num < 7; core_num++) {
				if ((core_mask & (1 << core_num))) {
					if (0 > start_cpu(core_num,
						test_decorated_load_dec_32,
						2, argp)) {
						APP_ERROR("start_cpu failed "
							"for core %d\n",
							core_num);
						return -1;
					}
				}
			}

			if (core_mask & (1 << 0))
				test_decorated_load_dec_32(2, argp);
			break;
		case TEST_4:
			for (core_num = 1; core_num < 7; core_num++) {
				if ((core_mask & (1 << core_num))) {
					if (0 > start_cpu(core_num,
						test_decorated_load_inc_32,
						2, argp)) {
						APP_ERROR("start_cpu failed "
							"for core %d\n",
							core_num);
						return -1;
					}
				}
			}

			if (core_mask & (1 << 0))
				test_decorated_load_inc_32(2, argp);
			break;
		case TEST_5:
			for (core_num = 1; core_num < 7; core_num++) {
				if ((core_mask & (1 << core_num))) {
					if (0 > start_cpu(core_num,
						test_decorated_store_32_acc_32,
						2, argp)) {
						APP_ERROR("start_cpu failed "
							"for core %d\n",
							core_num);
						return -1;
					}
				}
			}

			if (core_mask & (1 << 0))
				test_decorated_store_32_acc_32(2, argp);
			break;
		case TEST_6:
			for (core_num = 1; core_num < 7; core_num++) {
				if ((core_mask & (1 << core_num))) {
					if (0 > start_cpu(core_num,
						test_decorated_store_32_inc_acc_32,
						2, argp)) {
						APP_ERROR("start_cpu failed "
							"for core %d\n",
							core_num);
						return -1;
					}
				}
			}

			if (core_mask & (1 << 0))
				test_decorated_store_32_inc_acc_32(2, argp);
			break;
		case TEST_7:
			for (core_num = 1; core_num < 7; core_num++) {
				if ((core_mask & (1 << core_num))) {
					if (0 > start_cpu(core_num,
						test_decorated_notify_inc_32,
						2, argp)) {
						APP_ERROR("start_cpu failed "
							"for core %d\n",
							core_num);
						return -1;
					}
				}
			}

			if (core_mask & (1 << 0))
				test_decorated_notify_inc_32(2, argp);
			break;
		case TEST_8:
			for (core_num = 1; core_num < 7; core_num++) {
				if ((core_mask & (1 << core_num))) {
					if (0 > start_cpu(core_num,
						test_decorated_notify_clear_32,
						2, argp)) {
						APP_ERROR("start_cpu failed "
							"for core %d\n",
							core_num);
						return -1;
					}
				}
			}

			if (core_mask & (1 << 0))
				test_decorated_notify_clear_32(2, argp);
			break;

		default:
			break;
		}
	}

	if (curr_lwe_id == 0) {
		if (0 > barrier_destroy(&bar)) {
			APP_ERROR("barrier destroy failed\n");
			return -1;
		}
	}

	return 0;
}

int main(int argc, char *argv[])
{
	void *p_data = NULL;
	curr_lwe_id = get_lwe_id();

	if (curr_lwe_id == 0)
		APP_INFO("[Boot Core]: Welcome to DSPM Application");
	else
		APP_INFO("DSPM Application");

	if (curr_lwe_id != 0)
		return 0;

	if (curr_lwe_id == 0) {
		/*Add shell command for dspm*/
		if (0 != add_command("dspm", shell_cmd_dspm,
				"decorated storage performance measurement")) {
			APP_ERROR("add_command failed for dspm");
			return -1;
		}
	}

	while (1)
		swi_activate(&p_data);

	return 0;
}
