/** 
 \file fman_test.c
 \brief Basic FMAN Test Application
 */
/*
 * Copyright (C) 2008-2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <libfdt.h>
#include <crc64.h>
#include <timer_wheel.h>
#include "fman_test.h"
#include "net/annotations.h"
#include "tcp/tcp.h"
#include "udp/udp.h"
#include <dpa/dpa_eth.h>

__PERCPU struct dpa_buff_allocator *buff_allocator;
__PERCPU struct bman_pool *bpool[MAX_NUM_BMAN_POOLS];

/* FQID is hardcoded for Exact Match implementation.
 */
#define FQ_EXACT_BASE 0xB4
#define NUM_OF_FQ_FOR_EXACT_MATCH 3

#define TCP_PROTOCOL 0x06
#define UDP_PROTOCOL 0x11
#define VLAN_PROTOCOL 0x8100
#define IPV4_PROTOCOL 0x0800

int num_fq_close_pending;
static uint32_t recv_channel_map;
static volatile enum application_state fm_state = INVALID;

/* 3 Queues Allocated for Exact Match
 */
struct qman_fq *qman_fq_pcd_exact[NUM_OF_FQ_FOR_EXACT_MATCH];

static uint32_t frame_count;
static uint32_t frame_count_exact;

/**
 \brief DC Enqueue Rejection Notification Handler for FMAN Rx FQ
 \param[in] msg Message Ring entry to be processed
 \param[out] NULL
 */
static void fm_proc_dc_ern(struct qman_portal *qm, struct qman_fq *fq,
			   const struct qm_mr_entry *msg)
{
	APP_ERROR("%s: RC = %x, fqid = %x", __func__,\
		  msg->dcern.rc, msg->dcern.fqid);
	lwe_free_fd(buff_allocator, &msg->dcern.fd);
	return;
}

/**
 \brief Dequeue error handler for FMan Checker Application
	It shows the status bit of compound fd and drops the packet by freeing
	the buffer to BMAN.
 \param[in] struct qman_portal * Qman Portal Pointer
 \param[in] struct qman_fq * Pointer to a Qman frame Queue descriptor
 \param[in] struct qm_dqrr_entry * Dequeue Receive Ring entry to be processed
 \param[out] NULL
*/
static enum qman_cb_dqrr_result fm_proc_rx_err(struct qman_portal *qm,
					       struct qman_fq *fq,
					       const struct qm_dqrr_entry
					       *dqrr)
{
	APP_INFO("%s: Frame queue ID = %d, status = %x",\
		 __func__, dqrr->fqid, dqrr->fd.status);
	lwe_free_fd(buff_allocator, (struct qm_fd *)(&dqrr->fd));
	return qman_cb_dqrr_consume;
}

static enum qman_cb_dqrr_result fm_proc_rx_default(struct qman_portal *qm,
						   struct qman_fq *fq,
						   const struct qm_dqrr_entry
						   *dqrr)
{
	APP_INFO("%s: Frame queue ID = %d, status = %x",\
		 __func__, dqrr->fqid, dqrr->fd.status);
	APP_INFO("%s: OH port not implemented yet, configure PCD through FMC",
		 __func__);

	lwe_free_fd(buff_allocator, &(dqrr->fd));
	return qman_cb_dqrr_consume;
}

/**
 \brief Dequeue Handler for FMan Checker Application (Key Hash Check)
 \param[in] struct qman_portal * Qman Portal Pointer
 \param[in] struct qman_fq * Pointer to a Qman frame Queue descriptor
 \param[in] struct qm_dqrr_entry * Dequeue Receive Ring entry to be processed
 \param[out] NULL
 */
static enum qman_cb_dqrr_result fm_proc_rx_pcd(struct qman_portal *qm,
					       struct qman_fq *fq,
					       const struct qm_dqrr_entry
					       *dqrr)
{
	uint8_t *p_pkt;
	struct ethernet_header_t *eth_hdr;
	struct ip_header_t *ip_hdr = NULL;
	struct tcp_hdr_t *tcp_hdr = NULL;
	struct udp_hdr_t *udp_hdr = NULL;
	uint16_t vlan_tci;
	uint64_t hash;
	phys_addr_t addr;

	addr = ((phys_addr_t)(dqrr->fd.addr_hi) << 32) | dqrr->fd.addr_lo;
	p_pkt = ptov(addr) + dqrr->fd.offset;
	eth_hdr = (struct ethernet_header_t *)p_pkt;
	
	/*Printing Information of Incoming Packets. */
	frame_count = frame_count + 1;
	APP_INFO(" ");
	APP_INFO("Total number of packets received by all cores: %u",
		 frame_count);
	APP_INFO("Frame dequeued from FQ ID:\t 0x%x", dqrr->fqid);

	APP_DEBUG("Source MAC:\t\t\t" MAC_FMT, NMAC_STR(eth_hdr->source.bytes));
	APP_DEBUG("Destination MAC:\t\t\t" MAC_FMT,
		  NMAC_STR(eth_hdr->destination.bytes));
	APP_DEBUG("Ethernet type:\t\t\t 0x%04x", eth_hdr->proto);

	hash = crc64_init();

	APP_INFO("=====================================================");
	APP_INFO("FM KeyGen key element are:");
	if (eth_hdr->proto == IPV4_PROTOCOL) {
		ip_hdr = (struct ip_header_t *)(p_pkt + ETH_HDR_LEN);
		APP_INFO("Source IP:\t\t\t %d.%d.%d.%d(0x%x)",
			 IP_DOTTED_QUAD(ip_hdr->src_addr.word),
			 ip_hdr->src_addr.word);
		APP_INFO("Destination IP:\t\t %d.%d.%d.%d(0x%x)",
			 IP_DOTTED_QUAD(ip_hdr->dst_addr.word),
			 ip_hdr->dst_addr.word);

		hash = crc64_compute_word(ip_hdr->src_addr.word, hash);
		hash = crc64_compute_word(ip_hdr->dst_addr.word, hash);

		if (ip_hdr->proto == TCP_PROTOCOL) {
			tcp_hdr = (struct tcp_hdr_t *)
					(p_pkt + ETH_HDR_LEN
					+ IP_HEADER_LENGTH_NO_OPTIONS);

			/* FIXME: Validate TCP checksum */
			APP_INFO("IP protocol:\t\t\t 0x%02x", ip_hdr->proto);
			APP_INFO("TCP source port:\t\t 0x%04x", tcp_hdr->sport);
			APP_INFO("TCP dest port:\t\t\t 0x%04x", tcp_hdr->dport);

			hash = crc64_compute_byte(ip_hdr->proto, hash);
			hash = crc64_compute_hword(tcp_hdr->sport, hash);
			hash = crc64_compute_hword(tcp_hdr->dport, hash);
		} else if (ip_hdr->proto == UDP_PROTOCOL) {
			udp_hdr = (struct udp_hdr_t *)
					(p_pkt + ETH_HDR_LEN
					+ IP_HEADER_LENGTH_NO_OPTIONS);

			/* FIXME: Validate UDP checksum */
			APP_INFO("IP TOS:\t\t\t 0x%02x", ip_hdr->tos);
			APP_INFO("UDP source port:\t\t 0x%04x",
				 udp_hdr->sport_no);
			APP_INFO("UDP destination port:\t\t 0x%04x",
				 udp_hdr->dport_no);

			hash = crc64_compute_byte(ip_hdr->tos, hash);
			hash = crc64_compute_hword(udp_hdr->sport_no, hash);
			hash = crc64_compute_hword(udp_hdr->dport_no, hash);
		}

	} else if (eth_hdr->proto == VLAN_PROTOCOL) {
		vlan_tci = *((uint16_t *) (p_pkt + ETH_HDR_LEN));

		APP_INFO("Ethernet VLAN Tag:\t\t 0x%04x", vlan_tci);
		hash = crc64_compute_hword(vlan_tci, hash);

	} else {
		APP_INFO("Invalid Frame.");
		goto done;
	}
	APP_INFO("=====================================================");

	if ((uint16_t) (hash % HASH_MASK)
	    == (uint16_t) (dqrr->fqid % HASH_MASK))
		APP_INFO("Computed Key match found with FM KeyGen key");
	else {
		APP_ERROR("Key Match not found!!!!!");
		APP_ERROR("Computed Key = 0x%llx", hash);
		APP_ERROR("Keygen Key(masked) = 0x%x", dqrr->fqid % HASH_MASK);
		return qman_cb_dqrr_consume;
	}

done:
	lwe_free_fd(buff_allocator, &(dqrr->fd));
	return qman_cb_dqrr_consume;
}

/**
 \brief Dequeue Handler for FMan Checker Application (Exact match)
 \param[in] struct qman_portal * Qman Portal Pointer
 \param[in] struct qman_fq * Pointer to a Qman frame Queue descriptor
 \param[in] struct qm_dqrr_entry * Dequeue Receive Ring entry to be processed
 \param[out] NULL
 */
static enum qman_cb_dqrr_result fm_proc_rx_pcd_exact(struct qman_portal *qm,
						     struct qman_fq *fq,
						     const struct qm_dqrr_entry
						     *dqrr)
{
	uint8_t *p_pkt;
	struct ethernet_header_t *eth_hdr;
	struct ip_header_t *ip_hdr = NULL;
	struct tcp_hdr_t *tcp_hdr = NULL;
	struct udp_hdr_t *udp_hdr = NULL;
	uint16_t vlan_tci;
	phys_addr_t addr;
	
	addr = ((phys_addr_t)(dqrr->fd.addr_hi) << 32) | dqrr->fd.addr_lo;
	p_pkt = ptov(addr) + dqrr->fd.offset;
	
	eth_hdr = (struct ethernet_header_t *)p_pkt;

	/*Printing Information of Incoming Packets. */
	frame_count = frame_count + 1;
	frame_count_exact = frame_count_exact + 1;
	APP_INFO(" ");
	APP_INFO("Total number of packets received by all cores: %u",
		 frame_count);
	APP_INFO("Total number of packets for which exact match is found: %u",
		 frame_count_exact);
	APP_INFO("Frame dequeued from FQ ID:\t 0x%x", dqrr->fqid);

	APP_DEBUG("Source MAC:\t\t\t" MAC_FMT, NMAC_STR(eth_hdr->source.bytes));
	APP_DEBUG("Destination MAC:\t\t\t" MAC_FMT,
		  NMAC_STR(eth_hdr->destination.bytes));
	APP_DEBUG("Ethernet type:\t\t\t 0x%04x", eth_hdr->proto);

	APP_INFO("=====================================================");
	APP_INFO("FM KeyGen key element are:");
	if (eth_hdr->proto == IPV4_PROTOCOL) {
		ip_hdr = (struct ip_header_t *)(p_pkt + ETH_HDR_LEN);
		APP_INFO("Source IP:\t\t\t %d.%d.%d.%d(0x%x)",
			 IP_DOTTED_QUAD(ip_hdr->src_addr.word),
			 ip_hdr->src_addr.word);
		APP_INFO("Destination IP:\t\t %d.%d.%d.%d(0x%x)",
			 IP_DOTTED_QUAD(ip_hdr->dst_addr.word),
			 ip_hdr->dst_addr.word);

		if (ip_hdr->proto == TCP_PROTOCOL) {
			tcp_hdr = (struct tcp_hdr_t *)
					(p_pkt + ETH_HDR_LEN
					+ IP_HEADER_LENGTH_NO_OPTIONS);

			/* FIXME: Validate TCP checksum */
			APP_INFO("IP protocol:\t\t\t 0x%02x", ip_hdr->proto);
			APP_INFO("TCP source port:\t\t 0x%04x", tcp_hdr->sport);
			APP_INFO("TCP dest port:\t\t\t 0x%04x", tcp_hdr->dport);

		} else if (ip_hdr->proto == UDP_PROTOCOL) {
			udp_hdr = (struct udp_hdr_t *)
					(p_pkt + ETH_HDR_LEN
					+ IP_HEADER_LENGTH_NO_OPTIONS);

			/* FIXME: Validate UDP checksum */
			APP_INFO("IP TOS:\t\t\t 0x%02x", ip_hdr->tos);
			APP_INFO("UDP source port:\t\t 0x%04x",
				 udp_hdr->sport_no);
			APP_INFO("UDP destination port:\t\t 0x%04x",
				 udp_hdr->dport_no);
		}

	} else if (eth_hdr->proto == VLAN_PROTOCOL) {
		vlan_tci = *((uint16_t *) (p_pkt + ETH_HDR_LEN));

		APP_INFO("Ethernet VLAN Tag:\t\t 0x%04x", vlan_tci);

	} else {
		APP_INFO("Invalid Frame.");
		goto done;
	}
	APP_INFO("=====================================================");
	APP_INFO("Exact Match found by PCD");

done:
	lwe_free_fd(buff_allocator, &(dqrr->fd));
	return qman_cb_dqrr_consume;
}

/*This function is called by all cores*/
static int fm_common_init(void)
{
	int ecode;

	/* Initialise the timer framework for maximum possible timeout
	   supported by LWE. */
	if (0 != timer_framework_start()) {
		APP_ERROR("timer_framework_start Failed");
		return -1;
	}

	buff_allocator = dpa_allocator_create(&ecode);
	if (NULL == buff_allocator) {
		APP_ERROR("dpa_buff_allocator_create failed");
		return -1;
	}

	APP_INFO("Buffer allocator created...");
	return 0;
}

/*BMAN pool initialisation function*/
static int fm_bpools_init(void)
{
	struct bman_pool_params params;
	uint32_t bpid;

	/*Create BMAN pools */
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		/*Else create the pool object */
		params.flags = BMAN_POOL_FLAG_STOCKPILE;
		params.cb = NULL;
		params.cb_ctx = NULL;
		params.bpid = bpid;

		bpool[bpid] = bman_new_pool(&params);
		if (NULL == bpool[bpid]) {
			APP_ERROR("bman_new_pool failed for BPID: %u", bpid);
			return -1;
		}

		if (0 != dpa_allocator_add_bpool(buff_allocator,
						 dpa_bpools[bpid].size,
						 bpid, bpool[bpid])) {
			APP_ERROR("dpa_allocator_add_bpool failed");
			return -1;
		}

		APP_INFO("Added BPID: %u, Size: %u", bpid,
			 dpa_bpools[bpid].size);
	}

	return 0;
}

/*BMAN pool destroy function*/
static void fm_bpools_destroy(void)
{
	uint32_t bpid;

	/*Free the BMAN buffer pools */
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		bman_free_pool(bpool[bpid]);
		APP_INFO("BPID: %u destroyed", bpid);
	}

	/*Free the DPA allocator */
	dpa_allocator_free(buff_allocator);
	APP_INFO("DPA allocator freed");
}

/*
 * This function is used to move a FQ to OOS state, destroy it and free
 * the associated memory with FQ structure.
 */
static void close_fq(uint32_t swi_id, void *data)
{
	struct qman_fq *fq = data;
	uint32_t fqid;

	fqid = fq->fqid;
	if (0 > qman_oos_fq(fq)) {
		APP_ERROR("Moving FQID: %u to OOS failed", fq->fqid);
		free(fq);
		return;
	}

	qman_destroy_fq(fq, 0);
	free(fq);
	APP_INFO("Closed FQID: %u", fqid);
	num_fq_close_pending--;

}

/*This function waits till the FQ becomes empty and then closes it*/
static void wait_fq_empty(uint32_t swi_id, void *data)
{
	struct qman_fq *fq = data;
	uint32_t vdqcr;
	uint32_t flags;

	APP_INFO("Waiting for FQID: %u to become empty", fq->fqid);

	vdqcr = QM_VDQCR_PRECEDENCE_VDQCR | QM_VDQCR_NUMFRAMES_TILLEMPTY;

	/* NB: the use of QMAN_VOLATILE_FLAG_WAIT is prohibited in LWE
	 * applications. To avoid thrashing, we implement a throttled poll for
	 * completion. */
	if (qman_volatile_dequeue(fq, 0, vdqcr)) {
		APP_ERROR("qman_volatile_dequeue failure for FQID: %u",
			  fq->fqid);
		free(fq);
		return;
	}
loop:
	qman_fq_state(fq, NULL, &flags);
	if (flags & QMAN_FQ_STATE_VDQCR) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	/*Now the FQ is empty, close it */
	close_fq(SWI_ID_INV, data);
}

/* This function processes FQ state change notifications. This function is
 * called from ISR context. Thus this function must not make any blocking
 * and heavy weight APIs. When the FQ get retired and also gets empty, a
 * SWI is posted towards application's main loop for taking FQs OOS and
 * destroy them.
 */
static void fm_proc_fq_state_chg(struct qman_portal *qm,
				 struct qman_fq *fq,
				 const struct qm_mr_entry *msg)
{
	uint8_t verb;
	uint8_t fqs;

	/*Extract the response verb */
	verb = msg->verb & ~QM_MR_VERB_VBIT;

	if (QM_MR_VERB_FQRL == verb) {
		if (0 > swi_add(SWI_ID_INV, fq, close_fq, SWI_PRI_HIGH))
			APP_ERROR("Cannot add SWI for closing FQ");

		return;
	}

	fqs = msg->fq.fqs;

	if ((QM_MR_VERB_FQRN == verb) || (QM_MR_VERB_FQRNI == verb)) {
		if (fqs & QM_MR_FQS_NOTEMPTY) {
			APP_INFO("Retired FQID: %u not yet empty", fq->fqid);

			if (0 > swi_add(SWI_ID_INV, fq, wait_fq_empty,
					SWI_PRI_HIGH))
				APP_ERROR("Cannot add SWI for emptying FQ");
			return;
		}

		if (fqs & QM_MR_FQS_ORLPRESENT) {
			/*We need to wait till we get FQRL */
			APP_INFO("Retired FQID: %u has ORL list", fq->fqid);
			return;
		}

		/*Else FQ is empty and can be destroyed */
		if (0 > swi_add(SWI_ID_INV, fq, close_fq, SWI_PRI_HIGH)) {
			APP_ERROR("Cannot add SWI for closing FQ");
			return;
		}

		return;
	}

	APP_ERROR("Unexpected FQ state change verb: %d", verb);
}

/*This function waits for shutdown process to complete and also cleans up
per core resources*/
static void wait_for_shutdown_complete(void)
{
	void *p_data;
	uint32_t swi_id;

	while (1) {
		if (SHUTDOWN_COMPLETE == fm_state)
			break;

		/*This would dequeue any remanining notifications from QMAN */
		qman_poll();

		while (SWI_ID_INV != (swi_id = swi_activate(&p_data))) {
			switch (swi_id) {
			case SWI_ID_TIMER_EXPIRY:
				proc_timer_expiry((uint32_t) p_data);
				break;

			default:
				break;
			}
		}
	}

	if (0 != timer_framework_stop()) {
		APP_ERROR("timer_framework_stop failed");
		return;
	}

	fm_bpools_destroy();
	APP_INFO("Destroyed BMAN pools and buffer allocator");
}

static void fm_shutdown_complete(uint32_t timer_id, void *p_data)
{
	if (0 == num_fq_close_pending) {
		APP_INFO("Moving application state to SHUTDOWN_COMPLETE");
		fm_state = SHUTDOWN_COMPLETE;
		return;
	}

	APP_WARN("Num Tx FQ close pending: %d", num_fq_close_pending);
	APP_WARN("Waiting for some more time for FQs to retire");

	timer_id = reset_timer(timer_id, 0);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Resetting shutdown complete timer failed");
}

/*This function waits for Frame queues assigned to exact match to retire
*/
static void fm_shutdown_fq_exact(uint32_t timer_id, void *p_data)
{
	uint32_t flags;
	int ret;
	int j = 0;

	/*This will retire all the exact match frame queues */

	for (j = 0; j < NUM_OF_FQ_FOR_EXACT_MATCH; j++) {
		APP_INFO("Retiring Fq %u for PCD Exact Match..", j+1);
		ret = qman_retire_fq(qman_fq_pcd_exact[j], &flags);
		if (0 > ret) {
			APP_ERROR("qman_retire_fq failed for FQID: %u",
				  qman_fq_pcd_exact[j]->fqid);
			return;
		}

		num_fq_close_pending += 1;
	}

	/*
	 * If there are no Exact Match FQs which are pending to be closed,
	 * directly shutdown the application
	 */
	if (!num_fq_close_pending) {
		fm_shutdown_complete(INV_TIMER_ID, NULL);
		return;
	}

	/*Start a single shot timer to wait for all FQs to get retired */
	timer_id = start_timer(FMTEST_SHUTDOWN_TIMER_MS, false,
			       NULL, SWI_PRI_HIGH, fm_shutdown_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown complete timer failed");
	else
		APP_INFO("Shutdown complete timer started...");

}

static void fm_shutdown_rx_complete(uint32_t timer_id, void *p_data)
{
	if (num_fq_close_pending) {
		APP_WARN("Num Rx FQ close pending: %d", num_fq_close_pending);
		APP_WARN("Waiting for some more time for FQs to retire");

		timer_id = reset_timer(timer_id, 0);
		if (INV_TIMER_ID == timer_id)
			APP_ERROR("Reset shutdown Rx complete timer failed");
		return;
	}

	APP_INFO("Moving application state to SHUTDOWN_RX_COMPLETE");
	fm_state = SHUTDOWN_RX_COMPLETE;

	/* At this stage, all the traffic which was received has been
	 * processed and we can now move forward to retire Tx FQs
	 */
	num_fq_close_pending = dpa_eth_tx_fq_retire();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring Tx FQs failed");
		return;
	}

	APP_INFO("Number of Tx FQs close pending: %d", num_fq_close_pending);

	/*
	 * If there are no Tx FQs which are pending to be closed, directly
	 * shutdown the application
	 */
	if (!num_fq_close_pending) {
		fm_shutdown_fq_exact(INV_TIMER_ID, NULL);
		return;
	}

	/*Start a single shot timer to wait for all Tx FQs to get retired */
	timer_id = start_timer(FMTEST_SHUTDOWN_TIMER_MS, false,
			       NULL, SWI_PRI_HIGH, fm_shutdown_fq_exact, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown complete timer failed");
	else
		APP_INFO("Shutdown complete timer started...");
}

static void fm_proc_shutdown(void)
{
	uint32_t timer_id;

	APP_INFO("Processing shutdown command ..");
	/*Bootcore is responsible for partition cleanup */
	fm_state = SHUTDOWN_PENDING;

	/*Disable both Rx mac ports */
	if (-1 == dpa_eth_disable_ports(true, false))
		APP_ERROR("Disabling MAC-Rx port failed");
	else
		APP_INFO("Disabled MAC-Rx ports");

	/*Put all Rx frame queues in retired state */
	/*This would effectively disable the MAC-less Rx port */
	num_fq_close_pending = dpa_eth_rx_fq_retire();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring Rx FQs failed");
		return;
	}

	APP_INFO("Number of Rx FQs close pending: %d", num_fq_close_pending);

	/*If there is no FQ close pending, directly move to the next state */
	if (!num_fq_close_pending) {
		fm_shutdown_rx_complete(INV_TIMER_ID, 0);
		return;
	}

	/*Start a timer to complete Rx shutdown */
	timer_id = start_timer(FMTEST_SHUTDOWN_TIMER_MS,
			       false, NULL, SWI_PRI_HIGH,
			       fm_shutdown_rx_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown RX complete timer failed");
	else
		APP_INFO("Started shutdown Rx complete timer");
}

/**
 \brief FMAN Test entry function
        The first partition that comes up takes the role of the control partition. 
	Only control partition perform the creation and initializations of frame queues
	and enabling ports. 
	The per partition resources like qman portals are initialized 
	by each partitions. After the initializations are done, the partitions 
	enter into an infinite loop dequeueing frames from common pool channel 
	and printing the required information from them .
 \param[in] argc count of input parameters
 \param[in] argv Value of the parameters passed
 \param[out] int Return status of application
 */
int main(int argc, char *argv[])
{
	int32_t channel_id_lwe;
	uint32_t swi_id;
	void *p_data;

	uint32_t flags;
	int ret;
	uint32_t j;
	uint32_t fqid;
	struct qm_mcc_initfq opts;
	uint32_t ctx_lines;
	uint32_t data_lines;
	uint32_t ctx_a_excl;

	struct qman_fq_cb fm_rx_default_cb = { fm_proc_rx_default,
		NULL, fm_proc_dc_ern, fm_proc_fq_state_chg
	};

	struct qman_fq_cb fm_rx_pcd_cb = { fm_proc_rx_pcd,
		NULL, fm_proc_dc_ern, fm_proc_fq_state_chg
	};

	struct qman_fq_cb fm_rx_pcd_exact_cb = { fm_proc_rx_pcd_exact,
		NULL, fm_proc_dc_ern, fm_proc_fq_state_chg
	};

	struct qman_fq_cb fm_rx_err_cb = { fm_proc_rx_err,
		NULL, fm_proc_dc_ern, fm_proc_fq_state_chg
	};

	struct qman_fq_cb fm_tx_cb = { NULL,
		NULL, NULL, fm_proc_fq_state_chg
	};

	struct qman_fq_cb fm_tx_confirm_cb = { NULL,
		NULL, NULL, fm_proc_fq_state_chg
	};

	struct qman_fq_cb fm_tx_err_cb = { NULL,
		NULL, NULL, fm_proc_fq_state_chg
	};

	/*Boot core initialisation */
	if (0 == lwe_instance) {
		APP_INFO("Welcome to FMAN Test Application ...");
		fm_state = INITIALIZING;
	} else {
		while (INITIALIZING != fm_state) ;
	}

	/*Control core initialises global database */
	if (0 == lwe_instance) {
		if (0 >= dpa_eth_read(&recv_channel_map)) {
			APP_ERROR("Unable to read Eth Info from Devtree");
			return -EINVAL;
		}

		if (0 !=
		    dpa_eth_fq_create((struct qman_fq_cb *)&fm_rx_default_cb,
				      (struct qman_fq_cb *)&fm_rx_pcd_cb,
				      (struct qman_fq_cb *)&fm_rx_err_cb,
				      (struct qman_fq_cb *)&fm_tx_cb,
				      (struct qman_fq_cb *)&fm_tx_confirm_cb,
				      (struct qman_fq_cb *)&fm_tx_err_cb, 0)) {
			APP_ERROR("Unable to Create FQs");
			return -EINVAL;
		}

		flags = QMAN_FQ_FLAG_NO_ENQUEUE | QMAN_FQ_FLAG_LOCKED;

		for (j = 0; j < NUM_OF_FQ_FOR_EXACT_MATCH; j++) {
			/*3 Queues are created to handle exact match */
			fqid = FQ_EXACT_BASE + j;
			qman_fq_pcd_exact[j] = (struct qman_fq *)
			    malloc(sizeof(struct qman_fq));
			qman_fq_pcd_exact[j]->cb = fm_rx_pcd_exact_cb;
			qman_fq_pcd_exact[j]->fqid = fqid;
			ret = qman_create_fq(fqid, flags, qman_fq_pcd_exact[j]);
			if (unlikely(0 != ret)) {
				APP_ERROR("create_fqs failed for Rx PCD"
					" Exact Match %u", j + 1);
				return -1;
			}
			APP_INFO
			    ("Created Rx PCD Exact Match %u Base: %u, Count: 1",
			     j + 1, qman_fq_pcd_exact[j]->fqid);
		}

		APP_INFO("Num ethernet ports discovered: %u",
			 g_num_dpa_eth_ports);

		if (0 != dpa_eth_fq_init(64, 64, 64, NULL, NULL)) {
			APP_ERROR("Unable to initialize FQs");
			return -EINVAL;
		}

		/*3 Queues are initialised to handle exact match */
		flags = QMAN_INITFQ_FLAG_SCHED | QMAN_INITFQ_FLAG_LOCAL;
		opts.we_mask = QM_INITFQ_WE_DESTWQ |
		    QM_INITFQ_WE_CONTEXTA | QM_INITFQ_WE_FQCTRL;
		opts.fqd.fq_ctrl = QM_FQCTRL_CTXASTASHING |
		    QM_FQCTRL_LOCKINCACHE |
		    QM_FQCTRL_CPCSTASH | QM_FQCTRL_HOLDACTIVE;

		ctx_a_excl = QM_STASHING_EXCL_DATA;

		ctx_lines = GET_STASH_LINES(CACHE_LINE_SIZE);
		data_lines = GET_STASH_LINES(CACHE_LINE_SIZE);

		opts.fqd.context_a.stashing.exclusive = ctx_a_excl;
		opts.fqd.context_a.stashing.data_cl = data_lines;
		opts.fqd.context_a.stashing.context_cl = ctx_lines;

		for (j = 0; j < NUM_OF_FQ_FOR_EXACT_MATCH; j++) {
			if (0 !=
			    qman_init_fq(qman_fq_pcd_exact[j], flags, &opts)) {
				APP_ERROR("Unable to initialize FQs for Exact \
					 match %u", j+1);
				return -EINVAL;
			}

			APP_INFO("Init Rx Exact Match FQ %u, FQID: %u",
				 j + 1, qman_fq_pcd_exact[j]->fqid);
		}
	}
	/*Wait till DPA is Initialized */
	while (0 == g_num_dpa_eth_ports) ;

	if (0 > (channel_id_lwe = extract_qman_pool_channel(devtree_ptr))) {
		APP_ERROR("local_extract_qman_channel_id failed");
		return -EINVAL;
	} else
		APP_INFO("Pool Channel ID is :%u", channel_id_lwe);

	/*Common initialisation routine for all cores */
	if (0 > fm_common_init()) {
		APP_ERROR("fm_common_init failed");
		return -EINVAL;
	}

	/*Create BMAN pools */
	if (0 > fm_bpools_init()) {
		APP_ERROR("fm_bpool_init failed");
		return -EINVAL;
	}

	if (0 == lwe_instance) {
		if (unlikely(0 != dpa_eth_enable_ports(true))) {
			APP_ERROR("dpa_eth_enable_ports failed");
			return -1;
		}
		fm_state = RUNNING;
	} else {
		while (RUNNING != fm_state) ;
	}

	qman_static_dequeue_add(QM_SDQCR_CHANNELS_POOL(channel_id_lwe)
				| recv_channel_map);
	APP_INFO("Starting frame dequeue...");

	while (likely(1)) {
		if (unlikely(SHUTDOWN_PENDING == fm_state))
			break;

		qman_poll();

		while (unlikely(SWI_ID_INV !=
				(swi_id = swi_activate(&p_data)))) {

			switch (swi_id) {
			case SWI_ID_TIMER_EXPIRY:
				proc_timer_expiry((uint32_t) p_data);
				break;

			/*Shutdown SWI is received only on bootcore */
			case SWI_ID_SHUTDOWN:
				fm_proc_shutdown();
				break;

			default:
				break;
			}
		}
	}
	/*The following code is executed only in SHUTDOWN_PENDING state */
	/*Here, all cores wait for shutdown to complete */
	wait_for_shutdown_complete();
	APP_INFO("Shutdown completed on core: %u", lwe_instance);
	return 0;
}
