/**
\file
\brief      Main ethernet hub file. Contains startup routines
*/
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <inttypes.h>
#include <init.h>
#include <compat.h>
#include <dpa/dpa_eth.h>
#include <app_common.h>

#define NUM_ETH_PORTS	8
uint32_t num_fman_eth_ports;

struct {
	u32 channel;
	struct qman_fq fq[NUM_ETH_PORTS];
} port_fq_config[NUM_ETH_PORTS];

static int create_interconn_fqs(uint32_t port_id)
{
	u32 rx_fqid;
	uint32_t i;
	u32 flags;
	struct qman_fq *fq;

	flags = QMAN_FQ_FLAG_DYNAMIC_FQID |
		QMAN_FQ_FLAG_NO_ENQUEUE |
		QMAN_FQ_FLAG_TO_DCPORTAL;

	/*Create one hopping FQ for each other port*/
	APP_INFO("Creating hopping FQs for port ID: %u", port_id);

	for (i = 0 ; i < num_fman_eth_ports - 1; i++) {

		fq = &port_fq_config[port_id].fq[i];

		if (0 != qman_create_fq(0, flags, fq)) {
			APP_ERROR("qman_create_fq failed");
			return -1;
		}

		APP_INFO("Created FQID: %u, (hop count: %u)", fq->fqid, i);
	}

	/*Now create the default Rx FQ for this for 'i' hops*/
	rx_fqid = dpa_eth_ports[port_id].ingress.fq_start;

	flags = QMAN_FQ_FLAG_NO_ENQUEUE |
		QMAN_FQ_FLAG_TO_DCPORTAL;

	fq = &port_fq_config[port_id].fq[i];

	if (0 != qman_create_fq(rx_fqid, flags, fq)) {
		APP_ERROR("qman_create_fq failed");
		return -1;
	}

	APP_INFO("Created FQID: %u, (hop count: %u)", fq->fqid, i);

	return 0;
}


static int init_interconn_fqs(uint32_t port_id)
{
	struct qm_mcc_initfq opts;
	struct qman_fq *fq;
	uint32_t next_port_id;
	uint32_t alt_port_id;
	uint32_t egress_fqid;
	uint32_t conf_fqid;
	uint32_t i;
	u32 flags;

	if (port_id == num_fman_eth_ports - 1)
		next_port_id = 0;
	else
		next_port_id = port_id + 1;

	if (next_port_id == num_fman_eth_ports - 1)
		alt_port_id = 0;
	else
		alt_port_id = next_port_id + 1;

	APP_INFO("Initialising hop FQs for port ID: %u, Next port ID: %u",
			port_id, next_port_id);

	flags = QMAN_INITFQ_FLAG_SCHED;

	opts.we_mask = QM_INITFQ_WE_DESTWQ |
		QM_INITFQ_WE_CONTEXTA |
		QM_INITFQ_WE_CONTEXTB |
		QM_INITFQ_WE_FQCTRL;

	opts.fqd.fq_ctrl = QM_FQCTRL_LOCKINCACHE | QM_FQCTRL_CPCSTASH;
	opts.fqd.dest_wq = 4;
	opts.fqd.dest.channel = (uint16_t)port_fq_config[port_id].channel;

	opts.fqd.context_a.lo = 0x0;
	opts.fqd.context_a.hi = 0x80000000;

	for (i = 0 ; i < num_fman_eth_ports - 1; i++) {
		if (0 == i)
			/*Case of no confirmation queue*/
			conf_fqid = 0;
		else
			conf_fqid =
				port_fq_config[next_port_id].fq[i - 1].fqid;

		opts.fqd.context_b = conf_fqid;

		fq = &port_fq_config[port_id].fq[i];
		egress_fqid = fq->fqid;
		if (0 != qman_init_fq(fq, flags, &opts)) {
			APP_ERROR("qman_init_fq failed for FQID: %u",
					egress_fqid);
			return -1;
		}

		APP_INFO("Egress FQID: %u Chan: %u init, Confirm FQID: %u",
			egress_fqid, port_fq_config[port_id].channel,
			conf_fqid);
	}

	/*The default Rx FQ has to be initialised to be dequeued by next port*/
	conf_fqid = port_fq_config[alt_port_id].fq[i - 2].fqid;
	opts.fqd.context_b = conf_fqid;
	opts.fqd.dest.channel = (uint16_t)port_fq_config[next_port_id].channel;

	fq = &port_fq_config[port_id].fq[i];
	egress_fqid = fq->fqid;
	if (0 != qman_init_fq(fq, flags, &opts)) {
		APP_ERROR("qman_init_fq failed for FQID: %u", egress_fqid);
		return -1;
	}

	APP_INFO("Ingress FQID: %u Chan:%u init, Confirm FQID: %u",
		egress_fqid, port_fq_config[next_port_id].channel, conf_fqid);

	return 0;
}

/**
\brief   Main function of hub code.

\details Initialises the FQs required to make the ethernet hub functional.

\note    Should never return.
*/
int main(int argc, char *argv[])
{
	uint32_t recv_channel_map;
	uint32_t port_id;

	if (lwe_instance != 0) {
		APP_INFO("No DP software required. Returning...");
		return 0;
	}

	APP_INFO("Initialising the ethernet hub ...");

	if (0 >= dpa_eth_read(&recv_channel_map)) {
		APP_ERROR("dpa_eth_read failed");
		goto err_out;
	}

	/*Count number of fman-eth ports*/
	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		if (MAC_ETH != dpa_eth_ports[port_id].type)
			continue;

		port_fq_config[port_id].channel =
				dpa_eth_ports[port_id].egress.channel;

		APP_INFO("FMAN port ID: %u dequeues from channel: %u", port_id,
			port_fq_config[port_id].channel);

		num_fman_eth_ports++;
	}

	/* Create 'num_fman_eth_ports - 1' interconnect FQs for
	 * each fman-eth port*/
	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		if (MAC_ETH != dpa_eth_ports[port_id].type)
			continue;

		if (0 != create_interconn_fqs(port_id)) {
			APP_ERROR("create_interconn_fqs failed \
				for port: %u", port_id);
			goto err_out;
		}
	}

	APP_INFO("FMAN Port interconnect FQs created");

	/* Initialise above created interconnect FQs for each port*/
	for (port_id = 0; port_id < num_fman_eth_ports; port_id++) {
		if (0 != init_interconn_fqs(port_id)) {
			APP_ERROR("init_interconn_fqs failed \
				for port: %u",	port_id);
			goto err_out;
		}
	}

	APP_INFO("FMAN Port interconnect FQs initialised");

	if (0 != dpa_eth_enable_ports(true)) {
		APP_ERROR("dpa_eth_enable_ports failed");
		goto err_out;
	}

	APP_INFO("Hub initialisation completed ..");
	APP_INFO("No DP software required. Returning ...");

	return 0;

err_out:
	APP_ERROR("FATAL error, exiting\r\n");
	return -1;
}
