/**
 \file ipsecfwd.c
 \brief Basic IPSec Forwarding Application
 */
/*
 * Copyright (C) 2008-2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ipsecfwd.h"
#include <fsl_qman.h>
#include <timebase.h>
#include "barrier.h"
#include "ip/ip_forward.h"
#include "ip/ip_local.h"
#include "ipsec/ipsec_sec.h"
#ifdef PME_SWITCH
#include "ipsecfwd_scan.h"
#endif

#define CGR_SUPPORT
volatile struct lwe_ctrl_op_info *g_lwe_ctrl_sa_info;
volatile uint32_t *total_pkts_processed;

__PERCPU struct dpa_buff_allocator *buff_allocator_core, *buff_allocator;
__PERCPU struct dpa_buff_allocator *ipsec_buff_allocator;
__PERCPU uint32_t local_frames;
__PERCPU uint32_t curr_lwe_id;
__PERCPU bool dqrr_enter;
__PERCPU struct bman_pool *bpool[MAX_NUM_BMAN_POOLS];
__PERCPU uint32_t num_poll_before_shutdown;
__PERCPU uint32_t rx_errors;

int32_t pool_channel_id;
static uint32_t local_channel;
static bool infinite_lwe;
static uint64_t grand_total_cycles;

bool test_completed;
bool cleanup_completed;
uint32_t iface_subnet[IFACE_COUNT] = { 24, 29, 21, 22, 23, 25, 26, 27, 28 };
uint32_t local_node_count[IFACE_COUNT] = { 23, 23, 23, 23, 23, 23, 23, 23, 1 };
uint32_t recv_channel_map;
uint32_t initial_frame_count = INITIAL_FRAME_COUNT;
uint32_t local_hosts_ip[LINKLOCAL_NODES];
uint32_t iface_ip[LINKLOCAL_NODES];
uint32_t num_fq_free_pending;
uint32_t shutdown_doorbell_flag;
uint32_t ipsecfwd_lock;
uint32_t g_boot_lwe_core = INV_CORE_ID;
int32_t num_fq_close_pending;
struct dpa_bpool_t dpa_partition_bpools[MAX_NUM_BMAN_POOLS];
struct node_t local_nodes[LINKLOCAL_NODES];
struct node_t iface_nodes[IFACE_COUNT];
struct qman_fq *fq_destroy_pending[IPSEC_MAX_NUM_FQS] = { 0 };

#ifdef PME_SWITCH
volatile uint32_t *pre_pme_pkts = 0, *post_pme_pkts = 0;
volatile uint32_t *pme_pattern_pkts = 0;
#endif

stat32_t prev_ip_in_recv;

extern int32_t ipsecfwd_create_sa(struct lwe_ctrl_ipsec_info *ipsec_info);

static void ip_close_fq(uint32_t swi_id, void *data);
static void ip_wait_fq_empty(uint32_t swi_id, void *data);
static int destroy_split_key_fqs();
struct ip_stack_t stack;
struct ipsec_stack_t ipsec_stack;
struct lwe_barrier ipsecfwd_barrier;

/**
 \brief Event Supported is an array of event to be monitored
 */
struct core_pm_events events[MAX_PM_COUNTERS] = {
			{"Cycles         ", PM_CYCLE, 0},
			{"InstrComp      ", PM_INSTRUCTIONS_COMPLETED, 0},
			{"Debug Event 0  ", PM_DVT0_DETECTED, 0},
			{"Debug Event 1  ", PM_DVT1_DETECTED, 0}
};

extern void ipsec_sa_create(struct ipsec_tunnel_config_entry_t *config,
			    struct ipsec_stack_t *ipsec_stack,
			    struct ip_stack_t *ip_stack,
			    uint32_t mode, uint32_t fake_ip_addr);
int32_t g_tunnel_id = 0;
struct ipsec_tunnel_config_t *g_ipsec_tunnel_config;
volatile struct ipsa_context_info_t ipsa_context_info[MAX_SPA_PME_CNTX];

uint32_t ip_cleanup_info[IP_CLEANUP_MAX];
volatile enum application_state ipsec_app_state = INVALID;

typedef void (*cleanup) (uint32_t);
uint32_t num_close_pending;
bool done;
uint32_t g_sec_fq_count;
typedef void (*func_t) (uint32_t timer_id, void *p_data);
func_t shutdown_fnc;
static void ip_retire_fman_fq(uint32_t timer_id, void *p_data);
volatile bool flag_finite_reset;
volatile bool wait_flag;
uint32_t wait_timerid;

static int cmd_print_stats(int argc, char *argv[])
{
	printf("rx_errors: %u\n", rx_errors);
	return 0;
}

/* This function processes FQ state change notifications. This function is
 * called from ISR context. Thus this function must not make any blocking
 * and heavy weight APIs. When the FQ get retired and also gets empty, a
 * SWI is posted towards application's main loop for taking FQs OOS and
 * destroy them.
 */
void ip_fq_state_chg(struct qman_portal *qm,
		     struct qman_fq *fq, const struct qm_mr_entry *msg)
{
	uint8_t verb;
	uint8_t fqs;

	/*Extract the response verb */
	verb = msg->verb & ~QM_MR_VERB_VBIT;

	if (QM_MR_VERB_FQRL == verb) {
		if (0 > swi_add(SWI_ID_INV, fq, ip_close_fq, SWI_PRI_HIGH))
			APP_ERROR("Cannot add SWI for closing FQ");

		return;
	}

	fqs = msg->fq.fqs;

	if ((QM_MR_VERB_FQRN == verb) || (QM_MR_VERB_FQRNI == verb)) {
		if (fqs & QM_MR_FQS_NOTEMPTY) {
			APP_INFO("Retired FQID: %u not yet empty", fq->fqid);

			if (0 > swi_add(SWI_ID_INV, fq, ip_wait_fq_empty,
					SWI_PRI_HIGH))
				APP_ERROR("Cannot add SWI for emptying FQ");
			return;
		}

		if (fqs & QM_MR_FQS_ORLPRESENT) {
			/*We need to wait till we get FQRL */
			APP_INFO("Retired FQID: %u has ORL list", fq->fqid);
			return;
		}

		/*Else FQ is empty and can be destroyed */
		if (0 > swi_add(SWI_ID_INV, fq, ip_close_fq, SWI_PRI_HIGH)) {
			APP_ERROR("Cannot add SWI for closing FQ");
			return;
		}

		return;
	}

	APP_ERROR("Unexpected FQ state change verb: %d", verb);
}

/* This function processes bridging shutdown event*/
static void ip_process_shutdown(void)
{
	uint32_t timer_id;

	APP_INFO("Processing shutdown command ..");
	/*Bootcore is responsible for partition cleanup */
	ipsec_app_state = SHUTDOWN_PENDING;

	/*Disable both Rx mac ports */
	if (-1 == dpa_eth_disable_ports(true, false))
		APP_ERROR("Disabling MAC-Rx port failed");
	else
		APP_INFO("Disabled MAC-Rx ports");

	prev_ip_in_recv = ipsec_stack.ip_stack->ip_stats->ip_in_received;
	/*Start a timer to complete packet processing before shutdown */
	timer_id = start_timer(IP_SHUTDOWN_TIMER_MS * 2,
			       false, NULL, SWI_PRI_HIGH,
			       ip_retire_fman_fq, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown RX timer failed");
	else
		APP_INFO("Started shutdown Rx timer");
}

/* This function processes bridging shutdown event*/
static void ip_retire_fman_fq(uint32_t tmr_id, void *p_data)
{
	uint32_t timer_id;
	stat32_t total_out_packets;
	struct ip_statistics_t *ip_stats = ipsec_stack.ip_stack->ip_stats;

	total_out_packets = ip_stats->ip_out_forward
	    + ip_stats->ip_xmit_icmp_unreach_no_egress.words.lo
	    + ip_stats->ip_route_input_slow.words.lo
	    + ip_stats->ip_xmit_icmp_unreach_need_frag.words.lo
	    + ip_stats->ip_ttl_time_exceeded.words.lo
	    + ip_stats->ip_in_dropped.words.lo;
	if ((prev_ip_in_recv != ip_stats->ip_in_received)
	    || (ip_stats->ip_in_received != total_out_packets)) {
		prev_ip_in_recv = ip_stats->ip_in_received;

		/*Start a timer to complete packet processing before shutdown */
		timer_id = start_timer(IP_SHUTDOWN_TIMER_MS * 2,
				       false, NULL, SWI_PRI_HIGH,
				       ip_retire_fman_fq, NULL);
		if (INV_TIMER_ID == timer_id)
			APP_ERROR("Starting shutdown RX timer failed");
		else
			APP_INFO("Started shutdown Rx timer");

		return;
	}

	/*Put all Rx frame queues in retired state */
	/*This would effectively disable the MAC-less Rx port */
	num_fq_close_pending = dpa_eth_rx_fq_retire();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring Rx FQs failed");
		return;
	}

	APP_INFO("Number of Rx FQs close pending: %d", num_fq_close_pending);

	if (g_sec_fq_count)
		shutdown_fnc = ip_shutdown_sec_tx;
	else
		shutdown_fnc = ip_shutdown_sec_rx_complete;

	/*If there is no FQ close pending, directly move to the next state */
	if (!num_fq_close_pending) {
		shutdown_fnc(INV_TIMER_ID, 0);
		return;
	}

	/*Start a timer to complete FMAN RX shutdown */
	timer_id = start_timer(IP_SHUTDOWN_TIMER_MS,
			       false, NULL, SWI_PRI_HIGH, shutdown_fnc, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown RX timer failed");
	else
		APP_INFO("Started shutdown Rx timer");
}

/*This function waits till the FQ becomes empty and then closes it*/
static void ip_wait_fq_empty(uint32_t swi_id, void *data)
{
	struct qman_fq *fq = data;
	uint32_t vdqcr;
	uint32_t flags;

	APP_INFO("Waiting for FQID: %u to become empty", fq->fqid);

	vdqcr = QM_VDQCR_PRECEDENCE_VDQCR | QM_VDQCR_NUMFRAMES_TILLEMPTY;

	/* NB: the use of QMAN_VOLATILE_FLAG_WAIT is prohibited in LWE
	 * applications. To avoid thrashing, we implement a throttled poll for
	 * completion. */
	if (qman_volatile_dequeue(fq, 0, vdqcr)) {
		APP_ERROR("qman_volatile_dequeue failure for FQID: %u",
			fq->fqid);
		return;
	}
loop:
	qman_fq_state(fq, NULL, &flags);
	if (flags & QMAN_FQ_STATE_VDQCR) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	/*Now the FQ is empty, close it */
	ip_close_fq(SWI_ID_INV, data);
}

/*
 \bried This function is used to move a FQ to OOS state, destroy it and free
	the associated memory with FQ structure.
 */
static void ip_close_fq(uint32_t swi_id, void *data)
{
	struct qman_fq *fq = (struct qman_fq *)data;

	if (0 > qman_oos_fq(fq)) {
		APP_ERROR("Moving FQID: %lu to OOS failed", fq->fqid);
		return;
	}

	num_fq_close_pending--;

	fq_destroy_pending[num_fq_free_pending] = fq;
	num_close_pending--;
	APP_DEBUG("Number of fqs pending: %lu", num_close_pending);
	APP_DEBUG("Closed FQID: %lu", fq->fqid);
}

int ip_bpools_create(void)
{
	struct bman_pool_params params;
	uint32_t bpid;
	ipsec_buff_allocator = buff_allocator;
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_partition_bpools[bpid].size)
			continue;
		params.bpid = bpid;
		params.flags = BMAN_POOL_FLAG_STOCKPILE;
		params.cb = NULL;
		params.cb_ctx = NULL;
		bpool[bpid] = bman_new_pool(&params);

		if (unlikely(NULL == bpool[bpid])) {
			APP_ERROR("bman_new_pool failed");
			abort();
			return -EINVAL;
		}
		APP_INFO("DPA Allocator populated with BpoolID: %x", bpid);
		if (unlikely(0 != dpa_allocator_add_bpool(buff_allocator,
							  dpa_partition_bpools
							  [bpid].size, bpid,
							  bpool[bpid]))) {
			APP_ERROR("dpa_allocator_add_bpool failed");
			return -EINVAL;
		}

		if (!dpa_bpools[bpid].size) {
			if (unlikely
			    (0 !=
			     dpa_allocator_add_bpool(buff_allocator_core,
						     dpa_partition_bpools[bpid].
						     size, bpid,
						     bpool[bpid]))) {
				APP_ERROR("dpa_allocator_add_bpool failed");
				return -EINVAL;
			}
			APP_INFO("BufferPool %x is added to per core list",
				 bpid);
			ipsec_buff_allocator = buff_allocator_core;
		}
	}
	return 0;
}

void ip_cleanup_bpools_destroy(void)
{
	uint32_t bpid;

	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (dpa_partition_bpools[bpid].size) {
			bman_free_pool(bpool[bpid]);
			APP_INFO("BPID : %lu destroyed", bpid);
		}
	}

	/* Free the DPA allocator */
	dpa_allocator_free(buff_allocator_core);
	dpa_allocator_free(buff_allocator);
	APP_INFO("DPA allocator freed");
}

static void ip_cleanup_delete_ipsec_stack(uint32_t data)
{
	APP_INFO("%s", __func__);

	stats_free(ipsec_stack.stats);

	ipsec_tunnel_table_delete(ipsec_stack.itt);
	destroy_split_key_fqs();
}

static void ip_cleanup_delete_arp_cache(uint32_t data)
{
	uint32_t i, j, node_idx;
	struct net_dev_t *dev;
	const struct node_t *node;

	dev = stack.nt->device_head;
	for (j = 0, node_idx = 0; j < g_num_dpa_eth_ports; j++) {
		for (i = 0; i < local_node_count[j]; i++) {
			node = &local_nodes[node_idx];
			if (dev == NULL) {
				dev = stack.nt->device_head;
			}

			if (false == neigh_remove(stack.arp_table,
						  (void *)(node->ip.word),
						  stack.arp_table->proto_len)) {
				APP_INFO("Could not delete neighbor entry");
				goto _error;
			}

			node_idx++;
		}
		dev = dev->next;
	}
_error:
	APP_INFO("%s", __func__);
}

static void ip_cleanup_delete_device(uint32_t data)
{
	uint32_t port;
	struct net_dev_t *dev;
	struct ip_context_t *ctxt;

	for (port = 0; port < g_num_dpa_eth_ports; port++) {
		ctxt = stack.ctxt[port];
		dev = ctxt->fq_ctxt.dev;

		stats_free(dev->stats);
		free(dev);
		free(ctxt);
	}

	free(stack.nt);
	APP_INFO("%s", __func__);
}

static void ip_cleanup_delete_ip_stack(uint32_t data)
{
	stats_free(stack.ip_stats);
	free(stack.protos);
	ip_hooks_delete(stack.hooks);
	rc_delete(stack.rc);
	rt_delete(stack.rt);
	neigh_table_delete(stack.arp_table);

	APP_INFO("%s", __func__);
}

static void ip_cleanup_delete_ipsec_config(uint32_t data)
{
	ipsec_config_delete(g_ipsec_tunnel_config);
	APP_INFO("%s", __func__);
}

static void ip_cleanup_delete_mem_cache(uint32_t data)
{
	APP_INFO("%s", __func__);
}

/**
 \brief Cleanup Barrier created in application bringup
 */
static void ip_cleanup_barrier_destroy(uint32_t data)
{
	barrier_destroy(&ipsecfwd_barrier);
	APP_INFO("ip_cleanup_barrier_destroy()");
	return;
}

/**
 \brief Cleanup Stages of ipsecfwd hybrid application
 */
cleanup ip_cleanup_action[IP_CLEANUP_MAX] = {
#ifdef PME_SWITCH
	ipsa_cleanup_del_all_ctx,
#endif
	ip_cleanup_delete_ipsec_stack,
	ip_cleanup_delete_arp_cache,
	ip_cleanup_delete_device,
	ip_cleanup_delete_ip_stack,
	ip_cleanup_delete_ipsec_config,
	ip_cleanup_delete_mem_cache,
	ip_cleanup_barrier_destroy,
	ip_cleanup_bpools_destroy
};

/**
 \brief Creates Split key Job Queue Descriptor Buffer
 \param[IN] NULL
 \param[OUT] Split Key Job Queue Descriptor buffer
 */
void *create_split_key_sec_descriptor(void)
{
	struct ipsec_encap_descriptor_t *preheader_initdesc;

	preheader_initdesc = (void *)memalign(CACHE_LINE_SIZE,
					      sizeof(struct
						     ipsec_encap_descriptor_t));
	if (preheader_initdesc == NULL) {
		APP_ERROR("%s: No More Buffers left for Descriptor", __func__);
		return NULL;
	}
	memset(preheader_initdesc, 0, sizeof(struct ipsec_encap_descriptor_t));

	preheader_initdesc->prehdr.lo.field.offset = 1;

	return preheader_initdesc;
}

/**
 \brief Creates and initialized the FQs related to a tunnel
 \param[IN] NULL
 \param[OUT] 0 on success
 */
int32_t init_split_key_fqs(void)
{
	uint32_t flags;
	struct qman_fq *fq;
	struct qm_mcc_initfq opts;
	uint32_t ctx_a_excl;
	uint32_t ctx_a_len;
	void *ctxt_a;

	flags = QMAN_FQ_FLAG_NO_ENQUEUE | QMAN_FQ_FLAG_LOCKED;
	fq = (struct qman_fq *)memalign(CACHE_LINE_SIZE,
					sizeof(struct qman_fq) +
					sizeof(struct ipsec_context_t *));
	if (unlikely(NULL == fq)) {
		APP_ERROR("malloc failed in create_fqs for FQ ID: %lu",
			  KEY_SPLIT_FQ_FROM_SEC);
		return -ENOMEM;
	}

	memset(fq, 0, sizeof(struct qman_fq));
	g_splitkey_fq_from_sec = fq;
	fq->cb = ipfwd_split_key_cb;

	if (unlikely(0 != qman_create_fq(KEY_SPLIT_FQ_FROM_SEC, flags, fq))) {
		APP_ERROR("qman_create_fq failed for FQ ID: %lu",
			  KEY_SPLIT_FQ_FROM_SEC);
		return -EINVAL;
	}

	flags = QMAN_INITFQ_FLAG_SCHED | QMAN_INITFQ_FLAG_LOCAL;
	opts.we_mask =
	    QM_INITFQ_WE_DESTWQ | QM_INITFQ_WE_CONTEXTA | QM_INITFQ_WE_FQCTRL;
	opts.fqd.dest.wq = 1;
	opts.fqd.fq_ctrl = QM_FQCTRL_CTXASTASHING | QM_FQCTRL_HOLDACTIVE;
	ctx_a_excl = (QM_STASHING_EXCL_DATA | QM_STASHING_EXCL_CTX);
	ctx_a_len = (1 << 2) | 1;
	opts.fqd.context_a.hi = (ctx_a_excl << 24) | (ctx_a_len << 16);

	if (unlikely(0 != qman_init_fq(fq, flags, &opts))) {
		APP_ERROR("Unable to initialize ingress FQ for SEC4.0");
		return -1;
	}

	flags = QMAN_FQ_FLAG_LOCKED | QMAN_FQ_FLAG_TO_DCPORTAL;

	fq = (struct qman_fq *)memalign(CACHE_LINE_SIZE,
					sizeof(struct qman_fq));
	if (unlikely(NULL == fq)) {
		APP_ERROR("malloc failed in create_fqs for FQ ID");
		return -ENOMEM;
	}

	memset(fq, 0, sizeof(struct qman_fq));
	g_splitkey_fq_to_sec = fq;
	fq->cb = ipfwd_split_key_cb;

	if (unlikely(0 != qman_create_fq(KEY_SPLIT_FQ_TO_SEC, flags, fq))) {
		APP_ERROR("qman_create_fq failed for FQ ID");
		return -EINVAL;
	}

	ctxt_a = create_split_key_sec_descriptor();
	if (ctxt_a == NULL) {
		APP_ERROR("Unable to create job descriptor for split key");
		return -ENOMEM;
	}

	flags = QMAN_INITFQ_FLAG_SCHED;
	opts.we_mask = QM_INITFQ_WE_DESTWQ | QM_INITFQ_WE_CONTEXTA |
	    QM_INITFQ_WE_CONTEXTB;
	opts.fqd.context_a.hi = 0;
	opts.fqd.context_a.lo = APP_VTOP(ctxt_a);
	opts.fqd.context_b = KEY_SPLIT_FQ_FROM_SEC;
	opts.fqd.dest.channel = qm_channel_caam;
	opts.fqd.dest.wq = 0;

	if (unlikely(0 != qman_init_fq(fq, flags, &opts))) {
		APP_ERROR("Unable to Init CAAM Egress FQ");
		return -EINVAL;
	}
	return 0;
}

static int destroy_split_key_fqs()
{
	uint32_t flags;
	struct qman_fq *fq;

	fq = g_splitkey_fq_to_sec;

	if (qman_retire_fq(fq, &flags)) {
		APP_ERROR("qman_retire_fq failed for fq %d", fq->fqid);
		return -1;
	}
	if (0 > qman_oos_fq(fq)) {
		APP_ERROR("Moving FQID: %u to OOS failed",
			  g_splitkey_fq_from_sec->fqid);
		return -1;
	}
	qman_destroy_fq(fq, 0);
	free(fq);

	fq = g_splitkey_fq_from_sec;

	if (0 > qman_retire_fq(fq, &flags)) {
		APP_ERROR("qman_retire_fq failed for fq %d", fq->fqid);
		return -1;
	}
	if (0 > qman_oos_fq(fq)) {
		APP_ERROR("Moving FQID: %u to OOS failed",
			  g_splitkey_fq_from_sec->fqid);
		return -1;
	}
	qman_destroy_fq(fq, 0);
	free(fq);
	return 0;
}

/**
 \brief Adds a new Security association entry
 \param[out] lwe_ctrl_op_info contains SA parameters
 \return Integer status
 */
int32_t ipsecfwd_add_sa(struct lwe_ctrl_op_info * sa_info)
{
	return ipsecfwd_create_sa(&sa_info->app_info.ipsec_info);
}

/**
 \brief Deletes a Security Association
 \param[in] lwe_ctrl_op_info contains SA parameters
 \return Integer status
*/
int32_t ipsecfwd_del_sa(struct lwe_ctrl_op_info * sa_info)
{

	return 0;
}

/**
 \brief Get a Security Association Statistics
 \param[in] lwe_ctrl_op_info contains SA parameters
 \return Integer status
*/
int32_t ipsecfwd_get_sa_stats(struct lwe_ctrl_op_info * sa_info)
{
	struct rc_entry_t *entry = NULL;
	struct lwe_ctrl_ipsec_info *ipsec_info = &sa_info->app_info.ipsec_info;
	uint32_t encryptn_mode;

	APP_DEBUG("ipsecfwd_get_sa_stats: Enter");

	encryptn_mode =
	    (IPSEC_DIR_OUT ==
	     sa_info->app_info.ipsec_info.id.dir) ? ENCRYPT : DECRYPT;

	/* Route look up for the entry */
	entry = rc_entry_lookup(stack.rc,
				ipsec_info->sel.saddr,
				ipsec_info->sel.daddr, IP_HEADER_DEFAULT_TOS);
	if (entry == NULL) {
		APP_ERROR
		    ("rc_entry_lookup failed for decap src = 0x%x; dst = 0x%x",
		     ipsec_info->sel.saddr, ipsec_info->sel.daddr);
		return -1;
	}
	if (encryptn_mode == DECRYPT) {
		ipsec_info->stats.decap_pkts = entry->stats->hits.words.lo;
		ipsec_info->stats.decap_octets = entry->stats->octets.words.lo;

	} else {
		ipsec_info->stats.encap_pkts = entry->stats->hits.words.lo;
		ipsec_info->stats.encap_octets = entry->stats->octets.words.lo;
	}

	APP_DEBUG("ipsecfwd_get_sa_stats: Exit");
	return 0;
}

/**
 \brief Get a per interface Statistics
 \param[in] lwe_ctrl_op_info contains SA parameters
 \return Integer status
*/
int32_t ipsecfwd_get_if_stats(struct lwe_ctrl_op_info * sa_info)
{
	return 0;
}

/**
 \param[in] lwe_ctrl_op_info contains SA parameters
 \param[out] NULL
*/
int32_t ipsecfwd_get_sa_statslist(struct lwe_ctrl_op_info * sa_info)
{
	return 0;
}

/**
 \param[in] lwe_ctrl_op_info contains SA parameters
 \param[out] NULL
*/
int32_t ipsecfwd_get_if_statslist(struct lwe_ctrl_op_info * sa_info)
{
	return 0;
}

/**
 \brief Change Interface Configuration
 \param[out] lwe_ctrl_route_info contains Route parameters
 \return Integer status
 */
int32_t ipfwd_conf_intf(struct lwe_ctrl_op_info * route_info)
{
	APP_DEBUG("ipfwd_conf_intf: Enter");

	APP_DEBUG("ipfwd_conf_intf: Bitmask = 0x%x",
		  route_info->app_info.ip_info.intf_conf.bitmask);

	APP_DEBUG("ipfwd_conf_intf: Ifname = %s",
		  route_info->app_info.ip_info.intf_conf.ifname);
	APP_DEBUG("ipfwd_conf_intf: IPAddr = 0x%x",
		  route_info->app_info.ip_info.intf_conf.ip_addr);
	APP_DEBUG("ipfwd_conf_intf: MAC Addr = %x:%x:%x:%x:%x:%x",
		  route_info->app_info.ip_info.intf_conf.mac_addr[0],
		  route_info->app_info.ip_info.intf_conf.mac_addr[1],
		  route_info->app_info.ip_info.intf_conf.mac_addr[2],
		  route_info->app_info.ip_info.intf_conf.mac_addr[3],
		  route_info->app_info.ip_info.intf_conf.mac_addr[4],
		  route_info->app_info.ip_info.intf_conf.mac_addr[5]);

	APP_DEBUG("ipfwd_conf_intf: Exit");
	return 0;
}

int is_iface_ip(uint32_t ip_addr)
{
	int i;

	for (i = 0; i < g_num_dpa_eth_ports; i++) {
		if (iface_nodes[i].ip.word == ip_addr)
			return 0;
	}

	return -EINVAL;
}

/**
 \brief Gets interface node corresponding to an ip address
 \param[in] ip_addr IP Address
 \return    interface node, On success
	    NULL, 	    On failure
 */
struct node_t *ipfwd_get_iface_for_ip(uint32_t ip_addr)
{
	uint32_t port;

	for (port = 0; port < g_num_dpa_eth_ports; port++) {
		if ((iface_nodes[port].ip.word & 0xffffff00) ==
		    (ip_addr & 0xffffff00))
			break;
	}

	if (unlikely(port == g_num_dpa_eth_ports)) {
		APP_ERROR("%s: Exit: Failed: Not a valid IP addr");
		return NULL;
	}

	return &iface_nodes[port];
}

/**
 \brief Gets device pointer corresponding to an ip address
 \param[in] ip_addr IP Address
 */
struct net_dev_t *ipfwd_get_dev_for_ip(unsigned int ip_addr)
{
	uint32_t port, node = 0, node_idx;
	unsigned char *mac_addr;
	struct net_dev_t *dev;

	/* Check that the arp entry creation request is for a local node */
	for (port = 0, node_idx = 0; port < g_num_dpa_eth_ports; port++) {
		for (node = 0; node < local_node_count[port];
		     node++, node_idx++) {
			if (local_nodes[node_idx].ip.word == ip_addr)
				goto _TEMP;
		}
	}

	APP_DEBUG("%s: Not a Local Node", __func__);

_TEMP:

	/*
	 ** Find out the mac address of the iface node
	 ** corresponding to the local node
	 */
	for (port = 0; port < g_num_dpa_eth_ports; port++) {
		if ((iface_nodes[port].ip.word & 0xffffff00) ==
		    (ip_addr & 0xffffff00))
			break;
	}

	if (unlikely(port == g_num_dpa_eth_ports)) {
		APP_INFO("%s: Exit: Failed: Not in any iface subnet", __func__);
		return NULL;
	}

	/*
	 ** Finding the device ptr correspnding to the iface node
	 */
	mac_addr = (unsigned char *)&(iface_nodes[port].mac);
	dev = stack.nt->device_head;
	while (dev != NULL) {
		char *dev_mac = (char *)(dev->dev_addr);

		if (memcmp(dev_mac, mac_addr, 6) == 0)
			break;

		dev = dev->next;
	}

	return dev;
}

/**
 \brief Adds a new Route Cache entry
 \param[out] lwe_ctrl_route_info contains Route parameters
 \return Integer status
 */
int32_t ipfwd_add_route(struct lwe_ctrl_op_info * route_info)
{
	struct rc_entry_t *entry;
	struct rt_dest_t *dest;
	struct net_dev_t *dev = NULL;
	unsigned int gw_ipaddr = route_info->app_info.ip_info.gw_ipaddr;
#ifdef PME_SWITCH
	uint32_t pme_flow_id;

	pme_flow_id = route_info->app_info.ip_info.flow_id;
#endif

	APP_DEBUG("ipfwd_add_route: Enter");

	dest = rt_dest_alloc(stack.rt);
	if (dest == NULL) {
		APP_ERROR
		    ("Could not allocate route cache related data structure");
		return -1;
	}

	dest->next = NULL;
	dest->neighbor = neigh_lookup(stack.arp_table, gw_ipaddr,
				      stack.arp_table->proto_len);
	if (dest->neighbor == NULL) {
		APP_DEBUG
		    ("%s: Could not find neighbor entry for link-local address",
		     __func__);

		dev = ipfwd_get_dev_for_ip(gw_ipaddr);
		if (dev == NULL) {
			APP_ERROR("%s: not a valid gateway for any subnet",
				  __func__);
			return -1;
		}

		dest->neighbor = neigh_create(stack.arp_table);
		if (unlikely(!dest->neighbor)) {
			APP_ERROR("%s: Unable to create Neigh Entry", __func__);
			return -1;
		}

		if (NULL == neigh_init(stack.arp_table, dest->neighbor, dev,
				       (uint32_t *) & gw_ipaddr)) {
			APP_ERROR("%s: Unable to init Neigh Entry", __func__);
			return -1;
		}

		if (false == neigh_add(stack.arp_table, dest->neighbor)) {
			APP_ERROR("%s: Unable to add Neigh Entry", __func__);
			return -1;
		}
		/* MAC addr would be updated later through ARP request */

		APP_DEBUG("%s: Created neighbor entry, IP addr = %x",
			  __func__, gw_ipaddr);
	}

	dest->dev = dest->neighbor->dev;
#ifdef PME_SWITCH
	if (pme_flow_id) {
		if (!g_flow_to_entry[pme_flow_id]) {
			dest->scope = ROUTE_SCOPE_PME;
			dest->flow_id = pme_flow_id;
		} else {
			APP_ERROR
			    ("Route cache entry for flow_id %d	already present",
			     pme_flow_id);
			rt_dest_free(stack.rt, dest);
			return -1;
		}
	} else
#endif
		dest->scope = ROUTE_SCOPE_GLOBAL;

	entry = rc_create_entry(stack.rc);
	if (entry == NULL) {
		APP_ERROR("Could not allocate route cache entry");
		rt_dest_free(stack.rt, dest);
		return -1;
	}

	entry->saddr = route_info->app_info.ip_info.src_ipaddr;
	entry->daddr = route_info->app_info.ip_info.dst_ipaddr;

	entry->stats =
	    stats_memalign(CACHE_LINE_SIZE,
			   sizeof(struct rc_entry_statistics_t));
	if (entry->stats == NULL) {
		APP_ERROR("Unable to allocate route entry stats");
		return -ENOMEM;
	}
	memset(entry->stats, 0, sizeof(struct rc_entry_statistics_t));

	refcount_acquire(dest->neighbor->refcnt);
	entry->dest = dest;
	entry->tos = IP_TOS;

	if (rc_add_update_entry(stack.rc, entry) == false) {
		APP_INFO("Route cache entry updated");
		rc_free_entry(stack.rc, entry);
	}
#ifdef PME_SWITCH
	g_flow_to_entry[pme_flow_id] = dest;
#endif

	APP_DEBUG("ipfwd_add_route: Exit");
	return 0;
}

/**
 \brief Deletes a Route Cache entry
 \param[out] lwe_ctrl_route_info contains Route parameters
 \return Integer status
 */
int32_t ipfwd_del_route(struct lwe_ctrl_op_info * route_info)
{
	struct rt_dest_t *dest;
	APP_DEBUG("ipfwd_del_route: Enter");

	dest = rc_lookup(stack.rc,
			 route_info->app_info.ip_info.src_ipaddr,
			 route_info->app_info.ip_info.dst_ipaddr,
			 (uint8_t)route_info->app_info.ip_info.tos);
	if (dest == NULL) {
		APP_ERROR("Could not find route cache entry to be deleted");
		return -1;
	}
#ifdef PME_SWITCH
	if (dest->pme_ctx) {
		APP_ERROR
		    ("Could not delete route cache entry, pme ctx still valid");
		return -1;
	}
#endif

	refcount_release(dest->neighbor->refcnt);

	if (rc_remove_entry(stack.rc,
			    route_info->app_info.ip_info.src_ipaddr,
			    route_info->app_info.ip_info.dst_ipaddr,
			    (uint8_t)route_info->app_info.ip_info.tos)
				== false) {
		APP_ERROR("Could not delete route cache entry");
		return -1;
	}
#ifdef PME_SWITCH
	if (dest->flow_id)
		g_flow_to_entry[dest->flow_id] = 0;
#endif

	rt_dest_free(stack.rt, dest);
	APP_DEBUG("ipfwd_del_route: Exit");
	return 0;
}

/**
 \brief Adds a new Arp Cache entry
 \param[out] lwe_ctrl_route_info contains ARP parameters
 \return Integer status
 */
int32_t ipfwd_add_arp(struct lwe_ctrl_op_info * route_info)
{
	unsigned char *c = route_info->app_info.ip_info.mac_addr;
	unsigned int ip_addr = route_info->app_info.ip_info.src_ipaddr;
	struct net_dev_t *dev = NULL;
	struct neigh_t *n;

#if (LOG_LEVEL > 3)
	unsigned char *ip = (unsigned char *)&(ip_addr);
	APP_DEBUG("ipfwd_add_arp: Enter");

	APP_DEBUG("IP = %d.%d.%d.%d ; MAC = %x:%x:%x:%x:%x:%x", ip[0],
		  ip[1], ip[2], ip[3], c[0], c[1], c[2], c[3], c[4], c[5]);
#endif

	n = neigh_lookup(stack.arp_table, ip_addr,
				stack.arp_table->proto_len);

	if (n == NULL) {
		APP_DEBUG
		    ("%s: Could not find neighbor entry for link-local address",
		     __func__);

		dev = ipfwd_get_dev_for_ip(ip_addr);
		if (dev == NULL) {
			APP_DEBUG("ipfwd_add_arp: Exit: Failed");
			return -1;
		}

		n = neigh_create(stack.arp_table);
		if (unlikely(!n)) {
			APP_DEBUG("ipfwd_add_arp: Exit: Failed");
			return -1;
		}
		if (NULL == neigh_init(stack.arp_table, n, dev,
					(uint32_t *) &ip_addr)) {
			APP_ERROR("ipfwd_add_arp: Exit: Failed");
			return -1;
		}

		if (false == neigh_add(stack.arp_table, n)) {
			APP_ERROR("ipfwd_add_arp: Exit: Failed");
			return -1;
		}
	} else {
		n->neigh_state = NEIGH_STATE_UNKNOWN;
		if (route_info->app_info.ip_info.replace_entry) {
			if (false == neigh_replace(stack.arp_table, n)) {
				APP_ERROR("ipfwd_add_arp: Exit: Failed");
				return -1;
			}
		}
	}
	/* Update ARP cache entry */
	if (NULL == neigh_update(n, c, NEIGH_STATE_PERMANENT)) {
		APP_ERROR("ipfwd_add_arp: Exit: Failed");
		return -1;
	}

	APP_DEBUG("ipfwd_add_arp: Exit");
	return 0;
}

/**
 \brief Deletes an Arp Cache entry
 \param[out] lwe_ctrl_route_info contains ARP parameters
 \return Integer status
 */
int32_t ipfwd_del_arp(struct lwe_ctrl_op_info * route_info)
{
	struct neigh_t *neighbor = NULL;
	APP_DEBUG("ipfwd_del_arp: Enter");

	/*
	 ** Do a Neighbour LookUp for the entry to be deleted
	 */
	neighbor = neigh_lookup(stack.arp_table,
				(route_info->app_info.ip_info.src_ipaddr),
				stack.arp_table->proto_len);
	if (neighbor == NULL) {
		APP_ERROR
		    ("Could not find neighbor entry for link-local address");
		return -1;
	}

	/*
	 ** Find out if anyone is using this entry
	 */
	if (*(neighbor->refcnt) != 0) {
		APP_ERROR
		    ("Could not delete neighbor entry as it is being used");
		return -1;
	}

	/*
	 ** Delete the ARP Entry
	 */
	if (false == neigh_remove(stack.arp_table,
				  (void *)(route_info->app_info.ip_info.
					   src_ipaddr),
				  stack.arp_table->proto_len)) {
		APP_ERROR("Could not delete neighbor entry");
		return -1;
	}

	APP_DEBUG("ipfwd_del_arp: Exit");
	return 0;
}

/**
 \brief Gets Route Entry related stats
 \param[out] lwe_ctrl_route_info contains Route parameters
 \return Integer status
 */
int32_t ipfwd_stats_route(struct lwe_ctrl_op_info * route_info)
{
	struct rc_entry_t *entry = NULL;
	struct lwe_ctrl_route_stats *stats =
	    &(route_info->app_info.ip_info.stats);

	APP_DEBUG("ipfwd_stats_route: Enter");

	entry = rc_entry_lookup(stack.rc,
				route_info->app_info.ip_info.src_ipaddr,
				route_info->app_info.ip_info.dst_ipaddr,
				(uint8_t)route_info->app_info.ip_info.tos);
	if (entry == NULL)
		return -1;

	stats->hits.words.hi = entry->stats->hits.words.hi;
	stats->hits.words.lo = entry->stats->hits.words.lo;

	stats->misses.words.hi = entry->stats->octets.words.hi;
	stats->misses.words.lo = entry->stats->octets.words.lo;

	APP_DEBUG("ipfwd_stats_route: Exit");
	return 0;
}

/**
 \brief Gets System Stats related to the IPfwd Application
 \param[out] lwe_ctrl_route_info contains Route parameters
 \return Integer status
 */
int32_t ipfwd_sys_stats_route(struct lwe_ctrl_op_info * route_info)
{
	struct lwe_ctrl_route_stats *stats =
	    &(route_info->app_info.ip_info.stats);

	APP_DEBUG("ipfwd_sys_stats_route: Enter");

	stats->entry_count = stack.rc->stats->entry_count,
	    stats->stale_entries = stack.rc->stats->stale_entries;
	stats->new_entries = stack.rc->stats->new_entries;
	stats->removed_entries = stack.rc->stats->removed_entries;
	stats->hits.words.hi = stack.rc->stats->hits.words.hi;
	stats->hits.words.lo = stack.rc->stats->hits.words.lo;

	stats->misses.words.hi = stack.rc->stats->misses.words.hi;
	stats->misses.words.lo = stack.rc->stats->misses.words.lo;

	APP_DEBUG("ipfwd_sys_stats_route: Exit");
	return 0;
}

/**
 \brief Update application expected frame counter for its completion
	If frame_cnt specified is 0, then application would run indefinitely
	else would come out after printing stats after certain number of pkts.
 */
static int32_t ip_edit_num_cnt(struct lwe_ctrl_op_info *cp_info)
{
	infinite_lwe = (cp_info->app_info.ip_info.frame_cnt == 0) ?
	    true : false;

	initial_frame_count = cp_info->app_info.ip_info.frame_cnt;
	APP_INFO("Frame count changed to %d", initial_frame_count);
	if (infinite_lwe == true)
		APP_INFO("Application is going in infinite_mode");
	else
		APP_INFO("Application is going to finite_mode");

	return 0;
}

/**
 \brief Reset application stats
 */
static int32_t ip_reset_stats(struct lwe_ctrl_op_info *cp_info)
{
	memset(stack.ip_stats, 0, sizeof(struct ip_statistics_t));
	memset(ipsec_stack.stats, 0, sizeof(struct ipsec_pkt_statistics_t));
	memset(pme_stats, 0, sizeof(struct pme_stats_t));

	return 0;
}

/**
 \brief Message handler for message coming from Control plane
 \param[in] lwe_ctrl_op_info contains SA parameters
 \return NULL
*/
void ipsecfwd_process_req_from_cp(struct lwe_ctrl_op_info *sa_info)
{
	int32_t s32Result = 0;
	sa_info->result = LWE_CTRL_RSLT_FAILURE;

	APP_DEBUG("ipsecfwd_process_req_from_cp: Enter");

	switch (sa_info->msg_type) {
	case LWE_CTRL_CMD_TYPE_SAADD:
		APP_INFO("SA_ADD sa_info 0x%x", (uint32_t) sa_info);
		s32Result = ipsecfwd_add_sa(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_SADEL:
		s32Result = ipsecfwd_del_sa(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_SASTATS:
		s32Result = ipsecfwd_get_sa_stats(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_IFSTATS:
		s32Result = ipsecfwd_get_if_stats(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_SALIST:
		s32Result = ipsecfwd_get_sa_statslist(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_IFLIST:
		s32Result = ipsecfwd_get_if_statslist(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_ROUTE_ADD:
		s32Result = ipfwd_add_route(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_ROUTE_DEL:
		s32Result = ipfwd_del_route(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_ARP_ADD:
		s32Result = ipfwd_add_arp(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_ARP_DEL:
		s32Result = ipfwd_del_arp(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_IPFWD_SYS_STATS:
		s32Result = ipfwd_sys_stats_route(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_ROUTE_STATS:
		s32Result = ipfwd_stats_route(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_INTF_CONF_CHNG:
		s32Result = ipfwd_conf_intf(sa_info);
		break;
#ifdef PME_SWITCH
	case LWE_CTRL_CMD_TYPE_PME_ADD:
		s32Result = ipsa_add_ctx(sa_info);
		ip_cleanup_info[IPSA_CLEANUP_ADD_CTX]++;
		break;

	case LWE_CTRL_CMD_TYPE_PME_DEL:
		s32Result = ipsa_del_ctx(sa_info);
		break;
#endif
	case LWE_CTRL_CMD_TYPE_FRAMECNT_EDIT:
		s32Result = ip_edit_num_cnt(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_RESET_STATS:
		s32Result = ip_reset_stats(sa_info);
		break;

	case LWE_CTRL_CMD_TYPE_GO:
		s32Result = 0;
#ifdef PME_SWITCH
		g_all_context_initialized = 1;
#endif
		break;

	default:
		break;
	}

	if (s32Result == 0) {
		sa_info->result = LWE_CTRL_RSLT_SUCCESSFULL;
	} else {
		APP_ERROR("%s: CP Request can't be handled", __func__);
	}

	APP_DEBUG("ipsecfwd_process_req_from_cp: Exit");
	return;
}

/**
 \brief Initializes Receive context for IPSEC Forwarding app
 \param[in] struct net_dev * Netdev Pointer
 \param[in] struct ip_stack_t * ipstack pointer
 \param[out] Return initialized context
 */
static inline void
initialize_contexts(struct ip_context_t *ip_ctxt, struct net_dev_t *dev,
		    struct ip_stack_t *stack)
{
	ip_ctxt->fq_ctxt.handler = &ip_handler;
	ip_ctxt->fq_ctxt.dev = dev;
	ip_ctxt->stats = stack->ip_stats;
	ip_ctxt->hooks = stack->hooks;
	ip_ctxt->protos = stack->protos;
	ip_ctxt->rc = stack->rc;
}

/**
 \brief Initializes array of network local nodes
 \details Initializes array of network local nodes and assign it mac
	and ip addresses. Local nodes initialized are nodes directly
	connected to the interfaces
 \param[in] struct node_t * a Network Node
 \param[in] uint32_t Number of nodes to be created
 */
void create_local_nodes(struct node_t *arr)
{
	uint32_t port, node, node_idx;

	for (port = 0, node_idx = 0; port < g_num_dpa_eth_ports; port++) {
		for (node = 0; node < local_node_count[port];
		     node++, node_idx++) {
			arr[node_idx].mac.hilo.address_hi = ETHERNET_ADDR_MAGIC;
			arr[node_idx].mac.hilo.address_lo =
			    arr[node_idx].ip.word =
			    0xc0a80002 + (iface_subnet[port] << 8) + node;
		}
	}
}

/**
\brief Initializes array of network iface nodes
\details Assign mac and ip addresses to the interface nodes.
 Interface nodes corresponds to the interfaces
\param[in] struct node_t * a Network Node
\param[in] uint32_t Number of nodes to be created
*/
void create_iface_nodes(struct node_t *arr)
{
	uint32_t port, if_idx;

	for (port = 0, if_idx = 0; port < g_num_dpa_eth_ports; port++, if_idx++) {
		memcpy(arr[if_idx].mac.bytes, dpa_eth_ports[port].mac_addr, 6);
		arr[if_idx].ip.word = (0xc0a80001 + (iface_subnet[port] << 8));
		APP_INFO("PortID = %d is %s interface node with IP Address "
			 "%d.%d.%d.%d and MAC Address " MAC_FMT, port,
			 (dpa_eth_ports[port].type == MAC_LESS_ETH) ?
			 "MAC_LESS" : "FMAN",
			 arr[if_idx].ip.bytes[0], arr[if_idx].ip.bytes[1],
			 arr[if_idx].ip.bytes[2], arr[if_idx].ip.bytes[3],
			 NMAC_STR(arr[if_idx].mac.bytes));
	}
}

/*
 * Frame gets freed to bman pool
*/
void discard_handler(void *buf, uint8_t bpid)
{
	struct bm_buffer bman_buf;

	bman_buf.hi = 0;
	bman_buf.lo = (uint32_t) APP_VTOP(buf);
	bman_buf.bpid = bpid;

	if (dpa_allocator_free_buff(buff_allocator, &bman_buf))
		APP_ERROR("Failed to free buffer to bpid %x", bpid);
}

/**
 \brief Device Tx Initialization
 */
void dpa_dev_tx_init(struct dpa_dev_t *dev, struct dpa_fq_range_t *dpa_fq)
{
	uint32_t fq_idx;

	if (unlikely(dpa_fq->fq_count == 0)) {
		APP_ERROR("FQ Count is zero");
		return;
	}

	for (fq_idx = 0; fq_idx < dpa_fq->fq_count; fq_idx++) {
		dev->tx_fq[fq_idx] = dpa_fq->fq[fq_idx];
	}
}

/**
 \brief Device Rx Initialization
 */
void dpa_dev_rx_init(struct dpa_dev_t *dev, struct dpa_fq_range_t *dpa_fq,
		     struct ip_context_t *ctxt)
{
	uint32_t fq_idx;
	struct ip_fq_context_t *fq_ctxt;

	if (unlikely(dpa_fq->fq_count == 0)) {
		APP_ERROR("FQ Count is zero");
		return;
	}

	for (fq_idx = 0; fq_idx < dpa_fq->fq_count; fq_idx++) {
		dev->rx_fq[fq_idx] = dpa_fq->fq[fq_idx];
		fq_ctxt = (struct ip_fq_context_t *)(dpa_fq->fq[fq_idx]);
		fq_ctxt->ip_ctxt = ctxt;
	}
}

/**
 \brief Creates netdev Device Nodes
 \param[in] struct ip_stack_t * IPFwd stack structure
 \param[in] struct node_t * Link Node
 \param[in] uint32_t Number of network devices to be created
 \param[out] return status
 */
static int32_t
create_devices(struct ip_stack_t *stack, struct node_t *link_nodes)
{
	uint32_t port;
	struct net_dev_t *dev;
	struct ip_context_t *ctxt;

	stack->nt = net_dev_init();
	if (unlikely(!stack->nt)) {
		APP_ERROR("No memory available for neighbor table");
		return -ENOMEM;
	}
	for (port = 0; port < g_num_dpa_eth_ports; port++) {
		if (!
		    (ctxt =
		     memalign(CACHE_LINE_SIZE, sizeof(struct ip_context_t)))) {
			APP_ERROR("No Memory for IP context");
			return -ENOMEM;
		}

		if (unlikely((dev = dpa_dev_allocate(stack->nt)) == NULL)) {
			APP_ERROR("Unable to allocate net device Structure");
			free(ctxt);
			return -ENOMEM;
		}

		dpa_dev_init(dev);
		dev->set_ll_address(dev, &link_nodes[port].mac);
		dev->set_mtu(dev, IFACE_MTU);

		initialize_contexts(ctxt, dev, stack);
		if (dpa_eth_ports[port].type == MAC_LESS_ETH) {
			dpa_dev_rx_init((struct dpa_dev_t *)dev,
					&dpa_eth_ports[port].ingress, ctxt);
		} else {
			dpa_dev_rx_init((struct dpa_dev_t *)dev,
					&dpa_eth_ports[port].pcd_ingress, ctxt);
			dpa_dev_rx_init((struct dpa_dev_t *)dev,
					&dpa_eth_ports[port].ingress, ctxt);
		}
		stack->ctxt[port] = ctxt;
		dpa_dev_tx_init((struct dpa_dev_t *)dev,
				&dpa_eth_ports[port].egress);
		if (!net_dev_register(stack->nt, dev)) {
			APP_ERROR("%s: Netdev Register Failed", __func__);
			return -EINVAL;
		}
	}
	return 0;
}

/*
 \brief Peforms all global initialization required by IPsec
 \param[in] ipsec_stack is the global ipsec stack structure
 \param[out] NULL
 */
void initialize_ipsec_stack(struct ipsec_stack_t *ipsec_stack)
{
	struct ipsec_context_t *ipsec_ctxt;
	struct ip_context_t *ctxt;
	uint32_t i;
	struct bm_buffer bman_buf;

	ipsec_stack->ip_stack = &stack;
	ipsec_stack->itt = ipsec_tunnel_table_create();
	/* initialization virtual netdev table */
	ipsec_stack->stats =
	    stats_memalign(CACHE_LINE_SIZE, sizeof(*ipsec_stack->stats));
	if (ipsec_stack->stats == NULL) {
		APP_ERROR("Unable to allocate ipsec stats");
		return;
	}
	memset(ipsec_stack->stats, 0, sizeof(*ipsec_stack->stats));

	/* initialize the contextB for encap and decap returns */
	if (dpa_allocator_get_buff(ipsec_buff_allocator,
				   sizeof(struct ipsec_context_t),
				   &bman_buf) <= 0) {
		APP_ERROR("No More Buffers left");
		return;
	}
	ipsec_ctxt = APP_PTOV(bman_buf.lo);
	memset(ipsec_ctxt, 0, sizeof(*ipsec_ctxt));
	ipsec_ctxt->ip_ctxt.fq_ctxt.handler = &ipsec_encap_cb;
	ipsec_ctxt->ip_ctxt.stats = ipsec_stack->ip_stack->ip_stats;
	ipsec_ctxt->ip_ctxt.hooks = ipsec_stack->ip_stack->hooks;
	ipsec_ctxt->ip_ctxt.protos = ipsec_stack->ip_stack->protos;
	ipsec_ctxt->ip_ctxt.rc = ipsec_stack->ip_stack->rc;
	ipsec_ctxt->itt = ipsec_stack->itt;
	ipsec_ctxt->stats = ipsec_stack->stats;
	ipsec_stack->ipsec_ctxt = ipsec_ctxt;
	for (i = 0; i < g_num_dpa_eth_ports; i++) {
		ctxt = ipsec_stack->ip_stack->ctxt[i];
		ctxt->user_data = ipsec_ctxt;
	}
}

/**
 \brief Populate static arp entries
 \param[in] struct ip_stack_t * IPFwd stack structure
 \param[in] struct node_t * Link Node
 \param[out] 0 on success, otherwise -ve value
 */
int populate_arp_cache(struct ip_stack_t *stack, struct node_t *local_nodes)
{
	uint32_t i, j, node_idx;
	struct net_dev_t *dev;
	const struct node_t *node;

	dev = stack->nt->device_head;
	for (j = 0, node_idx = 0; j < g_num_dpa_eth_ports; j++) {
		for (i = 0; i < local_node_count[j]; i++) {
			node = &local_nodes[node_idx];
			if (dev == NULL) {
				dev = stack->nt->device_head;
			}

			if (0 > add_arp_entry(stack->arp_table, dev, node)) {
				APP_ERROR("%s: failed to add ARP entry",
					  __func__);
				return -EINVAL;
			}
			node_idx++;
		}
		dev = dev->next;
	}
	return 0;
}

/**
 \brief Initialize IP Stack
 \param[in] struct ip_stack_t * IPFwd Stack pointer
 \param[out] Return Status
 */
static int32_t initialize_ip_stack(struct ip_stack_t *stack)
{
	if (!(stack->arp_table = arp_table_create())) {
		APP_ERROR("Failed to create ARP Table");
		return -1;
	}
	if (!(neigh_table_init(stack->arp_table))) {
		APP_ERROR("Failed to init ARP Table");
		return -1;
	}
	if (!(stack->rt = rt_create())) {
		APP_ERROR("Failed in Route table initialized");
		return -1;
	}
	if (!(stack->rc = rc_create(IP_RC_EXPIRE_JIFFIES, IP_ADDRESS_BYTES))) {
		APP_ERROR("Failed in Route cache initialized");
		return -1;
	}
	if (!(stack->hooks = ip_hooks_create())) {
		APP_ERROR("Failed in IP Stack hooks initialized");
		return -1;
	}
	if (!(stack->protos = ip_protos_create())) {
		APP_ERROR("IP Stack L4 Protocols initialized");
		return -1;
	}
	if (!(stack->ip_stats = ipfwd_stats_init())) {
		APP_ERROR("Unable to allocate ip stats structure for stack");
		return -1;
	}
	memset(stack->ip_stats, 0, sizeof(struct ip_statistics_t));

	APP_DEBUG("IP Statistics initialized");
	return 0;
}

int32_t dpa_part_bpool_init()
{
	int32_t off = -1;
	int32_t ret = -1;
	uint32_t *bpool_cfg;
	uint32_t *bp_id;
	uint32_t size;
	uint32_t bpool_found = 0;

	while (1) {
		ret =
		    fdt_node_offset_by_compatible(devtree_ptr, off,
						  "fsl,p4080-bpool");
		if (ret < 0) {
			break;
		}
		off = ret;

		bpool_found++;

		/*Read the buffer-pool ID */
		bp_id = (uint32_t *) (fdt_getprop(devtree_ptr, ret, "fsl,bpid",
						  NULL));
		if (unlikely(NULL == bp_id)) {
			APP_ERROR("fdt_getprop failed for fsl,bpid");
			return -EINVAL;
		}

		bpool_cfg = (uint32_t *) fdt_getprop(devtree_ptr, ret,
						     "fsl,bpool-cfg", NULL);
		if (unlikely(NULL == bpool_cfg)) {
			APP_ERROR("fdt_getprop failed for fsl,bpool-cfg"
				  "for pool: %lu", *bp_id);
			return -EINVAL;
		}

		size = bpool_cfg[3];
		if (unlikely(0 == size)) {
			APP_ERROR
			    ("BPID: %lu configured with 0 size buffers",
			     *bp_id);
			return -EINVAL;
		}

		/*Already not initialised */
		if (0 == dpa_partition_bpools[*bp_id].size)
			dpa_partition_bpools[*bp_id].size = size;
	}

	if (bpool_found)
		return 0;

	return -EINVAL;
}

static inline void check_exit_criteria(void)
{
	if (unlikely(*total_pkts_processed >= initial_frame_count))
		test_completed = true;
}

/*This function waits for shutdown process to complete and also cleans up
per core resources*/
static void ip_wait_for_shutdown_complete(void)
{
	void *p_data;
	uint32_t swi_id;
	uint32_t i = 0;
	while (1) {
		if (SHUTDOWN_COMPLETE == ipsec_app_state)
			break;
		/*This would dequeue any remanining notifications from QMAN */
		qman_poll();
		while (SWI_ID_INV != (swi_id = swi_activate(&p_data))) {
			switch (swi_id) {
			case SWI_ID_TIMER_EXPIRY:
				proc_timer_expiry((uint32_t) p_data);
				break;
			default:
				break;
			}
		}

		rcu_synchronize_cpu(lwe_instance);
	}

	if (0 != timer_framework_stop()) {
		APP_ERROR("timer_framework_stop failed");
		return;
	}

	for (i = 0; i < num_fq_free_pending; i++) {
		qman_destroy_fq(fq_destroy_pending[i], 0);
		free(fq_destroy_pending[i]);
	}

	rcu_leave(lwe_instance);
	APP_INFO("Left RCU protection");
	/* Call all the remaining cleanup fn */
	for (i = IP_CLEANUP_INIT_IPSEC_STACK - 1; i < IP_CLEANUP_MAX - 1; i++)
		ip_cleanup_action[i]
		    (ip_cleanup_info[i]);
	APP_INFO("Destroyed BMAN pools and buffer allocator");
}

void ip_reset(void)
{
	uint32_t timer_id;
	stat32_t total_out_packets;
	struct ip_statistics_t *ip_stats = ipsec_stack.ip_stack->ip_stats;

	total_out_packets = ip_stats->ip_out_forward
	    + ip_stats->ip_xmit_icmp_unreach_no_egress.words.lo
	    + ip_stats->ip_route_input_slow.words.lo
	    + ip_stats->ip_xmit_icmp_unreach_need_frag.words.lo
	    + ip_stats->ip_ttl_time_exceeded.words.lo
	    + ip_stats->ip_in_dropped.words.lo;

	if ((prev_ip_in_recv != ip_stats->ip_in_received)
	    || (ip_stats->ip_in_received != total_out_packets)) {
		prev_ip_in_recv = ip_stats->ip_in_received;

		/*Start a timer to complete packet processing before reset */
		timer_id = start_timer(IP_SHUTDOWN_TIMER_MS * 2,
				       false, NULL, SWI_PRI_HIGH,
				       ip_reset, NULL);
		if (INV_TIMER_ID == timer_id)
			APP_ERROR("Starting timer for process packets failed");
		else
			APP_INFO("Started timer for process packets");

		return;
	} else
		ipsec_app_state = RESET_COMPLETE;

}

/* This function processes reset event*/
static void ip_process_reset(void)
{
	uint32_t timer_id;

	APP_INFO("Processing reset command ..");
	ipsec_app_state = RESET_PENDING;

	/*Disable both Rx mac ports */
	if (-1 == dpa_eth_disable_ports(true, false))
		APP_ERROR("Disabling MAC-Rx port failed");
	else
		APP_INFO("Disabled MAC-Rx ports");

	prev_ip_in_recv = ipsec_stack.ip_stack->ip_stats->ip_in_received;
	/*Start a timer to complete packet processing before reset */
	timer_id = start_timer(IP_SHUTDOWN_TIMER_MS * 2,
			       false, NULL, SWI_PRI_HIGH, ip_reset, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting process packets timer failed");
	else
		APP_INFO("Started timer for processing packets");
}

static void ip_wait_complete(void)
{
	APP_INFO("Waited for a few min, now exiting");
	wait_flag = 1;
}

/* This function processes wait event */
static void ip_timer_start(void)
{
	/*Start a timer to wait for any CP command */
	wait_timerid = start_timer(IP_WAIT_TIMER_MS * 2,
				   false, NULL, SWI_PRI_HIGH,
				   ip_wait_complete, NULL);
	if (INV_TIMER_ID == wait_timerid)
		APP_ERROR("Starting wait timer failed");
	else
		APP_INFO("Started wait timer");
}

int standby;
int primary_is_ok = 1;

#ifdef HOT_STANDBY_ENABLE

/* FQ memory offset on whiteboard */
#define FQ_OFFSET      0x02000000

volatile int primary_hb;
uint32_t qman_fq_ptr;

void init_shared_fq_ptr()
{
	struct dpa_fq_range_t *fq_range;
	int port_id, count;

	qman_fq_ptr = (uint32_t) & _whiteboard_start + FQ_OFFSET;
	APP_INFO("fq start is 0x%x", qman_fq_ptr);

	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {

		fq_range = &dpa_eth_ports[port_id].egress;
		for (count = 0; count < fq_range->fq_count; count++) {
			fq_range->fq[count] = (struct qman_fq *)qman_fq_ptr;
			qman_fq_ptr += sizeof(struct qman_fq) + sizeof(struct
								       ip_context_t);
		}
		fq_range = &dpa_eth_ports[port_id].ingress;
		for (count = 0; count < fq_range->fq_count; count++) {
			fq_range->fq[count] = (struct qman_fq *)qman_fq_ptr;
			qman_fq_ptr += sizeof(struct qman_fq) + sizeof(struct
								       ip_context_t);
		}
		fq_range = &dpa_eth_ports[port_id].ingress_err;
		for (count = 0; count < fq_range->fq_count; count++) {
			fq_range->fq[count] = (struct qman_fq *)qman_fq_ptr;
			qman_fq_ptr += sizeof(struct qman_fq) + sizeof(struct
								       ip_context_t);
		}
		fq_range = &dpa_eth_ports[port_id].pcd_ingress;
		for (count = 0; count < fq_range->fq_count; count++) {
			fq_range->fq[count] = (struct qman_fq *)qman_fq_ptr;
			qman_fq_ptr += sizeof(struct qman_fq) + sizeof(struct
								       ip_context_t);
		}
		fq_range = &dpa_eth_ports[port_id].egress_err;
		for (count = 0; count < fq_range->fq_count; count++) {
			fq_range->fq[count] = (struct qman_fq *)qman_fq_ptr;
			qman_fq_ptr += sizeof(struct qman_fq) + sizeof(struct
								       ip_context_t);
		}
		fq_range = &dpa_eth_ports[port_id].egress_confirm;
		for (count = 0; count < fq_range->fq_count; count++) {
			fq_range->fq[count] = (struct qman_fq *)qman_fq_ptr;
			qman_fq_ptr += sizeof(struct qman_fq) + sizeof(struct
								       ip_context_t);
		}
	}
	return;
}

void identify_partition()
{
	char *label = (char *)(fdt_getprop(devtree_ptr, 0, "label", NULL));

	if (strcmp(label, "p3-lwe") == 0) {
		standby = 0;
		APP_INFO("\nPrimary Partition\n");
	} else if (strcmp(label, "p4-lwe") == 0) {
		standby = 1;
		APP_INFO("\nStandby Partition\n");
	} else {
		APP_ERROR("unknown label for the partition");
	}
}

uint32_t *bc_id;

void get_byte_channel_id()
{
	int off;

	off = fdt_node_offset_by_compatible(devtree_ptr, -1,
					    "fsl,hv-byte-channel-handle");
	bc_id = (uint32_t *) (fdt_getprop(devtree_ptr, off, "reg", NULL));
	if (bc_id == NULL)
		APP_ERROR("byte channel not found");
}

/* Time interval in seconds at which standby checks if primary is alive */
#define HEARTBEAT	3

/* Number of responses to standby after which primary stops itself */
#define PRIMARY_ALIVE_COUNT	10

void is_primary_alive()
{
	static int num_of_retries;
	int tx_count, ret;
	int rx_count = 0;
	char buf[] = "heartbeat check";

	tx_count = strlen(buf) + 1;

	ret = fh_byte_channel_send(*bc_id, &tx_count, buf);
	if (0 == ret && tx_count != 0) {
		APP_INFO("Message sent to Primary: %s", buf);
		udelay(HEARTBEAT * 1000 * 1000);
		ret = fh_byte_channel_poll(*bc_id, &rx_count, &tx_count);
		if (rx_count != 0) {
			int retval;
			retval = fh_byte_channel_receive(*bc_id, &rx_count,
							 buf);
			if (retval == 0) {
				APP_INFO("Response from Primary partition is:"
					 " %s", buf);
				num_of_retries = 0;
			} else {
				APP_ERROR(" fh_byte_channel_receive failed..");
			}
		} else {
			APP_INFO("No response from Primary. Re-trying...");
			++num_of_retries;
			if (num_of_retries >= 3) {
				APP_INFO("Primary is not alive."
					 " Starting Standby...");
				primary_is_ok = 0;
				return;
			}
		}
	} else {
		APP_INFO("send failed...Num of char sent %d", tx_count);
	}
}
#endif
int ipsecfwd_init_fqs(void)
{
	struct td_param pcd_td_param;
	struct td_param *pcd_td_ptr = NULL;
#ifdef CGR_SUPPORT
	pcd_td_param.mant = 1; /**< Mantissa is 1 */
	pcd_td_param.exp = 11; /**< Exponent is 11 */
	pcd_td_ptr = &pcd_td_param;
	/* Allow 4k Frames in CG */
	pcd_td_param.flag = QM_FQCTRL_CGE;
	pcd_td_param.cgr_cb = NULL;
	pcd_td_param.cgr_id = (uint8_t)get_lwe_id(); /* First CGR ID */
	if (!init_cgr(pcd_td_ptr)) {
		APP_ERROR("Unabled to create Congestion Group");
		return -EINVAL;
       	}
#endif
	if (0 != dpa_eth_fq_init(64, 64, 64, NULL, pcd_td_ptr)) {
		APP_ERROR("Unable to initialize FQs");
		return -EINVAL;
	}
	return 0;
}

/**
 \brief Main function of IPSEC FWD Application
 \details Once entered from LWE Initialization, this ipsecfwd code is in
	a infinite loop looking for ingress frames for forwarding/tunneling
 \param[in] int Argument count
 \param[in] char *argv[] Argument list pointer
 \return This function only returns if forwarding application fails.
	Otherwise this goes in a infinite loop
 */
int main(int argc, char *argv[])
{
	struct ipsec_tunnel_config_t *ipsec_tunnel_config;
	int ecode = 0;
	uint32_t underruns = 0, swi_id, id;
	volatile struct lwe_ctrl_op_info *sa_info;
#ifdef PME_SWITCH
	int count;
	struct spa_pme_ctx_stats_t *pme_ctx_stats;
#endif
	void *p_data;
	uint32_t pkt_processed_till_now;
	bool shutdn_in_proc = 0, reset_in_proc = 0;

	ipsec_init();
	curr_lwe_id = get_lwe_id();
	spin_lock(&ipsecfwd_lock);
	g_lwe_ctrl_sa_info = (struct lwe_ctrl_op_info *)&_whiteboard_start;
	sa_info = g_lwe_ctrl_sa_info;
	if (INV_CORE_ID == g_boot_lwe_core) {
		APP_INFO
		    ("%d: Welcome to FSL IPSec Forwarding Application!",
		     curr_lwe_id);
		add_command("print_stats", cmd_print_stats, "Print app stats");

#ifdef DEBUG_PROFILE
		/* Initialize System Debug Space */
		init_dcsr();
#endif
		g_boot_lwe_core = curr_lwe_id;
		infinite_lwe = true;
		grand_total_cycles = 0;
		test_completed = false;
		cleanup_completed = false;
		done = false;
		num_fq_free_pending = 0;
		flag_finite_reset = 0;
		wait_flag = 0;
		memset((void *)sa_info, 0, sizeof(g_lwe_ctrl_sa_info));
		if (0 >
		    (pool_channel_id =
		     extract_qman_pool_channel(devtree_ptr))) {
			APP_ERROR("local_extract_qman_channel_id failed");
			abort();
			return -EINVAL;
		}
		local_channel = pool_channel_id;
		pool_channel_id = pool_channel_id + qm_channel_pool1 -1;
	} else {
		APP_INFO("IPSec Forwarding Got partition ID %x", curr_lwe_id);
	}

	APP_INFO("Initial Frame count is %x", initial_frame_count);
	spin_unlock(&ipsecfwd_lock);
	markpoint(0);

	if (!(buff_allocator = dpa_allocator_create(&ecode))) {
		APP_ERROR("Global Buffer Allocator Initialization Failure");
		abort();
		return -ENOMEM;
	}

	if (!(buff_allocator_core = dpa_allocator_create(&ecode))) {
		APP_ERROR("Core Buffer Allocator Initialization Failure");
		abort();
		return -ENOMEM;
	}

	if (curr_lwe_id == g_boot_lwe_core) {
		if (barrier_init(&ipsecfwd_barrier, get_online_core_mask())) {
			APP_INFO("Barrier initialization failed");
			return -EINVAL;
		}
		ip_cleanup_info[IP_CLEANUP_BARRIER_INIT]++;

		APP_INFO("LWE Mask is %x", get_online_core_mask());

		memset(dpa_partition_bpools, 0,
		       MAX_NUM_BMAN_POOLS * sizeof(struct dpa_bpool_t));
		if (unlikely(0 > dpa_part_bpool_init())) {
			APP_ERROR("Parition BPOOL Initialization failed");
			return -EINVAL;
		}
#ifdef HOT_STANDBY_ENABLE
		identify_partition();
		get_byte_channel_id();

		while (standby == 1 && primary_is_ok == 1)
			is_primary_alive();

#endif

		if (0 >= dpa_eth_read(&recv_channel_map)) {
			APP_ERROR("Unable to read Eth Info from Devtree");
			return -EINVAL;
		}
#ifdef HOT_STANDBY_ENABLE
		init_shared_fq_ptr();
#endif
		if (0 !=
		    dpa_eth_fq_create((struct qman_fq_cb *)&ipfwd_rx_cb,
				      (struct qman_fq_cb *)&ipfwd_rx_cb_pcd,
				      (struct qman_fq_cb *)&ipfwd_rx_cb_err,
				      (struct qman_fq_cb *)&ipfwd_tx_cb,
				      (struct qman_fq_cb *)&ipfwd_tx_cb_confirm,
				      (struct qman_fq_cb *)&ipfwd_tx_cb_err,
				      sizeof(struct ip_context_t *))) {
			APP_ERROR("Unable to Create FQs");
			return -EINVAL;
		}
		if (standby == 0) {
			ipsecfwd_init_fqs();
		}

#ifdef PME_SWITCH
		pme_stats = stats_memalign(CACHE_LINE_SIZE,
					   sizeof(struct pme_stats_t));
		memset(pme_stats, 0, sizeof(struct pme_stats_t));
		pre_pme_pkts = &(pme_stats->g_total_frames);
		post_pme_pkts = &(pme_stats->g_post_pme_pkts);
		pme_pattern_pkts = &(pme_stats->g_pme_pattern_pkts);
		/* Workaround for dynamic FQ allocation in PME
		 * Earlier it was failing for PME flows greater
		 * than 120 */
		for (id = FQID_BASE; id <= FQID_RANGE; id++)
			qm_fq_free(id);
#endif
	}
	/* Wait Till DPA is Initialized */
	while (0 == g_num_dpa_eth_ports) ;
	while (standby == 1 && primary_is_ok == 1) ;

	/*Control partition has already extracted frame pool info */
	ip_bpools_create();
	num_poll_before_shutdown = 0;
	if (curr_lwe_id == g_boot_lwe_core) {
		if (unlikely(NULL == mem_cache_init())) {
			APP_ERROR("Cache Creation error");
			return -ENOMEM;
		}
		ip_cleanup_info[IP_CLEANUP_MEM_CACHE_INIT]++;

		ipsec_tunnel_config = ipsec_config_create();
		ip_cleanup_info[IP_CLEANUP_IPSEC_CONFIG_CREATE]++;

		g_ipsec_tunnel_config = ipsec_tunnel_config;
		APP_DEBUG("IPSEC tunnel configuration Created");
		markpoint(30);
		if (initialize_ip_stack(&stack)) {
			APP_ERROR("Error Initializing IP Stack");
			return -ENOMEM;
		}
		ip_cleanup_info[IP_CLEANUP_INIT_IP_STACK]++;

		markpoint(29);
		create_iface_nodes(iface_nodes);
		create_local_nodes(local_nodes);
		markpoint(28);
		if (create_devices(&stack, iface_nodes)) {
			APP_ERROR("Unable to Create Devices");
			return -ENOMEM;
		}
		ip_cleanup_info[IP_CLEANUP_DEVICE_CREATE]++;

		markpoint(27);
		populate_arp_cache(&stack, local_nodes);
		ip_cleanup_info[IP_CLEANUP_ARP_CACHE_CREATE]++;

		APP_INFO
		    ("ARP Cache Populated, Stack pointer is %x and its size = %x",
		     &stack, sizeof(stack));
		markpoint(26);
		initialize_ipsec_stack(&ipsec_stack);
		ip_cleanup_info[IP_CLEANUP_INIT_IPSEC_STACK]++;

		markpoint(24);
		APP_INFO("IPsec Forward test with Control Path");
		markpoint(23);
		if (standby == 0) {
			if (init_split_key_fqs()) {
				APP_ERROR("Unable to initialize Split key FQs");
				return -EINVAL;
			}
		}

		shutdown_doorbell_flag = 0;
		ipsec_app_state = RUNNING;
	}
#ifndef INFINITE_IPSECFWD
	if (0 != timer_framework_start()) {
		APP_ERROR("timer_framework_start Failed");
		return -1;
	}
#endif
	APP_INFO("Frames to be recv on channel map: 0x%x", recv_channel_map);
	qman_static_dequeue_add(recv_channel_map |
				QM_SDQCR_CHANNELS_POOL(local_channel));

	markpoint(3);
conf_mode_loop:
	if (curr_lwe_id == g_boot_lwe_core) {
		APP_INFO("In configuration mode, waiting for CP command");
		while (1) {
			if (sa_info->state == LWE_CTRL_CMD_STATE_BUSY) {
				ipsecfwd_process_req_from_cp(sa_info);
				if (sa_info->msg_type == LWE_CTRL_CMD_TYPE_GO) {
					sa_info->state =
					    LWE_CTRL_CMD_STATE_IDLE;
#ifdef IPSECFWD_HYBRID_GENERATOR
					if (unlikely(0 !=
						     dpa_eth_enable_ports
						     (true))) {
#else
					if (unlikely(0 !=
						     dpa_eth_enable_ports
						     (false))) {
#endif
						APP_ERROR
						    ("dpa_eth_enable_ports Failed");
						return -EINVAL;
					}
#ifdef DEBUG_PROFILE
				configure_dcsr(true, true, PROF_DDR_CPC);
#endif

					break;
				}
				sa_info->state = LWE_CTRL_CMD_STATE_IDLE;
			}
		}
	}
#ifdef PME_SWITCH
	else {
		do {
			qman_poll();
			if (g_all_context_initialized)
				break;
		} while (1);
	}
#endif
	barrier_sync(&ipsecfwd_barrier);
#ifdef DEBUG_PROFILE
	core_pm_configure(events, MAX_PM_COUNTERS);
#endif
	if (curr_lwe_id == g_boot_lwe_core) {
		if (flag_finite_reset)
			flag_finite_reset = 0;
		ipsec_app_state = RUNNING;
	}
	ipsecfwd_performance_enter();
	markpoint(4);

	/* If running from automation scripts, checkpoint should be taken here */
	MAGIC_BREAKPOINT;

	total_pkts_processed =
	    (uint32_t *) & (ipsec_stack.ip_stack->ip_stats->ip_out_forward);
	atb_start = get_atb();
	if ((curr_lwe_id == g_boot_lwe_core)
	    || (curr_lwe_id == LWE_BOOT_CORE)) {
		while (1) {
			markpoint(5);
#ifdef HOT_STANDBY_ENABLE
			if (standby == 0) {
				if (curr_lwe_id == g_boot_lwe_core) {
					int rx_count = 0;
					int tx_count = 0;
					char buf[20];
					int ret;
					ret = fh_byte_channel_poll(*bc_id,
								   &rx_count,
								   &tx_count);
					if (0 == ret && rx_count != 0) {
						int retval;
						retval =
						    fh_byte_channel_receive
						    (*bc_id, &rx_count, buf);
						if (retval == 0) {
							APP_INFO
							    ("Message from Standby"
							     ": %s", buf);
							strcpy(buf, "ok");
							tx_count =
							    strlen(buf) + 1;
							fh_byte_channel_send
							    (*bc_id, &tx_count,
							     buf);
							++primary_hb;
						} else {
							APP_ERROR
							    ("fh_byte_channel_"
							     "receive failed");
						}
					}
				}
				if (primary_hb >= PRIMARY_ALIVE_COUNT) {
					APP_INFO
					    ("Stopping one core in Primary");
					while (1) ;
				}
			}
#endif
			qman_poll();
			markpoint(6);
			if (likely(dqrr_enter == true)) {
				dqrr_enter = false;
			} else {
				if (unlikely(infinite_lwe == false)) {
					check_exit_criteria();
					if (unlikely(test_completed == true))
						break;
#ifndef INFINITE_IPSECFWD
				} else {
					while (unlikely(SWI_ID_INV !=
							(swi_id =
							 swi_activate
							 (&p_data)))) {
						switch (swi_id) {
						case SWI_ID_TIMER_EXPIRY:
							proc_timer_expiry((uint32_t)
									  p_data);
							break;
							/*Shutdown SWI is received only on bootcore */
						case SWI_ID_SHUTDOWN:
							if (curr_lwe_id !=
							    g_boot_lwe_core) {
								shutdown_doorbell_flag
								    = 1;
							} else
								ip_process_shutdown
								    ();
							break;
						default:
							break;
						}
					}

					if (ipsec_app_state ==
					    SHUTDOWN_COMPLETE) {
						num_poll_before_shutdown++;
						if (IP_DQRR_MAXFILL ==
						    num_poll_before_shutdown)
							break;
					}

					if (ipsec_app_state ==
					    RESET_COMPLETE) {
						APP_INFO
						    ("Reset process done for infinite");
						goto conf_mode_loop;
					}

					if (curr_lwe_id == g_boot_lwe_core) {
						if (unlikely(shutdown_doorbell_flag == 1)) {
							APP_INFO
							    ("Shutdown Req came in");
							shutdown_doorbell_flag =
							    0;
							ip_process_shutdown();
						}
						if ((sa_info->state !=
						     LWE_CTRL_CMD_STATE_IDLE)
						    && (sa_info->msg_type ==
							LWE_CTRL_CMD_TYPE_RESET))
						{
							APP_INFO
							    ("Reset Req came in");
							sa_info->result =
							    LWE_CTRL_RSLT_SUCCESSFULL;
							sa_info->state =
							    LWE_CTRL_CMD_STATE_IDLE;
							if (infinite_lwe ==
							    false) {
								flag_finite_reset
								    = 1;
								APP_INFO
								    ("Reset finite mode");
								pkt_processed_till_now
								    =
								    *total_pkts_processed;
								break;
							} else {
								APP_INFO
								    ("Reset infinitemode");
								ip_process_reset
								    ();
							}
						} else if ((sa_info->state !=
							    LWE_CTRL_CMD_STATE_IDLE)
							   && (sa_info->
							       msg_type ==
							       LWE_CTRL_CMD_TYPE_SASTATS
							       ||
							       LWE_CTRL_CMD_TYPE_IPFWD_SYS_STATS
							       ||
							       LWE_CTRL_CMD_TYPE_ROUTE_STATS
							       ||
							       LWE_CTRL_CMD_TYPE_IFSTATS
							       ||
							       LWE_CTRL_CMD_TYPE_IFLIST))
						{
							ipsecfwd_process_req_from_cp
							    (sa_info);
							sa_info->state =
							    LWE_CTRL_CMD_STATE_IDLE;
						}
					}
					if ((curr_lwe_id == LWE_BOOT_CORE)
					    && flag_finite_reset)
						break;
#endif
				}
				markpoint(4);
				underruns++;
			}

		}
	} else {
		while (1) {
			markpoint(5);
#ifdef HOT_STANDBY_ENABLE
			if (standby == 0 && primary_hb >= PRIMARY_ALIVE_COUNT) {
				APP_INFO("Stopping one of secondary core in"
					 " Primary part");
				while (1) ;
			}
#endif
			qman_poll();
			markpoint(6);
			if (likely(dqrr_enter == true)) {
				dqrr_enter = false;
#ifndef INFINITE_IPSECFWD
			} else {
				markpoint(4);
				underruns++;
				if (unlikely(infinite_lwe == false)) {
					check_exit_criteria();
					if (unlikely(test_completed == true))
						break;
				} else {
					if (unlikely(flag_finite_reset))
						break;

					if (unlikely(ipsec_app_state ==
					    SHUTDOWN_COMPLETE)) {
						num_poll_before_shutdown++;
						if (IP_DQRR_MAXFILL ==
						    num_poll_before_shutdown)
							break;
					}
					if (unlikely(ipsec_app_state == RESET_COMPLETE)) {
						APP_INFO
						    ("Reset process done for infinite");
						goto conf_mode_loop;
					}
				}
#endif
			}
		}
	}

	total_cycles = get_atb() - atb_start;
#ifdef DEBUG_PROFILE
	APP_INFO("Total cycles : %lld", total_cycles);
	APP_INFO("[Core %d]: Total Underruns = %d, \
		Frames Processed = %d", curr_lwe_id, (underruns - 1), local_frames);
#endif
	if (flag_finite_reset) {
		if (curr_lwe_id == g_boot_lwe_core) {
			ip_process_reset();
			while (1) {
				qman_poll();
				while (unlikely(SWI_ID_INV !=
						(swi_id =
						 swi_activate(&p_data)))) {
					switch (swi_id) {
					case SWI_ID_TIMER_EXPIRY:
						proc_timer_expiry((uint32_t)
								  p_data);
						break;
					default:
						break;
					}
				}
				if (ipsec_app_state == RESET_COMPLETE) {
					APP_INFO("Reset process done finite");
					break;
				}
			}
		} else {
			while (1) {
				qman_poll();
				if (ipsec_app_state == RESET_COMPLETE)
					break;
			}
		}
	}
	spin_lock(&ipsecfwd_lock);
	grand_total_cycles += total_cycles;
	spin_unlock(&ipsecfwd_lock);
	barrier_sync(&ipsecfwd_barrier);
	if (curr_lwe_id == LWE_BOOT_CORE) {
		if (flag_finite_reset)
			APP_INFO("%d packets processed in %lld total cycles, \
				Cycles/Frame = %lld", pkt_processed_till_now, grand_total_cycles, grand_total_cycles / pkt_processed_till_now);
		else
			APP_INFO("%d packets processed in %lld total cycles, \
			Cycles/Frame = %lld", *total_pkts_processed, grand_total_cycles, grand_total_cycles / (*total_pkts_processed));
#ifdef PME_SWITCH
		for (count = 0; count < MAX_SPA_PME_CNTX; count++) {
			if (ipsa_context_info[count].valid) {
				pme_ctx_stats =
				    ipsa_context_info[count].pme_ctx_stats;
				APP_INFO
				    ("Flow id %d : Total scanned pkts %d \t"
				     " Number of matches = %d ", count,
				     pme_ctx_stats->post_scan_frames,
				     pme_ctx_stats->match_count);
			}
		}
		APP_INFO("Total packets sent to PME: %d", *pre_pme_pkts);
		APP_INFO("Total packets processed by PME: %d", *post_pme_pkts);
		APP_INFO("Total matches found: %d", *pme_pattern_pkts);
#endif
		APP_INFO("Test Passed");
	}
	barrier_sync(&ipsecfwd_barrier);
	if (flag_finite_reset) {
		if (curr_lwe_id == g_boot_lwe_core) {
			grand_total_cycles = 0;
			test_completed = false;
		}
		goto conf_mode_loop;
	}
	markpoint(31);
	if (test_completed == true) {
		if (curr_lwe_id == g_boot_lwe_core)
			ip_timer_start();
		if ((curr_lwe_id == g_boot_lwe_core)
		    || (curr_lwe_id == LWE_BOOT_CORE)) {
			APP_INFO("Waiting for reset or shutdown command");
			while (1) {
				qman_poll();
				while (unlikely(SWI_ID_INV !=
						(swi_id =
						 swi_activate(&p_data)))) {
					switch (swi_id) {
					case SWI_ID_TIMER_EXPIRY:
						proc_timer_expiry((uint32_t)
								  p_data);
						break;
						/*Shutdown SWI is received only on bootcore */
					case SWI_ID_SHUTDOWN:
						if (curr_lwe_id !=
						    g_boot_lwe_core) {
							shutdown_doorbell_flag =
							    1;
						} else {
							wait_flag = 0;
							shutdn_in_proc = 1;
							if (0 !=
							    stop_timer
							    (wait_timerid))
								APP_ERROR
								    ("Stoptimerfail");
							else
								APP_INFO
								    ("Stopped timer");
							ip_process_shutdown();
						}
						break;
					default:
						break;
					}
				}
				if (curr_lwe_id == g_boot_lwe_core) {
					if (shutdown_doorbell_flag == 1) {
						APP_INFO
						    ("Shutdown Req came in");
						shutdown_doorbell_flag = 0;
						wait_flag = 0;
						shutdn_in_proc = 1;
						if (0 !=
						    stop_timer(wait_timerid))
							APP_ERROR
							    ("Stop timer fail");
						else
							APP_INFO
							    ("Stopped timer");
						ip_process_shutdown();
					}
					if (wait_flag && !shutdn_in_proc
					    && !reset_in_proc) {
						wait_flag = 0;
						ip_process_shutdown();
					}

					if ((sa_info->state !=
					     LWE_CTRL_CMD_STATE_IDLE)
					    && (sa_info->msg_type ==
						LWE_CTRL_CMD_TYPE_RESET)) {
						APP_INFO("Reset Req came in");
						sa_info->result =
						    LWE_CTRL_RSLT_SUCCESSFULL;
						sa_info->state =
						    LWE_CTRL_CMD_STATE_IDLE;
						reset_in_proc = 1;
						wait_flag = 0;
						if (0 !=
						    stop_timer(wait_timerid))
							APP_ERROR
							    ("Stop timer fail");
						else
							APP_INFO
							    ("Stopped timer");
						if (test_completed == true)
							test_completed = false;
						ip_process_reset();
					}
				}
				if (ipsec_app_state == SHUTDOWN_COMPLETE) {
					APP_INFO("Shutdown process at exit");
					break;
				}
				if (ipsec_app_state == RESET_COMPLETE) {
					APP_INFO
					    ("Reset done going to conf mode");
					grand_total_cycles = 0;
					reset_in_proc = 0;
					goto conf_mode_loop;
				}
			}
		} else {
			while (1) {
				qman_poll();
				if (ipsec_app_state == RESET_COMPLETE)
					goto conf_mode_loop;
				if (ipsec_app_state == SHUTDOWN_COMPLETE)
					break;
			}
		}
	}
	if (curr_lwe_id != g_boot_lwe_core) {
		while (done != true)
			qman_poll();
	} else {
		ip_wait_for_shutdown_complete();
		done = true;
	}

	APP_INFO("Shutdown completed on core: %u", lwe_instance);
	return 0;
}
