/**
 \ipsecfwd_handler.c
 \brief IPSec Forwarding Application Handlers
 */
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compat.h"
#include "ipsecfwd.h"
#include <dpa_alloc.h>
#include <dpa/dpa_eth.h>
#include <fsl_qman.h>
#include <timebase.h>
#include "barrier.h"
#include <swi.h>
#include "pme/ipsa_pme.h"
#include "ip/ip_forward.h"
#include "ip/ip_local.h"
#include "arp/arp.h"
#include "ipsec/ipsec_sec.h"

extern __PERCPU uint32_t rx_errors;

void ip_fq_state_chg(struct qman_portal *qm, struct qman_fq *fq,
			const struct qm_mr_entry *msg);
/**
 \brief Enqueue Rejection Notification Handler for SEC Tx FQ
 \param[in] msg Message Ring entry to be processed
 \param[out] NULL
 */
static void ipsecfwd_ern_handler(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	APP_ERROR("%s: RC = %x, seqnum = %x", __func__,\
			msg->ern.rc, msg->ern.seqnum);
	lwe_free_fd(buff_allocator, &msg->ern.fd);
	return;
}

/**
 \brief Enqueue Rejection Notification Handler for FMAN Rx FQs
 \param[in] msg Message Ring entry to be processed
 \param[out] NULL
 */
static void ipfwd_dc_ern_handler(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	APP_ERROR("%s: RC = %x, FQID = %x", __func__,\
		 msg->dcern.rc, msg->dcern.fqid);
	lwe_free_fd(buff_allocator, &msg->dcern.fd);
	return;
}

/**
 \brief Enqueue Rejection Notification Handler for FMAN Tx FQs
 \param[in] msg Message Ring entry to be processed
 \param[out] NULL
 */
static void ipfwd_ern_handler(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	APP_ERROR("%s: RC = %x, seqnum = %x", __func__,\
		 msg->ern.rc, msg->ern.seqnum);
	lwe_free_fd(buff_allocator, &msg->ern.fd);
	return;
}

/**
 \brief Error Frame Dequeue Handler for IPSEC FWD Application
 \param[in] dqrr Dequeue Receive Ring entry to be processed
 \param[out] NULL
 */
static enum qman_cb_dqrr_result dqrr_handler_tx_err(struct qman_portal *qm,
						   struct qman_fq *fq,
						   const struct qm_dqrr_entry
						   *dqrr)
{
	APP_ERROR("In %s: Status Returned is %x on FQID %x", __func__, \
			dqrr->fd.status, dqrr->fqid);

	lwe_free_fd(buff_allocator, (struct qm_fd *)(&dqrr->fd));
	return qman_cb_dqrr_consume;
}

/**
 \brief Error Frame Dequeue Handler for IPSEC FWD Application
 \param[in] dqrr Dequeue Receive Ring entry to be processed
 \param[out] NULL
 */
static enum qman_cb_dqrr_result dqrr_handler_tx_confirm(struct qman_portal *qm,
						struct qman_fq *fq,
						const struct qm_dqrr_entry
						*dqrr)
{
	APP_INFO("In %s: Status Returned is %x on FQID %x", __func__, \
			dqrr->fd.status, dqrr->fqid);

	lwe_free_fd(buff_allocator, (struct qm_fd *)(&dqrr->fd));
	return qman_cb_dqrr_consume;
}

/**
 \brief Error Frame Dequeue Handler for IPSEC FWD Application
 \param[in] dqrr Dequeue Receive Ring entry to be processed
 \param[out] NULL
 */

static enum qman_cb_dqrr_result dqrr_entry_handler_err(struct qman_portal *qm,
						   struct qman_fq *fq,
						   const struct qm_dqrr_entry
						   *dqrr)
{
	APP_DEBUG("In %s: Status Returned is %x, on FQID = %x", __func__, \
		dqrr->fd.status, dqrr->fqid);
	lwe_free_fd(buff_allocator, (struct qm_fd *)(&dqrr->fd));
	rx_errors++;

	return qman_cb_dqrr_consume;
}

/**
 \brief Take Snapshot of all perfmon counters every 2M frames
 */
void core_perfmon_snapshot(void)
{
	local_frames++;
	if (get_lwe_id() != g_boot_lwe_core)
		return;

	if (unlikely(!(local_frames & 0x1FFFFF)))
		system_snapshot();
}
/**
 \brief Dequeue Handler for IPSEC FWD Application
 \param[in] struct qm_dqrr_entry * Dequeue Receive Ring entry to be processed
 \param[out] NULL
 */
static enum qman_cb_dqrr_result dqrr_entry_handler_pcd(struct qman_portal *qm,
						   struct qman_fq *fq,
						   const struct qm_dqrr_entry
						   *dqrr)
{
	struct annotations_t *notes;
	struct ip_fq_context_t *ip_fq_ctxt = (struct ip_fq_context_t *)fq;
	struct fq_context_t *context =
	    (struct fq_context_t *)(ip_fq_ctxt->ip_ctxt);
	struct sg_entry_t *sg;
	uint8_t *data;

	if (likely(dqrr_enter == false))
		dqrr_enter = true;

	/** Following qman_fq is my context */
	switch (dqrr->fd.format) {
	case qm_fd_contig:
		notes = (void *)APP_PTOV(dqrr->fd.addr_lo);
		data = (uint8_t *)notes + dqrr->fd.offset;
#ifdef DEBUG_PROFILE
		mtspr(SPR_DEVENT, DVT0_DETECTED);
		core_perfmon_snapshot();
#endif
		break;
	case qm_fd_compound:
		sg = (struct sg_entry_t *)
		   APP_PTOV(dqrr->fd.addr_lo);
		notes = APP_PTOV(sg->addr_lo);
		data = (uint8_t *)notes + sg->offset;

#ifdef LWE_APPS_DEBUG
		if (!notes) {
			APP_ERROR("Got Null Output Buffer from SEC");
			return qman_cb_dqrr_consume;
		}
#endif
		break;
	case qm_fd_sg:
		sg = (struct sg_entry_t *)
		   APP_PTOV(dqrr->fd.addr_lo + dqrr->fd.offset);
		notes = APP_PTOV(dqrr->fd.addr_lo);
		data = (uint8_t *)APP_PTOV(sg->addr_lo + sg->offset);
		break;
	default:
		APP_ERROR("Unsupported format packet came");
		goto done;
	}
	notes->fd = (struct qm_fd *)(&(dqrr->fd));

	context->handler(context, notes, data);
done:

	return qman_cb_dqrr_consume;
}

/**
 \brief Dequeue Handler for IPSEC FWD Application default FQ
 \param[in] struct qm_dqrr_entry * Dequeue Receive Ring entry to be processed
 \param[out] NULL
 */
static enum qman_cb_dqrr_result dqrr_entry_handler(struct qman_portal *qm,
						   struct qman_fq *fq,
						   const struct qm_dqrr_entry
						   *dqrr)
{
	struct annotations_t *notes;
	struct ip_fq_context_t *ip_fq_ctxt = (struct ip_fq_context_t *)fq;
	struct fq_context_t *context =
	    (struct fq_context_t *)(ip_fq_ctxt->ip_ctxt);
	struct sg_entry_t *sg;
	uint8_t *data;
	struct ethernet_header_t *eth_hdr;

	if (likely(dqrr_enter == false))
		dqrr_enter = true;

	/** Following qman_fq is my context */
	switch (dqrr->fd.format) {
	case qm_fd_contig:
		notes = (void *)APP_PTOV(dqrr->fd.addr_lo);
		data = (uint8_t *)notes + dqrr->fd.offset;
		break;
	case qm_fd_compound:
		sg = (struct sg_entry_t *)
		   APP_PTOV(dqrr->fd.addr_lo);
		notes = APP_PTOV(sg->addr_lo);
		data = (uint8_t *)notes + sg->offset;

#ifdef LWE_APPS_DEBUG
		if (!notes) {
			APP_ERROR("Got Null Output Buffer from SEC");
			return qman_cb_dqrr_consume;
		}
#endif
		break;
	case qm_fd_sg:
		sg = (struct sg_entry_t *)
		   APP_PTOV(dqrr->fd.addr_lo + dqrr->fd.offset);
		notes = APP_PTOV(dqrr->fd.addr_lo);
		data = (uint8_t *)APP_PTOV(sg->addr_lo + sg->offset);
		break;
	default:
		APP_ERROR("Unsupported format packet came");
		goto done;
	}
	notes->fd = (struct qm_fd *)(&(dqrr->fd));

	eth_hdr =  (struct ethernet_header_t *) data;
	if (ETH_P_ARP == eth_hdr->proto)
		arp_handler(notes, data);
	else
		context->handler(context, notes, data);
done:

	return qman_cb_dqrr_consume;
}

/**
 \brief Handler for Split Key generated by SEC Block
 */
static enum qman_cb_dqrr_result dqrr_split_key_handler(struct qman_portal *qm,
					struct qman_fq *fq,
					const struct qm_dqrr_entry *dqrr)
{
	g_key_split_flag = 1;
	return qman_cb_dqrr_consume;
}

/**
 \brief Dummy function for unsupported handlers
 */
static void my_cb_notimplemented(struct qman_portal *qm, struct qman_fq *fq,
				 const struct qm_mr_entry *msg)
{
	APP_INFO("In %s", __func__);
}

/**
 \brief Handler for SEC Generated Split Key
 */
const struct qman_fq_cb ipfwd_split_key_cb = {
	.dqrr = dqrr_split_key_handler,
	.ern = my_cb_notimplemented,
	.dc_ern = my_cb_notimplemented,
	.fqs = my_cb_notimplemented
};

/**
 \brief RX Error Callback Handler
 */
const struct qman_fq_cb ipfwd_rx_cb_err = {
	.dqrr = dqrr_entry_handler_err,
	.ern = my_cb_notimplemented,
	.dc_ern = ipfwd_dc_ern_handler,
	.fqs = ip_fq_state_chg
};

/**
 \brief TX Error Callback Handler
 */
const struct qman_fq_cb ipfwd_tx_cb_err = {
	.dqrr = dqrr_handler_tx_err,
	.ern = my_cb_notimplemented,
	.dc_ern = my_cb_notimplemented,
	.fqs = ip_fq_state_chg
};

/**
 \brief TX confirm Callback Handler
 */
const struct qman_fq_cb ipfwd_tx_cb_confirm = {
	.dqrr = dqrr_handler_tx_confirm,
	.ern = my_cb_notimplemented,
	.dc_ern = my_cb_notimplemented,
	.fqs = ip_fq_state_chg
};

/**
 \brief IPFwd PCD FQ Callback Handler
 */
const struct qman_fq_cb ipfwd_rx_cb_pcd = {
	.dqrr = dqrr_entry_handler_pcd,
	.ern = my_cb_notimplemented,
	.dc_ern = ipfwd_dc_ern_handler,
	.fqs = ip_fq_state_chg
};

/**
 \brief IPFwd Callback Handler
 */
const struct qman_fq_cb ipfwd_rx_cb = {
	.dqrr = dqrr_entry_handler,
	.ern = my_cb_notimplemented,
	.dc_ern = ipfwd_dc_ern_handler,
	.fqs = ip_fq_state_chg
};

/**
 \brief IPSec Transmit FQ Callback Handler
 */
const struct qman_fq_cb ipsecfwd_tx_cb = {
	.dqrr = my_cb_notimplemented,
	.ern = ipsecfwd_ern_handler,
	.dc_ern = my_cb_notimplemented,
	.fqs = ip_fq_state_chg
};

/**
 \brief IPFwd Transmit FQ Callback Handler
 */
const struct qman_fq_cb ipfwd_tx_cb = {
	.dqrr = my_cb_notimplemented,
	.ern = ipfwd_ern_handler,
	.dc_ern = my_cb_notimplemented,
	.fqs = ip_fq_state_chg
};
