/**
 \ipsecfwd_init.c
 \brief IPSec Forwarding Application Init
 */
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compat.h"
#include "ipsecfwd.h"
#include <dpa_alloc.h>
#include <dpa/dpa_eth.h>
#include <fsl_qman.h>
#include <timebase.h>
#include "barrier.h"
#include <swi.h>
#include "pme/ipsa_pme.h"
#include "ip/ip_forward.h"
#include "ip/ip_local.h"
#include "arp/arp.h"
#include "ipsec/ipsec_sec.h"
#include "pme/common_pme.h"

__PERCPU uint64_t atb_start;
__PERCPU uint64_t total_cycles;
static struct qman_fq *g_fq_from_sec[IPSEC_TUNNEL_ENTRIES];
static struct qman_fq *g_fq_to_sec[IPSEC_TUNNEL_ENTRIES];

void ip_shutdown_complete(uint32_t timer_id, void *p_data)
{
	if (0 == num_fq_close_pending) {
		APP_INFO("Moving application state to SHUTDOWN_COMPLETE");
		ipsec_app_state = SHUTDOWN_COMPLETE;
		return;
	}

	APP_WARN("Num Rx FQ close pending: %d", num_fq_close_pending);
	APP_WARN("Waiting for some more time for FQs to retire");

	timer_id = reset_timer(timer_id, 0);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Resetting shutdown complete timer failed");
}

void ipsec_init(void)
{
	buff_allocator_core = NULL;
	buff_allocator = NULL;
	ipsec_buff_allocator = NULL;
	atb_start = 0;
	total_cycles = 0;
	local_frames = 0;
	dqrr_enter = false;
}

/**
 \brief Creates and initialized the FQs related to a tunnel
 \param[out] entry IpSec tunnel entry
 \param[out] ctxt IpSec Context
 \param[out] ctxt_a Context A for the FQ
 \param[out] sec_fq FQ Id for the Queue towards SEC
 \return Integer status
 */
int32_t init_ipsec_fqs(struct ipsec_tunnel_t *entry,
			struct ipsec_context_t *ctxt, void *ctxt_a,
			uint32_t sec_fq)
{
	uint32_t flags;
	struct qman_fq *fq;
	struct ip_fq_context_t *ip_fq_ctxt;
	struct qm_mcc_initfq opts;
	uint32_t ctx_a_excl;
	uint32_t ctx_a_len;

	flags = QMAN_FQ_FLAG_NO_ENQUEUE | QMAN_FQ_FLAG_LOCKED;

	fq = (struct qman_fq *)memalign(CACHE_LINE_SIZE,
					sizeof(struct qman_fq) +
					sizeof(struct ipsec_context_t *));
	if (unlikely(NULL == fq)) {
		APP_ERROR("malloc failed in create_fqs for FQ ID: %lu",
			  sec_fq + 1);
		return -ENOMEM;
	}
	/* Rx Callback Handler */
	fq->cb = ipfwd_rx_cb_pcd;
	g_fq_from_sec[g_sec_fq_count] = fq;

	if (unlikely(0 != qman_create_fq(sec_fq + 1, flags, fq))) {
		APP_ERROR("qman_create_fq failed for FQ ID: %lu", sec_fq + 1);
		return -1;
	}

	flags = QMAN_INITFQ_FLAG_SCHED;
	opts.we_mask =
	    QM_INITFQ_WE_DESTWQ | QM_INITFQ_WE_CONTEXTA | QM_INITFQ_WE_FQCTRL;
	opts.fqd.fq_ctrl = QM_FQCTRL_CTXASTASHING | QM_FQCTRL_HOLDACTIVE;

	ctx_a_excl = (QM_STASHING_EXCL_DATA | QM_STASHING_EXCL_CTX);

	ctx_a_len = (1 << 2) | 1;

	opts.fqd.context_a.hi = (ctx_a_excl << 24) | (ctx_a_len << 16);

	opts.fqd.dest.channel = (uint16_t)pool_channel_id;
	/* Post-CAAM WQ priority changed from 0 to 2 */
	opts.fqd.dest.wq = 2;

	if (unlikely(0 != qman_init_fq(fq, flags, &opts))) {
		APP_ERROR("Unable to initialize ingress FQ for SEC4.0");
		return -1;
	}
	ip_fq_ctxt = (struct ip_fq_context_t *)fq;
	ip_fq_ctxt->ip_ctxt = (struct ip_context_t *)ctxt;

	flags = QMAN_FQ_FLAG_LOCKED | QMAN_FQ_FLAG_TO_DCPORTAL;

	fq = (struct qman_fq *)memalign(CACHE_LINE_SIZE,
					sizeof(struct qman_fq));
	if (unlikely(NULL == fq)) {
		APP_ERROR("malloc failed in create_fqs for FQ ID");
		return -1;
	}

	/* Tx Callback Handlers */
	fq->cb = ipsecfwd_tx_cb;
	if (unlikely(0 != qman_create_fq(sec_fq, flags, fq))) {
		APP_ERROR("qman_create_fq failed for FQ ID");
		return -1;
	}

	flags = QMAN_INITFQ_FLAG_SCHED;
	opts.we_mask = QM_INITFQ_WE_DESTWQ | QM_INITFQ_WE_CONTEXTA |
	    QM_INITFQ_WE_CONTEXTB;
	opts.fqd.context_a.hi = 0;
	opts.fqd.context_a.lo = APP_VTOP(ctxt_a);
	opts.fqd.context_b = sec_fq + 1;
	opts.fqd.dest.channel = qm_channel_caam;
	opts.fqd.dest.wq = 0;

	if (unlikely(0 != qman_init_fq(fq, flags, &opts))) {
		APP_ERROR("Unable to Init CAAM Egress FQ");
		return -EINVAL;
	}
	entry->qm_fq_to_sec = fq;
	g_fq_to_sec[g_sec_fq_count] = fq;
	g_sec_fq_count++;

	return 0;
}

static int32_t retire_sec_rx_fqs(void)
{
	uint32_t fq_cnt;
	uint32_t flags;

	for (fq_cnt = 0; fq_cnt < g_sec_fq_count; fq_cnt++) {
		APP_DEBUG("SEC FQ cnt %d", fq_cnt);
		if (0 > qman_retire_fq(g_fq_from_sec[fq_cnt], &flags)) {
			APP_ERROR("qman_retire_fq failed for fq %d",
				  g_fq_from_sec[fq_cnt]->fqid);
			return -1;
		}
	}
	return fq_cnt;
}

static int32_t retire_sec_tx_fqs(void)
{
	uint32_t fq_cnt;
	uint32_t flags;

	for (fq_cnt = 0; fq_cnt < g_sec_fq_count; fq_cnt++) {
		APP_DEBUG("SEC FQ cnt %d", fq_cnt);
		if (0 > qman_retire_fq(g_fq_to_sec[fq_cnt], &flags)) {
			APP_ERROR("qman_retire_fq failed for fq %d",
				  g_fq_to_sec[fq_cnt]->fqid);
			return -1;
		}
	}
	return fq_cnt;
}

void ip_shutdown_sec_rx_complete(uint32_t timer_id, void *p_data)
{
	if (num_fq_close_pending) {
		APP_WARN("Num SEC Rx FQ close pending: %d",
			 num_fq_close_pending);
		APP_WARN("Waiting for some more time for FQs to retire");

		timer_id = reset_timer(timer_id, 0);
		if (INV_TIMER_ID == timer_id)
			APP_ERROR("Reset shutdown SEC Rx timer failed");
		return;
	}
	APP_INFO("Moving application state to SHUTDOWN_RX_COMPLETE");
	ipsec_app_state = SHUTDOWN_SEC_RX_COMPLETE;
	/* At this stage, all the traffic which was received has been
	 * processed and we can now move forward to retire Tx FQs
	 */
	num_fq_close_pending = dpa_eth_tx_fq_retire();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring Tx FQs failed");
		return;
	}

	APP_INFO("Number of FMAN Tx FQs close pending: %d",
		 num_fq_close_pending);

	/*
	 * If there are no Tx FQs which are pending to be closed, call
	 * shutdown complete
	 */
	if (!num_fq_close_pending) {
		ip_shutdown_complete(INV_TIMER_ID, NULL);
		return;
	}

	/*Start a single shot timer to wait for all Tx FQs to get retired */
	timer_id = start_timer(IP_SHUTDOWN_TIMER_MS, false,
			       NULL, SWI_PRI_HIGH, ip_shutdown_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown complete timer failed");
	else
		APP_INFO("Shutdown complete timer started...");
}

void ip_shutdown_sec_tx_complete(uint32_t timer_id, void *p_data)
{
	if (num_fq_close_pending) {
		APP_WARN("Num SEC Tx FQ close pending: %d",
			 num_fq_close_pending);
		APP_WARN("Waiting for some more time for FQs to retire");

		timer_id = reset_timer(timer_id, 0);
		if (INV_TIMER_ID == timer_id)
			APP_ERROR("Reset shutdown SEC Tx timer failed");
		return;
	}
	APP_INFO("Moving application state to SHUTDOWN_SEC_TX_COMPLETE");
	ipsec_app_state = SHUTDOWN_SEC_TX_COMPLETE;
	/* At this stage, all the traffic which was received has been
	 * processed and we can now move forward to retire SEC Rx FQs
	 */
	num_fq_close_pending = retire_sec_rx_fqs();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring SEC Rx FQs failed");
		return;
	}

	APP_INFO("Number of SEC Rx FQs close pending: %d",
		 num_fq_close_pending);

	/*
	 * If there are no SEC Rx FQs which are pending to be closed, directly
	 * move to the next state
	 */
	if (!num_fq_close_pending) {
		ip_shutdown_sec_rx_complete(INV_TIMER_ID, NULL);
		return;
	}

	/*Start a single shot timer to wait for all SEC Rx FQs to get retired */
	timer_id = start_timer(IP_SHUTDOWN_TIMER_MS, false,
			       NULL, SWI_PRI_HIGH,
			       ip_shutdown_sec_rx_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting RX shutdown timer failed");
	else
		APP_INFO("Shutdown RX timer started...");
}

void ip_shutdown_sec_tx(uint32_t timer_id, void *p_data)
{
	APP_INFO("Processing shutdown SEC TX command ..");
	if (num_fq_close_pending) {
		APP_WARN("Num Rx FQ close pending: %d", num_fq_close_pending);
		APP_WARN("Waiting for some more time for FQs to retire");

		timer_id = reset_timer(timer_id, 0);
		if (INV_TIMER_ID == timer_id)
			APP_ERROR("Reset shutdown Rx complete timer failed");
		return;
	}
	APP_INFO("Moving application state to SHUTDOWN_TX_COMPLETE");
	ipsec_app_state = SHUTDOWN_TX_COMPLETE;
	/*Put all SEC Tx frame queues in retired state */
	num_fq_close_pending = retire_sec_tx_fqs();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring SEC Tx FQs failed");
		return;
	}

	APP_INFO("Number of SEC Tx FQs close pending: %d",
		 num_fq_close_pending);

	/*If there is no SEC TX FQ close pending, directly move to
	 * the next state */
	if (!num_fq_close_pending) {
		ip_shutdown_sec_tx_complete(INV_TIMER_ID, 0);
		return;
	}

	/*Start a single shot timer to wait for all SEC Tx FQs to get retired */
	timer_id = start_timer(IP_SHUTDOWN_TIMER_MS, false,
			       NULL, SWI_PRI_HIGH,
			       ip_shutdown_sec_tx_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown SEC TX timer failed");
	else
		APP_INFO("Shutdown SEC TX timer started...");
}
