/**
 \file ipsecfwd_scan.c
 \brief Basic IPSec Forwarding Scan Application library
 */
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compat.h"
#include "ipsecfwd.h"
#include <dpa_alloc.h>
#include <dpa/dpa_eth.h>
#include <fsl_qman.h>
#include <timebase.h>
#include "barrier.h"
#include "ip/ip_forward.h"
#include "ip/ip_local.h"

uint32_t g_total_flow = 0;
struct rt_dest_t *g_flow_to_entry[MAX_SPA_PME_CNTX];
struct pme_stats_t *pme_stats;
uint8_t g_all_context_initialized = 0;

struct disable_token {
	struct pme_ctx_ctrl_token driver_token;
	int done;
};

static void disable_cb(struct pme_ctx *ctx, const struct qm_fd *fd,
		       struct pme_ctx_ctrl_token *__token)
{
	struct disable_token *token = (void *)__token;
	token->done = 1;
}

static int blocking_pme_disable(struct pme_ctx *ctx)
{
	int ret;
	struct disable_token *token = malloc(sizeof(*token));
	if (!token)
		return -ENOMEM;
	memset(token, 0, sizeof(*token));
	token->driver_token.cb = disable_cb;
	/* NB: the use of PME_CTX_OP_WAIT is prohibited in LWE applications. */
	ret = pme_ctx_disable(ctx, 0, &token->driver_token);
	if (ret <= 0)
		return ret;
	while (!token->done)
		qman_poll();
	free(token);
	return 0;
}

int32_t ipsa_cleanup_del_all_ctx(void)
{
	uint32_t flow_id;
	struct rt_dest_t *dest;
	struct spa_pme_ctx_stats_t *pme_ctx_stats;

	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if (ipsa_context_info[flow_id].valid) {
			dest = g_flow_to_entry[flow_id];
			pme_ctx_stats =
			    ipsa_context_info[flow_id].pme_ctx_stats;
			APP_INFO("PME DBG : valid ctx is %d", flow_id);
			/* Disable and then destroy the PME context */
			if (blocking_pme_disable(dest->pme_ctx)) {
				APP_INFO
				    ("ERROR:  PME context disable"
				     " failed for flow id %x", flow_id);
				return -1;
			}
			pme_ctx_finish(dest->pme_ctx);
			ipsa_context_info[flow_id].valid = 0;
			free(dest->pme_ctx);
			stats_free(pme_ctx_stats);
			g_total_flow--;
			APP_INFO("PME DBG : After del ctx total flows %d",
				 g_total_flow);
		}
	}
	stats_free(pme_stats);
	return 0;
}

int32_t ipsa_create_compd_fd_outbuf(struct qm_fd *fd,
				    struct qm_fd *old_fd,
				    struct ip_header_t *ip_hdr, uint32_t size)
{
	struct sg_entry_t *sg;
	struct sg_entry_t *sg_inframe;
	struct bm_buffer bman_buf_out;
	struct bm_buffer bman_buf_sg;
	uint32_t *out_buf;

	if (unlikely
	    (dpa_allocator_get_buff(ipsec_buff_allocator,
				    size, &bman_buf_out) <= 0)) {
		APP_ERROR("%s: Couldn't return Buffer", __func__);
		return -ENOMEM;
	}

	if (unlikely
	    (dpa_allocator_get_buff(ipsec_buff_allocator,
				    2 * sizeof(struct sg_entry_t),
				    &bman_buf_sg) <= 0)) {
		APP_ERROR("%s: Couldn't return Buffer", __func__);
		if (0 >
		    dpa_allocator_free_buff(ipsec_buff_allocator,
					    &bman_buf_out)) {
			APP_ERROR(" dpa allocator free out buff fails");
			return -1;
		}
		return -ENOMEM;
	}
	sg = (struct sg_entry_t *)APP_PTOV(bman_buf_sg.lo);

	/* output buffer */
	sg->addr_hi = 0;
	sg->addr_lo = (uint32_t) bman_buf_out.lo;
	sg->length = dpa_partition_bpools[bman_buf_out.bpid].size;
	sg->offset = 0;
	sg->bpid = bman_buf_out.bpid;
	sg->extension = 0;
	sg->final = 0;
	sg->reserved_offset = 0;
	sg->reserved_zero = 0;
	sg->reserved_zero2 = 0;

	/* fill first byte of output buffer with a magic number */
	out_buf = (uint32_t *) APP_PTOV(sg->addr_lo);
	*out_buf = SPA_MAGIC_VALUE;

	/* input buffer */
	sg++;
	sg->addr_hi = 0;
	sg->addr_lo = old_fd->addr_lo;
	if (old_fd->format == qm_fd_sg) {
		sg->length = old_fd->length20;
		sg->offset = old_fd->offset;
		sg->extension = 1;
		sg_inframe = (struct sg_entry_t *)
		    APP_PTOV(sg->addr_lo + old_fd->offset);
		sg_inframe->length -=
		    (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
		sg_inframe->offset +=
		    (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
	} else {
		sg->length = ip_hdr->total_len - IP_HEADER_LENGTH_NO_OPTIONS;
		sg->offset =
		    old_fd->offset + ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS;
		sg->extension = 0;
	}
	sg->bpid = old_fd->bpid;
	sg->final = 1;
	sg->reserved_offset = 0;
	sg->reserved_zero = 0;
	sg->reserved_zero2 = 0;

	fd->addr_hi = 0;
	fd->addr_lo = bman_buf_sg.lo;
	fd->_format1 = qm_fd_compound;
	fd->cong_weight = 0;
	fd->bpid = bman_buf_sg.bpid;
	fd->cmd = 0;

	return 0;
}

/**
 \brief Creates compd fd without output buffer
 \param[inout] fd frame dsecriptor pointer
 \param[in] ip_hdr IP header pointer
 \param[in] size size of the output buffer
 \return Integer status
 */
int32_t ipsa_create_compd_fd(struct qm_fd *fd,
			     struct qm_fd *old_fd, struct ip_header_t *ip_hdr,
			     uint8_t bpid)
{
	struct sg_entry_t *sg, *sg_inframe;
	struct bm_buffer bman_buf_sg;

	if (unlikely
	    (dpa_allocator_get_buff(ipsec_buff_allocator,
				    2 * sizeof(struct sg_entry_t),
				    &bman_buf_sg) <= 0)) {
		APP_ERROR("%s: Couldn't return Buffer", __func__);
		return -ENOMEM;
	}

	sg = (struct sg_entry_t *)APP_PTOV(bman_buf_sg.lo);

	/* output buffer NULL */
	sg->addr_hi = 0;
	sg->addr_lo = 0;
	sg->length = 0;
	sg->offset = 0;
	sg->bpid = bpid;
	sg->final = 0;

	/* input buffer */
	sg++;
	sg->addr_hi = 0;
	sg->addr_lo = old_fd->addr_lo;
	if (old_fd->format == qm_fd_sg) {
		sg->length = old_fd->length20;
		sg->offset = old_fd->offset;
		sg->extension = 1;
		sg_inframe = (struct sg_entry_t *)
		    APP_PTOV(sg->addr_lo + old_fd->offset);
		sg_inframe->length -=
		    (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
		sg_inframe->offset +=
		    (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
	} else {
		sg->length = ip_hdr->total_len - IP_HEADER_LENGTH_NO_OPTIONS;
		sg->offset =
		    old_fd->offset + ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS;
		sg->extension = 0;
	}
	sg->bpid = old_fd->bpid;
	sg->final = 1;
	sg->reserved_offset = 0;
	sg->reserved_zero = 0;
	sg->reserved_zero2 = 0;

	fd->addr_hi = 0;
	fd->addr_lo = bman_buf_sg.lo;
	fd->_format1 = qm_fd_compound;
	fd->cong_weight = 0;
	fd->bpid = bman_buf_sg.bpid;
	fd->cmd = 0;

	return 0;
}

enum IP_STATUS ipsa_send_simple(struct ip_context_t *ctxt,
				struct annotations_t *notes,
				struct ip_header_t *ip_hdr_ptr)
{
	struct qm_fd *fd;
	struct qm_fd fd2;
	struct ip_header_t *ip_hdr = ip_hdr_ptr;
	struct sg_entry_t *sg;

	APP_DEBUG("PME DBG:pme_send_ simple entered");
	fd = notes->fd;
	memcpy(&fd2, fd, sizeof(struct qm_fd));
	fd = &fd2;

	if (fd->format == qm_fd_sg) {
		sg = (struct sg_entry_t *)
		    APP_PTOV(fd->addr_lo + fd->offset);
		sg->length -= (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
		sg->offset += (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
	} else {
		fd->length20 = ip_hdr->total_len - IP_HEADER_LENGTH_NO_OPTIONS;
		fd->offset += ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS;
	}
	notes->fd = fd;
	if (unlikely(ipsa_scan(ctxt, notes, ip_hdr_ptr) < 0)) {
		APP_ERROR("%d :Scan call failed", __func__);
		/* We may have to drop original pkt */
		decorated_notify_inc_64(&(ctxt->stats->ip_in_dropped));
		lwe_free_fd(ipsec_buff_allocator, notes->fd);
		return IP_STATUS_DROP;
	}
	APP_DEBUG("PME DBG:pme_send_ simple exit");

	return IP_STATUS_ACCEPT;
}

enum IP_STATUS ipsa_send_compd_outbuf(struct ip_context_t *ctxt,
				      struct annotations_t *notes,
				      struct ip_header_t *ip_hdr)
{
	struct qm_fd *fd;
	struct qm_fd fd2;
	uint32_t buffer_sz;
	uint32_t flow_id;

	flow_id = notes->dest->flow_id;
	buffer_sz = ipsa_context_info[flow_id].buffer_sz;
	fd = notes->fd;

	APP_DEBUG("PME DBG:pme_send_compd outbuf");
	/* Change simple to compd fd with outbuf non NULL to send it to PME */
	if (unlikely
		(ipsa_create_compd_fd_outbuf(&fd2, fd, ip_hdr, buffer_sz))) {
		APP_ERROR("%s :Unable to Allocate compound FD", __func__);
		/* We may have to drop original pkt */
		decorated_notify_inc_64(&(ctxt->stats->ip_in_dropped));
		lwe_free_fd(ipsec_buff_allocator, notes->fd);
		return IP_STATUS_DROP;
	}
	notes->fd = &fd2;
	if (unlikely(ipsa_scan(ctxt, notes, ip_hdr) < 0)) {
		APP_ERROR("%d :Scan call failed", __func__);
		/* We may have to drop original pkt */
		decorated_notify_inc_64(&(ctxt->stats->ip_in_dropped));
		lwe_free_fd(ipsec_buff_allocator, notes->fd);
		return IP_STATUS_DROP;
	}
	return IP_STATUS_ACCEPT;
}

enum IP_STATUS ipsa_send_compd(struct ip_context_t *ctxt,
			       struct annotations_t *notes,
			       struct ip_header_t *ip_hdr)
{
	struct qm_fd *fd;
	struct qm_fd fd2;
	uint8_t bpid;
	uint32_t flow_id;

	flow_id = notes->dest->flow_id;
	bpid = ipsa_context_info[flow_id].bpid;
	fd = notes->fd;

	/* Change simple to compd fd with outbuf NULL to send it to PME */
	if (unlikely(ipsa_create_compd_fd(&fd2, fd, ip_hdr, bpid))) {
		APP_ERROR("%s :Unable to Allocate compound FD", __func__);
		/* We may have to drop original pkt */
		decorated_notify_inc_64(&(ctxt->stats->ip_in_dropped));
		lwe_free_fd(ipsec_buff_allocator, notes->fd);
		return IP_STATUS_DROP;
	}
	notes->fd = &fd2;
	if (unlikely(ipsa_scan(ctxt, notes, ip_hdr) < 0)) {
		APP_ERROR("%d :Scan call failed", __func__);
		/* We may have to drop original pkt */
		decorated_notify_inc_64(&(ctxt->stats->ip_in_dropped));
		lwe_free_fd(ipsec_buff_allocator, notes->fd);
		return IP_STATUS_DROP;
	}
	return IP_STATUS_ACCEPT;
}

int32_t ipsa_scan(struct ip_context_t *ctxt,
		  struct annotations_t *notes, struct ip_header_t *ip_hdr)
{
	struct pme_ctx *ctx;
	uint32_t args;
	struct ipsa_pme_token_t *pme_token;
	struct rt_dest_t *dest = NULL;
	int32_t ret = 0;
	uint32_t flow_id;
	uint32_t flags = 0;

	dest = notes->dest;
	flow_id = dest->flow_id;

	APP_DEBUG("PME DBG:ipsa scan 1");

	ctx = dest->pme_ctx;
	pme_token =
	    (struct ipsa_pme_token_t *)malloc(sizeof(struct ipsa_pme_token_t));
	if (unlikely(pme_token == NULL)) {
		APP_ERROR("Could not do malloc for token");
		return -1;
	}
	pme_token->flow_id = flow_id;
	pme_token->ip_ctxt = ctxt;
	flags = ipsa_context_info[flow_id].scan_flags;

	/* Setting the scan flags */
	if (!ipsa_context_info[flow_id].first_pkt) {
		flags |= PME_CMD_SCAN_SR;
		ipsa_context_info[flow_id].first_pkt = 1;
	}

	args = PME_SCAN_ARGS(flags, ipsa_context_info[flow_id].set,
			     ipsa_context_info[flow_id].subset);

	ret = pme_ctx_scan(ctx, 0, notes->fd,
			   args, (struct pme_ctx_token *)pme_token);
	if (unlikely(ret < 0)) {
		APP_ERROR(" Error in pme ctx scan");
		free(pme_token);
		return ret;
	}
	decorated_notify_inc_32(&(pme_stats->g_total_frames));
	APP_DEBUG("PME DBG : Total Frames going for scan %d",
		  pme_stats->g_total_frames);

	return 0;
}

void ipsa_create_simple_fd(struct qm_fd *simple_fd, struct qm_fd *compound_fd)
{
	struct sg_entry_t *sg, *sg_inframe;
	void *out_buf;
	struct annotations_t *new_notes;

	APP_DEBUG("PME DBG: create simple fd");
	sg = (struct sg_entry_t *)APP_PTOV(compound_fd->addr_lo);
	sg++;
	simple_fd->addr_hi = 0;
	simple_fd->addr_lo = sg->addr_lo;
	simple_fd->bpid = sg->bpid;
	if (sg->extension == 1) {
		simple_fd->format = qm_fd_sg;
		simple_fd->offset = (uint16_t)sg->offset;
		simple_fd->length20 = sg->length;
		sg_inframe = (struct sg_entry_t *)
		    APP_PTOV(sg->addr_lo + sg->offset);
		sg_inframe->length +=
		    (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
		sg_inframe->offset -=
		    (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
	} else {
		simple_fd->format = qm_fd_contig;
		simple_fd->offset = (uint16_t)sg->offset - (ETH_HDR_LEN +
						  IP_HEADER_LENGTH_NO_OPTIONS);
		simple_fd->length20 = sg->length + ETH_HDR_LEN +
		    IP_HEADER_LENGTH_NO_OPTIONS;
	}
	sg--;
	new_notes = (struct annotations_t *)APP_PTOV(simple_fd->addr_lo);
	new_notes->fd = simple_fd;

	/* Free Output buffer */
	out_buf = (void *)APP_PTOV(sg->addr_lo);
	discard_handler(out_buf, sg->bpid);

	/* Free SG buffer */
	discard_handler(sg, compound_fd->bpid);
}

void ipsa_ctx_rx_cb_simple(struct pme_ctx *pme_ctx, const struct qm_fd *fd,
			   struct pme_ctx_token *spa_pme_token)
{
	struct ip_context_t *ctxt;
	struct annotations_t *ip_notes;
	struct ip_header_t *ip_hdr;
	uint32_t flow_id;
	uint8_t res_flags;
	struct rt_dest_t *dest;
	struct spa_pme_ctx_stats_t *pme_ctx_stats;
	struct qm_fd new_fd;
	struct ipsa_pme_token_t *pme_ctx_token =
	    (struct ipsa_pme_token_t *)spa_pme_token;
	struct sg_entry_t *sg;

	memcpy(&new_fd, fd, sizeof(struct qm_fd));

	APP_DEBUG("PME DBG : Simple cb");
	ctxt = pme_ctx_token->ip_ctxt;
	flow_id = pme_ctx_token->flow_id;
	pme_ctx_stats = ipsa_context_info[flow_id].pme_ctx_stats;

	/* Increment post scan frames */
	decorated_notify_inc_32(&pme_ctx_stats->post_scan_frames);
	APP_DEBUG("PME DBG : Post scan Frames per ctx %d",
		  pme_ctx_stats->post_scan_frames);
	decorated_notify_inc_32(&(pme_stats->g_post_pme_pkts));
	APP_DEBUG("PME DBG : Post pme Frames for all ctx %d",
		  pme_stats->g_post_pme_pkts);

	res_flags = pme_fd_res_flags(fd);

	if (res_flags & PME_STATUS_TRUNCATED) {
		APP_DEBUG("PME DBG: simple fd has truncated bit");
		decorated_notify_inc_32(&(pme_ctx_stats->match_count));
		APP_DEBUG("PME DBG : match cnt per ctx %d",
			  pme_ctx_stats->match_count);
		decorated_notify_inc_32(&(pme_stats->g_pme_pattern_pkts));
		APP_DEBUG("PME DBG : Total matches for all ctx %d",
			  pme_stats->g_pme_pattern_pkts);
	}

	APP_DEBUG("PME DBG: pme cb simple");
	ip_notes = APP_PTOV(new_fd.addr_lo);
	ip_notes->fd = &new_fd;
	/* get notes and context information so that we can hand
	 * over the processing to the remainder of IPv4 processing
	 * chain
	 */
	if (ip_notes->fd->format == qm_fd_sg) {
		sg = (struct sg_entry_t *)
		    APP_PTOV(ip_notes->fd->addr_lo + ip_notes->fd->offset);
		ip_hdr = APP_PTOV(sg->addr_lo) + sg->offset -
		    IP_HEADER_LENGTH_NO_OPTIONS;
		sg->length += (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
		sg->offset -= (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
	} else {
		ip_hdr = (void *)((uint8_t *) ip_notes + new_fd.offset -
				  IP_HEADER_LENGTH_NO_OPTIONS);
		ip_notes->fd->length20 += ETH_HDR_LEN +
		    IP_HEADER_LENGTH_NO_OPTIONS;
		ip_notes->fd->offset = new_fd.offset -
		    (ETH_HDR_LEN + IP_HEADER_LENGTH_NO_OPTIONS);
	}

	dest = ip_notes->dest;
	switch (dest->scope) {
	case ROUTE_SCOPE_PME:
	case ROUTE_SCOPE_DECAP | ROUTE_SCOPE_PME:
		APP_DEBUG("PME DBG:...simple scope global");
		ip_forward(ctxt, ip_notes, ip_hdr);
		break;
	case ROUTE_SCOPE_PME | ROUTE_SCOPE_ENCAP:
		APP_DEBUG("PME DBG:...simple scope pme encap");
		ipsec_encap_send(ctxt, ip_notes, ip_hdr);
		break;
	default:
		ip_local_deliver(ctxt, ip_notes, ip_hdr);
	}

	/* Free Token + frame buffer memory */
	free(pme_ctx_token);

	APP_DEBUG("PME DBG: pme cb simple exit");

	return;
}

void ipsa_ctx_rx_cb_compd_outbuf(struct pme_ctx *pme_ctx,
				 const struct qm_fd *fd,
				 struct pme_ctx_token *spa_pme_token)
{
	struct ip_context_t *ctxt;
	struct annotations_t *ip_notes;
	struct ip_header_t *ip_hdr;
	uint32_t flow_id = 0;
	struct qm_fd simple_fd;
	struct sg_entry_t *sg, *sg_inframe;
	uint32_t *out_buf = NULL;
	uint8_t res_flags;
	struct rt_dest_t *dest;
	struct spa_pme_ctx_stats_t *pme_ctx_stats;
	struct qm_fd *compound_fd;
	struct ipsa_pme_token_t *pme_ctx_token =
	    (struct ipsa_pme_token_t *)spa_pme_token;

	compound_fd = (struct qm_fd *)fd;
	ctxt = pme_ctx_token->ip_ctxt;
	flow_id = pme_ctx_token->flow_id;
	pme_ctx_stats = ipsa_context_info[flow_id].pme_ctx_stats;

	/* Increment post scan frames */
	decorated_notify_inc_32(&(pme_ctx_stats->post_scan_frames));
	APP_DEBUG("PME DBG : Post scan Frames per ctx %d",
		  pme_ctx_stats->post_scan_frames);
	decorated_notify_inc_32(&(pme_stats->g_post_pme_pkts));
	APP_DEBUG("PME DBG : Post pme Frames for all ctx %d",
		  pme_stats->g_post_pme_pkts);

	sg = (struct sg_entry_t *)
	    APP_PTOV(compound_fd->addr_lo);
	out_buf = (uint32_t *) APP_PTOV(sg->addr_lo);
	if (*out_buf != SPA_MAGIC_VALUE) {
		/* Update match count */
		decorated_notify_inc_32(&(pme_ctx_stats->match_count));
		APP_DEBUG("PME DBG : match cnt per ctx %d",
			  pme_ctx_stats->match_count);
		decorated_notify_inc_32(&(pme_stats->g_pme_pattern_pkts));
		APP_DEBUG("PME DBG : Total matches for all ctx %d",
			  pme_stats->g_pme_pattern_pkts);
	}

	res_flags = pme_fd_res_flags(fd);

	if (res_flags & PME_STATUS_TRUNCATED)
		APP_DEBUG("PME DBG: compd  fd outbuf has truncated bit");

	/* Convert compound fd to simple to send it to FMAN */
	ipsa_create_simple_fd(&simple_fd, compound_fd);
	simple_fd.cmd = 0;
	ip_notes = (struct annotations_t *)APP_PTOV(simple_fd.addr_lo);
	if (simple_fd.format == qm_fd_sg) {
		sg_inframe = (struct sg_entry_t *)
		    APP_PTOV(simple_fd.addr_lo + simple_fd.offset);
		ip_hdr = sg_inframe + ETH_HDR_LEN;
	} else {
		ip_hdr = (void *)((uint8_t *) ip_notes + simple_fd.offset +
				  ETH_HDR_LEN);
	}

	dest = ip_notes->dest;
	switch (dest->scope) {
	case ROUTE_SCOPE_PME:
	case ROUTE_SCOPE_DECAP | ROUTE_SCOPE_PME:
		APP_DEBUG("PME DBG:...compd scope global");
		ip_forward(ctxt, ip_notes, ip_hdr);
		break;
	case ROUTE_SCOPE_PME | ROUTE_SCOPE_ENCAP:
		APP_DEBUG("PME DBG:...compd scope pme encap");
		ipsec_encap_send(ctxt, ip_notes, ip_hdr);
		break;
	default:
		ip_local_deliver(ctxt, ip_notes, ip_hdr);
	}

	/* Free Token + frame buffer memory */
	free(pme_ctx_token);
	return;
}

void ipsa_ctx_rx_cb_compd(struct pme_ctx *pme_ctx, const struct qm_fd *fd,
			  struct pme_ctx_token *spa_pme_token)
{
	struct ip_context_t *ctxt;
	struct annotations_t *ip_notes;
	struct ip_header_t *ip_hdr;
	uint32_t flow_id = 0;
	struct qm_fd simple_fd;
	struct sg_entry_t *sg, *sg_inframe;
	uint32_t *out_buf;
	struct rt_dest_t *dest;
	struct spa_pme_ctx_stats_t *pme_ctx_stats;
	struct qm_fd *compound_fd;
	struct ipsa_pme_token_t *pme_ctx_token =
	    (struct ipsa_pme_token_t *)spa_pme_token;

	compound_fd = (struct qm_fd *)fd;
	ctxt = pme_ctx_token->ip_ctxt;
	flow_id = pme_ctx_token->flow_id;
	pme_ctx_stats = ipsa_context_info[flow_id].pme_ctx_stats;

	/* Increment post scan frames */
	decorated_notify_inc_32(&(pme_ctx_stats->post_scan_frames));
	APP_DEBUG("PME DBG : Post scan Frames per ctx %d",
		  pme_ctx_stats->post_scan_frames);
	decorated_notify_inc_32(&(pme_stats->g_post_pme_pkts));
	APP_DEBUG("PME DBG : Post pme Frames for all ctx %d",
		  pme_stats->g_post_pme_pkts);

	APP_DEBUG("PME DBG: pme cb compd");
	sg = (struct sg_entry_t *)
	    APP_PTOV(compound_fd->addr_lo);
	out_buf = (uint32_t *) APP_PTOV(sg->addr_lo);
	if (out_buf) {
		APP_DEBUG("PME DBG: comp fd has out buf now");
		/* Update match count */
		decorated_notify_inc_32(&(pme_ctx_stats->match_count));
		APP_DEBUG("PME DBG : match cnt per ctx %d",
			  pme_ctx_stats->match_count);
		decorated_notify_inc_32(&(pme_stats->g_pme_pattern_pkts));
		APP_DEBUG("PME DBG : Total matches for all ctx %d",
			  pme_stats->g_pme_pattern_pkts);
	}

	/* Convert compound fd to simple to send it to FMAN */
	ipsa_create_simple_fd(&simple_fd, compound_fd);
	simple_fd.cmd = 0;
	ip_notes = APP_PTOV(simple_fd.addr_lo);
	ip_notes->fd = &simple_fd;

	ip_notes = (struct annotations_t *)APP_PTOV(simple_fd.addr_lo);
	if (simple_fd.format == qm_fd_sg) {
		sg_inframe = (struct sg_entry_t *)
		    APP_PTOV(simple_fd.addr_lo + simple_fd.offset);
		ip_hdr = sg_inframe + ETH_HDR_LEN;
	} else {
		ip_hdr = (void *)((uint8_t *) ip_notes + simple_fd.offset +
				  ETH_HDR_LEN);
	}

	dest = ip_notes->dest;
	switch (dest->scope) {
	case ROUTE_SCOPE_PME:
	case ROUTE_SCOPE_DECAP | ROUTE_SCOPE_PME:
		APP_DEBUG("PME DBG:...scope global");
		ip_forward(ctxt, ip_notes, ip_hdr);
		break;
	case ROUTE_SCOPE_PME | ROUTE_SCOPE_ENCAP:
		APP_DEBUG("PME DBG:...scope pme encap");
		ipsec_encap_send(ctxt, ip_notes, ip_hdr);
		break;
	default:
		ip_local_deliver(ctxt, ip_notes, ip_hdr);
	}

	/* Free Token + frame buffer memory */
	free(pme_ctx_token);
	return;
}

/**
 \brief Enqueue Rejection Notification Handler for PME Tx FQ
 \param[in] msg Message Ring entry to be processed
 \param[out] NULL
 */
static void ipsa_ctx_cb_ern(struct pme_ctx *ctx, const struct qm_mr_entry *msg,
			    struct pme_ctx_token *token)
{
	struct ipsa_pme_token_t *pme_ctx_token =
	    (struct ipsa_pme_token_t *)token;

	APP_ERROR("%s: RC = %x, seqnum = %x", __func__,
		  msg->ern.rc, msg->ern.seqnum);
	free(pme_ctx_token);
	/* All rejected enqueued frames are freed */
	lwe_free_fd(buff_allocator, &(msg->ern.fd));
	return;
}

static void update_flow_cb(struct pme_ctx *ctx, const struct qm_fd *fd,
			   struct pme_ctx_ctrl_token *token)
{
	free(token);
}

int32_t ipsa_ctxt_init(uint32_t flow_id,
		       struct lwe_ctrl_spa_cfg_info *spa_cfg_info)
{
	struct rt_dest_t *dest;
	uint32_t flags = 0;
	struct pme_flow params;
	struct pme_ctx *pme_ctx;
	struct spa_pme_ctx_stats_t *pme_ctx_stats;
	struct pme_ctx_ctrl_token *token;

	token = memalign(CACHE_LINE_SIZE, sizeof(*token));
	pme_ctx = (struct pme_ctx *)memalign(CACHE_LINE_SIZE,
					     sizeof(struct pme_ctx));
	if (unlikely(!pme_ctx || !token)) {
		APP_ERROR("Could not do malloc");
		free(pme_ctx);
		free(token);
		return 1;
	}
	token->cb = update_flow_cb;

	pme_ctx_stats =
	    (struct spa_pme_ctx_stats_t *)stats_memalign(CACHE_LINE_SIZE,
					 sizeof(struct spa_pme_ctx_stats_t));
	if (unlikely(pme_ctx_stats == NULL)) {
		APP_ERROR("Could not do stats_memalign");
		free(pme_ctx);
		free(token);
		return 1;
	}

	memset(pme_ctx_stats, 0, sizeof(struct spa_pme_ctx_stats_t));
	memset(pme_ctx, 0, sizeof(struct pme_ctx));

	dest = g_flow_to_entry[flow_id];
	dest->pme_ctx = pme_ctx;
	ipsa_context_info[flow_id].pme_ctx_stats = pme_ctx_stats;
	dest->pme_ctx->ern_cb = ipsa_ctx_cb_ern;

	/* Initialize the callback handler */
	if (spa_cfg_info->report_verbosity == REPORT_DISABLED) {
		/* Send simple fd */
		dest->pme_ctx->cb = ipsa_ctx_rx_cb_simple;
		ipsa_context_info[flow_id].func = ipsa_send_simple;
	} else {
		if (spa_cfg_info->output_buff_frm_app) {
			/* Send compd fd with o/p buff */
			ipsa_context_info[flow_id].func =
			    ipsa_send_compd_outbuf;
			/*Scan callback for compd fd with o/p buff */
			dest->pme_ctx->cb = ipsa_ctx_rx_cb_compd_outbuf;
		} else {
			/* Send compd fd without o/p buff */
			ipsa_context_info[flow_id].func = ipsa_send_compd;
			/* Scan callback for compd fd without o/p buff */
			dest->pme_ctx->cb = ipsa_ctx_rx_cb_compd;
		}
	}

	/* Configure the PME mode flags */
	if (!(spa_cfg_info->flag & SET_SCAN_MODE_FLOW))
		flags |= PME_CTX_FLAG_DIRECT;

	if (unlikely(pme_ctx_init(dest->pme_ctx, flags,
				  ipsa_context_info[flow_id].bpid,
				  (uint8_t)spa_cfg_info->qos_in,
				  (uint8_t)spa_cfg_info->qos_out,
				  pool_channel_id, NULL))) {
		APP_ERROR(" Error in pme ctx init");
		stats_free(pme_ctx_stats);
		free(pme_ctx);
		free(token);
		return 1;
	}

	/* enable the context */
	if (unlikely(pme_ctx_enable(dest->pme_ctx))) {
		pme_ctx_finish(dest->pme_ctx);
		APP_ERROR("Error in pme ctx enable");
		stats_free(pme_ctx_stats);
		free(pme_ctx);
		free(token);
		return 1;
	}

	/* If direct mode than no need to update below params */
	if (!(spa_cfg_info->flag & SET_SCAN_MODE_FLOW)) {
		ipsa_context_info[flow_id].ipsa_pme_ctx = dest->pme_ctx;
		free(token);
		return 0;
	}

	/* initialize all params to ZERO */
	pme_sw_flow_init(&params);
	flags = 0;

	params.sos = 1;
	params.sessionid = spa_cfg_info->session_id;
	flags |= PME_CMD_FCW_SEQ;

	params.clim = (uint16_t)spa_cfg_info->compare_limit;
	params.mlim = (uint16_t)spa_cfg_info->match_limit;
	flags |= PME_CMD_FCW_DXE;
	if (spa_cfg_info->flag & SET_RESIDUE_ENABLE) {
		flags |= PME_CMD_FCW_RES;
		params.ren = 1;
	}

	if (spa_cfg_info->report_verbosity == REPORT_DISABLED) {
		if ((spa_cfg_info->flag & SET_SCAN_MODE_FLOW)) {
			params.esee = 0;
			flags |= PME_CMD_FCW_SRE;
		}
	} else {
		params.srvm = (uint8_t)spa_cfg_info->report_verbosity;
		params.esee = 1;
		flags |= PME_CMD_FCW_SRE;
	}
	if (unlikely(pme_ctx_ctrl_update_flow(dest->pme_ctx, flags, &params,
					      token))) {
		APP_ERROR("Error in PME flow updation");
		return 1;
	}
	ipsa_context_info[flow_id].ipsa_pme_ctx = dest->pme_ctx;

	return 0;
}

int32_t ipsa_add_ctx(struct lwe_ctrl_op_info *route_info)
{
	uint32_t flow_id;
	uint32_t report_sz = 0;
	uint32_t flags = 0;
	int n;
	uint32_t bpsize;
	uint32_t bpid;
	struct lwe_ctrl_spa_cfg_info *spa_cfg_info =
	    &(route_info->app_info.pme_info);

	flow_id = spa_cfg_info->flow_id;

	if (ipsa_context_info[flow_id].valid) {
		APP_DEBUG(" Flow id %s is already initialized", flow_id);
		return -1;
	}

	if (!g_flow_to_entry[flow_id]) {
		APP_DEBUG("No route added for Flow id %s", flow_id);
		return -1;
	}

	if (unlikely(g_total_flow >= MAX_SPA_PME_CNTX))
		return -1;

	if (spa_cfg_info->report_verbosity == REPORT_DISABLED) {
		/* Get Buffer Pool ID for maximum buffer size
		 * to give some valid bpid */
		if (0 > get_bpid_bpsize(MAX_BUFF_SZ, &bpid, &bpsize)) {
			APP_ERROR(" Bpools not initialised");
			return -1;
		}
		ipsa_context_info[flow_id].bpid = bpid;
	} else {
		if (!(spa_cfg_info->flag & SET_SCAN_MODE_FLOW))
			spa_cfg_info->match_limit = 0xffff;
		/* Get output buffer size from verbosity and mlimit */
		report_sz = get_report_size(spa_cfg_info->report_verbosity,
					    spa_cfg_info->match_limit);
		APP_DEBUG("PME DBG: rep sz is %d", report_sz);
		/* Get Buffer Pool ID for output buffer */
		if (0 > get_bpid_bpsize(report_sz, &bpid, &bpsize)) {
			APP_ERROR(" Bpools not initialised");
			return -1;
		}
		APP_DEBUG("PME DBG: buffer sz is %d", bpsize);
		ipsa_context_info[flow_id].buffer_sz = bpsize;
		ipsa_context_info[flow_id].bpid = bpid;
	}

	/* Setting the scan flags */
	n = spa_cfg_info->report_verbosity;
	if (n != REPORT_DISABLED)
		flags |= PME_CMD_SCAN_SRVM(n);

	if (spa_cfg_info->flag & SET_SCAN_MODE_FLOW)
		ipsa_context_info[flow_id].first_pkt = 0;
	else {
		flags |= PME_CMD_SCAN_SR | PME_CMD_SCAN_E;
		ipsa_context_info[flow_id].first_pkt = 1;
	}

	ipsa_context_info[flow_id].scan_flags = flags;
	ipsa_context_info[flow_id].set = spa_cfg_info->set;
	ipsa_context_info[flow_id].subset = spa_cfg_info->subset;

	/* Initialize the pme ctx for this flow id */
	if (ipsa_ctxt_init(flow_id, spa_cfg_info))
		return -1;

	ipsa_context_info[flow_id].valid = 1;
	g_total_flow++;
	APP_DEBUG("PME DBG : total flows %d", g_total_flow);
	return 0;
}

int32_t ipsa_del_ctx(struct lwe_ctrl_op_info *route_info)
{
	uint32_t flow_id;
	struct rt_dest_t *dest;
	struct lwe_ctrl_spa_cfg_info *spa_cfg_info =
	    &(route_info->app_info.pme_info);
	struct spa_pme_ctx_stats_t *pme_ctx_stats;

	flow_id = spa_cfg_info->flow_id;
	dest = g_flow_to_entry[flow_id];
	pme_ctx_stats = ipsa_context_info[flow_id].pme_ctx_stats;

	if (!ipsa_context_info[flow_id].valid) {
		APP_ERROR("Context for flow id %s doesn't exist", flow_id);
		return -1;
	}

	/* Disable and then destroy the PME context */
	if (blocking_pme_disable(dest->pme_ctx)) {
		APP_ERROR
		    ("ERROR:  PME context disable failed for flow id %x",
		     flow_id);
		return -1;
	}

	pme_ctx_finish(dest->pme_ctx);

	ipsa_context_info[flow_id].valid = 0;
	free(dest->pme_ctx);
	stats_free(pme_ctx_stats);
	g_total_flow--;
	APP_DEBUG("PME DBG : After del ctx total flows %d", g_total_flow);
	return 0;
}
