/**
 \file ipsecfwd_scan.h
 \brief Basic IPSec Forwarding Scan Application library
 */
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef __IPSECFWD_SCAN_H__
#define __IPSECFWD_SCAN_H__
#include "ipsecfwd.h"

extern uint32_t g_total_flow;
extern struct rt_dest_t *g_flow_to_entry[MAX_SPA_PME_CNTX];
extern struct pme_stats_t *pme_stats;
extern uint8_t g_all_context_initialized;

/**
 \brief Creates compd fd with output buffer allocated
 \param[inout] fd frame dsecriptor pointer
 \param[in] ip_hdr IP header pointer
 \param[in] size size of the output buffer
 \return Integer status
 */
extern int32_t ipsa_create_compd_fd_outbuf(struct qm_fd *, struct qm_fd *,
					   struct ip_header_t *, uint32_t);

/**
 \brief Creates compd fd without output buffer
 \param[inout] fd frame dsecriptor pointer
 \param[in] ip_hdr IP header pointer
 \param[in] size size of the output buffer
 \return Integer status
 */
extern int32_t ipsa_create_compd_fd(struct qm_fd *,
				    struct qm_fd *, struct ip_header_t *,
				    uint8_t);

/**
 \brief Sends PME scan request for simple fd
 \param[in] ctxt IP context pointer
 \param[in] notes pointer to Annotations structure
 \param[in] ip_hdr_ptr IP header pointer
 \return Integer status
*/
extern enum IP_STATUS ipsa_send_simple(struct ip_context_t *,
				       struct annotations_t *,
				       struct ip_header_t *);

/**
 \brief Sends PME scan request for compound fd with output buffer
 \param[in] ctxt IP context pointer
 \param[in] notes pointer to Annotations structure
 \param[in] ip_hdr_ptr IP header pointer
 \return Integer status
 */
extern enum IP_STATUS ipsa_send_compd_outbuf(struct ip_context_t *,
					     struct annotations_t *,
					     struct ip_header_t *);

/**
 \brief Sends PME scan request for compound fd without output buffer
 \param[in] ctxt IP context pointer
 \param[in] notes pointer to Annotations structure
 \param[in] ip_hdr_ptr IP header pointer
 \return Integer status
 */
extern enum IP_STATUS ipsa_send_compd(struct ip_context_t *,
				      struct annotations_t *,
				      struct ip_header_t *);

/**
 \brief PME scan call
 \param[in] - Context for pme.
 \param[in] - struct annotations_t *notes pointer.
 \param[in] - struct ip_header_t *ip_hdr pointer.
 \param[out] - integer status
 */
extern int32_t ipsa_scan(struct ip_context_t *,
			 struct annotations_t *, struct ip_header_t *);

/**
 \brief Convert compound fd to simple
 \param[out] simple_fd pointer to simple fd
 \param[in] compound_fd pointer to compound fd
 \return None
 */
extern void ipsa_create_simple_fd(struct qm_fd *, struct qm_fd *);

/**
 \brief Scan call back handler for simple fd
 \param[in] pme_ctx Context for pme
 \param[in] fd Frame descriptor
 \param[in] spa_pme_token pointer to PME context token
 \return Integer status
 */
extern void ipsa_ctx_rx_cb_simple(struct pme_ctx *, const struct qm_fd *,
				  struct pme_ctx_token *);

/**
 \brief Scan call back handler for compound fd with output buffer from
 Application
 \param[in] pme_ctx Context for pme
 \param[in] fd Frame descriptor
 \param[in] spa_pme_token pointer to PME context token
 \return Integer status
 */
extern void ipsa_ctx_rx_cb_compd_outbuf(struct pme_ctx *,
					const struct qm_fd *,
					struct pme_ctx_token *);

/**
 \brief Scan call back handler for compound fd with output buffer from PME
 \param[in] pme_ctx Context for pme
 \param[in] fd Frame descriptor
 \param[in] spa_pme_token pointer to PME context token
 \return Integer status
 */
extern void ipsa_ctx_rx_cb_compd(struct pme_ctx *, const struct qm_fd *,
				 struct pme_ctx_token *);

/**
 \brief Enable & initialize the PME context.
 \param[in] flow_id id of particular flow.
 \param[in] spa_cfg_info contains whiteboard cfg
 \return Integer status
 */
extern int32_t ipsa_ctxt_init(uint32_t, struct lwe_ctrl_spa_cfg_info *);

/**
 \brief Configure for initializing the PME context.
 \param[in] lwe_ctrl_info contains PME parameters.
 \return Integer status.
 */
extern int32_t ipsa_add_ctx(struct lwe_ctrl_op_info *);

/**
 \brief Disable & delete the PME context.
 \param[in] lwe_ctrl_info contains PME parameters.
 \return Integer status.
 */
extern int32_t ipsa_del_ctx(struct lwe_ctrl_op_info *);

/**
 \brief Cleanup routine for all PME Contexts
 */
extern int32_t ipsa_cleanup_del_all_ctx(void);
#endif
