#!/bin/sh
# File updated simics checkpoint with 4GB memory in uboot.
# Enables line interleaving in checkpoint for performance
# 
#  Copyright (C) 2009 Freescale Semiconductor, Inc.
# 
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
# NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

myname=`basename $0`

set -- `getopt 'lpbs' $*`

usage_and_exit() {
  echo "Usage:  $0 [-l][-p][-b][-s] <checkpoint>"
  echo "  -l:  enable line interleaving (also default behavior)"
  echo "  -p:  enable page interleaving"
  echo "  -b:  enable bank interleaving"
  echo "  -s:  enable superbank interleaving"
  exit 1
}

# Initialize variables to default:  line interleaving enabled, all others off.
do_line=1
do_page=0
do_bank=0
do_superbank=0

for opt in $*; do
  case $opt in
    -l) do_line=1; shift;;
    -p) do_page=1; do_line=0; shift;;
    -b) do_bank=1; do_line=0; shift;;
    -s) do_superbank=1; do_line=0; shift;;
    --) shift;;	# Reached end of options.
  esac
done

if [ `expr $do_line + $do_page + $do_bank + $do_superbank` -ne 1 ]; then
  echo "ERROR:  more than one interleaving option specified?"
  usage_and_exit
fi

CHKPT=$1

# Make sure we got an argument, and that the file exists.
if [ -z ${CHKPT} -o ! -f ${CHKPT} ]; then
	usage_and_exit
fi

if [ ! -f ${CHKPT} ]; then
	echo "${myname}:  checkpoint '${CHKPT}' does not exist?"
	exit 1
fi

# If there are any CS_CONFIG registers that contain 0x80010102,
# alter them to 0x80014102, for DDR3 with 8 banks per CS.
awk -v progname=$myname -v do_line=$do_line -v do_page=$do_page -v do_bank=$do_bank -v do_superbank=$do_superbank '
/OBJECT/ { obj = $2; }

# Here is how we should alter things.  Note that this may need
#   to be smarter to handle things when our memory config changes.
# Line: CS_CONFIG: (0xA0004202, 0, 0, 0);
# Page: CS_CONFIG: (0xA1004202, 0, 0, 0)
# Bank: CS_CONFIG: (0xA2004202, 0, 0, 0)
# Superbank: CS_CONFIG: (0xA3004202, 0, 0, 0)

## Enable Cache Line Interleaving
## Alter addr_LAWAR
/0, 0, 0x80000000, 0/ {
  sub("0x80000000","0");
  }
/ 0, 0, 0, 0x8110001e, 0x8100001e/ {
  sub("0x8110001e, 0x8100001e","0x81400020, 0");
  }
/0x81004019, 0x81005019, 0x81006016/ {
  sub("0x81004019, 0x81005019, 0x81006016","0x81404019, 0x81405019, 0x81406016");
  }
/0x83c00014, 0x81001010, 0x8100201c, 0x8100301c/ {
  sub("0x81001010, 0x8100201c, 0x8100301c","0x81401010, 0x8140201c, 0x8140301c");
  }
/CS_CONFIG: \(0x80334301, 0x80004301/ {
  if (do_line) {
     printf "NOTE:  %s:  Forcibly altered checkpoint to enable Cache Line Interleaving '%s'\n", progname, obj > "/dev/stderr";
     sub("0x80334301, 0x80004301","0xA0004302, 0");
  } else if (do_page) {
     printf "NOTE:  %s:  Forcibly altered checkpoint to enable Page Interleaving '%s'\n", progname, obj > "/dev/stderr";
     sub("0x80334301, 0x80004301","0xA1004302, 0");
  } else if (do_bank) {
     printf "NOTE:  %s:  Forcibly altered checkpoint to enable Bank Interleaving '%s'\n", progname, obj > "/dev/stderr";
     sub("0x80334301, 0x80004301","0xA2004302, 0");
  } else if (do_superbank) {
     printf "NOTE:  %s:  Forcibly altered checkpoint to enable Superbank Interleaving '%s'\n", progname, obj > "/dev/stderr";
     sub("0x80334301, 0x80004301","0xA3004302, 0");
  }
  matches++;
  }
/CS_BNDS:/ {
  sub("63","0x000000FF");
  sub("0x40007f","0");
  sub("0x8000bf","0x000000FF");
  sub("0xc000ff","0");
  }
# Always print the line, which may have been altered above.
{
  print;
}
END {
  # Verify that we had at least one match, otherwise Things Are Bad.
  if (matches == 0) {
    printf "ERROR:  %s:  Could not find a matching CS_CONFIG line to alter!  Expected CS_CONFIG to hold 0x80004102.\n", progname > "/dev/stderr";
    exit (1);
  }
}
' ${CHKPT} > ${CHKPT}.fix
if [ $? -eq 0 ]; then
  mv ${CHKPT}.fix ${CHKPT}
  echo "${myname}:  Added interleaving to checkpoint file ${CHKPT}"
  echo "${myname}:  Please ensure that sim.num_interleave_controllers is set to 2...."
  exit 0
else
  echo "ERROR:  Due to above errors, original file ${CHKPT} is unmodified.  Attempted changes (if any) are in ${CHKPT}.fix."
  # Do nothing.
  exit 1
fi
