#!/bin/sh
# 
#  Copyright (C) 2009 Freescale Semiconductor, Inc.
# 
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
# NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

only_nonzeros() {
  awk '($(NF) != 0) {print}'
}

usage_and_exit() {
cat 1>&2 <<EOF
Usage:  $0 [-p <num_pkt_events>[,<num_pkt_events>]] file [file ...]

Process all (or last num_pkt_events) events in the specified file(s),
and output stats.  Note that only the rate stats honor the -p
flag.

If the more-complicated XX,YY value is given, it will only honor
the YY events starting after event number XX.  (tail -n +XX | head -n YY)
EOF
exit 1
}


# If we are given a -p flag for the last P packets, honor it.
TEMP=`getopt "p:" $*`
if [ $? != 0 ]; then usage_and_exit; fi
eval set -- "$TEMP"

for opt in $*; do
  case $opt in
    -p) 
      if expr index $2 ',' >/dev/null; then
        # Got the complex XX,YY format.
	pkt_first_arg=`echo $2 | sed -e 's/,.*$//'`
	pkt_second_arg=`echo $2 | sed -e 's/.*,//'`
      else
        pkt_tail=$2;
      fi
      shift 2;;
    --) shift;;
  esac
done

for f in $*; do
   # If the file is not a regular file, then first cp it to a
   # temp file.
   if [ ! -f $f ]; then
   set -x
     cp $f /tmp/summ_pkts.$$.out

     f=/tmp/summ_pkts.$$.out
     remove_me="${remove_me} $f"
   fi
   # Check if there is a stats.after file.
   statfile=`echo $f | sed -e 's/.out$/.stats.after/'`
   # If no file, use the output file.
   if [ ! -f ${statfile} ]; then statfile=$f; fi
   printf "=== Packet summary for $f ===\n"
   # See if it is running at 600MHz or 800MHz etc.
   sys_mhz=`grep ^sim.system_clock_freq_mhz $f | awk '{print $(NF)}'`
   qman_mhz=`grep ^sim.clock_ratio $f | awk -v sys_mhz=$sys_mhz '
   {
     ratio=$(NF);
     # Split by colon.
     split(ratio, parts, ":");
     print sys_mhz * parts[3] / parts[2];
     }
   '`
   echo "*** Assuming Platform is at 800 MHz, QMan is at 400 MHz."

   # If the test did not complete, print a warning.
   if grep "Test Passed" $f >/dev/null; then
      : # Do nothing.
   else
      echo "WARNING:  No SUCCESS line detected -- simulation still in progress, or failed?"
      exit
   fi

   printf "fgen inbound: \n"
   grep fgen..inbound_packets $statfile | only_nonzeros

   printf "fman inbound: \n"
   grep fman..inbound $statfile | only_nonzeros
   printf "===============\n"

   # This is only printed in a side-copy of libfm.so, so if it
   # is there, print it, but if not, don't even print the
   # header because a 0 would be misleading.
   if grep "FMan. dropping a packet due to" $f >/dev/null; then
      for fman in 0 1; do
         printf "fman${fman} RxFIFO drops:  "
         grep "FMan${fman} dropping a packet due to" $f | wc -l
      done
   fi

   # Extract qman trace for faster processing
   qf=`echo $f | sed -e 's/.out$/.qmantrace/'`
   if [ "$qf" = "$f" ]; then
     # If the given filename doesn't have .out in the name, then
     # we better be careful, because $qf will be the same, and
     # the command below will clobber our input!
     qf=/tmp/summ_pkts.$$.qmantrace
     remove_me="${remove_me} $qf"
   fi
   grep "qman\]" $f > $qf
   if [ "$pkt_tail" != "" ]; then
     mv $qf ${qf}.full
     tail -n $pkt_tail ${qf}.full > $qf
   fi
   if [ "$pkt_first_arg" != "" ]; then
     mv $qf ${qf}.full
     tail -n +${pkt_first_arg} ${qf}.full | head -n ${pkt_second_arg} > $qf
   fi

   printf "fgen outbound: \n"
   grep fgen..outbound_packets $statfile | only_nonzeros

   printf "fman outbound: \n"
   grep fman..outbound $statfile | only_nonzeros

   awk -v qman_mhz=$qman_mhz '
   BEGIN {
      first_sw_dq=1;
   }
   #/enqueue event on DC portal 0/ {
   ($4=="enqueue" && $7=="DC" && $9=="0") {
      dc0_eq++;
      if(dc0_eq==1){
         cyc_first_dc0_eq = $3;
      }
      cyc_last_dc0_eq = $3;
   }
   #/enqueue event on DC portal 1/ {
   ($4=="enqueue" && $7=="DC" && $9=="1") {
      dc1_eq++;
      if(dc1_eq==1){
         cyc_first_dc1_eq = $3;
      }
      cyc_last_dc1_eq = $3;
   }
   #/enqueue event on DC portal 2/ {
   ($4=="enqueue" && $7=="DC" && $9=="2") {
      dc2_eq++;
      if(dc2_eq==1){
         cyc_first_dc2_eq = $3;
      }
      cyc_last_dc2_eq = $3;
   }
   #/enqueue event on SW/ {
   ($4=="enqueue" && $7=="SW") {
      sw_eq++;
      if(sw_eq==1){
         cyc_first_sw_eq = $3;
      }
      cyc_last_sw_eq = $3;
      bklog_post_sw_eq = dc0_dq + dc1_dq;

   }
   #/dequeue event on DC portal 0/ {
   ($4=="dequeue" && $7=="DC" && $9=="0") {
      dc0_dq++;
      if(dc0_dq==1){
         cyc_first_dc0_dq = $3;
      }
      cyc_last_dc0_dq = $3;
   }
   #/dequeue event on DC portal 1/ {
   ($4=="dequeue" && $7=="DC" && $9=="1") {
      dc1_dq++;
      if(dc1_dq==1){
         cyc_first_dc1_dq = $3;
      }
      cyc_last_dc1_dq = $3;
   }
   #/dequeue event on DC portal 2/ {
   ($4=="dequeue" && $7=="DC" && $9=="2") {
      dc2_dq++;
      if(dc2_dq==1){
         cyc_first_dc2_dq = $3;
         bklog_pre_dc2_dq = sw_eq;
      }
      cyc_last_dc2_dq = $3;
   }
   #/dequeue event on SW/ {
   ($4=="dequeue" && $7=="SW") {
      sw_dq++;
      if(sw_dq==1){
         cyc_first_sw_dq = $3;
         bklog_pre_sw_dq = dc0_eq+dc1_eq; 
      }
      cyc_last_sw_dq = $3;
   }
    # Always calculate current num packets in system.
   {
      if(bklog_post_sw_eq > max_bklog_post_sw_eq){max_bklog_post_sw_eq = bklog_post_sw_eq;}
      pkts_in_part = dc0_eq + dc1_eq - dc0_dq - dc1_dq;
      if (pkts_in_part > max_pkts_in_part) {
         max_pkts_in_part = pkts_in_part;
      }
      pkts_in_pre_sw = dc0_eq + dc1_eq - sw_eq; 
      if (pkts_in_pre_sw > max_pkts_in_pre_sw) {
         max_pkts_in_pre_sw = pkts_in_pre_sw;
      }
      pkts_in_pre_caam = sw_eq - dc2_dq;
      if (pkts_in_pre_caam > max_pkts_in_pre_caam) {
         max_pkts_in_pre_caam = pkts_in_pre_caam;
      }
      pkts_in_pre_egress = sw_eq - dc0_dq - dc1_dq;
      if (pkts_in_pre_egress > max_pkts_in_pre_egress) {
         max_pkts_in_pre_egress = pkts_in_pre_egress;
      }
#      if (pkts_in_part > 100 + last_pkt_in_thresh) {
#         #printf "After DC enqueue #%d, %d frames in part.\n", dcnqs, pkts_in_part;
#         last_pkt_in_thresh = pkts_in_part;
#      }
   }
function print_mpps_gbps(prompt, num, delta_cycs) {
	qman_mhz=400;
# QMan runs at 400MHz, hence the 400 factor to get from packets/cycles into
# packets/second.
# For small packets, there are 12+8 bytes of overhead, so a Mpps
# rate needs to be multiplied by 84*8 to turn it into Gbps.
       gbps_factor = (64+12+8)*8/1000;
       if (delta_cycs) {
	  mpps = num/delta_cycs * qman_mhz;
       } else mpps = 0;
       gbps = mpps*gbps_factor;
       # For Ethernet, ignore IFG and preamble.
       eth_factor = 64*8/1000;
       eth_gbps = mpps*eth_factor;
       printf "%-12s rate:  %6d in %8d QMan cycs = %5.4f Mpps (%5.4f raw Gbps, %5.4f effective Ethernet Gbps for 64-byte frames)\n", prompt, num, delta_cycs, mpps, gbps, eth_gbps;
}

   END {
      printf "===============\n"
      if(dc0_eq>0){printf "FMan0 \t Enqueues: %d\n", dc0_eq;}
      if(dc1_eq>0){printf "FMan1 \t Enqueues: %d\n", dc1_eq;}
      if(sw_dq>0){printf "SW \t Dequeues: %d\n", sw_dq;}
      if(dc2_dq>0){printf "CAAM \t Dequeues: %d\n", dc2_dq;}
      if(dc2_eq>0){printf "CAAM \t Enqueues: %d\n", dc2_eq;}
      if(sw_eq>0){printf "SW \t Enqueues: %d\n", sw_eq;}
      if(dc0_dq>0){printf "FMan0 \t Dequeues: %d\n", dc0_dq;}
      if(dc1_dq>0){printf "FMan1 \t Dequeues: %d\n", dc1_dq;}
      printf "===============\n"
      printf "DC enqueues before first SW   dequeue:  %5d\n", bklog_pre_sw_dq;
      printf "DC enqueues  after first SW   dequeue:  %5d\n", dc0_eq + dc1_eq - bklog_pre_sw_dq;
      if(dc2_dq>0){printf "SW enqueues before first CAAM dequeue:  %5d\n", bklog_pre_dc2_dq;}
      if(dc2_dq>0){printf "SW enqueues  after first CAAM dequeue:  %5d\n", sw_eq - bklog_pre_dc2_dq;}
      printf "DC dequeues before last SW   enqueue:   %5d\n", bklog_post_sw_eq;
      printf "DC dequeues  after last SW   enqueue:   %5d\n", dc0_dq + dc1_dq - bklog_post_sw_eq;
      printf "===============\n"
      print_mpps_gbps("FMan0 enqueue", dc0_eq, cyc_last_dc0_eq - cyc_first_dc0_eq);
      print_mpps_gbps("FMan1 enqueue", dc1_eq, cyc_last_dc1_eq - cyc_first_dc1_eq);
      if(dc2_dq>0){print_mpps_gbps("CAAM  dequeue", dc2_dq, cyc_last_dc2_dq - cyc_first_dc2_dq);}
      print_mpps_gbps("SW    dequeue", sw_dq, cyc_last_sw_dq - cyc_first_sw_dq);
      print_mpps_gbps("SW    enqueue", sw_eq, cyc_last_sw_eq - cyc_first_sw_eq);
      if(dc2_eq>0){print_mpps_gbps("CAAM  enqueue", dc2_eq, cyc_last_dc2_eq - cyc_first_dc2_eq);}
      print_mpps_gbps("FMan0 dequeue", dc0_dq, cyc_last_dc0_dq - cyc_first_dc0_dq);
      print_mpps_gbps("FMan1 dequeue", dc1_dq, cyc_last_dc1_dq - cyc_first_dc1_dq);


      printf "===============\n"
      printf "Max packet backlog pre-SW:  %d\n", max_pkts_in_pre_sw;
      if(dc2_dq>0){printf "Max packet backlog pre-CAAM:  %d\n", max_pkts_in_pre_caam;}
      printf "Max packet backlog post-SW/pre-egress:  %d\n", max_pkts_in_pre_egress;
      printf "Max packet backlog in system:  %d\n", max_pkts_in_part;
   }
   ' $qf

   printf "Dai-style egress rate (steady-state only): ";
      # Measure packets only in steady state -- from when first
      # frame egresses (pipe has been filled) until last ingress
      # (we aren't yet in final draining process yet).
      # First, grab the time of the first frame out.
      first_cyc=`grep 'dequeue.event.on.DC.portal.[01]' $f | head -1 | awk '{print $3}'`
      # And the last ingress.
      last_cyc=`grep 'enqueue.event.on.DC.portal.[01]' $f | tail -1 | awk '{print $3}'`
      printf "%d - %d = %d\n" $last_cyc $first_cyc `echo $last_cyc - $first_cyc | bc -q`
      # Now only grep the egress events between the two.
      frames=`grep -A100000000 "cycle $first_cyc" $f | grep -B1000000000 "cycle $last_cyc" | grep -c 'dequeue.event.on.DC.portal.[01]'`
      awk -v qman_mhz=$qman_mhz -v frames=$frames -v first_cyc=$first_cyc -v last_cyc=$last_cyc 'BEGIN {
        delta=last_cyc-first_cyc;
        printf "Dai-style egress rate:  %6d in %9d QMan cycs = %5.2f Mpps\n", frames, delta, frames/delta * 400
      }'


   # Check for asserts etc.
   check_exp="simulation.state.has.been.corrupted|Assertion|tbassert_msg"
   if egrep $check_exp $f >/dev/null; then
     printf "WARNING:  An assertion occurred in this run:\n" 1>&2
     egrep $check_exp $f
   fi


done

if [ "$remove_me" != "" ]; then
  rm ${remove_me}
fi
