/**
 \file arp.c
 */
/*
 * Copyright (C) 2007-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdint.h>
#include "arp/arp.h"
#include "app_common.h"
#include "ether.h"
#include "dpa/dpa_eth.h"
#include <sg_api.h>
#include "ipsec/ipsec_common.h"
#include "net/neigh.h"


uint32_t arp_lock;
extern struct dpa_buff_allocator *buff_allocator;

struct neigh_table_t arp_table = {
	.proto_len = 4,
	.constructor = arp_constructor,
	.config = {
		   .base_reachable_timeout = 30,
		   .reachable_timeout = 30,
		   .retrans_timeout = 1,
		   .quiesce_timeout = 5,
		   .solicit_queue_len = 1}
};

struct neigh_table_t *arp_table_create()
{
	struct neigh_table_t *table;

	table = &arp_table;
	return table;
}

uint32_t arp_handle_request(struct ethernet_header_t * eth_hdr,
			    struct node_t * node)
{
	struct arp_header_t *arp_header;

	arp_header =
	    (struct arp_header_t *)((uint32_t) eth_hdr +
				    sizeof(struct ethernet_header_t));
	if (memcmp(&arp_header->arp_targetip, &node->ip.word, IP_ADDRESS_BYTES)) {
		return -1;
	}

	memcpy(&arp_header->arp_targetip, &arp_header->arp_senderip,
	       IP_ADDRESS_BYTES);
	memcpy(&arp_header->arp_senderip, &node->ip.word, IP_ADDRESS_BYTES);
	arp_header->arp_opcode = ARPOP_REPLY;
	memcpy(eth_hdr->destination.bytes, eth_hdr->source.bytes, ETH_ADDR_LEN);
	memcpy(eth_hdr->source.bytes, node->mac.bytes, ETH_ADDR_LEN);
	memcpy(arp_header->arp_targetaddr.bytes,
	       eth_hdr->destination.bytes, ETH_ADDR_LEN);
	memcpy(arp_header->arp_senderaddr.bytes, eth_hdr->source.bytes,
	       ETH_ADDR_LEN);
	return 0;
}

int add_arp_entry(struct neigh_table_t *arp_table, struct net_dev_t *dev,
			struct node_t *node)
{
	struct neigh_t *n;

	n = neigh_create(arp_table);
	if (NULL == n) {
		APP_ERROR("%s: Unable to create Neigh Entry", __func__);
		return -EINVAL;
	}

	if (NULL == dev) {
		dev = ipfwd_get_dev_for_ip((unsigned int)(node->ip.word));
		if (NULL == dev) {
			APP_ERROR("%s: failed to get device", __func__);
			return -EINVAL;
		}
	}

	if (NULL == neigh_init(arp_table, n, dev,
				(uint32_t *) &node->ip.word)) {
		APP_ERROR("%s: Unable to init Neigh Entry", __func__);
		return -EINVAL;
	}
	if (false == neigh_add(arp_table, n)) {
		APP_ERROR("%s: Unable to add Neigh Entry", __func__);
		return -EINVAL;
	}
	if (NULL ==  neigh_update(n, (uint8_t *) node->mac.bytes,
				NEIGH_STATE_PERMANENT)) {
		APP_ERROR("%s: Unable to update Neigh Entry", __func__);
		return -EINVAL;
	}

	return 0;
}

void arp_handler(struct annotations_t *notes, void *data)
{
	struct arp_header_t *arp_hdr;
	struct neigh_t *n;
	struct node_t new_node;
	struct net_dev_t *dev;
	int merge_flag = 0;

	arp_hdr = (struct arp_header_t *)((uint8_t *) data + ETH_HDR_LEN);
	dev = ipfwd_get_dev_for_ip(arp_hdr->arp_targetip);

	if (unlikely(!dev))
		return;

	if (arp_hdr->arp_opcode == ARPOP_REPLY)
		APP_INFO("Got ARP reply from IP %x", arp_hdr->arp_senderip);

	spin_lock(&arp_lock);
	n = neigh_lookup(stack.arp_table,
			arp_hdr->arp_senderip, IP_ADDRESS_BYTES);
	if (n) {
		if (arp_hdr->arp_opcode == ARPOP_REPLY) {
			/* stop retransmit timer */
			if (NEIGH_STATE_PENDING == n->neigh_state) {
				if (0 != stop_timer(n->retransmit_timer)) {
					APP_ERROR
					    ("%s: stopping timer 0x%x failed",
						__func__, n->retransmit_timer);
					return;
				} else
					APP_INFO("%s: timer 0x%x stopped...",
						__func__, n->retransmit_timer);
			/* Send first packet */
			dev->xmit(dev, &n->fd, NULL);
			}
		}

		/* Update ARP cache entry */
		n->neigh_state = NEIGH_STATE_UNKNOWN;
		if (NULL == neigh_update(n, arp_hdr->arp_senderaddr.bytes,
						NEIGH_STATE_PERMANENT)) {
			APP_ERROR("%s: unable to update neigh entry",
				__func__);
			spin_unlock(&arp_lock);
			return;
		}
		merge_flag = 1;
	}

	if (is_iface_ip(arp_hdr->arp_targetip)) {
		APP_ERROR("%s: Target IP is not for own interface", __func__);
		lwe_free_fd(buff_allocator, notes->fd);
		spin_unlock(&arp_lock);
		return;
	}

	if (!merge_flag) {
		memcpy(new_node.mac.bytes,
			(uint8_t *) &arp_hdr->arp_senderaddr, ETH_ADDR_LEN);
		memcpy(new_node.ip.word,
			(uint8_t *) &arp_hdr->arp_senderip, IP_ADDRESS_BYTES);
		if (0 > add_arp_entry(stack.arp_table, NULL, &new_node)) {
			APP_ERROR("%s: failed to add ARP entry", __func__);
			lwe_free_fd(buff_allocator, notes->fd);
			spin_unlock(&arp_lock);
			return;
		}
	}

	spin_unlock(&arp_lock);

	if (arp_hdr->arp_opcode == ARPOP_REQUEST) {
		APP_INFO("Got ARP request from IP 0x%x", arp_hdr->arp_senderip);

		memcpy(new_node.mac.bytes, dev->dev_addr, ETH_ADDR_LEN);
		memcpy((uint8_t *)&new_node.ip.word,
			(uint8_t *)&arp_hdr->arp_targetip, IP_ADDRESS_BYTES);
		arp_handle_request((struct ethernet_header_t *) data,
				&new_node);
		dev->xmit(dev, notes->fd, NULL);
		APP_INFO("Sent ARP reply for IP 0x%x", arp_hdr->arp_targetip);
	} else {
		lwe_free_fd(buff_allocator, notes->fd);
	}

	return;
}

int arp_send_request(struct net_dev_t *dev, uint32_t target_ip)
{
	struct arp_header_t *arp_header;
	struct node_t *target_iface_node;
	struct bm_buffer bman_buf;
	struct qm_fd fd;
	struct ethernet_header_t *eth_hdr;
	phys_addr_t addr = 0;
	uint32_t len;

	len = ETH_HDR_LEN + ARP_HDR_LEN + ETH_FCS_LEN;
	if (0 >= dpa_allocator_get_buff(buff_allocator, len, &bman_buf)) {
		APP_ERROR("%s: couldn't allocate buf of size %d",
			__func__, len);
		return -ENOMEM;
	}

	fd.addr_hi = (uint8_t) bman_buf.hi;
	fd.addr_lo = bman_buf.lo;
	fd.bpid = bman_buf.bpid;
	fd.format = qm_fd_contig;
	fd.offset = 0;
	fd.length20 = len;

	addr = fd.addr_hi;
	addr = (addr << 32) | (uint32_t) bman_buf.lo;
	eth_hdr = ptov(addr);

	memset(eth_hdr->destination.bytes, ETH_DST_BROADCAST, ETH_ADDR_LEN);
	memcpy(eth_hdr->source.bytes, &dpa_eth_ports[dev->ifindex].mac_addr,
		ETH_ADDR_LEN);
	eth_hdr->proto = ETH_P_ARP;

	arp_header =
	    (struct arp_header_t *)((uint8_t *) eth_hdr +
				sizeof(struct ethernet_header_t));
	arp_header->arp_hrd = ARP_HTYPE_ETH;
	arp_header->arp_proto = ARP_PTYPE_IP;
	arp_header->arp_hrdlen = ARP_HLEN_ETH;
	arp_header->arp_protolen = ARP_PLEN_IP;
	arp_header->arp_opcode = ARPOP_REQUEST;

	target_iface_node = ipfwd_get_iface_for_ip(target_ip);
	if (!target_iface_node) {
		lwe_free_fd(buff_allocator, &fd);
		return -ENODEV;
	}

	memcpy(arp_header->arp_senderaddr.bytes, eth_hdr->source.bytes,
						ETH_ADDR_LEN);
	arp_header->arp_senderip = target_iface_node->ip.word;
	memset(arp_header->arp_targetaddr.bytes, 0, ETH_ADDR_LEN);
	arp_header->arp_targetip = target_ip;

	APP_INFO("Sending ARP request for IP %x", target_ip);
	dev->xmit(dev, &fd, NULL);

	return 0;
}

void arp_solicit(struct neigh_t *n, void *annotations, void *ll_payload)
{
	struct annotations_t *notes = annotations;
	decorated_notify_inc_64(&n->nt->stats->solicit_errors);
	lwe_free_fd(buff_allocator, notes->fd);
}

void arp_error_handler(struct neigh_t *n, void *annotations, void *ll_payload)
{
	struct annotations_t *notes = annotations;
	decorated_notify_inc_64(&n->nt->stats->protocol_errors);
	lwe_free_fd(buff_allocator, notes->fd);
}

void arp_constructor(struct neigh_t *n)
{
	n->funcs->solicit = &arp_solicit;
	n->funcs->error_handler = &arp_error_handler;
}

