/**
 \file dpa_bman.c          
 \brief Contains routines for buffer pools.
 */
/*
 * Copyright (C) 2009-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <app_libfdt.h>
#include <misc.h>
#include <compat.h>
#include <mmap.h>
#include <app_common.h>
#include "dpa_alloc.h"
#include <dpa/dpa_bman.h>

struct dpa_bpool_t dpa_bpools[MAX_NUM_BMAN_POOLS];

/**
 \brief Returns bpool of largest size
 \param[out] Largest buffer size available
 \param[out] Bpid carrying largest buffers
 \return 0 on success and -1 on failure
 */
int32_t get_maxsize_bpool(uint32_t * bpsize, uint8_t * bpid)
{
	uint8_t bpoolid;

	*bpid = 0;
	for (bpoolid = 0; bpoolid < MAX_NUM_BMAN_POOLS; bpoolid++)
		if (dpa_bpools[bpoolid].size > dpa_bpools[*bpid].size)
			*bpid = bpoolid;

	if (dpa_bpools[*bpid].size == 0)
		return -1;

	*bpsize = dpa_bpools[*bpid].size;
	return 0;
}

/**
 \brief Returns bpool, and size of the buffer for the a given data size
 \param[in] size Size of the Data for which the buffer is needed
 \param[out] bpid Id of the Bpool from which the buffer can be allocated for the data
 \param[out] bpsize Buffer Size that will be allocated from the Bpid Pool 
 \return 0 if successfull, else -1
 */
int get_bpid_bpsize(uint32_t size, uint32_t * bpid, uint32_t * bpsize)
{
	uint32_t bpoolid;
	uint32_t temp = 0;

	APP_DEBUG(" get_bpid_bpsize(size =  %x )", size);
	for (bpoolid = 0; bpoolid < MAX_NUM_BMAN_POOLS; bpoolid++) {
		if (dpa_bpools[bpoolid].size > dpa_bpools[temp].size)
			temp = bpoolid;

		if (dpa_bpools[bpoolid].size < size)
			continue;

		*bpid = bpoolid;
		*bpsize = dpa_bpools[bpoolid].size;
		return 0;
	}

	APP_DEBUG(" selected bpid  = %x", temp);
	if (dpa_bpools[temp].size == 0)
		return -1;

	*bpid = temp;
	*bpsize = dpa_bpools[temp].size;
	return 0;
}

/**
 \brief Reads the available Bpools - Ids, and their sizes from the device tree
 \return 0 if successfull, else -1
 */
int dpa_bpool_init(void)
{
	int off = -1;
	int ret = -1;
	uint32_t *bpool_cfg;
	uint32_t *bp_id;
	uint32_t size;
	uint32_t bpool_found = 0;

	while (1) {
		ret =
		    fdt_node_offset_by_compatible(devtree_ptr, off,
						  "fsl,bpool");
		if (ret < 0)
			break;

		off = ret;

		bpool_found++;

		/*Read the buffer-pool ID */
		bp_id = (uint32_t *) (fdt_getprop(devtree_ptr, ret, "fsl,bpid",
						  NULL));
		if (unlikely(NULL == bp_id)) {
			APP_ERROR("fdt_getprop failed for fsl,bpid");
			return -1;
		}

		bpool_cfg = (uint32_t *) fdt_getprop(devtree_ptr, ret,
						     "fsl,bpool-cfg", NULL);
		if (unlikely(NULL == bpool_cfg)) {
			APP_ERROR("fdt_getprop failed for fsl,bpool-cfg"
				  "for pool: %u", *bp_id);
			return -1;
		}

		size = bpool_cfg[3];
		if (unlikely(0 == size)) {
			APP_ERROR
			    ("BPID: %u configured with 0 size buffers",
			     *bp_id);
			return -1;
		}

		/*Already not initialised */
		if (0 == dpa_bpools[*bp_id].size)
			dpa_bpools[*bp_id].size = size;
	}

	if (bpool_found)
		return 0;

	return -1;
}
