/**
 \file dpa_eth.c          
 \brief Contains routines for reading DPA nodes from DTS.
 */
/*
 * Copyright (C) 2009-2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <app_libfdt.h>
#include <misc.h>
#include <compat.h>
#include <fsl_qman.h>
#include <mmap.h>
#include <app_common.h>
#include <dpa/dpa_eth.h>
#include <limits.h>

struct dpa_port_t dpa_eth_ports[MAX_NUM_PORTS];
volatile uint32_t g_num_dpa_eth_ports = 0;

static inline void fm_mac_out(struct fman_mac *fm_mac,
			      uint32_t reg_off, uint32_t val)
{
	out_be32((uint32_t *)((uintptr_t) fm_mac + reg_off), val);
}

static inline uint32_t fm_mac_in(const struct fman_mac *fm_mac,
				 uint32_t reg_off)
{
	return in_be32((uint32_t *)((uintptr_t) fm_mac + reg_off));
}

/**
\brief Enables a ethernet port

\details Ethernet ports are disabled when LWE partition comes up.
	This function enables a given port with given promiscuous mode.

\param[in]	port_id - Port ID to be enabled.
\param[in]	is_promisc - Whether to put port in promiscuous mode

\retval 0 - SUCCESS
\retval -1 - Failure
*/
static
int enable_mac_port(const uint32_t port_id, const bool is_promisc)
{
	uint32_t *reg;
	uint32_t *fixed_link;
	uint32_t page_size;
	uint32_t maccfg;
	uint32_t ecntrl;
	uint32_t speed;
	uint32_t duplex;
	int ret;
	int lenp;
	int mac_node;
	phys_addr_t pa;
	void *va;

	mac_node = dpa_eth_ports[port_id].mac_node;

	reg = (uint32_t *) fdt_getprop(devtree_ptr, mac_node, "reg", NULL);
	if (NULL == reg) {
		APP_ERROR("fdt_getprop failed for reg");
		return -1;
	}

	fixed_link = (uint32_t *) fdt_getprop(devtree_ptr, mac_node,
						"fixed-link", &lenp);
	if (NULL == fixed_link) {
		APP_ERROR("fdt_getprop failed for fixed-link");
		return -1;
	}

	if (20 != lenp) {
		APP_ERROR("Invalid fixed-link property length: %d", lenp);
		return -1;
	}

	duplex = fixed_link[1];
	speed = fixed_link[2];

	ret = fdt_node_check_compatible(devtree_ptr, mac_node,
					"fsl,fman-1g-mac");
	if (unlikely((ret < 0) || (ret > 1))) {
		APP_ERROR("fdt_node_check_compatible failed");
		return -1;
	}

	page_size = reg[3];
	pa = ((uint64_t)reg[0] << 32) | reg[1];

	if (0 > dev_mmap(pa, page_size, page_size,
			 TLB_MAS2_IO, TLB_MAS3_KERN, &va)) {
		APP_ERROR("enable_mac_port(): dev_mmap failed");
		return -1;
	}

	if (likely(0 == ret)) {
		if (true == is_promisc) {
			uint32_t rctrl;

			rctrl = fm_mac_in((struct fman_mac *)va, RCTRL_OFFSET);

			/*Enable promisc mode */
			rctrl = rctrl | RCTRL_MPROM | RCTRL_UPROM;
			fm_mac_out((struct fman_mac *)va, RCTRL_OFFSET, rctrl);
		}

		/*Set the speed and duplex*/
		maccfg = fm_mac_in((struct fman_mac *)va, MACCFG2_OFFSET);
		maccfg = maccfg & 0xFFFFF0FF;

		if (1000 == speed)
			maccfg = maccfg | 0x200;
		else
			maccfg = maccfg | 0x100;

		if (1 == duplex)
			maccfg = maccfg | 0x1;
		else
			maccfg = maccfg & ~(0x1);

		fm_mac_out((struct fman_mac *)va, MACCFG2_OFFSET, maccfg);

		/*Turn OFF the RMM bit*/
		ecntrl = fm_mac_in((struct fman_mac *)va, ECNTRL_OFFSET);
		ecntrl = ecntrl & 0xFFFFFFF7;
		fm_mac_out((struct fman_mac *)va, ECNTRL_OFFSET, ecntrl);

		/*Read the old value of MAC config register */
		maccfg = fm_mac_in((struct fman_mac *)va, MACCFG1_OFFSET);

		/*Enable TX, RX enable */
		maccfg = maccfg | MACCFG1_RX_EN | MACCFG1_TX_EN;

		fm_mac_out((struct fman_mac *)va, MACCFG1_OFFSET, maccfg);

		APP_INFO("Enabled 1G MAC (promisc = %u), port id: %u",
			 is_promisc, port_id);

		APP_INFO("Speed: %u Mbps, Duplex: %s",
				speed, (duplex ? "Full" : "Half"));

		if (0 > dev_unmap(va))
			printf("enable_mac_port(): dev_unmap failed");
		return 0;
	}

	/*Else, ret == 1 and it may be 10G ethernet */
	ret = fdt_node_check_compatible(devtree_ptr, mac_node,
					"fsl,fman-10g-mac");
	if (likely(0 == ret)) {
		maccfg = fm_mac_in((struct fman_mac *)va,
				   TGEC_CMD_CTRL_REG_OFFSET);

		if (true == is_promisc) {
			maccfg = maccfg | CMD_CFG_PROMIS_EN;
		}

		maccfg = maccfg | CMD_CFG_RX_EN | CMD_CFG_TX_EN;

		fm_mac_out((struct fman_mac *)va, TGEC_CMD_CTRL_REG_OFFSET,
			   maccfg);

		APP_INFO("Enabled 10G MAC (promisc = %u), port id: %u",
			 is_promisc, port_id);

		APP_INFO("Speed: %u Mbps, Duplex: %s",
				speed, (duplex ? "Full" : "Half"));

		if (0 > dev_unmap(va))
			printf("enable_mac_port(): dev_unmap failed");

		return 0;
	}

	if (0 > dev_unmap(va))
		printf("enable_mac_port(): dev_unmap failed");

	return -1;
}

int loopback_mac_port(const uint32_t port_id)
{
	uint32_t *reg;
	uint32_t page_size;
	uint32_t maccfg;
	uint32_t rctrl;

	int ret;
	int mac_node;
	phys_addr_t pa;
	void *va;

	mac_node = dpa_eth_ports[port_id].mac_node;

	reg = (uint32_t *) fdt_getprop(devtree_ptr, mac_node, "reg", NULL);
	if (NULL == reg) {
		APP_ERROR("fdt_getprop failed for reg");
		return -1;
	}

	ret = fdt_node_check_compatible(devtree_ptr, mac_node,
					"fsl,fman-1g-mac");
	if (unlikely((ret < 0) || (ret > 1))) {
		APP_ERROR("fdt_node_check_compatible failed");
		return -1;
	}

	page_size = reg[3];
	pa = ((uint64_t)reg[0]<<32) | reg[1];

	if (0 > dev_mmap(pa, page_size, page_size,
			 TLB_MAS2_IO, TLB_MAS3_KERN, &va)) {
		APP_ERROR("enable_mac_port(): dev_mmap failed");
		return -1;
	}

	if (likely(0 == ret)) {		/*Gigabit ethernet*/

		/*Enable promisc mode first*/
		rctrl = fm_mac_in((struct fman_mac *)va, RCTRL_OFFSET);
		rctrl = rctrl | RCTRL_MPROM | RCTRL_UPROM;
		fm_mac_out((struct fman_mac *)va, RCTRL_OFFSET, rctrl);

		/*Read the old value of MAC config register */
		maccfg = fm_mac_in((struct fman_mac *)va, MACCFG1_OFFSET);

		/*Enable loopback */
		maccfg = maccfg | MACCFG1_LOOPBACK;
		fm_mac_out((struct fman_mac *)va, MACCFG1_OFFSET, maccfg);

		APP_INFO("Enabled loopback on 1G MAC, port id: %u", port_id);

		if (0 > dev_unmap(va))
			printf("enable_mac_port(): dev_unmap failed");
		return 0;
	}

	/*Else, ret == 1 and it may be 10G ethernet */
	ret = fdt_node_check_compatible(devtree_ptr, mac_node,
					"fsl,fman-10g-mac");
	if (likely(0 == ret)) {
		maccfg = fm_mac_in((struct fman_mac *)va,
				   TGEC_CMD_CTRL_REG_OFFSET);

		maccfg = maccfg | CMD_CFG_PROMIS_EN;

		maccfg = maccfg | CMD_CFG_LOOPBACK_EN;

		fm_mac_out((struct fman_mac *)va, TGEC_CMD_CTRL_REG_OFFSET,
			   maccfg);

		APP_INFO("Enabled loopback on 10G MAC, port id: %u", port_id);

		if (0 > dev_unmap(va))
			printf("enable_mac_port(): dev_unmap failed");

		return 0;
	}

	if (0 > dev_unmap(va))
		printf("enable_mac_port(): dev_unmap failed");

	return -1;

}

int shell_cmd_loopback(int argc, char *argv[])
{
	uint32_t port_id;

	if (argc < 2) {
		printf("loopback <port_id>\n");
		return -1;
	}

	errno = 0;
	port_id = strtoul((const char *)argv[1], NULL, 0);
	if ((ERANGE == errno && ULONG_MAX == port_id) ||
				(0 != errno && 0 == port_id)) {
		printf("strtoul failed converting port_id\n");
		return -1;
	}

	if (port_id >= g_num_dpa_eth_ports) {
		printf("Invalid port ID: %u\n", port_id);
		return -1;
	}

	if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
		APP_INFO("Cannot enable loopback on mac-less port ID: %u\n",
			port_id);
		return -1;
	}

	if (0 > loopback_mac_port(port_id)) {
		printf("loopback_mac_port failed for port ID: %u\n", port_id);
		return -1;
	}

	printf("Enabled promiscous + loopback mode on Port ID: %u\n", port_id);

	return 0;
}

/**
\brief Disables a ethernet port

\details This function disables a given ethernet port.

\param[in] port_id - Port ID to be disabled.

\retval 0 - SUCCESS
\retval -1 - Failure
*/
static
int disable_mac_port(const uint32_t port_id,
		const bool disable_rx, const bool disable_tx)
{
	const uint32_t *reg;
	uint32_t page_size;
	uint32_t maccfg;
	int ret;
	int mac_node;
	phys_addr_t pa;
	void *va;
	uint32_t rctrl;
	int retval;
	int lenp;

	mac_node = dpa_eth_ports[port_id].mac_node;

	reg = (uint32_t *) fdt_getprop(devtree_ptr, mac_node, "reg", &lenp);
	if (unlikely(NULL == reg)) {
		APP_ERROR("fdt_getprop failed for reg");
		return -1;
	}

	if (unlikely((4 * sizeof(uint32_t)) != lenp)) {
		APP_ERROR("Invalid length of reg property: %d", lenp);
		return -1;
	}

	ret = fdt_node_check_compatible(devtree_ptr, mac_node,
					"fsl,fman-1g-mac");
	if (unlikely((ret < 0) || (ret > 1))) {
		APP_ERROR("fdt_node_check_compatible failed");
		return -1;
	}

	page_size = reg[3];
	pa = ((uint64_t)reg[0] << 32) | reg[1];

	if (unlikely(0 > dev_mmap(pa, page_size, page_size,
			 TLB_MAS2_IO, TLB_MAS3_KERN, &va))) {
		APP_ERROR("%s: dev_mmap failed", __func__);
		return -1;
	}

	if (likely(0 == ret)) {
		rctrl = fm_mac_in((const struct fman_mac *)va, RCTRL_OFFSET);

		/*Disable promisc mode*/
		rctrl = rctrl & ~RCTRL_MPROM & ~RCTRL_UPROM;
		fm_mac_out((struct fman_mac *)va, RCTRL_OFFSET, rctrl);

		/*Read the old value of MAC config register */
		maccfg = fm_mac_in((const struct fman_mac *)va, MACCFG1_OFFSET);

		if (true == disable_rx)
			maccfg = maccfg & ~MACCFG1_RX_EN;
		if (true == disable_tx)
			maccfg = maccfg & ~MACCFG1_TX_EN;

		fm_mac_out((struct fman_mac *)va, MACCFG1_OFFSET, maccfg);

		APP_INFO("Disabled 1G MAC port id: %u", port_id);

		retval = 0;
		goto unmap;
	}

	/*Else, ret == 1 and it may be 10G ethernet */
	ret = fdt_node_check_compatible(devtree_ptr, mac_node,
					"fsl,fman-10g-mac");
	if (likely(0 == ret)) {
		maccfg = fm_mac_in((const struct fman_mac *)va,
				TGEC_CMD_CTRL_REG_OFFSET);

		/*Disable promiscuous mode*/
		maccfg = maccfg & ~CMD_CFG_PROMIS_EN;

		if (true == disable_rx)
			maccfg = maccfg & ~CMD_CFG_RX_EN;

		if (true == disable_tx)
			maccfg = maccfg & ~CMD_CFG_TX_EN;

		fm_mac_out((struct fman_mac *)va,
				TGEC_CMD_CTRL_REG_OFFSET, maccfg);

		APP_INFO("Disabled 10G MAC port id: %u", port_id);

		retval = 0;
		goto unmap;
	}

	retval = -1;
unmap:
	if (unlikely(0 > dev_unmap(va)))
		printf("%s: dev_unmap failed", __func__);

	return retval;
}

/**
\brief  Reads buffer pools associated with a given port.

\details Reads buffer pools associated with a given port.

\param[in]	dpa_node - Device tree handle for port 
\param[in]	port_id - Port ID to read 

\return Negative - Failure
  	Non-Negative - Number of bpools discovered
*/
int dpa_eth_bpool_read(const int dpa_node, const uint32_t port_id)
{
	uint32_t index;
	uint32_t *bp_id;
	uint32_t *bpool_cfg;
	uint32_t size;
	int bpool_node;
	int bpool_len;

	/*Read the list of bpools for this port */
	const uint32_t *bpool_handle =
		(const uint32_t *)fdt_getprop(devtree_ptr, dpa_node,
						"fsl,bman-buffer-pools",
						&bpool_len);
	if (NULL == bpool_handle) {
		APP_ERROR("fdt_getprop failed for fsl,bman-buffer-pools");
		return -1;
	}

	for (index = 0; index < bpool_len / sizeof(uint32_t); index++) {
		bpool_node = fdt_node_offset_by_phandle(devtree_ptr,
							bpool_handle[index]);
		if (unlikely(0 > bpool_node)) {
			APP_ERROR
			    ("fdt_node_offset_by_phandle failed for bpools");
			return -1;
		}

		/*Read the buffer-pool ID */
		bp_id =
		    (uint32_t
		     *) (fdt_getprop(devtree_ptr, bpool_node, "fsl,bpid",
				     NULL));
		if (unlikely(NULL == bp_id)) {
			APP_ERROR("fdt_getprop failed for fsl,bpid");
			return -1;
		}

		bpool_cfg = (uint32_t *) fdt_getprop(devtree_ptr, bpool_node,
						     "fsl,bpool-cfg", NULL);
		if (unlikely(NULL == bpool_cfg)) {
			APP_ERROR
			("fdt_getprop failed for fsl,bpool-cfg for pool: %u",
			*bp_id);
			return -1;
		}

		dpa_eth_ports[port_id].bpid[index] = (uint8_t) (*bp_id);

		size = bpool_cfg[3];
		if (unlikely(0 == size)) {
			APP_ERROR
			    ("BPID: %u configured with 0 size buffers",
			     *bp_id);
			return -1;
		}

		/*Already not initialised */
		if (0 == dpa_bpools[*bp_id].size) {
			dpa_bpools[*bp_id].size = size;
			continue;
		}

		/*Already init */
		if (unlikely(size != dpa_bpools[*bp_id].size)) {
			APP_ERROR("BPID: %u already init with size: %u",
				  *bp_id, dpa_bpools[*bp_id].size);
			return -1;
		}
	}

	return index;		/*Number of bpools configured for this port */
}

/**
\brief  Reads a DPA port for FQs and bpools.

\details Reads a DPA port for FQs and bpools.

\param[in]	dpa_node - Device tree handle for port 
\param[in]	port_id - Port ID to read 

\return Negative - Failure
  	Positive - Channel map to receive frames on
*/
int dpa_eth_node_read(const int dpa_node, const uint32_t port_id)
{
	uint32_t *rx_frame_queue;
	uint32_t *tx_frame_queue;
	uint32_t *channel_phandle;
	uint32_t *rx_channel_id;
	uint32_t *tx_channel_id;
	uint32_t *fman_mac;
	uint32_t *port_handles;
	uint8_t *mac_addr;

	int channel_node;
	int mac_node;
	int tx_port_node;
	int mac_len;
	int num_bpools;

	num_bpools = dpa_eth_bpool_read(dpa_node, port_id);
	if (num_bpools <= 0) {
		APP_ERROR("dpa_eth_bpool_read failed");
		return -1;
	}

	APP_INFO("Number of bpools used: %d", num_bpools);

	/*Get ethernet MAC associated with DPA port */
	fman_mac = (uint32_t *) fdt_getprop(devtree_ptr, dpa_node,
					    "fsl,fman-mac", NULL);
	if (NULL == fman_mac) {
		APP_INFO("Port id: %u is MAC less ethernet device", port_id);

		dpa_eth_ports[port_id].type = MAC_LESS_ETH;

		/*Read the MAC address */
		mac_addr = (uint8_t *) fdt_getprop(devtree_ptr, dpa_node,
						   "local-mac-address",
						   &mac_len);

		mac_node = -1;	/*For warning removal */

	} else {
		dpa_eth_ports[port_id].type = MAC_ETH;

		mac_node = fdt_node_offset_by_phandle(devtree_ptr, *fman_mac);
		if (0 > mac_node) {
			APP_ERROR
			    ("fdt_node_offset_by_phandle failed for fman_mac");
			return -1;
		}

		dpa_eth_ports[port_id].mac_node = mac_node;

		/*Read the MAC address */
		mac_addr = (uint8_t *) fdt_getprop(devtree_ptr, mac_node,
						   "local-mac-address",
						   &mac_len);
	}

	if (NULL == mac_addr) {
		APP_ERROR("fdt_getprop failed for local-mac-address");
		return -1;
	}

	if (mac_len != ETH_ALEN) {
		APP_ERROR("Invalid MAC address length: %d", mac_len);
		return -1;
	}

	memcpy(&dpa_eth_ports[port_id].mac_addr[0], mac_addr, ETH_ALEN);

	APP_INFO("MAC address for port id: %u = " MAC_FMT,
		 port_id, NMAC_STR(mac_addr));

	/*Allocate frame RX queues here */
	rx_frame_queue = (uint32_t *) fdt_getprop(devtree_ptr, dpa_node,
						  "fsl,qman-frame-queues-rx",
						  NULL);
	if (NULL == rx_frame_queue) {
		APP_ERROR("fdt_getprop failed for fsl,qman-frame-queues-rx");
		return -1;
	}

	channel_phandle = (uint32_t *) fdt_getprop(devtree_ptr, dpa_node,
						   "fsl,qman-channel", NULL);
	if (NULL == channel_phandle) {
		APP_ERROR("fdt_getprop failed for fsl,qman-channel");
		return -1;
	}

	channel_node = fdt_node_offset_by_phandle(devtree_ptr,
						  *channel_phandle);
	if (0 > channel_node) {
		APP_ERROR
		    ("fdt_node_offset_by_phandle failed for channel_phandle");
		return -1;
	}

	rx_channel_id = (uint32_t *) fdt_getprop(devtree_ptr, channel_node,
						 "fsl,qman-channel-id", NULL);
	if (NULL == rx_channel_id) {
		APP_ERROR("fdt_getprop failed for fsl,qman-channel-id");
		return -1;
	}

	dpa_eth_ports[port_id].ingress.channel = *rx_channel_id;
	dpa_eth_ports[port_id].ingress.work_queue = 3;

	if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
		dpa_eth_ports[port_id].ingress.fq_start = rx_frame_queue[0];
		dpa_eth_ports[port_id].ingress.fq_count = rx_frame_queue[1];
	} else {
		dpa_eth_ports[port_id].ingress_err.fq_start = rx_frame_queue[0];
		dpa_eth_ports[port_id].ingress_err.fq_count = rx_frame_queue[1];

		dpa_eth_ports[port_id].ingress.fq_start = rx_frame_queue[2];
		dpa_eth_ports[port_id].ingress.fq_count = rx_frame_queue[3];

		dpa_eth_ports[port_id].pcd_ingress.fq_start = rx_frame_queue[4];
		dpa_eth_ports[port_id].pcd_ingress.fq_count = rx_frame_queue[5];

		dpa_eth_ports[port_id].pcd_ingress.channel = *rx_channel_id;
		/* Pre-CAAM WQ priority changed from 1 to 3 */
		dpa_eth_ports[port_id].pcd_ingress.work_queue = 3;

		dpa_eth_ports[port_id].ingress_err.channel = *rx_channel_id;
		dpa_eth_ports[port_id].ingress_err.work_queue = 1;
	}

	/*Read transmit frame queues */
	tx_frame_queue = (uint32_t *) fdt_getprop(devtree_ptr, dpa_node,
						  "fsl,qman-frame-queues-tx",
						  NULL);
	if (NULL == tx_frame_queue) {
		APP_ERROR("fdt_getprop failed for fsl,qman-frame-queues-tx");
		return -1;
	}

	if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
		dpa_eth_ports[port_id].egress.fq_start = tx_frame_queue[0];
		dpa_eth_ports[port_id].egress.fq_count = tx_frame_queue[1];

		/*Transmit FQs not needed to be initialized by MAC_LESS_ETH */
		APP_INFO("Egress FQs for MAC-less, Base: %u, Count: %u",
			 dpa_eth_ports[port_id].egress.fq_start,
			 dpa_eth_ports[port_id].egress.fq_count);

		return dpa_eth_ports[port_id].ingress.channel;
	} else {
		dpa_eth_ports[port_id].egress_err.fq_start = tx_frame_queue[0];
		dpa_eth_ports[port_id].egress_err.fq_count = tx_frame_queue[1];

		dpa_eth_ports[port_id].egress_confirm.fq_start =
		    tx_frame_queue[2];
		dpa_eth_ports[port_id].egress_confirm.fq_count =
		    tx_frame_queue[3];

		dpa_eth_ports[port_id].egress.fq_start = tx_frame_queue[4];
		dpa_eth_ports[port_id].egress.fq_count = tx_frame_queue[5];
	}

	port_handles = (uint32_t *) fdt_getprop(devtree_ptr, mac_node,
						"fsl,port-handles", NULL);
	if (NULL == port_handles) {
		APP_ERROR("fdt_getprop failed for fsl,port-handles");
		return -1;
	}

	/*Reading TX port handle for egress frame queue */
	tx_port_node = fdt_node_offset_by_phandle(devtree_ptr, port_handles[1]);
	if (0 > tx_port_node) {
		APP_ERROR
		    ("fdt_node_offset_by_phandle failed for tx port_handle");
		return -1;
	}

	tx_channel_id = (uint32_t *) fdt_getprop(devtree_ptr, tx_port_node,
						 "fsl,qman-channel-id", NULL);
	if (NULL == tx_channel_id) {
		APP_ERROR("fdt_getprop failed for fsl,qman-channel-id");
		return -1;
	}

	/*Port TX frame queues */
	dpa_eth_ports[port_id].egress.channel = *tx_channel_id;
	dpa_eth_ports[port_id].egress.work_queue = 1;	/*Check later */

	/*Port TX confirm frame queues */
	dpa_eth_ports[port_id].egress_confirm.channel = *rx_channel_id;
	dpa_eth_ports[port_id].egress_confirm.work_queue = 1;	/*Check later */

	/*Port TX err frame queues */
	dpa_eth_ports[port_id].egress_err.channel = *rx_channel_id;
	dpa_eth_ports[port_id].egress_err.work_queue = 1;	/*Check later */

	return dpa_eth_ports[port_id].ingress.channel;
}

/**
\brief Reads the frame queues as needed by DPA

\details Creates ingress and egress frame queues

\return Negative - Failure
  	Non-Negative - Success, number of ports initialised

\note   Called only once from only one partition (i.e. control partition)
*/
int dpa_eth_read(uint32_t * recv_channel_map)
{
	int dpa_offset;
	uint32_t port_id;
	uint32_t num_enabled_ports;
	int32_t recv_channel_id;
	const char *status;
	int lenp;

	memset(&dpa_eth_ports, 0, sizeof(dpa_eth_ports));
	memset(&dpa_bpools, 0, sizeof(dpa_bpools));

	/*Now read all the DPA nodes */
	dpa_offset = -1;
	*recv_channel_map = 0;
	num_enabled_ports = 0;
	for (port_id = 0; port_id < MAX_NUM_PORTS; port_id++) {
		dpa_offset = fdt_node_offset_by_compatible(devtree_ptr,
							   dpa_offset,
							   "fsl,dpa-ethernet");
		if (0 > dpa_offset) {
			/*Not even one LWEDPA node not present */
			if (0 == port_id) {
				APP_ERROR("fdt_node_offset_by_compatible \
					failed for fsl,dpa-ethernet");
				return -1;
			} else {
				/*No more DPA nodes present */
				break;
			}
		}

		status = (const char *)
			fdt_getprop(devtree_ptr, dpa_offset, "status", &lenp);
		if (NULL == status) {
			if (-FDT_ERR_NOTFOUND != lenp) {
				APP_WARN("fdt_getprop failed: %s",
 fdt_strerror(lenp));
				return -1;
			}

		} else {
			if (0 == strcmp(status, "disabled"))
				APP_INFO("Skipping disabled port");
			else
				printf("Unexpected port state: %s", status);
			continue;
		}

		recv_channel_id = dpa_eth_node_read(dpa_offset,
						num_enabled_ports);
		if (0 > recv_channel_id) {
			APP_ERROR("init_dpa_frame_queues failed");
			return -1;
		}

		if (recv_channel_id >= qm_channel_swportal0 &&
		    recv_channel_id <= qm_channel_swportal9) {
			recv_channel_id = 0;	/*Dedicated channel case */
		} else if (recv_channel_id >= qm_channel_pool1 &&
			   recv_channel_id <= qm_channel_pool15) {
			/*Pool channels start from 1 */
			recv_channel_id = recv_channel_id -
			    qm_channel_pool1 + 1;
		} else {
			APP_ERROR("Invalid recv channel number: %d",
				  recv_channel_id);
			return -1;
		}

		*recv_channel_map = *recv_channel_map |
		    QM_SDQCR_CHANNELS_POOL(recv_channel_id);

		num_enabled_ports++;
	}

	if (num_enabled_ports == MAX_NUM_PORTS)
		printf("No more space left for more ports !!!");

	APP_INFO("Number of ports enabled : %u", num_enabled_ports);

	g_num_dpa_eth_ports = num_enabled_ports; /*Number of ports enabled */

	return g_num_dpa_eth_ports;
}

/**
\brief Creates the frame queues associated with DPA ethernet ports

\details Memory is allocated for FQ structures and FQs are created.

\param[in]	rx_default_cb - Callback function for default Rx FQs
\param[in]	rx_pcd_cb - Callback function for PCD Rx FQs
\param[in]	rx_err_cb - Callback function for RX err FQs
\param[in]	tx_cb - Callback function for Tx FQs
\param[in]	tx_confirm_cb - Callback function for Tx confirm FQs
\param[in]	tx_err_cb - Callback function for Tx err FQs
\param[in]	priv_data_size - Private data size appended to each FQ 
				structure.

\retval 0 - SUCCESS
\retval -1 - Failure

\note   Called only once from all core of partition
*/
int dpa_eth_fq_create(struct qman_fq_cb *rx_default_cb,
		      struct qman_fq_cb *rx_pcd_cb,
		      struct qman_fq_cb *rx_err_cb,
		      struct qman_fq_cb *tx_cb,
		      struct qman_fq_cb *tx_confirm_cb,
		      struct qman_fq_cb *tx_err_cb, uint32_t priv_data_size)
{
	uint32_t port_id;
	uint32_t flags;
	int ret;

	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		APP_INFO("Creating FQs for port id: %u", port_id);

		flags = QMAN_FQ_FLAG_LOCKED | QMAN_FQ_FLAG_TO_DCPORTAL;

		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type)
			flags = flags | QMAN_FQ_FLAG_NO_MODIFY;

		ret = create_fqs(&dpa_eth_ports[port_id].egress, flags,
				 tx_cb, "Tx default", priv_data_size);
		if (unlikely(0 != ret)) {
			APP_ERROR("create_fqs failed for Tx");
			return -1;
		}

		flags = QMAN_FQ_FLAG_NO_ENQUEUE | QMAN_FQ_FLAG_LOCKED;
		ret = create_fqs(&dpa_eth_ports[port_id].ingress, flags,
				 rx_default_cb, "Rx default", priv_data_size);
		if (unlikely(0 != ret)) {
			APP_ERROR("create_fqs failed for Rx default");
			return -1;
		}

		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
			APP_INFO("Skip for MAC-less port: %u", port_id);
			continue;
		}

		ret = create_fqs(&dpa_eth_ports[port_id].ingress_err, flags,
				 rx_err_cb, "Rx err", priv_data_size);
		if (unlikely(0 != ret)) {
			APP_ERROR("create_fqs failed for Rx err");
			return -1;
		}

		ret = create_fqs(&dpa_eth_ports[port_id].pcd_ingress, flags,
				 rx_pcd_cb, "Rx PCD", priv_data_size);
		if (unlikely(0 != ret)) {
			APP_ERROR("create_fqs failed for Rx PCD");
			return -1;
		}

		ret = create_fqs(&dpa_eth_ports[port_id].egress_err, flags,
				 tx_err_cb, "Tx err", priv_data_size);
		if (unlikely(0 != ret)) {
			APP_ERROR("create_fqs failed for Tx err");
			return -1;
		}

		ret = create_fqs(&dpa_eth_ports[port_id].egress_confirm, flags,
				 tx_confirm_cb, "Tx confirm", priv_data_size);
		if (unlikely(0 != ret)) {
			APP_ERROR("create_fqs failed Tx confirm");
			return -1;
		}
	}

	return 0;
}

/**
\brief Initialises the frame queues associated with DPA ethernet ports

\details RX queues are initialised with stashing ON.
	 TX queues are initialised without stashing.
	 For interpartition ethernet only, RX queues are initialised. 

\param[in]	data_stash_size - Stashing size for frame data
\param[in]	ann_stash_size - Stashing size for frame annotation
\param[in]	ctx_stash_size - Stashing size for FQ context

\retval 0 - SUCCESS
\retval -1 - Failure

\note   Called only once from only one partition (i.e. control partition)
*/
int dpa_eth_fq_init(uint32_t data_stash_size,
		uint32_t ann_stash_size, uint32_t ctx_stash_size,
		struct qman_orp_pcd *pcd_orp, struct td_param *pcd_td)
{
	uint32_t port_id;
	uint32_t flags;

	uint32_t ctx_lines;
	uint32_t ann_lines;
	uint32_t data_lines;
	uint32_t ctx_a_excl;

	struct qm_mcc_initfq opts;

	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		APP_INFO("Initializing FQs for port id: %u", port_id);

		flags = QMAN_INITFQ_FLAG_SCHED;
		opts.we_mask = QM_INITFQ_WE_DESTWQ |
		    QM_INITFQ_WE_CONTEXTA | QM_INITFQ_WE_FQCTRL;
		opts.fqd.fq_ctrl =
			 QM_FQCTRL_CTXASTASHING | QM_FQCTRL_LOCKINCACHE |
				QM_FQCTRL_CPCSTASH;
		if (NULL != pcd_orp) {
			opts.we_mask |= QM_INITFQ_WE_ORPC ;
			opts.fqd.fq_ctrl |= QM_FQCTRL_ORP; 
			opts.fqd.orprws = pcd_orp->orprws;
			opts.fqd.oa = pcd_orp->oa;
			opts.fqd.olws = pcd_orp->olws;
		} else {
#ifdef HOT_STANDBY_ENABLE
			opts.fqd.fq_ctrl |= QM_FQCTRL_AVOIDBLOCK;
#endif
		}

		ctx_a_excl = 
			QM_STASHING_EXCL_ANNOTATION | QM_STASHING_EXCL_DATA;

		ctx_lines = GET_STASH_LINES(ctx_stash_size);
		data_lines = GET_STASH_LINES(data_stash_size);
		ann_lines = GET_STASH_LINES(ann_stash_size);

		opts.fqd.context_a.stashing.exclusive = ctx_a_excl;
		opts.fqd.context_a.stashing.annotation_cl = ann_lines;
		opts.fqd.context_a.stashing.data_cl = data_lines;
		opts.fqd.context_a.stashing.context_cl = ctx_lines;

		/* No need to set the address in context_a.
		 * This is set by driver itself */
		if (0 != init_fqs(&dpa_eth_ports[port_id].ingress, flags,
				  &opts, "Rx default")) {
			APP_ERROR("init_fqs failed for Rx default FQs");
			return -1;
		}

		/*For MAC-less ports, no other FQ needs to be initialized */
		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
			APP_INFO("Skip FQ-init for MAC-less port: %u",
				 port_id);
			continue;
		}

		/*Check if tail drop is required on PCD FQs*/
		if (NULL != pcd_td) {
			opts.fqd.fq_ctrl |= pcd_td->flag;
			if (pcd_td->flag == QM_FQCTRL_TDE) {
				opts.we_mask |= QM_INITFQ_WE_TDTHRESH;
				opts.fqd.td.exp = pcd_td->exp;
				opts.fqd.td.mant = pcd_td->mant;
				APP_INFO("Init PCD FQs with TD mant = %hu exp = %hu", \
					pcd_td->mant, pcd_td->exp);
			} else if (pcd_td->flag == QM_FQCTRL_CGE) {
				opts.we_mask |= QM_INITFQ_WE_CGID;
				/* Currently hardcoding all ingress pcd FQs to single CGR */
				opts.fqd.cgid = pcd_td->cgr_id;
				APP_INFO("Enabling Congestion Group %x", opts.fqd.cgid);
			}
		}

		opts.fqd.fq_ctrl |= QM_FQCTRL_HOLDACTIVE;
		if (0 != init_fqs(&dpa_eth_ports[port_id].pcd_ingress, flags,
				  &opts, "Rx PCD")) {
			APP_ERROR("init_fqs failed for Rx PCD FQs");
			return -1;
		}

		/*We do not need stashing for following FQs */
		opts.fqd.fq_ctrl = 0;//QM_FQCTRL_AVOIDBLOCK;
#ifdef HOT_STANDBY_ENABLE
		opts.fqd.fq_ctrl = QM_FQCTRL_AVOIDBLOCK;
#endif
		opts.we_mask = QM_INITFQ_WE_DESTWQ;
		if (0 != init_fqs(&dpa_eth_ports[port_id].ingress_err, flags,
				  &opts, "Rx err")) {
			APP_ERROR("init_fqs failed for Rx err FQs");
			return -1;
		}

		if (0 != init_fqs(&dpa_eth_ports[port_id].egress_err, flags,
				  &opts, "Tx err")) {
			APP_ERROR("init_fqs failed for Tx err FQs");
			return -1;
		}

		if (0 != init_fqs(&dpa_eth_ports[port_id].egress_confirm, flags,
				  &opts, "Tx confirm")) {
			APP_ERROR("init_fqs failed for Tx confirm FQs");
			return -1;
		}

		opts.fqd.fq_ctrl = QM_FQCTRL_PREFERINCACHE;
		opts.we_mask = QM_INITFQ_WE_DESTWQ | QM_INITFQ_WE_CONTEXTA |
		    QM_INITFQ_WE_CONTEXTB;
		opts.fqd.context_a.hi = 0x80000000;
		opts.fqd.context_a.lo = 0x0;
		opts.fqd.context_b = 0;	/*By default no confirmation needed */
		if (0 != init_fqs(&dpa_eth_ports[port_id].egress, flags,
				  &opts, "Tx")) {
			APP_ERROR("init_fqs failed for Tx FQs");
			return -1;
		}
	}

	return 0;
}

/**
\brief Retires the Rx FQs associated with each port
\retval +ve - Success, number of FQs which are not marked OOS.
\retval -1 - Failure
*/
int dpa_eth_rx_fq_retire(void)
{
	uint32_t port_id;
	int rem = 0;
	int ret;

	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		APP_INFO("Retiring Rx FQs for port : %u", port_id);

		ret = retire_fqs(&dpa_eth_ports[port_id].ingress);
		if (0 > ret) {
			APP_ERROR("retire_fqs failed for default Rx FQs");
			return -1;
		}

		rem = rem + ret;

		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type)
			continue;

		ret = retire_fqs(&dpa_eth_ports[port_id].pcd_ingress);
		if (0 > ret) {
			APP_ERROR("retire_fqs failed for PCD Rx FQs");
			return -1;
		}

		rem = rem + ret;

		ret = retire_fqs(&dpa_eth_ports[port_id].ingress_err);
		if (0 > ret) {
			APP_ERROR("retire_fqs failed for Rx err FQs");
			return -1;
		}

		rem = rem + ret;
	}

	return rem;
}

/**
\brief Retires the Tx FQs associated with each port
\retval 0 - SUCCESS
\retval -1 - Failure
*/
int dpa_eth_tx_fq_retire()
{
	uint32_t port_id;
	int rem = 0;
	int ret;

	/*Here we cannot retire Tx FQ for MAC-less port*/
	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		APP_INFO("Retiring Tx FQs for port : %u", port_id);

		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
			APP_INFO("Destroying Tx FQs of MAC-less port");
			destroy_fqs(&dpa_eth_ports[port_id].egress);
			continue;
		}

		ret = retire_fqs(&dpa_eth_ports[port_id].egress);
		if (0 > ret) {
			APP_ERROR("retire_fqs failed for Tx default FQs");
			return -1;
		}

		rem = rem + ret;

		ret = retire_fqs(&dpa_eth_ports[port_id].egress_err);
		if (0 > ret) {
			APP_ERROR("retire_fqs failed for Tx err FQs");
			return -1;
		}

		rem = rem + ret;

		ret = retire_fqs(&dpa_eth_ports[port_id].egress_confirm);
		if (0 > ret) {
			APP_ERROR("retire_fqs failed for Tx confirm FQs");
			return -1;
		}

		rem = rem + ret;
	}

	return rem;
}


/**
\brief Enables all ethernet ports

\details Ethernet ports are disabled when LWE partition comes up.
	This function enables them so that traffic can be sent or received.

\retval 0 - SUCCESS
\retval -1 - Failure

\note   Called only once from only one partition (i.e. control partition)
*/
int dpa_eth_enable_ports(const bool is_promisc)
{
	uint32_t port_id;

	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
			APP_INFO("Skipping MAC enable for MAC-less port");
			continue;
		}

		if (unlikely(0 != enable_mac_port(port_id, is_promisc))) {
			APP_ERROR("enable_mac_port failed for port_id: %u",
				  port_id);
			return -1;
		}
	}

	return 0;
}

/**
\brief Disables all ethernet ports

\details This function disables ethernet ports so that traffic from them stops.

\retval 0 - SUCCESS
\retval -1 - Failure
*/
int dpa_eth_disable_ports(const bool disable_rx, const bool disable_tx)
{
	uint32_t port_id;

	assert((true == disable_rx) || (true == disable_tx));

	for (port_id = 0; port_id < g_num_dpa_eth_ports; port_id++) {
		if (MAC_LESS_ETH == dpa_eth_ports[port_id].type) {
			APP_INFO("Skip disable MAC-less port_id: %u", port_id);
			continue;
		}

		if (unlikely(0 != disable_mac_port(port_id, disable_rx,
								disable_tx))) {
			APP_ERROR("Disabling mac port failed for port_id: %u",
				  port_id);
			return -1;
		}
	}

	return 0;
}
