/**
 \file dpa_qman.c
 \brief Contains routines for reading/creating and delelting qman FQ.
 */
/*
 * Copyright (C) 2009-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <app_libfdt.h>
#include <misc.h>
#include <compat.h>
#include <fsl_qman.h>
#include <mmap.h>
#include <cache.h>
#include <app_common.h>
#include <dpa/dpa_qman.h>

/**
\brief  Frame queues initialisation function

\details Inits a number of frame queues as requested.

\return -1 - Failure
  	0 - Success
*/

int init_fqs(struct dpa_fq_range_t *dpa_fq,
	     uint32_t flags, struct qm_mcc_initfq *opts, const char *fq_type)
{
	uint32_t count;
	uint32_t fq_id;

	for (count = 0; count < dpa_fq->fq_count; count++) {

		opts->fqd.dest.channel = (uint16_t)dpa_fq->channel;
		opts->fqd.dest.wq = (uint16_t)dpa_fq->work_queue;

		if (unlikely(0 != qman_init_fq(dpa_fq->fq[count], flags, opts))) {
			fq_id = dpa_fq->fq_start + count;
			APP_ERROR("qm_init_fq failed for fq_id: %u", fq_id);
			return -1;
		}
	}

	APP_INFO("Init %s FQs Base: %u, Count: %u, "
		 "Channel: %u, WQ: %u", fq_type,
		 dpa_fq->fq_start, dpa_fq->fq_count,
		 dpa_fq->channel, dpa_fq->work_queue);

	return 0;
}

/**
\brief  Frame queues creation function

\details Creates a number of frame queues as requested.

\return Negative - Failure
  	Non-Negative - Success
*/
int create_fqs(struct dpa_fq_range_t *fq_range, const uint32_t flags,
		struct qman_fq_cb *cb, const char *fq_type,
		uint32_t priv_data_size)
{
	struct qman_fq *fq;
	uint32_t fq_id;
	uint32_t count;

	fq_id = fq_range->fq_start;

	for (count = 0; count < fq_range->fq_count; count++) {
#ifndef HOT_STANDBY_ENABLE
		fq = (struct qman_fq *)memalign(CACHE_LINE_SIZE, sizeof(struct qman_fq) +
					      priv_data_size);
		if (unlikely(NULL == fq)) {
			APP_ERROR("malloc failed in create_fqs for FQ ID: %u",
			     fq_id);
			return -1;
		}
#else
		fq = fq_range->fq[count];
#endif
		fq->cb = *cb;

		if (unlikely(0 != qman_create_fq(fq_id, flags, fq))) {
			APP_ERROR("qman_create_fq failed for FQ ID: %u",
					fq_id);
			return -1;
		}

		fq_range->fq[count] = fq;

		fq_id++;
	}

	APP_INFO("Created %s Base: %u, Count: %u", fq_type,
		 fq_range->fq_start, fq_range->fq_count);

	return 0;
}

/**
 \brief Destroys a FQ and free up associated memory
 */
void destroy_fqs(const struct dpa_fq_range_t *fq_range)
{
	uint32_t count;

	for (count = 0; count < fq_range->fq_count; count++) {
		qman_destroy_fq(fq_range->fq[count], 0);
		free(fq_range->fq[count]);
	}
}

/**
 \brief Retires the frame queues.
 \details If FQ is retired immediately, it is marked OOS and destroyed.
	If retirement is pending, caller needs to move FQ in OOS and free
	it on receiving FQRN.
 \param[in] fq_range - FQ range to free
 \return Negative - Failure
	Non-Negative - Number of FQs which are not marked OOS
 */
int retire_fqs(const struct dpa_fq_range_t *fq_range)
{
	uint32_t fqid;
	uint32_t count;
	uint32_t flags;
	int ret;
	struct qman_fq *fq;

	for (count = 0; count < fq_range->fq_count; count++) {
		fqid = fq_range->fq_start + count;
		fq = fq_range->fq[count];

		ret = qman_retire_fq(fq, &flags);
		if (0 > ret) {
			APP_ERROR("qman_retire_fq failed for FQID: %u", fqid);
			return -1;
		}
	}

	return count;
}

/**
 \brief Extract qman pool channel offset.
 \param[in] const void * Device tree Pointer
 \return Negative - Failure
  	 Non-Negative - Success, denotes pool channel ID
 */
int32_t extract_qman_pool_channel(const void *devtree_ptr)
{
	uint32_t *pool_channel;
	uint32_t *pool_channel_id, pool_offset;
	int node = -1;

	int pool_channel_node;

	if ((node =
	     fdt_node_offset_by_compatible(devtree_ptr, node,
					   "fsl,qman-portal")) < 0) {
		APP_ERROR("Without dpa: fdt_node_offset_by_compatible failed");
		return -1;
	}

	pool_channel = (uint32_t *) fdt_getprop(devtree_ptr, node,
						"fsl,qman-pool-channels", NULL);
	if (NULL == pool_channel) {
		APP_ERROR("fdt_getprop failed for fsl,qman-pool-channels");
		return -1;
	}

	pool_channel_node = fdt_node_offset_by_phandle(devtree_ptr,
							*pool_channel);
	if (0 > pool_channel_node) {
		APP_ERROR("fdt_node_offset_by_phandle failed for pool_channel");
		return -1;
	}

	pool_channel_id =
	    (uint32_t *) fdt_getprop(devtree_ptr, pool_channel_node,
				     "fsl,qman-channel-id", NULL);
	if (NULL == pool_channel_id) {
		APP_ERROR("fdt_getprop failed for fsl, pool qman-channel-id");
		return -1;
	}

	pool_offset = *pool_channel_id;

	if (pool_offset <= qm_channel_swportal9) {
		pool_offset = 0;	/*Dedicated channel case */
	} else if (pool_offset >= qm_channel_pool1
		   && pool_offset <= qm_channel_pool15) {
		/*Pool channels start from 1 */
		pool_offset = pool_offset - qm_channel_pool1 + 1;
	} else {
		APP_ERROR("Invalid pool channel number: %d", pool_offset);
		return -1;
	}

	return pool_offset;
}

/**
 \brief Initialize Congestion Group Record
 \param[in] Congestion parameters for Congestion group initialization
 \param[in] Congestion Group ID
 \param[out] Congestion group Record
 */
struct qman_cgr *init_cgr(struct td_param *cgr_param)
{
	struct qm_mcc_initcgr cgr_state;
	struct qman_cgr *cgr = NULL;
	uint32_t res = 0;

	cgr = (struct qman_cgr *)memalign(CACHE_LINE_SIZE, sizeof(struct qman_cgr));
	if(cgr == NULL) {
		APP_ERROR("%s: Memory for CGR not available", __func__);
		return NULL;
	}
	memset(cgr, 0, sizeof(struct qman_cgr));

	cgr->cgrid = cgr_param->cgr_id;
	memset(&cgr_state, 0, sizeof(struct qm_mcc_initcgr));
	APP_INFO("%s: Congestion Record ID being created is %x", __func__,\
		 cgr_param->cgr_id);
	cgr_state.we_mask = QM_CGR_WE_CS_THRES | QM_CGR_WE_CSTD_EN |
			 QM_CGR_WE_MODE;
#ifdef CGR_NOTIFY_REQUIRED
	cgr->cb = cgr_param->cgr_cb;
	cgr_state.we_mask |= QM_CGR_WE_CSCN_EN;
	cgr_state.cgr.cscn_en = 1;
#endif
	cgr_state.cgr.mode = QMAN_CGR_MODE_FRAME;
	cgr_state.cgr.cstd_en = 1;
	cgr_state.cgr.cs_thres.TA = cgr_param->mant;
	cgr_state.cgr.cs_thres.Tn = cgr_param->exp;
	res = qman_create_cgr(cgr, QMAN_CGR_FLAG_USE_INIT, &cgr_state);
	if (res) {
		APP_ERROR("%s: Error Creating CGR", __func__);
		free(cgr);
		return NULL;
	}

	return cgr;
}
