/**
 \file dpa_qman.h
 \brief Header file defining structure of DPA Qman nodes.
 */
/*
 * Copyright (C) 2009, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __DPA_QMAN_H__
#define __DPA_QMAN_H__

#include <stdbool.h>
#include <fsl_qman.h>

#ifdef __cplusplus
	extern "C" {
#endif

#define MAX_NUM_FQ	1024

/**
 \brief DPA FQ Range Info
*/
struct dpa_fq_range_t {
	uint32_t fq_start;		/**< Starting FQ ID */
	uint32_t work_queue;		/**< Work Queue */
	uint32_t channel;		/**< Channel */
	uint32_t fq_count;		/**< FQ Count */

	struct qman_fq *fq[MAX_NUM_FQ];	/**< Array of QMan FQ structure */
};

struct qman_orp_pcd {
	uint8_t reserved:2;
	uint8_t orprws:3; /**< Order Reservation Window Size 32*2^orprws */
	uint8_t oa:1; /**< Autoadvance of NESN Enabled */
	/**< ORWS Values
	0 = Disabled. Late arrivals always rejected.
	1 = Window size is 32 frames.
	2 = Window size is the same as the ORP restoration
		window size configured in the ORPRWS field.
	3 = Window size is 8192 frames. Late arrivals always accepted.
	*/
	uint8_t olws:2;
	uint8_t reserved2[3];
};

/**
 \brief  Structure type for passing tail drop params
 \detail The number of bytes of data that can be in FQ is: TD_MANT * 2^TD_EXP.
         Also total number of bytes should be below: 0xE0000000
*/
struct td_param {
	uint16_t __reserved1:3;
	uint16_t mant:8;
	uint16_t exp:5;
	uint16_t flag;
	uint8_t cgr_id;
	qman_cb_cgr cgr_cb;
} __packed;

/**
 \brief Extract qman pool channel.
        Reads the device tree for Qman channel information. 
 \param[in] const void * Device tree Pointer
 \return Negative - Failure
  	 Non-Negative - Success, denotes pool channel ID
 \note Called from every lwe partition for their own pool channel information
 */
int32_t extract_qman_pool_channel(const void *devtree_ptr);

int init_fqs(struct dpa_fq_range_t *dpa_fq,
	     uint32_t flags, struct qm_mcc_initfq *opts, const char *fq_type);

int create_fqs(struct dpa_fq_range_t *fq_range, const uint32_t flags,
	       struct qman_fq_cb *cb, const char *fq_type,
	       uint32_t priv_data_size);

int retire_fqs(const struct dpa_fq_range_t *fq_range);

void destroy_fqs(const struct dpa_fq_range_t *fq_range);

/**
 \brief Initialize Congestion Group Record
 \param[in] Congestion parameters for Congestion group initialization
 \param[out] Congestion group Record
 */
struct qman_cgr *init_cgr(struct td_param *cgr_param);
#ifdef __cplusplus
	}
#endif

#endif
