/**
 \file ip_forward.c
 \brief Implements forwarding function if routelookup is successful
 */
/*
 * Copyright (C) 2007-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdint.h>
#include "app_common.h"
#include <hv_misc.h>
#include "ip/ip_common.h"
#include "ip/ip_context.h"
#include "ip/ip_forward.h"
#include "ip/ip_output.h"
#include "ip/ip_hooks.h"
#include "net/rt.h"
#include "net/net_dev.h"
#include "net/annotations.h"

extern struct dpa_buff_allocator *buff_allocator;
enum IP_STATUS ip_forward(struct ip_context_t *ctxt,
			  struct annotations_t *notes,
			  struct ip_header_t *ip_hdr)
{
	struct net_dev_t *dev;

	markpoint(11);
	dev = notes->dest->dev;
	if (likely(ip_hdr->ttl > 1)) {
		ip_hdr->ttl -= 1;
		ip_hdr->hdr_chksum += 0x100;
		if (unlikely((ip_hdr->hdr_chksum & 0xff00) == 0))
			ip_hdr->hdr_chksum += 0x1;

	} else {
		/* send_icmp(ip_hdr, ICMP_MSG_TIME_EXCEEDED); */
		decorated_notify_inc_64(&(ctxt->stats->ip_ttl_time_exceeded));
		lwe_free_fd(buff_allocator, notes->fd);
		return IP_STATUS_DROP;
	}

	/* If we have not dropped yet, and if the interface MTU is smaller than
	   the frame, and we can't fragment, drop the frame.  This calculation
	   is uneccessarily in the "DROP" path if the above tests fail - should
	   not do it if status is not ACCEPT.
	 */
	if (unlikely(notes->fd->length20 > dev->mtu) &&
	    unlikely(ip_hdr->frag.bits.df == 1)) {
		APP_ERROR("%s: Dropping pkt, mtu exceeded|fragmented",
			  __func__);
		decorated_notify_inc_64(&
					(ctxt->stats->
					 ip_xmit_icmp_unreach_need_frag));
		lwe_free_fd(buff_allocator, notes->fd);
		return IP_STATUS_DROP;
	}

	/* If we have not dropped it yet, and source == dest, send redirect */
	if (unlikely(dev->ifindex == notes->parse.port_id)) {
		/* send_icmp(ip_hdr, ICMP_MSG_REDIRECT); */
		decorated_notify_inc_64(&
					(ctxt->stats->
					 ip_xmit_icmp_redir_in_eq_out));
	}

	/* If we have not dropped yet, execute all of the forwarding hooks */
	return exec_hook(ctxt->hooks, IP_HOOK_FORWARD, ctxt, notes,
			 ip_hdr, &ip_forward_finish, SOURCE_POST_FMAN);
}

enum IP_STATUS ip_forward_finish(struct ip_context_t *ctxt,
				 struct annotations_t *notes,
				 struct ip_header_t *ip_hdr)
{
	markpoint(12);
	decorated_notify_inc_32(&(ctxt->stats->ip_out_forward));
	return ip_send(ctxt, notes, ip_hdr);
}
