/**
 \file ip_route.c
 \brief IPv4 Route lookup is done for forwarding/encap/decap decision.
 */
/*
 * Copyright (C) 2007-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <stdbool.h>
#include "app_common.h"
#include "statistics.h"
#include "ip/ip_route.h"
#include "ip/ip_forward.h"
#include "ip/ip_local.h"
#include "ip/ip_rc.h"
#include "ip/ip_protos.h"
#include "net/rt.h"
#include "net/annotations.h"
#include "net/frame_handler.h"
#include "ipsec/ipsec_encap.h"
#include "ipsec/ipsec_decap.h"
#include "pme/ipsa_pme.h"

enum IP_STATUS ip_route_input(struct ip_context_t *ctxt,
			      struct annotations_t *notes,
			      struct ip_header_t *ip_hdr, enum state source)
{
	enum IP_STATUS retval = IP_STATUS_DROP;
	struct ipsec_esp_hdr_t *esp_hdr;
	struct ipsec_context_t *ipsec_ctxt;
	markpoint(9);

	switch (source) {
	case SOURCE_POST_FMAN:
	{
		struct rc_entry_t *entry;
		if ((ip_hdr->proto == IP_PROTO_ESP)) {
			/* Pkt may be for decryption */
			retval = ipsec_decap_send(ctxt, notes, ip_hdr);

			if (retval != IP_STATUS_HOLD)
				return retval;
		}
		entry = rc_entry_fast_lookup(ctxt->rc,
					ip_hdr->src_addr.word,
					ip_hdr->dst_addr.word,
					ip_hdr->tos,
					RC_BUCKET_INDEX(notes));
		APP_DEBUG("Hash index= %x", RC_BUCKET_INDEX(notes));

		 /* needed for IPSec */
		if (entry == NULL) {
			entry = rc_entry_lookup(ctxt->rc,
					ip_hdr->src_addr.word,
					ip_hdr->dst_addr.word,
					ip_hdr->tos);
			if (entry == NULL) {
				APP_DEBUG("Fast Lookup Failed, going slow \
				   for Src = 0x%x; Dest = 0x%x; TOS = 0x%x",
				   ip_hdr->src_addr.word,
				   ip_hdr->dst_addr.word, ip_hdr->tos);
				retval =
					ip_route_input_slow(ctxt, notes, ip_hdr);
				return retval;
			}
		}

		notes->dest = entry->dest;
		decorated_notify_inc_64(&(entry->stats->hits));
		retval = ip_route_finish(ctxt, notes, ip_hdr);
	}
		break;
	case SOURCE_POST_ENCAP:
	{
		struct ipsec_tunnel_t *entry;
		esp_hdr =
			(struct ipsec_esp_hdr_t *)((uint32_t) ip_hdr +
					   sizeof(struct ip_header_t));
		APP_DEBUG("%s: Pkt src = %x, dst = %x and spi = %x", __func__,
		ip_hdr->src_addr.word, ip_hdr->dst_addr.word, esp_hdr->spi);
		ipsec_ctxt = (struct ipsec_context_t *)ctxt;
		retval = ipsec_lookup_tunnel_entry(&entry,
			(struct ipsec_tunnel_table_t *)
			ipsec_ctxt->itt,
			ip_hdr->src_addr.word,
			ip_hdr->dst_addr.word,
			esp_hdr->spi);
		if (unlikely(retval == false)) {
			APP_ERROR("%s: Couldn't find tunnel", __func__);
			decorated_notify_inc_32(&(ipsec_ctxt->stats->dropped));
			return IP_STATUS_HOLD;
		}

		notes->dest = entry->dest;
		retval = ip_route_finish(ctxt, notes, ip_hdr);
	}
	break;
	case SOURCE_POST_DECAP:
	{
		struct rc_entry_t *entry = rc_entry_lookup(ctxt->rc,
					ip_hdr->src_addr.word,
					ip_hdr->dst_addr.word,
					ip_hdr->tos);
		if (entry == NULL) {
			APP_DEBUG("Fast Lookup Failed, going slow for Src = \
				0x%x; Dest = 0x%x; TOS = 0x%x",
				ip_hdr->src_addr.word,
				ip_hdr->dst_addr.word, ip_hdr->tos);
			retval =
				ip_route_input_slow(ctxt, notes, ip_hdr);
			return retval;
		}
		notes->dest = entry->dest;
		decorated_notify_inc_64(&(entry->stats->hits));
		retval = ip_route_finish(ctxt, notes, ip_hdr);
	}
		break;
	default:
		APP_ERROR("Invalid Case of routing");
		break;
	}

	return retval;
}

enum IP_STATUS ip_route_input_slow(struct ip_context_t *ctxt,
				   struct annotations_t *notes,
				   struct ip_header_t *ip_hdr __UNUSED)
{
	decorated_notify_inc_64(&(ctxt->stats->ip_route_input_slow));
	lwe_free_fd(buff_allocator, notes->fd);
	return IP_STATUS_DROP;
}

enum IP_STATUS ip_route_finish(struct ip_context_t *ctxt,
				struct annotations_t *notes,
				struct ip_header_t *ip_hdr)
{
	struct rt_dest_t *dest;

	markpoint(10);

	dest = notes->dest;
	switch (dest->scope) {
	case ROUTE_SCOPE_GLOBAL:
	case ROUTE_SCOPE_DECAP:
		if (likely(dest->dev)) {
			return ip_forward(ctxt, notes, ip_hdr);
		} else {
			decorated_notify_inc_64(&
						(ctxt->
						 stats->ip_xmit_icmp_unreach_no_egress));
		}
		break;
	case ROUTE_SCOPE_PME | ROUTE_SCOPE_ENCAP:
	case ROUTE_SCOPE_PME:
		APP_DEBUG("In route scope pme tunnel");
		if (dest->pme_ctx)
			return ipsa_context_info[dest->flow_id].func(ctxt,
								     notes,
								     ip_hdr);
		break;
	case ROUTE_SCOPE_ENCAP:
		return ipsec_encap_send(ctxt, notes, ip_hdr);
	case ROUTE_SCOPE_LOCAL:
		return ip_local_deliver(ctxt, notes, ip_hdr);
	}
	lwe_free_fd(buff_allocator, notes->fd);
	return IP_STATUS_DROP;
}
