/**
 \file ipsec.c
 \brief Implements a simple, fast cache for looking up IPSec tunnels. 
 * Copyright (C) 2008-2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdint.h>
#include <stdbool.h>
#include <misc.h>
#include "ipsec/ipsec.h"
#include "ipsec/ipsec_common.h"
#include "ipsec/ipsec_decap.h"
#include "ipsec/ipsec_sec.h"
#include "net/frame_desc.h"
#include "mm/mem_cache.h"
#include "mmap.h"
#include "ip/ip.h"
#include "ip/ip_protos.h"
#include "ip/ip_hooks.h"
#include "ethernet/eth.h"
#include "net/net_dev.h"
#include "ip/ip_output.h"

static bool __ipsec_lookup(struct ipsec_tunnel_t **entry,
			   struct ipsec_tunnel_table_t *itt,
			   uint32_t tunnel_src_ip, uint32_t tunnel_dst_ip,
			   uint32_t spi, struct ipsec_bucket_t *bucket);

static struct ipsec_bucket_t *__ipsec_find_bucket(struct
						  ipsec_tunnel_table_t
						  *itt, uint32_t tunnel_src_ip,
						  uint32_t tunnel_dst_ip,
						  uint32_t spi);

static struct ipsec_tunnel_t **__ipsec_find_entry(struct ipsec_bucket_t
						  *bucket,
						  uint32_t tunnel_src_ip,
						  uint32_t tunnel_dst_ip,
						  uint32_t spi);

void __ipsec_free_entry(void *entry, void *ctxt);

struct ipsec_tunnel_table_t *ipsec_tunnel_table_create(void)
{
	struct ipsec_tunnel_table_t *itt;
	uint32_t entries;
	struct ipsec_bucket_t *bucket;
	int ctr;

	itt = memalign(CACHE_LINE_SIZE, sizeof(struct ipsec_tunnel_table_t));
	if (itt == NULL) {
		return NULL;
	}

	itt->stats =
	    stats_memalign(CACHE_LINE_SIZE, sizeof(struct ipsec_statistics_t));

	if (itt->stats == NULL) {
		free(itt);
		return NULL;
	}
	memset(itt->stats, 0, sizeof(struct ipsec_statistics_t));

	itt->free_entries = mem_cache_create(sizeof(struct ipsec_tunnel_t),
					     IPSEC_TUNNEL_ENTRIES);

	if (itt->free_entries == NULL) {
		stats_free(itt->stats);
		free(itt);
		return NULL;
	}

	entries = mem_cache_refill(itt->free_entries, IPSEC_TUNNEL_ENTRIES);
	if (entries != IPSEC_TUNNEL_ENTRIES) {
		free(itt->free_entries);
		stats_free(itt->stats);
		free(itt);
		return NULL;
	}

	for (ctr = 0; ctr < IPSEC_TUNNEL_ENTRIES; ctr++) {
		bucket = &(itt->buckets[ctr]);
		bucket->head_entry = NULL;
		bucket->id = ctr;
		bucket->wlock = SPIN_LOCK_UNLOCKED;
	}

	wmb();

	return itt;
}

void ipsec_tunnel_table_delete(struct ipsec_tunnel_table_t *itt)
{
	/*TBD: */
#if 0
	uint32_t entries;
	entries = mem_cache_refill(itt->free_entries, IPSEC_TUNNEL_ENTRIES);
	if (entries != IPSEC_TUNNEL_ENTRIES) {
		return NULL;
	}

	itt->free_entries = mem_cache_create(sizeof(struct ipsec_tunnel_t),
					     IPSEC_TUNNEL_ENTRIES);

	if (itt->free_entries == NULL) {
		return NULL;
	}
#endif
	stats_free(itt->stats);
	free(itt);
}

struct ipsec_tunnel_t *ipsec_create_tunnel_entry(struct
						 ipsec_tunnel_table_t *itt)
{
	return mem_cache_alloc(itt->free_entries);
}

bool ipsec_lookup_tunnel_entry(struct ipsec_tunnel_t ** entry,
			       struct ipsec_tunnel_table_t * itt,
			       uint32_t tunnel_src_ip, uint32_t tunnel_dst_ip,
			       uint32_t spi)
{
	struct ipsec_bucket_t *bucket;
	bool retval;

	bucket = __ipsec_find_bucket(itt, tunnel_src_ip, tunnel_dst_ip, spi);

	retval =
	    __ipsec_lookup(entry, itt, tunnel_src_ip, tunnel_dst_ip, spi,
			   bucket);

	return retval;
}

bool ipsec_tunnel_fast_lookup(struct ipsec_tunnel_t ** entry,
			      struct ipsec_tunnel_table_t * itt,
			      uint32_t tunnel_src_ip, uint32_t tunnel_dst_ip,
			      uint32_t spi, uint32_t index)
{
	struct ipsec_bucket_t *bucket =
	    &(itt->buckets[index & IPSEC_HASH_MASK]);
	bool retval;

	retval =
	    __ipsec_lookup(entry, itt, tunnel_src_ip, tunnel_dst_ip, spi,
			   bucket);
	return retval;
}

bool ipsec_add_tunnel_entry(struct ipsec_tunnel_table_t * itt,
			    struct ipsec_tunnel_t * new_entry)
{
	struct ipsec_bucket_t *bucket;
	struct ipsec_tunnel_t **entry_ptr;
	struct ipsec_tunnel_t *prev_entry;
	uint32_t count;
	bool rc;

	rc = false;

	count = itt->stats->entry_count;

	if (count < IPSEC_TUNNEL_ENTRIES) {
		bucket =
		    __ipsec_find_bucket(itt, new_entry->tunnel_src_addr,
					new_entry->tunnel_dst_addr,
					new_entry->spi);
#ifdef IPSEC_RCU_ENABLE
		rcu_read_lock();
#endif
		spin_lock(&(bucket->wlock));
		entry_ptr =
		    __ipsec_find_entry(bucket, new_entry->tunnel_src_addr,
				       new_entry->tunnel_dst_addr,
				       new_entry->spi);
#ifdef IPSEC_RCU_ENABLE
		prev_entry = rcu_dereference(*entry_ptr);
#else
		prev_entry = *entry_ptr;
#endif
		/* Entry not already in table, insert */
		if (prev_entry == NULL) {
			new_entry->next = NULL;
			new_entry->valid = true;

#ifdef IPSEC_RCU_ENABLE
			rcu_assign_pointer(*entry_ptr, new_entry);
#else
			*entry_ptr = new_entry;
#endif
			rc = true;
		}
		spin_unlock(&(bucket->wlock));
#ifdef IPSEC_RCU_ENABLE
		rcu_read_unlock();
#endif

	}

	if (rc == true) {
		/* TODO: need to call api to id source ip to 
		 * route cache as ipsec here */
		decorated_notify_inc_32(&(itt->stats->new_entries));
		decorated_notify_inc_32(&(itt->stats->entry_count));
	}
	return rc;
}

bool ipsec_remove_tunnel_entry(struct ipsec_tunnel_table_t * itt,
			       uint32_t tunnel_src_ip, uint32_t tunnel_dst_ip,
			       uint32_t spi)
{
	struct ipsec_tunnel_t **entry_ptr;
	struct ipsec_tunnel_t *entry;
	struct ipsec_bucket_t *bucket;
	bool rc;

	rc = false;
	bucket = __ipsec_find_bucket(itt, tunnel_src_ip, tunnel_dst_ip, spi);

#ifdef IPSEC_RCU_ENABLE
	rcu_read_lock();
#endif
	spin_lock(&(bucket->wlock));
	entry_ptr =
	    __ipsec_find_entry(bucket, tunnel_src_ip, tunnel_dst_ip, spi);
#ifdef IPSEC_RCU_ENABLE
	entry = rcu_dereference(*entry_ptr);
#else
	entry = *entry_ptr;
#endif
	if (entry != NULL) {
#ifdef IPSEC_RCU_ENABLE
		rcu_assign_pointer(*entry_ptr, entry->next);
#else
		*entry_ptr = entry->next;
#endif
	}
	spin_unlock(&(bucket->wlock));
#ifdef IPSEC_RCU_ENABLE
	rcu_read_unlock();
#endif

	if (entry != NULL) {
		rc = true;
		itt->stats->entry_count--;
		decorated_notify_inc_32(&(itt->stats->removed_entries));
#ifdef IPSEC_RCU_ENABLE
		rcu_free(&__ipsec_free_entry, entry, itt);
#else
		__ipsec_free_entry(entry, itt);
#endif
	}

	return rc;
}

static bool __ipsec_lookup(struct ipsec_tunnel_t **entry,
			   struct ipsec_tunnel_table_t *itt,
			   uint32_t tunnel_src_ip, uint32_t tunnel_dst_ip,
			   uint32_t spi, struct ipsec_bucket_t *bucket)
{
	struct ipsec_tunnel_t **entry_ptr;
	struct ipsec_tunnel_t *result;
	bool rc;

	rc = false;

#ifdef IPSEC_RCU_ENABLE
	rcu_read_lock();
#endif
	entry_ptr =
	    __ipsec_find_entry(bucket, tunnel_src_ip, tunnel_dst_ip, spi);
#ifdef IPSEC_RCU_ENABLE
	result = rcu_dereference(*entry_ptr);
#else
	result = *entry_ptr;
#endif
	if (result != NULL) {
		*entry = result;
		decorated_notify_inc_64(&(itt->stats->table_hits));
		rc = true;
	} else {
		decorated_notify_inc_64(&(itt->stats->table_misses));
	}
#ifdef IPSEC_RCU_ENABLE
	rcu_read_unlock();
#endif

	return rc;
}

static struct ipsec_bucket_t *__ipsec_find_bucket(struct
						  ipsec_tunnel_table_t *itt,
						  uint32_t tunnel_src_ip,
						  uint32_t tunnel_dst_ip,
						  uint32_t spi)
{
	uint32_t hash;

	hash = compute_decap_hash(tunnel_src_ip, tunnel_dst_ip, spi);
	return &(itt->buckets[hash]);
}

static struct ipsec_tunnel_t **__ipsec_find_entry(struct ipsec_bucket_t
						  *bucket,
						  uint32_t tunnel_src_ip,
						  uint32_t tunnel_dst_ip,
						  uint32_t spi)
{
	struct ipsec_tunnel_t **entry_ptr;
	struct ipsec_tunnel_t *entry;

	entry_ptr = &(bucket->head_entry);
#ifdef IPSEC_RCU_ENABLE
	entry = rcu_dereference(*entry_ptr);
#else
	entry = *entry_ptr;
#endif
	while (entry != NULL) {
		if ((entry->valid == true) && (entry->spi == spi)
		    && (entry->tunnel_src_addr == tunnel_src_ip)
		    && (entry->tunnel_dst_addr == tunnel_dst_ip)) {
			break;
		} else {
			entry_ptr = &(entry->next);
		}
#ifdef IPSEC_RCU_ENABLE
		entry = rcu_dereference(*entry_ptr);
#else
		entry = *entry_ptr;
#endif
	}

	return entry_ptr;
}

void __ipsec_free_entry(void *entry, void *ctxt)
{
	struct ipsec_tunnel_table_t *itt;

	itt = ctxt;
	mem_cache_free(itt->free_entries, entry);
	itt->stats->entry_count--;
}

void ipsec_free_entry(void *ctxt, void *entry)
{
	struct ipsec_tunnel_table_t *itt;

	itt = ctxt;
	mem_cache_free(itt->free_entries, entry);
}
