/**
 \file ipsec_cpdp.h
 \brief Implements a simple, fast cache for looking up IPSec tunnels. 
 */
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef LIB_IPSEC_IPSEC_CPDP_H
#define LIB_IPSEC_IPSEC_CPDP_H

struct lwe_ctrl_sa_algo {

/* Different types of Algos - Authentication and Encryption*/
#define AES_CBC 		0
#define TRIP_DES_CBC 		1

#define HMAC_SHA1 		0
	unsigned int alg_type;			/**< Algo type - HMAC_SHA1 or AES_CBC */

	unsigned int alg_key_len;		/**< Length of the key in bits */

#define LWE_CTRL_MAX_ALGO_KEY_LEN	48
	char alg_key[LWE_CTRL_MAX_ALGO_KEY_LEN];			/**< Key */
  char *alg_key_ptr;
};

/**< SA info structure*/
struct lwe_ctrl_sa_id {
	unsigned int saddr;					/**< Src IP of Tunnel end point */
	unsigned int daddr;					/**< Dest IP of Tunnel end point */
	unsigned int spi;						/**< SPI */

#define IPSEC_PROTO_ESP 50
	unsigned char proto;					/**< Protocol Type */

#define IPSEC_DIR_IN		1
#define IPSEC_DIR_OUT	2
	unsigned char dir;
};

/**< Selector info structure */
struct lwe_ctrl_sa_selector {
	unsigned int saddr;					/**< Src IP of Host */
	unsigned int daddr;					/**< Dest IP of Host */
};

/**
 \brief Statistics Data
 */
struct lwe_ctrl_ipsec_stats {
	unsigned int decap_pkts;  /**< Decrypted pkt count */
	unsigned int encap_pkts;  /**< Encrypted pkt count */
	unsigned int decap_octets; /**< Decrypted bytes count */
	unsigned int encap_octets; /**< Encrypted bytes count */
	unsigned int decap_errored_pkts; /**Decrypted erroneous pkt count */
	unsigned int encap_errored_pkts; /** Encrypted erroneous pkts count */
};

/**
 \brief	The Structure is used in External API - ipsecfwd_create_sa
 \details	For configuring the SA with Tunnel End Points, Host IPs, 
  		Authentication Algo info, and Encryption Algo info
 */
struct lwe_ctrl_ipsec_info {
	struct lwe_ctrl_sa_id id;		/**< SA Info */
	struct lwe_ctrl_sa_selector sel;	/**< Selector Info */
	struct lwe_ctrl_sa_algo aalg;	/**< Authentication Algo info */
	struct lwe_ctrl_sa_algo ealg;	/**< Encryption Algo info */
	struct lwe_ctrl_ipsec_stats stats;
	unsigned int flow_id; /**< PME Flow Id */
};

/**
 \brief	Structure for maintaining Stats related to Route Cache
 */
struct lwe_ctrl_route_stats {
	uint32_t entry_count;		/**< Number of enteries in the Route Cache*/
	uint32_t stale_entries;	/**< Number of stale enteries in the Route Cache*/
	uint32_t new_entries;		/**< Number of enteries added in the Route Cache*/
	uint32_t removed_entries;/**< Number of enteries removed in the Route Cache*/
	union stat64_t hits;		/**< Number of times the LookUp succeeded*/
	union stat64_t misses;	/**< Number of times the LookUp failed*/
};

/**
 \brief	Structure for configuring an Interface
 */
struct lwe_ctrl_intf_conf {
	uint32_t ip_addr; /**< IP Address */
	uint32_t mtu; /**< MTU */
	uint32_t enable; /**< Enabling the interface */
	unsigned char mac_addr[6];  /**< MAC Address of the interface */
	char ifname[10];  /**< Interface Name */
#define LWE_CTRL_PARAM_BMASK_IFNAME		(1 << 0)
#define LWE_CTRL_PARAM_BMASK_IPADDR		(1 << 1)
#define LWE_CTRL_PARAM_BMASK_MACADDR		(1 << 2)
#define LWE_CTRL_PARAM_BMASK_MTU			(1 << 3)
#define LWE_CTRL_PARAM_BMASK_ENABLE		(1 << 4)
#define LWE_CTRL_PARAM_MAX_INTF_BIT_NO			5

#define LWE_CTRL_INTF_CONF_MDTR_PARAM_MAP (LWE_CTRL_PARAM_BMASK_IFNAME)

	uint32_t bitmask;
};

struct lwe_ctrl_ip_info {
	unsigned int src_ipaddr;			/**<Source IP Address>*/
	unsigned int dst_ipaddr;			/**<Destination IP Address>*/
	unsigned int gw_ipaddr;				/**<Gateway IP Address>*/
	unsigned int tos;				/**<Gateway IP Address>*/
	unsigned char mac_addr[6];	/**< Mac Address */
	unsigned int flow_id; /**< Flow Id */
	unsigned int frame_cnt; /**<Frame Count */
	unsigned int replace_entry;  /**< Used for overwriting an existing ARP entry */
	struct lwe_ctrl_route_stats stats;  /**< Route Stats */
	struct lwe_ctrl_intf_conf intf_conf; /**< Interface Configuration */
};

#define PME_MAX_FILE_SIZE				(1024 * 10)

/**
 \brief	Structure to configure a PME flow for IP Packet Scan App 
 */
struct lwe_ctrl_spa_cfg_info {
	uint32_t flow_id;	 /**< Flow Id */
	uint8_t set;	     /**< Set Id (0 - 255) in the PME Database */
	uint16_t subset;	/**< Subset Id (0  0xff00) in the PME Database */

#define SET_SCAN_MODE_FLOW 		(1<<0)
#define SET_RESIDUE_ENABLE 		(1<<1)
#define SET_SCAN_FN_TYPE_BLOCK 		(1<<2)
#define SET_SCAN_MODE_DIRECT 		(~(1<<0))
#define SET_RESIDUE_DISABLE 		(~(1<<1))
#define SET_SCAN_FN_TYPE_NON_BLOCK 	(~(1<<2))

	uint32_t flag;	      /**< Flag for setting Scan mode type, residue enabling, and scan fn type */

	uint32_t report_verbosity;	  /**< Report Verbosity */
	uint32_t session_id;	    /**< session Id in the SRE associated with this PME Context (Flow) */
	uint32_t input_buff_chunk;	  /**< Output Buffer Size for the report */
	uint32_t qos_in;	/**< Qos - priority associated with the Q towards PME */
	uint32_t qos_out;	 /**< Qos - priority associated with the Q from PME */
	uint32_t match_limit;	     /**< Max number of matches to be made on this flow */
	uint32_t compare_limit;	       /**< Max number of comparisons to be made on this flow */
	uint32_t output_buff_frm_app;	     /**< Allocate buffer in app for output */
	uint32_t input_buff_size;	 /**< Input Buffer Size of the generated packet */
	uint32_t size_of_file;	    /**< File Size */
	uint8_t file[PME_MAX_FILE_SIZE];
};

/**
 \brief	App info structure used for configuring IPsec/ Ipfwd/ PME Apps 
 */
union lwe_ctrl_app_info {
	struct lwe_ctrl_ipsec_info ipsec_info;	/**< IPsec Info structure */
	struct lwe_ctrl_ip_info ip_info;	/**< IPfwd Info structure */
	struct lwe_ctrl_spa_cfg_info pme_info;	/**< PME Info structure */
};

/**
 \brief	Structure used for communicating between CP and DP. 
        The structure is kept on the White Board. 
 */
struct lwe_ctrl_op_info {

#define LWE_CTRL_CMD_STATE_IDLE 0
#define LWE_CTRL_CMD_STATE_BUSY 1
	unsigned int state;				/**< State of Command */

#define LWE_CTRL_CMD_TYPE_SAADD		1
#define LWE_CTRL_CMD_TYPE_SADEL		2
#define LWE_CTRL_CMD_TYPE_SASTATS		3
#define LWE_CTRL_CMD_TYPE_IFSTATS		4
#define LWE_CTRL_CMD_TYPE_SALIST		5
#define LWE_CTRL_CMD_TYPE_IFLIST		6
#define LWE_CTRL_CMD_TYPE_ROUTE_ADD		7
#define LWE_CTRL_CMD_TYPE_ROUTE_DEL		8
#define LWE_CTRL_CMD_TYPE_ROUTE_STATS		9
#define LWE_CTRL_CMD_TYPE_IPFWD_SYS_STATS		10
#define LWE_CTRL_CMD_TYPE_INTF_CONF_CHNG		11
#define LWE_CTRL_CMD_TYPE_ARP_ADD			12
#define LWE_CTRL_CMD_TYPE_ARP_DEL			13
#define LWE_CTRL_CMD_TYPE_PME_ADD			14
#define LWE_CTRL_CMD_TYPE_PME_DEL			15
#define LWE_CTRL_CMD_TYPE_FRAMECNT_EDIT		16
#define LWE_CTRL_CMD_TYPE_GO			17
#define LWE_CTRL_CMD_TYPE_RESET			18
#define LWE_CTRL_CMD_TYPE_RESET_STATS		19

	unsigned int msg_type;			/**<Type of Request>*/

#define LWE_CTRL_RSLT_SUCCESSFULL		1
#define LWE_CTRL_RSLT_FAILURE		0
	unsigned int result;				/**<Result - Successful, Failure>*/

	union lwe_ctrl_app_info app_info;  /**< App info structure used for configuring IPsec/ Ipfwd/ PME Apps */
};

extern volatile struct lwe_ctrl_op_info g_sLweCtrlSaInfo;

#endif /* ifndef LIB_IPSEC_IPSEC_CPDP_H */
