/**
 \file ipsec_decap.c
 \brief IPsec decapsulation processing
 */
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <stdbool.h>
#include "app_common.h"
#include "ipsec/ipsec.h"
#include "ipsec_sec.h"
#include "net/frame_desc.h"
#include "mm/mem_cache.h"
#include "mmap.h"
#include "ip/ip.h"
#include "ip/ip_protos.h"
#include "ip/ip_rc.h"
#include "ip/ip_hooks.h"
#include "ethernet/eth.h"
#include "net/net_dev.h"
#include "ip/ip_output.h"
#include "ip/ip_accept.h"
#include "ipsec/ipsec_init.h"

enum IP_STATUS ipsec_decap_send(struct ip_context_t *ip_ctxt,
				struct annotations_t *notes, void *ip_hdr_ptr)
{
	struct ipsec_tunnel_t *entry;
	bool retval = false;
	struct ipsec_esp_hdr_t *esp_hdr;
	struct ip_header_t *ip_hdr = ip_hdr_ptr;
	struct ipsec_context_t *ipsec_ctxt;
	struct qm_fd *fd, fd2;
	uint32_t ret;

	esp_hdr =
	    (struct ipsec_esp_hdr_t *)((char *) ip_hdr +
				       sizeof(struct ip_header_t));
	APP_DEBUG("%s: Pkt src = %x, dst = %x and spi = %x", __func__,
		  ip_hdr->src_addr.word, ip_hdr->dst_addr.word, esp_hdr->spi);
	ipsec_ctxt = (struct ipsec_context_t *)ip_ctxt->user_data;
	retval = ipsec_tunnel_fast_lookup(&entry,
					  (struct ipsec_tunnel_table_t *)
					  ipsec_ctxt->itt,
					  ip_hdr->src_addr.word,
					  ip_hdr->dst_addr.word, esp_hdr->spi,
					  RC_BUCKET_INDEX(notes));
	if (unlikely(retval == false)) {
		APP_INFO("Packet not in fast tunnel_list");
		retval = ipsec_lookup_tunnel_entry(&entry,
						   (struct ipsec_tunnel_table_t
						    *)
						   ipsec_ctxt->itt,
						   ip_hdr->src_addr.word,
						   ip_hdr->dst_addr.word,
						   esp_hdr->spi);
		if (unlikely(retval == false)) {
			APP_ERROR("%s: Couldn't find tunnel", __func__);
			decorated_notify_inc_32(&(ipsec_ctxt->stats->dropped));
			return IP_STATUS_HOLD;
		}
	}
	fd = notes->fd;

	/* note to self, does data come in at ip_hdr or at mac_hdr? */
	if (unlikely(ipsec_create_compound_fd(&fd2, fd, ip_hdr, DECRYPT))) {
		APP_DEBUG("%s :Unable to Allocate compound FD", __func__);
		/* We may have to drop original pkt */
		decorated_notify_inc_64(&(ip_ctxt->stats->ip_in_dropped));
		discard_handler(notes, notes->fd->bpid);
		return IP_STATUS_DROP;
	}

	fd = &fd2;
	/* update statistics that enqueue to sec is done for decap */
	decorated_notify_inc_32(&(ipsec_ctxt->stats->decap_pre_sec));

loop:
	/* enqueue frame to SEC4.0 for Encap*/
	ret = qman_enqueue(entry->qm_fq_to_sec, fd, 0);

	if (unlikely(ret)) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	return IP_STATUS_ACCEPT;
}

void ipsec_decap_cb(struct fq_context_t *fq_ctxt,
		    struct annotations_t *notes, void *data __UNUSED)
{
	struct qm_fd simple_fd;
	struct sg_entry_t *sg;
	struct ip_header_t *ip_hdr;
	struct annotations_t *ip_notes;
	struct ipsec_context_t *ctxt = (struct ipsec_context_t *)fq_ctxt;
	struct qm_fd *compound_fd = notes->fd;
	uint8_t padlen = 0;

	markpoint(21);
	/* Error Returned by SEC Block, Refer to Section 1.5.2.9 of BG */
	if (unlikely(compound_fd->status)) {
		APP_ERROR("SEC Block returned Status = %x", compound_fd->status);
		return;
	}

	ipsec_create_simple_fd(&simple_fd, compound_fd, DECRYPT);

	ip_notes = APP_PTOV(simple_fd.addr_lo);
	decorated_notify_inc_32(&(ctxt->stats->decap_post_sec));

	/* get notes and context information so that we can hand
	   over the processing to the remainder of IPv4 processing
	   chain
	 */
	if (unlikely(qm_fd_contig == simple_fd.format)) {
		ip_hdr = (void *)((uint8_t *) ip_notes + simple_fd.offset);

		padlen = *(uint8_t *)((uint8_t *)ip_hdr +
				ip_notes->fd->length20 - PADLEN_OFFSET)
				 + PADLEN_OFFSET;
		if (unlikely(padlen > IPSEC_MAX_HMAC_KEY_SIZE_BYTES)) {
			APP_ERROR("Bad Padlen = %d", padlen);
			if (simple_fd.bpid)
				lwe_free_fd(buff_allocator, &simple_fd);
			else
				APP_ERROR("Bad BPID in %s", __func__);
			return;
		}
	} else {
		sg = (struct sg_entry_t *)
			APP_PTOV(simple_fd.addr_lo + simple_fd.offset);
		ip_hdr = (void *) APP_PTOV(sg->addr_lo + sg->offset);
		do {
			sg++;
		} while (!sg->final);
		padlen = *(uint8_t *)((uint8_t *)
				APP_PTOV(sg->addr_lo + sg->offset)
				+ sg->length - PADLEN_OFFSET) + PADLEN_OFFSET;
		if (padlen == sg->length) {
			sg--;
			sg->final = 1;
		} else
			sg->length -= padlen;
	}
	ip_notes->fd->length20 -= padlen;

	if (likely(qm_fd_contig == simple_fd.format))
		ip_notes->fd->offset -= ETH_HDR_LEN;
	else if (qm_fd_sg == simple_fd.format) {
		sg = (struct sg_entry_t *)
			APP_PTOV(simple_fd.addr_lo + simple_fd.offset);
		sg->offset -= ETH_HDR_LEN;
		sg->length += ETH_HDR_LEN;
	}
	ip_notes->fd->length20 += ETH_HDR_LEN;

	/* hand over the packet to IP layer for egress transmission */
	ip_accept_preparsed((struct ip_context_t *)ctxt, ip_notes, ip_hdr,
		 SOURCE_POST_DECAP);
}

void ipsec_decap_initialize_fqs(struct ipsec_tunnel_t *entry,
				struct ipsec_context_t *ctxt, uint32_t sec_fq,
				struct lwe_ctrl_sa_algo *ealg,
				struct lwe_ctrl_sa_algo *aalg)
{
	void *ctxtA;
	struct ipsec_context_t *decap_ctxt;

	/* IPSEC FQ Context */
	if (!(decap_ctxt = memalign
	      (CACHE_LINE_SIZE, sizeof(struct ipsec_context_t)))) {
		APP_ERROR("%s: No More Buffers left", __func__);
		return;
	}

	memcpy(decap_ctxt, ctxt, sizeof(struct ipsec_context_t));
	decap_ctxt->ip_ctxt.fq_ctxt.handler = &ipsec_decap_cb;
	decap_ctxt->ip_ctxt.user_data = decap_ctxt;
	if (!(ctxtA = create_decapsulation_sec_descriptor(entry, ealg, aalg))) {
		APP_ERROR("Unable to allocate buffer for descriptor");
		return;
	}
	if (init_ipsec_fqs(entry, decap_ctxt, ctxtA, sec_fq)) {
		APP_ERROR("Failed to Initialize encap Context");
		return;
	}
}
