/**
 \file ipsec_encap.c
 \brief IPsec encapsulation processing
 */
/*
 * Copyright (C) 2008-2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdbool.h>
#include "app_common.h"
#include "ipsec_sec.h"
#include "net/frame_desc.h"
#include "mm/mem_cache.h"
#include "mmap.h"
#include "ip/ip_rc.h"
#include "ip/ip_protos.h"
#include "ip/ip_hooks.h"
#include "ethernet/eth.h"
#include "net/net_dev.h"
#include "ip/ip_output.h"
#include "ip/ip_accept.h"
#include "ip/ip_forward.h"

#define FM_FD_CMD_RPD  0x40000000	/* Read Prepended Data */
#define FM_FD_CMD_DTC  0x10000000	/* Do TCP Checksum */
#define FM_L3_PARSE_RESULT_IPV4	0x8000

enum IP_STATUS ipsec_encap_send(struct ip_context_t *ip_ctxt,
				struct annotations_t *notes, void *ip_hdr_ptr)
{
	struct ip_header_t *ip_hdr = ip_hdr_ptr;
	struct ipsec_tunnel_t *entry = notes->dest->tunnel;
	struct qm_fd *fd = notes->fd, fd2;
	uint32_t ret;

	if (unlikely(ipsec_create_compound_fd(&fd2, fd, ip_hdr, ENCRYPT))) {
		/* We may have to drop original pkt */
		decorated_notify_inc_64(&(ip_ctxt->stats->ip_in_dropped));
		lwe_free_fd(buff_allocator, fd);
		return IP_STATUS_DROP;
	}

	fd = &fd2;
#ifdef INFINITE_IPSECFWD
	struct ipsec_context_t *ipsec_ctxt =
	    (struct ipsec_context_t *)ip_ctxt->user_data;
	decorated_notify_inc_32(&(ipsec_ctxt->stats->encap_pre_sec));
#endif
loop:
	ret = qman_enqueue(entry->qm_fq_to_sec, fd, 0);

	if (unlikely(ret)) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	return IP_STATUS_STOLEN;
}

void ipsec_encap_cb(struct fq_context_t *fq_ctxt,
		    struct annotations_t *notes, void *data __UNUSED)
{
	struct qm_fd simple_fd;
	struct ip_header_t *ip_hdr;
	struct annotations_t *ip_notes;
	struct ipsec_context_t *ctxt = (struct ipsec_context_t *)fq_ctxt;
	struct qm_fd *compound_fd = notes->fd;

	markpoint(21);
	if (unlikely(compound_fd->status)) {
		APP_DEBUG("Non-Zero Status from SEC Block %x", compound_fd->status);
		lwe_free_fd(buff_allocator, compound_fd);
		return;
	}
	ipsec_create_simple_fd(&simple_fd, compound_fd, ENCRYPT);
	ip_notes = APP_PTOV(simple_fd.addr_lo);

	ip_hdr = (void *)((uint8_t *) ip_notes + simple_fd.offset);
#ifndef ENABLE_CKSUM_SEC
	ip_notes->parse.l3r = FM_L3_PARSE_RESULT_IPV4;
	ip_notes->parse.ipO = ETH_HDR_LEN;
	simple_fd.cmd = FM_FD_CMD_RPD | FM_FD_CMD_DTC;
#endif
	ip_notes->fd->offset -= ETH_HDR_LEN;
	ip_notes->fd->length20 += ETH_HDR_LEN;
	decorated_notify_inc_32(&(ctxt->stats->encap_post_sec));
	ip_accept_preparsed((struct ip_context_t *)ctxt, ip_notes, ip_hdr,
		SOURCE_POST_ENCAP);
}

void ipsec_encap_initialize_fqs(struct ipsec_tunnel_t *entry,
				struct ip_header_t *ip_hdr,
				struct ipsec_context_t *ctxt, uint32_t sec_fq,
				struct lwe_ctrl_sa_algo *ealg,
				struct lwe_ctrl_sa_algo *aalg)
{
	void *ctxtA;

	if (!
	    (ctxtA =
	     create_encapsulation_sec_descriptor(entry, ip_hdr,
						 IP_PROTO_TCP, ealg, aalg))) {
		APP_ERROR("Error creating descriptor");
		return;
	}

	if (init_ipsec_fqs(entry, ctxt, ctxtA, sec_fq)) {
		APP_ERROR("Failed to Initialize encap Context");
		return;
	}
}
