/**
 \file ipsec_init.c
 \brief IPsec tunnels Initialization Routines
 */
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdint.h>
#include <stdbool.h>
#include <misc.h>
#include "ipsec/ipsec.h"
#include "ipsec/ipsec_sec.h"
#include "ipsec/ipsec_init.h"
#include "ipsec/ipsec_common.h"
#include "ipsec/ipsec_encap.h"
#include "ipsec/ipsec_decap.h"
#include "net/frame_desc.h"
#include "mm/mem_cache.h"
#include "mmap.h"
#include "ip/ip.h"
#include "ip/ip_rc.h"
#include "ip/ip_protos.h"
#include "ip/ip_hooks.h"
#include "ethernet/eth.h"
#include "net/net_dev.h"
#include "ip/ip_output.h"
#include "net/neigh.h"
#include "app_common.h"
#include "desc.h"
#include "pme/ipsa_pme.h"

#define ENTRIES 1024
#define ENTRIES_POOL_SIZE (ENTRIES << 1)

int32_t g_key_split_flag;
struct qman_fq *g_splitkey_fq_from_sec;
struct qman_fq *g_splitkey_fq_to_sec;


extern struct ipsec_tunnel_config_t *g_ipsec_tunnel_config;
extern int32_t g_tunnel_id;
extern struct ipsec_stack_t ipsec_stack;
uint32_t ipsec_flow_id;

struct ipsec_tunnel_config_t *ipsec_config_create(void)
{
	struct ipsec_tunnel_config_t *config;
	uint32_t entries;

	config =
	    memalign(CACHE_LINE_SIZE, sizeof(struct ipsec_tunnel_config_t));
	if (config == NULL) {
		return NULL;
	}

	config->free_entries =
	    mem_cache_create(sizeof(struct ipsec_tunnel_config_entry_t),
			     ENTRIES_POOL_SIZE);
	if (config->free_entries == NULL) {
		free(config);
		return NULL;
	}

	entries = mem_cache_refill(config->free_entries, ENTRIES_POOL_SIZE);
	if (entries != ENTRIES_POOL_SIZE) {
		free(config->free_entries);
		free(config);
		return NULL;
	}

	wmb();
	return config;
}


void ipsec_config_delete(struct ipsec_tunnel_config_t *config)
{
#if 0
	uint32_t entries;

	entries = mem_cache_refill(config->free_entries, ENTRIES_POOL_SIZE);
	if (entries != ENTRIES_POOL_SIZE) {
		return NULL;
	}

	config->free_entries =
	    mem_cache_create(sizeof(struct ipsec_tunnel_config_entry_t),
			     ENTRIES_POOL_SIZE);
	if (config->free_entries == NULL) {
		free(config);
		return NULL;
	}
#endif
	free(config);

	return;
}

struct ipsec_tunnel_config_entry_t *ipsec_config_entry_create(struct
							      ipsec_tunnel_config_t
							      *config)
{
	return mem_cache_alloc(config->free_entries);
}

/**
 \brief Create a new Security association entry
 \param[out] ipsec_info contains SA parameters - SA Id, selector info,
	      encryption algo info, authentication algo info
 \return Integer status
 */
int32_t ipsecfwd_create_sa(struct lwe_ctrl_ipsec_info * ipsec_info)
{

	struct ipsec_tunnel_config_entry_t *ipsec_tunnel_config_entry;
	struct ipsec_tunnel_config_t *ipsec_tunnel_config =
		g_ipsec_tunnel_config;
	uint32_t *next_hop_addr;
	uint32_t encryption_mode;
	int32_t ret;

	APP_DEBUG("ipsecfwd_create_sa: Enter");

	ipsec_tunnel_config_entry =
	    ipsec_config_entry_create(ipsec_tunnel_config);

	if (!ipsec_tunnel_config_entry) {
		APP_ERROR("%s: Unable to allocate tunnel config entry",
			  __func__);
		return -ENOMEM;
	}

	encryption_mode =
	    (IPSEC_DIR_OUT == ipsec_info->id.dir) ? ENCRYPT : DECRYPT;
#ifdef PME_SWITCH
	ipsec_flow_id = ipsec_info->flow_id;
#endif
	ipsec_tunnel_config_entry->src_ip = ipsec_info->sel.saddr;
	ipsec_tunnel_config_entry->dst_ip = ipsec_info->sel.daddr;
	ipsec_tunnel_config_entry->tunnel_src_ip_addr = ipsec_info->id.saddr;
	ipsec_tunnel_config_entry->tunnel_dst_ip_addr = ipsec_info->id.daddr;

	memcpy(&ipsec_tunnel_config_entry->enc_key[0],
		ipsec_info->ealg.alg_key, ipsec_info->ealg.alg_key_len);

	memcpy(&ipsec_tunnel_config_entry->auth_key[0],
		ipsec_info->aalg.alg_key, ipsec_info->aalg.alg_key_len);

	ipsec_tunnel_config_entry->tunnel_id = g_tunnel_id;
	ipsec_tunnel_config_entry->spi = ipsec_info->id.spi;

	if (encryption_mode == ENCRYPT) {
		ipsec_tunnel_config_entry->seq_num = 0;
		next_hop_addr = &ipsec_tunnel_config_entry->tunnel_dst_ip_addr;
	} else {
		ipsec_tunnel_config_entry->seq_num = 1;	/* per generated data for decap */
		next_hop_addr = &ipsec_tunnel_config_entry->dst_ip;
	}
	ret = ipsec_tunnel_create(ipsec_tunnel_config_entry,
			    &ipsec_stack, next_hop_addr, encryption_mode,
			    &ipsec_info->ealg, &ipsec_info->aalg);
	if (ret == 0)
		g_tunnel_id++;

	APP_DEBUG("ipsecfwd_create_sa: Exit");
	return ret;
}

/**
 \brief Prepare all data structures to make the DUT ready for
	processing IPsec encapsulation.
 \param[in] config Pointer to IPsec config entry
 \param[in] next_hop_addr Next Hop Address
 \param[in] ipsec_stack Ipsec Stack pointer
 \param[in] ealg Encryption Algo Info
 \param[in] aalg Authentication Algo Info
 \return Status
 */
int32_t ipsec_tunnel_encap_init(struct ipsec_tunnel_config_entry_t *config,
			     uint32_t * next_hop_addr,
			     struct ipsec_stack_t *ipsec_stack,
			     struct lwe_ctrl_sa_algo *ealg,
			     struct lwe_ctrl_sa_algo *aalg)
{
	struct ipsec_tunnel_t *entry;
	bool tunnel_route;
	struct ip_header_t *outer_ip_hdr;
	uint32_t sec_fq;
	uint8_t scope;

	sec_fq = 20480 + (config->tunnel_id * 2);

	/* initialize tunnel entry and add to tunnel table */
	if (!(entry = ipsec_create_tunnel_entry(ipsec_stack->itt))) {
		APP_ERROR("Unable to Allocate tunnel entry");
		return -ENOMEM;
	}

	entry->tunnel_id = config->tunnel_id;
	entry->src_addr = config->src_ip;
	entry->dst_addr = config->dst_ip;
	entry->tunnel_src_addr = config->tunnel_src_ip_addr;
	entry->tunnel_dst_addr = config->tunnel_dst_ip_addr;
	entry->spi = config->spi;
	entry->seq_num = config->seq_num;
	entry->esp_dec_key = (uint8_t *) ealg->alg_key;
	entry->esp_auth_key = (uint8_t *) aalg->alg_key;
	APP_DEBUG("SPI Value is %x", config->spi);

	if (false == ipsec_add_tunnel_entry(ipsec_stack->itt, entry)) {
		ipsec_free_entry(ipsec_stack->itt, entry);
		APP_ERROR("Tunnel Entry Couldn't be added");
		return -EINVAL;
	}
	/* create outer ip, initialize fq - pdb, fq contexts etc. */
	if (!
	    (outer_ip_hdr =
	     memalign(CACHE_LINE_SIZE, sizeof(struct ip_header_t)))) {
		APP_ERROR("%s: Unable to allocate memory for outer ip header",
			  __func__);
		return -ENOMEM;
	}
	ipsec_build_outer_ip_hdr(outer_ip_hdr, &config->tunnel_src_ip_addr,
				 &config->tunnel_dst_ip_addr);
	ipsec_encap_initialize_fqs(entry, outer_ip_hdr,
				   ipsec_stack->ipsec_ctxt, sec_fq, ealg, aalg);
	/* populate route cache */
	tunnel_route = true;
	scope = ROUTE_SCOPE_ENCAP;

#ifdef PME_SWITCH
	if (ipsec_flow_id)
		scope |= ROUTE_SCOPE_PME;
#endif

	populate_route_cache(config->src_ip, config->dst_ip,
			     *next_hop_addr, ipsec_stack, tunnel_route, entry,
			     false, scope);
	tunnel_route = false;
	scope = ROUTE_SCOPE_GLOBAL;
	/* entry for each remote to remote host-pair */
	populate_route_cache(config->tunnel_src_ip_addr,
			     config->tunnel_dst_ip_addr,
			     *next_hop_addr, ipsec_stack, tunnel_route, entry,
			     false, scope);
	free(outer_ip_hdr);
	return 0;
}

/**
 \brief Prepare all data structures to make the DUT ready for
	processing IPsec decapsulation.
 \param[in] config Pointer to IPsec config entry
 \param[in] next_hop_addr Next Hop Address
 \param[in] ipsec_stack Ipsec Stack pointer
 \param[in] ealg Encryption Algo Info
 \param[in] aalg Authentication Algo Info
 \return Status
 */
int32_t ipsec_tunnel_decap_init(struct ipsec_tunnel_config_entry_t *config,
			     uint32_t * next_hop_addr,
			     struct ipsec_stack_t *ipsec_stack,
			     struct lwe_ctrl_sa_algo *ealg,
			     struct lwe_ctrl_sa_algo *aalg)
{
	struct ipsec_tunnel_t *entry;
	bool tunnel_route;
	uint32_t sec_fq;
	uint8_t scope;

	/* obtain FQs going-into SEC40 and coming-out of SEC40 */
	sec_fq = 20480 + (config->tunnel_id * 2);

	/* initialize tunnel entry and add to tunnel table */
	/* initialize tunnel entry and add to tunnel table */
	if (!(entry = ipsec_create_tunnel_entry(ipsec_stack->itt))) {
		APP_ERROR("Unable to Allocate tunnel entry");
		return -ENOMEM;
	}
	entry->tunnel_id = config->tunnel_id;
	entry->src_addr = config->src_ip;
	entry->dst_addr = config->dst_ip;
	entry->tunnel_src_addr = config->tunnel_src_ip_addr;
	entry->tunnel_dst_addr = config->tunnel_dst_ip_addr;
	entry->spi = config->spi;
	entry->seq_num = config->seq_num;
	entry->esp_dec_key = (uint8_t *) ealg->alg_key;
	entry->esp_auth_key = (uint8_t *) aalg->alg_key;
	APP_DEBUG("Added Decap Tunnel src = %x, dst = %x, spi = %x",
		 config->tunnel_src_ip_addr, config->tunnel_dst_ip_addr,
		 config->spi);

	if (false == ipsec_add_tunnel_entry(ipsec_stack->itt, entry)) {
		ipsec_free_entry(ipsec_stack->itt, entry);
		APP_ERROR("Tunnel Entry Couldn't be added");
		return -ENOMEM;
	}

	/* create the pdb based on tunnel sa information */
	ipsec_decap_initialize_fqs(entry, ipsec_stack->ipsec_ctxt, sec_fq, ealg,
				   aalg);

	/* populate route cache */
	tunnel_route = true;
	scope = ROUTE_SCOPE_DECAP;
	/* entry for each remote to remote host-pair */
	populate_route_cache(config->tunnel_src_ip_addr, 
			     config->tunnel_dst_ip_addr, *next_hop_addr,
			     ipsec_stack, tunnel_route, entry, false, scope);

	/* populate route cache */
	tunnel_route = false;
#ifdef PME_SWITCH
	if (ipsec_flow_id)
		scope = ROUTE_SCOPE_PME;
#endif
	/* entry for each remote to remote host-pair */
	populate_route_cache(config->src_ip, config->dst_ip, *next_hop_addr,
			     ipsec_stack, tunnel_route, entry, true, scope);
	return 0;
}
int32_t ipsec_tunnel_create(struct ipsec_tunnel_config_entry_t *config,
			 struct ipsec_stack_t *ipsec_stack,
			 uint32_t * next_hop_addr, uint32_t mode,
			 struct lwe_ctrl_sa_algo *ealg,
			 struct lwe_ctrl_sa_algo *aalg)
{
	struct sg_entry_t *sg;
	void *split_key, *alg_key, *job_desc;
	struct qm_fd fd;
	uint16_t bufsize;

	job_desc = memalign(CACHE_LINE_SIZE, 256);
	if (job_desc == NULL) {
		APP_ERROR("No More Buffers left for Job Desc");
		return -ENOMEM;
	}
	memset(job_desc, 0, 256);

	split_key = memalign(CACHE_LINE_SIZE, 60);
	if (split_key == NULL) {
		APP_ERROR("No More Buffers left for split key");
		free(job_desc);
		return -ENOMEM;
	}
	memset(split_key, 0, 60);

	alg_key = memalign(CACHE_LINE_SIZE, 60);
	if (alg_key == NULL) {
		APP_ERROR("No More Buffers left for Auth Algo key");
		free(job_desc);
		free(split_key);
		return -ENOMEM;
	}

	bufsize = 256;
	memcpy(alg_key, aalg->alg_key, 20);
	APP_DEBUG("ipsec_tunnel_create: before cnstr_jobdesc_mdsplitkey");
	if (0 != cnstr_jobdesc_mdsplitkey(job_desc, &bufsize,
				alg_key, OP_ALG_ALGSEL_SHA1,
				split_key)) {
		APP_ERROR("Unable to create Job descriptor");
		free(job_desc);
		free(split_key);
		return -EINVAL;
	}		

	sg = memalign(CACHE_LINE_SIZE, 2*sizeof(struct sg_entry_t));
	if (sg == NULL) {
		APP_ERROR("No More Buffers left for Auth Algo key");
		free(job_desc);
		free(split_key);
		free(alg_key);
		return -ENOMEM;
	}
	memset(sg, 0, 2*sizeof(struct sg_entry_t));
	sg->addr_hi = 0;
	sg->addr_lo = (uint32_t) APP_VTOP(split_key);
	sg->offset = 0;
	sg->length = 60;
	sg->bpid = 0;
	sg->extension = 0;
	sg->final = 0;
	sg->reserved_offset = 0;
	sg->reserved_zero = 0;
	sg->reserved_zero2 = 0;

	/* Create Job Desc */
	APP_DEBUG("After cnstr_jobdesc_mdsplitkey: %d", bufsize);

	/* input buffer */
	sg++;
	sg->addr_hi = 0;
	sg->addr_lo = APP_VTOP(job_desc);
	sg->length = bufsize * 4;
	sg->offset = 0;
	sg->bpid = 0;
	sg->extension = 0;
	sg->final = 1;
	sg--;
	fd.addr_hi = 0;
	fd.addr_lo = APP_VTOP(sg);
	fd.bpid = 0;
	fd._format1 = qm_fd_compound;
	fd.cong_weight = 0;
	fd.cmd = 0;

	/* Enqueue on the FQ */
	/* NB: the use of QMAN_ENQUEUE_FLAG_WAIT is prohibited in LWE
	 * applications. To avoid thrashing, we implement a throttled poll for
	 * retrying the enqueue. */
loop:
	if (qman_enqueue(g_splitkey_fq_to_sec, &fd, 0)) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	/* Dequeue the reponse */
	g_key_split_flag = 0;
	while (g_key_split_flag == 0)
		qman_poll();

	aalg->alg_key_ptr = APP_VTOP(split_key);
	aalg->alg_key_len = 40;
	free(alg_key);
	free(job_desc);
	free(sg);
	if (mode == ENCRYPT) {
		return ipsec_tunnel_encap_init(config, next_hop_addr,
					ipsec_stack, ealg, aalg);
	}
	return ipsec_tunnel_decap_init(config, next_hop_addr,
					ipsec_stack, ealg, aalg);
}

/* This routine has been modified to support ipsec traffic */
void populate_route_cache(uint32_t src_ip,
			  uint32_t dst_ip,
			  uint32_t hop_ip,
			  struct ipsec_stack_t *ipsec_stack,
			  bool tunnel_route, struct ipsec_tunnel_t *tunnel,
			  bool add_entry, uint8_t scope)
{
	struct rc_entry_t *entry;
	struct rt_dest_t *dest;
	struct rt_dest_t *dest_prev;

	if (!(dest = rt_dest_alloc(ipsec_stack->ip_stack->rt))) {
		APP_ERROR("%s: Unable to allocate route table entry",
			  __func__);
		return;
	}

	dest->next = NULL;
	APP_DEBUG
	    ("Creating Route Cache for src = %x and destination %x, hop_ip = %x",
	     src_ip, dst_ip, hop_ip);
	dest->scope = scope;
#ifdef PME_SWITCH
	if (ipsec_flow_id) {
		if ((scope == (ROUTE_SCOPE_ENCAP | ROUTE_SCOPE_PME)) ||
			(scope == ROUTE_SCOPE_PME)) {
			dest->flow_id = ipsec_flow_id;
			if (g_flow_to_entry[ipsec_flow_id]) {
				APP_INFO
			    	("Route cache entry for flow_id %d already present",
			     	ipsec_flow_id);
				dest_prev = g_flow_to_entry[ipsec_flow_id];
				dest->pme_ctx = dest_prev->pme_ctx;
			}
		}
	}
#endif
	if (tunnel_route == true) {
		dest->dev = NULL;
		dest->neighbor = NULL;
		dest->tunnel = tunnel;	/* Required for encap tunnel */
		if (scope == ROUTE_SCOPE_DECAP)
			return;
	} else {
		dest->neighbor =
		    neigh_lookup(ipsec_stack->ip_stack->arp_table,
				 hop_ip,
				 ipsec_stack->ip_stack->arp_table->proto_len);
		if (dest->neighbor == NULL) {
			APP_INFO("Neighbor not found");
			return;
		}
		dest->dev = dest->neighbor->dev;
		tunnel->dest = dest;
		if (add_entry == false)
			return;
	}

	if (!(entry = rc_create_entry(ipsec_stack->ip_stack->rc))) {
		APP_ERROR("No free RC entry available");
		rt_dest_try_free(ipsec_stack->ip_stack->rt, dest);
		return;
	}
	entry->saddr = src_ip;
	entry->daddr = dst_ip;

	if (!(entry->stats =
		stats_memalign(CACHE_LINE_SIZE,
			sizeof(struct rc_entry_statistics_t)))) {
		APP_ERROR("%s: Unable to allocate memory for route stats",
			__func__);
		rt_dest_try_free(ipsec_stack->ip_stack->rt, dest);
		return;
	}
	memset(entry->stats, 0, sizeof(struct rc_entry_statistics_t));

	refcount_acquire(dest->refcnt);
	entry->dest = dest;
	entry->tos = IP_HEADER_DEFAULT_TOS;
	if (rc_add_update_entry(ipsec_stack->ip_stack->rc, entry) == false) {
		APP_INFO("Route cache entry updated");
		rc_free_entry(ipsec_stack->ip_stack->rc, entry);
	}
#ifdef PME_SWITCH
	if (ipsec_flow_id) {
		if ((scope == (ROUTE_SCOPE_ENCAP | ROUTE_SCOPE_PME)) ||
			(scope == ROUTE_SCOPE_PME))
				g_flow_to_entry[ipsec_flow_id] = dest;
	}
#endif
}
