/**
 \file ipsec_sec.c
 \brief SEC40 descriptor creation for IPSEC encapsulation and decapsulation
 */
/*
 * Copyright (C) 2008-2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "ipsec_sec.h"
#include "mmap.h"
#include "net/frame_desc.h"
#include "app_common.h"
#include "dcl.h"
#include "dpa/dpa_bman.h"

extern struct dpa_buff_allocator *ipsec_buff_allocator;

/**
 \brief Initializes the SEC40 descriptor
	for encapsulation with preheader and Initialization descriptor
 \param[in] sa Pointer to the SA Info structure
 \param[in] outer_ip_header Pointer to the IP Header 
 \param[in] next_header
 \param[in] ealg Encryption Algo Info
 \param[in] aalg Authentication Algo Info
 \return Pointer to the IPSec Encap SEC40 descriptor structure
 */
void *create_encapsulation_sec_descriptor(struct ipsec_tunnel_t *sa,
					  struct ip_header_t *outer_ip_header,
					  uint8_t next_header,
					  struct lwe_ctrl_sa_algo *ealg,
					  struct lwe_ctrl_sa_algo *aalg)
{
	struct ipsec_encap_descriptor_t *preheader_initdesc;
	uint16_t desc_len;
	uint32_t bpool_maxsize;
	uint8_t maxsize_bpool_id;
	struct bm_buffer bman_buf;
	struct ip_header_t iphdr;
	struct ipsec_encap_pdb pdb;
	struct cipherparams cipher;
	struct authparams auth;
	unsigned char *buff_start = NULL;
	int stat;

	if (ipsec_get_maxsize_bpool(&bpool_maxsize, &maxsize_bpool_id)) {
		APP_ERROR("%s: Did you forgot to add buffers in dts",
			  __func__);
		return NULL;
	}

	if (unlikely
	    (dpa_allocator_get_buff
	     (ipsec_buff_allocator, sizeof(struct ipsec_encap_descriptor_t),
	      &bman_buf) <= 0)) {
		APP_ERROR("%s: No More Buffers left", __func__);
		return NULL;
	}
	preheader_initdesc = (void *)APP_PTOV(bman_buf.lo);
	memset(preheader_initdesc, 0, sizeof(struct ipsec_encap_descriptor_t));

	desc_len = (sizeof(struct ipsec_encap_descriptor_t) -
		    sizeof(struct preheader_t)) / sizeof(uint32_t);

	APP_DEBUG("Desc Len in %s is %x", __func__, desc_len);

	buff_start =
	    (unsigned char *)preheader_initdesc + sizeof(struct preheader_t);
	memcpy(&iphdr, outer_ip_header, sizeof(iphdr));
	memset(&pdb, 0, sizeof(struct ipsec_encap_pdb));
	pdb.seq_num = sa->seq_num;
	pdb.spi = sa->spi;
	pdb.ip_hdr_len = 20;
	pdb.options = PDBOPTS_ESPCBC_TUNNEL | PDBOPTS_ESPCBC_INCIPHDR |
			PDBOPTS_ESPCBC_IPHDRSRC | PDBOPTS_ESPCBC_IVSRC;
	pdb.hmo_cbc.dttl = 1;
#ifdef ENABLE_CKSUM_SEC
	pdb.options |= PDBOPTS_ESPCBC_CKSUM;
#endif
	switch (ealg->alg_type) {
	case AES_CBC:
		cipher.algtype = CIPHER_TYPE_IPSEC_AESCBC;
		break;
	case TRIP_DES_CBC:
		cipher.algtype = CIPHER_TYPE_IPSEC_3DESCBC;
		break;
	default:
		APP_ERROR("%s: Unsupported Encryption Algorithm", __func__);
		return NULL;
	}
	cipher.key = ealg->alg_key;
	cipher.keylen = ealg->alg_key_len * 8;	/* Encryption keysize in bits */

	if (aalg->alg_type == HMAC_SHA1) {
		auth.algtype = AUTH_TYPE_IPSEC_SHA1HMAC_96;
	} else {
		APP_ERROR("%s: Unsupported Authentication Algorithm",
			  __func__);
		return NULL;
	}
	auth.key = aalg->alg_key_ptr;
	auth.keylen = aalg->alg_key_len * 8;	/* SHA1 keysize in bits */


	/* Now construct */
	stat = cnstr_shdsc_ipsec_encap((uint32_t *) buff_start, &desc_len,
						 &pdb, (uint8_t *)&iphdr,
						 &cipher, &auth);

	if (stat) {
		return NULL;
	};

	*(buff_start + 5) = next_header;

	preheader_initdesc->prehdr.hi.field.idlen = desc_len;
	preheader_initdesc->prehdr.lo.field.offset = 1;
	preheader_initdesc->prehdr.lo.field.pool_id = maxsize_bpool_id;
	preheader_initdesc->prehdr.lo.field.pool_buffer_size =
		(uint16_t)bpool_maxsize;

	HEXDUMP(preheader_initdesc, (sizeof(struct ipsec_encap_descriptor_t)));

	return preheader_initdesc;
}

/**
 \brief Initializes the SEC40 descriptor for Decapsulation with preheader and
	Initialization descriptor
 \param[in] sa Pointer to the SA Info structure
 \param[in] ealg Encryption Algo Info
 \param[in] aalg Authentication Algo Info
 \return Pointer to the IPSec Encap SEC40 descriptor structure
 */
void
*create_decapsulation_sec_descriptor(struct ipsec_tunnel_t *sa,
				     struct lwe_ctrl_sa_algo *ealg,
				     struct lwe_ctrl_sa_algo *aalg)
{
	struct ipsec_decap_descriptor_t *preheader_initdesc;
	uint16_t desc_len;
	struct bm_buffer bman_buf;
	uint32_t bpool_maxsize;
	uint8_t maxsize_bpool_id;
	struct ipsec_decap_pdb pdb;
	struct cipherparams cipher;
	struct authparams auth;
	unsigned char *buff_start = NULL;

	int stat;

	if (ipsec_get_maxsize_bpool(&bpool_maxsize, &maxsize_bpool_id)) {
		APP_ERROR("%s: Did you forget to add buffers in dts",
			  __func__);
		return NULL;
	}

	if (dpa_allocator_get_buff
	    (ipsec_buff_allocator, sizeof(struct ipsec_decap_descriptor_t),
	     &bman_buf) <= 0) {
		APP_ERROR("No More Buffers left");
		return NULL;
	}
	preheader_initdesc = (void *)APP_PTOV(bman_buf.lo);
	memset(preheader_initdesc, 0, sizeof(struct ipsec_decap_descriptor_t));
	desc_len = (sizeof(struct ipsec_decap_descriptor_t) -
		    sizeof(struct preheader_t)) / sizeof(uint32_t);

	buff_start =
	    (unsigned char *)preheader_initdesc + sizeof(struct preheader_t);

	memset(&pdb, 0, sizeof(struct ipsec_decap_pdb));
	pdb.seq_num = sa->seq_num;
	pdb.ip_hdr_len = 20;
	pdb.options = PDBOPTS_ESPCBC_TUNNEL | PDBOPTS_ESPCBC_OUTFMT |
			 PDBOPTS_ESPCBC_ARSNONE;
	switch (ealg->alg_type) {
	case AES_CBC:
		cipher.algtype = CIPHER_TYPE_IPSEC_AESCBC;
		break;
	case TRIP_DES_CBC:
		cipher.algtype = CIPHER_TYPE_IPSEC_3DESCBC;
		break;
	default:
		APP_ERROR("%s: Unsupported Encryption Algorithm", __func__);
		return NULL;
	}
	cipher.key = ealg->alg_key;
	cipher.keylen = ealg->alg_key_len * 8;	/* Decryption keysize in bits */

	if (aalg->alg_type == HMAC_SHA1) {
		auth.algtype = AUTH_TYPE_IPSEC_SHA1HMAC_96;
	} else {
		APP_ERROR("%s: Unsupported Authentication Algorithm",
			  __func__);
		return NULL;
	}
	auth.key = aalg->alg_key_ptr;
	auth.keylen = aalg->alg_key_len * 8;	/* SHA1 keysize in bits */

	/* Now construct */
	stat = cnstr_shdsc_ipsec_decap((uint32_t *) buff_start,
					       &desc_len,
					       &pdb, &cipher, &auth);
	if (stat) {
		return NULL;
	}

	preheader_initdesc->prehdr.hi.field.idlen = desc_len;
	/* 1 burst length to be reserved */
	preheader_initdesc->prehdr.lo.field.offset = 1;
	preheader_initdesc->prehdr.lo.field.pool_id = maxsize_bpool_id;
	preheader_initdesc->prehdr.lo.field.pool_buffer_size =
		(uint16_t)bpool_maxsize;

	HEXDUMP(preheader_initdesc, (sizeof(struct ipsec_decap_descriptor_t)));
	return preheader_initdesc;
}
