/**
 \file common_pme.h
 \brief Header file defining structure and hash defines for PME.
 */
/*
 * Copyright (C) 2009-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __COMMON_PME_H__
#define __COMMON_PME_H__

#include <compat.h>
#include <stdlib.h>
#include "init.h"
#include "app_common.h"
#include "fsl_pme.h"

#define INV_PARTITION_NUM 0
#define MAX_SPA_PME_CNTX 1024
#define MAX_RULE_ID	63
#define SRE_RULE_ID_MASK	0x3f
#define SPA_MAGIC_VALUE 0x12344321

#define SIMPLE_FD 	0
#define COMPOUND_FD 	1

#define REPORT_DISABLED 4

enum pme_report_type {
	pme_simple_match_type = 0x01,
	pme_verbose_match_type = 0x02,

	pme_stateless_rule = 0x03,
	pme_dual_state_rule_no_context = 0x04,
	pme_dual_state_rule_with_context = 0x05,
	pme_dual_state_rule_with_context_verbose = 0x06,
	pme_multi_state_rule = 0x07,
	pme_multi_state_rule_verbose = 0x08,

	pme_record_type_mask = 0x0f,

	pme_rule_report_type = 0x70,

	pme_end_of_report_type = 0x80,

	pme_inconclusive_right_record_type_class = 0x10,
	pme_inconclusive_left_record_type_class = 0x20,
	pme_full_match_record_type_class = 0x40,
};

struct pme_report_simple_match_record_t {
	uint8_t type; /* bit 4-7 = pme_simple_match_type */
	uint8_t match_length;	/* Length of matching bytes. ($N) */
	uint8_t unit_offset[6];	/* Scan unit offset in 48 bits, relative to
				the head of stream. Starts at 0. ($Si) */
	uint32_t match_byte_offset;/* Last byte of the match, relative to the
				scan unit. Offset starts at 1 for the
				first byte. ($M) */
	uint32_t exp_tag; /* Tag of matching expression ($T) */
} __attribute__((packed));

struct pme_report_verbose_match_record_t {
	/* Byte 0 bit 4-7 = pme_verbose_match_type
	Byte 1-16 : Same as the simple match report */
	struct pme_report_simple_match_record_t simple_match_record;
	/* Byte 17-20 : last stream position of */
	/* Line Break character ($Ob) */
	uint32_t last_line_break;
	/* Byte 21-24 : last stream position of */
	/* Extended ASCII character ($Ox) */
	uint32_t last_ext_char;
	/* Byte 25 : captured $X size ($Xn) */
	uint8_t  captured_x_size;
	/* Byte 26 : captured $Y size ($Yn) */
	uint8_t  captured_y_size;
	/* Byte 27:5 bits of 0, full match bit ($I[5]),
	inconclusive left indicator ($I[6]),
	inconclusive right indicator ($I[7]) */
	uint8_t inconclusive_detail;
	/* Byte 28-35 : captured $X (0 if no capture) */
	uint64_t captured_x;
	/* Byte 36-43 : captured $Y (0 if no capture) */
	uint64_t captured_y;
} __attribute__((packed));

struct pme_report_stateless_match_record_t {
	uint8_t type;	/* pme_stateless_rule */
} __attribute__((packed));

struct pme_report_verbose2_match_record_t {
	uint8_t type;	/* pme_dual_state_rule_no_context/
					pme_dual_state_rule_with_context/
					pme_multi_state_rule */
	uint16_t		rule_number;
	uint8_t			rule_state;
} __attribute__((packed));

struct pme_report_verbose3_match_record_t {
	uint8_t type; /* pme_multi_state_rule_verbose /
				pme_dual_state_rule_with_context_verbose */
	uint16_t rule_number;
	uint8_t rule_state;
	uint8_t rule_context[16]; /* Rule context (left aligned 16 bytes
				  reported even if rule only uses 2 bytes,
				  4 bytes or 8 bytes). */
	uint8_t global_context[16]; /* Global Context */
	uint16_t session_flags; /* Session Persistent Flags. */
	uint16_t volatile_flags; /* Volatile Flags */
} __attribute__((packed));

struct pme_report_rule_record_t {
	uint8_t type;	/* pme_rule_report_type */
	uint8_t	unused;
	uint16_t length;	/* total report length
					(including type) */
	uint32_t reserved;
	uint32_t rule_id;
} __attribute__((packed));

/**
 \brief stats structure
 */
struct simple_report_t {
	uint32_t frames_scanned;
	uint32_t frames_with_atleast_one_pattern;
	uint32_t frames_with_fail_response;
};

struct detail_report_t {
	uint32_t patterns_count_in_all_frames;
	uint32_t sre_rule_hit_count[MAX_RULE_ID];
	uint32_t truncated_report_count;
};

struct spa_pme_ctx_stats_t {
       uint32_t post_scan_frames;
       uint32_t match_count;
       uint32_t fail_response;
};

struct spa_pme_ctxt_stats_t {
	struct simple_report_t simple_report;
	struct detail_report_t detail_report;
};

/**
 \brief Token structure used to send with scan request to PME
 */
struct spa_pme_ctx_token_t {
	struct pme_ctx_token pme_ctx_token;
	uint32_t flow_id;
	uint32_t args;
	void *ip_ctx;
	void *notes;
};

/**
 \brief PME context information
 */
struct spa_context_info_t {
	bool valid;		/**< Context is initialised or not */
	bool retired_non_empty;
	enum qman_fq_state fq_state;
	uint32_t lock;		/**< For updating stats */
	uint32_t bpid;		/**< Buffer pool id */
	uint32_t number_of_frames;	/**< Number of scanned data buffers */
	uint32_t fq_id;		/**< frame queue id on which initial
				   frames needed to be enqueued */
	struct pme_ctx *spa_pme_ctx;	/**< pme_ctx allocated from bpool */
	struct spa_pme_ctxt_stats_t *pme_ctx_stats;	/**< per context stats */
	struct qman_fq *qman_fq;
};

/**
 \brief Scan callback function
 */
struct scan_cb_func_t {
	void (*scan_cb_simple) (struct pme_ctx *, const struct qm_fd *,
				struct pme_ctx_token *);
							 /**< Fn Ptr for Scan callback for Simple FD input data */
	void (*scan_cb_cmpd_buff) (struct pme_ctx *, const struct qm_fd *,
				   struct pme_ctx_token *);
							    /**< Fn Ptr for Scan callback for input data in Compound FD along with output buffer*/
	void (*scan_cb_cmpd_bpid) (struct pme_ctx *, const struct qm_fd *,
				   struct pme_ctx_token *);
							    /**< Fn Ptr for Scan callback for input data in Compound FD */
};

struct frame_info_t {
	uint32_t *buff;
	uint32_t format;
	uint32_t bpid;
};

struct pme_bpool_t {
	uint32_t size;
};
extern struct scan_cb_func_t scan_cb_func;
extern struct spa_context_info_t spa_context_info[];
extern uint32_t block_scan;
extern struct dpa_buff_allocator *buff_allocator;

extern int free_bpool_buff(void *, uint8_t);
#endif
