/**
 \file wb.c
 \brief White Board Protocol for usage with single CP, multiple DPs
 */
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "wb.h"

struct wb_part_map wb_part[] =  {
	{"p1-linux", 0},
	{"p2-linux", 1},
	{"p3-lwe", 2},
	{"p4-lwe", 3},
	{"", 0xffff}
};

#ifdef WB_LINUX
uint32_t wb_get_time_ms(struct timeval *tv, struct timezone *tz)
{
	memset(tv, 0, sizeof(struct timeval));
	gettimeofday(tv, tz);
	return (tv->tv_sec * 1000000 + tv->tv_usec) / 1000;
}
#endif

uint32_t wb_get_id_from_part_name(char *part_name)
{
	struct wb_part_map *part_map = wb_part;
	int i = 0;

	while (part_map[i].partition_id != 0xffff) {
		if (!(strcmp(part_name, part_map[i].label)))
			return part_map[i].partition_id;
		i++;
	}

	return part_map[i].partition_id;
}

int32_t wb_get_addr_from_id(void *wb_id, struct wb **wb)
{
#ifdef WB_LINUX
	int fd;
	void *map_addr;

	fd = open("/dev/mem", O_RDWR);
	if (fd < 0) {
		printf("\nERROR: failed in opening /dev/mem");
		return WB_FAILURE;
	}

	map_addr = mmap(0, SIZE, PROT_READ | PROT_WRITE,
			MAP_SHARED, fd, (off_t)wb_id);
	if (map_addr == MAP_FAILED) {
		close(fd);
		printf("\nERROR: mmapping the file");
		return WB_FAILURE;
	}

	*wb = map_addr;
#else
	*wb = wb_id;
#endif
	return WB_SUCCESS;
}


/**
 * \brief	Register for using the White Board
 *
 * \details	Called by any partition which wants to
 *		use a particular white board.
 *
 * \param[in]	partition_id -
 * \param[in]	cb_func  -
 * \param[out]  handle  -
 * \param[in]	wb_id -
 *
 * \return
 *
 */
uint32_t wb_register(uint32_t partition_id,
	uint32_t (*wb_cb)(struct wb_handle *handle, struct wb_msg *msg),
	struct wb_handle *handle, void * wb_id)
{
	struct wb *wb = NULL;
	/* Validate wb_id*/

	/* Extract the white board address from the wb_id */
	wb_get_addr_from_id(wb_id, &wb);

#if 0
	/* Check if partition already registered */
	if (wb->valid_partitions_mask & (1 << partition_id))
		return WB_FAILURE;
#endif

	/* Fill handle with parameters */
	handle->wb_cb = wb_cb;
	handle->partition_id = partition_id;
	handle->wb_id = wb_id;
	handle->wb = wb;

	/* Update the partition usage mask for the WB */
	wb->valid_partitions_mask |= (1 << partition_id);
	return WB_SUCCESS;
}


/**
 * \brief	UnRegister for the White Board usage
 *
 * \details	Called by any partition which wants to
 *		stop using the white board
 *
 * \param[in]  handle  -
 *
 * \return
 *
 */
uint32_t wb_unregister(struct wb_handle *handle)
{
	struct wb *wb;

	/* Extract the white board address from the wb_id */
	wb = handle->wb;

	/* Check if partition registered or not*/
	if (!(wb->valid_partitions_mask & (1 << handle->partition_id)))
		return WB_FAILURE;

	/* Update parition usage mask for the WB */
	wb->valid_partitions_mask &= ~(1 << handle->partition_id);

#ifdef WB_LINUX
	munmap(handle->wb_id, SIZE);
#endif
	return WB_SUCCESS;
}


/**
 * \brief	Acquire Lock before writing onto White Board
 *
 * \details	Called to Acquire Lock before writing onto White Board,
 *		and for getting the start of the WB from where one can write.
 *
 * \param[in]  handle  -
 * \param[out]	msg -
 *
 * \return
 *
 */
uint32_t wb_acquire_lock(struct wb_handle *handle, struct wb_msg **msg)
{
	/* */
	*msg = &(handle->wb->msg);

	return WB_SUCCESS;
}


/**
 * \brief	Release Lock after writing onto White Board
 *
 * \details	Called to release Lock after writing onto White Board,
 *		and getting the response back.
 *
 * \param[in]  handle  -
 * \param[out]	msg -
 *
 * \return	void
 *
 */
void wb_release_lock(struct wb_handle *handle, struct wb_msg **msg)
{
	*msg = NULL;
	return;
}


/**
 * \brief	Check for any request/ message on the WB
 *
 * \details	Called by  a partition to see if it received
 *		any message on the WB
 *
 * \param[out]  handle  -
 *
 * \return	void
 *
 */
void wb_poll(struct wb_handle *handle)
{
	struct wb *wb;

	/* Extract the white board address from the wb_id */
	wb = handle->wb;

	/* Validate of this partition can poll on this WB */
	if (!(wb->valid_partitions_mask & (1 << handle->partition_id)))
		return;

	/* Check on the WB for any request for this partition */
	if ((wb->msg.dst_partition[handle->partition_id]) &&
		(wb->msg.result_partition[handle->partition_id] == 0)) {
		if (handle->wb_cb)
			handle->wb_cb(handle, &wb->msg);
	}

	return;
}

/**
 * \brief	Sending message/ request via WB to another partition
 * \details	Called by partition to send a message to
 *		another partition via WB
 * \param[in]	msg -
 * \param[in]	timeout_ms  -
 * \return
 */
int32_t wb_send(struct wb_msg *msg, uint32_t partition_mask,
		uint32_t timeout_ms)
{
	int64_t timeout = 0;
	uint32_t resp_partition_mask = 0;
	uint32_t i = 0;

	/* Get the Dst Mask from the msg */
	for (i = 0; i < WB_MAX_PARTITIONS; i++) {
		if (partition_mask & (1 << i))
			msg->dst_partition[i] = 1;
	}

	/* Check on WB for response in a loop */
	if (timeout_ms == 0) {
		while (1) {
			for (i = 0; i < WB_MAX_PARTITIONS; i++) {
				if (msg->result_partition[i]) {
					resp_partition_mask |= (1<<i);
					msg->dst_partition[i] = 0;
				}
			}
			if (resp_partition_mask == partition_mask)
				return WB_SUCCESS;
		}
	} else {
#ifdef WB_LINUX
		struct timeval tv;
		struct timezone tz;

		tz.tz_minuteswest = 0;
		tz.tz_dsttime = 0;
		timeout = wb_get_time_ms(&tv, &tz) + timeout_ms;
#else
		timeout = get_timebase() + (timebase_freq / 1000 * timeout_ms);
#endif
		do {
			for (i = 0; i < WB_MAX_PARTITIONS; i++) {
				if (msg->result_partition[i]) {
					resp_partition_mask |= (1<<i);
					 msg->dst_partition[i] = 0;
				}
			}

			if (resp_partition_mask == partition_mask)
				return WB_SUCCESS;
#ifdef WB_LINUX
		} while (timeout > wb_get_time_ms(&tv, &tz));
#else
		} while (timeout > get_timebase());
#endif
	}

	/* Check for timout having passed */
	return WB_TIMEOUT;
}
