/**
 \file wb.h
 \brief White Board Protocol for usage with single CP, multiple DPs
 */
/*
 * Copyright (C) 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef WB_H
#define WB_H

#include <stdint.h>
#include <stdbool.h>
#ifdef WB_LINUX
#include <string.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>
#include <sys/time.h>
#else
#include "misc.h"
#include "compiler.h"
#include "timebase.h"
#include <string.h>
#endif

#define WB_MAX_PARTITIONS 8
#define WB_MAX_DATA_SIZE  0

#ifdef WB_LINUX
#define SIZE 64000000
#endif

#define WB_SUCCESS	1
#define WB_FAILURE	2
#define WB_TIMEOUT	3


struct wb_msg {
	uint32_t src_partition;  /* Partition ID of the partition
					generating/ issuing a request */
	uint32_t dst_partition[WB_MAX_PARTITIONS]; /* Partition ID array
					of destination partitions to which
						the request needs to be sent */
	uint32_t msg_type; /* Message Type */
	uint32_t msg_size; /* Size of the message starting from
			.data. field */
	uint32_t result_partition[WB_MAX_PARTITIONS]; /* Array for
			Result/ response of the destination partitions */
	char data[WB_MAX_DATA_SIZE];
};

struct wb {
	uint32_t valid_partitions_mask; /* Set of partitions which have
				registered with this WB, and can use it */
	uint32_t wb_lock;
	struct wb_msg msg; /* WB message */
};

struct wb_handle {
	void *wb_id; /* WB ID in case there are more than one WBs */
	uint32_t partition_id; /* Partition ID of the App using it */

	/* Callback function registered by the App to be called
	when a Request is received for it on WB */
	uint32_t (*wb_cb)(struct wb_handle *handle, struct wb_msg *msg) ;
	struct wb *wb; /* WB address */
};

struct wb_part_map {
	char label[10];
	uint32_t partition_id;
};

#define WB_GET_MSG_TYPE(wb_msg) (wb_msg->msg_type)
#define WB_SET_MSG_TYPE(wb_msg, value)	(wb_msg->msg_type = value)
#define WB_SET_RSLT(wb_handle, result)	\
	wb_handle.wb->msg.result_partition[wb_handle.partition_id] = result
#define WB_RESET_RSLT(wb_msg, partition_id)	\
	(wb_msg->result_partition[partition_id] = 0)
#define WB_GET_RSLT(wb_msg, partition_id)	\
	(wb_msg->result_partition[partition_id])

uint32_t wb_get_id_from_part_name(char *part_name);

uint32_t wb_register(uint32_t partition_id,
	uint32_t (*wb_cb)(struct wb_handle *handle, struct wb_msg *msg),
	struct wb_handle *handle, void * wb_id);

uint32_t wb_unregister(struct wb_handle *handle);

uint32_t wb_acquire_lock(struct wb_handle *handle, struct wb_msg **msg);

void wb_release_lock(struct wb_handle *handle, struct wb_msg **msg);

void wb_poll(struct wb_handle *handle);

int32_t wb_send(struct wb_msg *msg, uint32_t partition_mask,
	uint32_t timeout_ms);

#endif

