/** 
\file	pktwire.c
\brief	Basic Packet wire (Echo back mechanism)
*/
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <pktwire.h>
#include <barrier.h>
#include <app_common.h>
#include <dpa/dpa_eth.h>
#include <net/frame_desc.h>
#include <timebase.h>
#include <timer_wheel.h>
#include <ethernet/eth.h>
#include <shell.h>

#undef ORP_ENABLED

/* Tail Drop is enabled in pktwire 
 * comment following to disable Tail drop
 */
#define TAIL_DROP
#define CGR_SUPPORT

/* PktWire is working in Echomode where IP SRC/DST
 * and MAC SRC/DST are swapped. To see pktwire working
 * in loop-back mode without header manipulation, comment
 * ECHOMODE
 */
#define ECHOMODE

/** dpa buffer allocator object per core */
__PERCPU struct dpa_buff_allocator *buff_allocator;
__PERCPU struct bman_pool *bpool[MAX_NUM_BMAN_POOLS];

__PERCPU struct pkt_annot *pkt_annotations;

__PERCPU uint32_t rx_errors;


uint32_t recv_channel_map;

int num_fq_close_pending;
static volatile enum application_state pktwr_state = INVALID;

/**
\brief		Send packet back to same interface

\param[in]	qm_fd - Frame descriptor to be send out

\param[out]	port - One of the FM port to send packet out to

\return		0 on success, -ve value for failure
*/
static void pktwr_send_out(struct qm_fd *p_fd, uint8_t port_index,
					const struct qm_dqrr_entry *dqrr)
{
	struct qman_fq *p_fq;
	uint32_t ret;

	p_fq = dpa_eth_ports[port_index].egress.fq[0];
loop:
#ifndef ORP_ENABLED
	ret = qman_enqueue(p_fq, p_fd, QMAN_ENQUEUE_FLAG_DCA |
				QMAN_ENQUEUE_FLAG_DCA_PTR(dqrr));
#else
	ret = qman_enqueue_orp(p_fq, p_fd, QMAN_ENQUEUE_FLAG_DCA |
				QMAN_ENQUEUE_FLAG_DCA_PTR(dqrr),
				pkt_annotations->orp_fq,
				pkt_annotations->seq);
#endif

	if (unlikely(ret)) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	APP_DEBUG("Sent packet to Fman");
}

/**
\brief		Calculates Port Information from Frame Queue ID

\param[in]	fq_id - Frame Queue Identifier

\return		Port ID of incoming frame
*/
static uint32_t fqid_to_port(uint32_t fq_id)
{
	uint32_t fq_start;
	uint32_t fq_end;
	uint32_t port_index;

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		fq_start = dpa_eth_ports[port_index].pcd_ingress.fq_start;
		fq_end = fq_start + dpa_eth_ports[port_index].pcd_ingress.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end)
			return port_index;
	}

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		fq_start = dpa_eth_ports[port_index].ingress.fq_start;
		fq_end = fq_start + dpa_eth_ports[port_index].ingress.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end)
			return port_index;
	}

	return INV_PORT;
}

/**
\brief		Switch IP and MAC address in frame and send it out

\param[in]	dqrr - Dequeue Receive Ring entry to be processed

\param[out]	None
*/
static enum qman_cb_dqrr_result pktwr_proc_rx_default(struct qman_portal *p,
						   struct qman_fq *fq,
						   const struct
						   qm_dqrr_entry *dqrr)
{
	uint8_t *p_pkt;
#ifdef ECHOMODE
	struct ip_header_t *ip_hdr = NULL;
	uint32_t tmp;
#endif
	uint32_t src_port;
	struct sg_entry_t *sgentry;
	phys_addr_t addr = 0;
	uint32_t offset;

	if (likely(dqrr->fd.format == 0)) {
		APP_DEBUG("%s: received packet is of contiguous format",
			 __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo;
		offset = dqrr->fd.offset;
	} else if (unlikely(dqrr->fd.format == QM_FD_FORMAT_SG)) {
		APP_DEBUG("%s: received packet is of sg format", __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo + dqrr->fd.offset;
		sgentry = (struct sg_entry_t *)ptov(addr);

		addr = sgentry->addr_hi;
		addr = (addr << 32) + sgentry->addr_lo;
		offset = sgentry->offset;
	} else {
		APP_ERROR("Unsupported Frame Format");
		return qman_cb_dqrr_consume;
	}

	pkt_annotations = ptov(addr);
	p_pkt = (uint8_t *)pkt_annotations + offset;
	src_port = fqid_to_port(dqrr->fqid);
	if (unlikely(INV_PORT == src_port)) {
		APP_ERROR("Frame from unknown port, Dropping it.");
		lwe_free_fd(buff_allocator, &dqrr->fd);
		return qman_cb_dqrr_consume;
	}

#ifdef ORP_ENABLED
	pkt_annotations->orp_fq = fq;
	pkt_annotations->seq = dqrr->seqnum;
#endif

#ifdef ECHOMODE
	ip_hdr = (struct ip_header_t *)(p_pkt + ETH_HDR_LEN);
	ether_header_swap((struct ethernet_header_t *)p_pkt);
	tmp = ip_hdr->dst_addr.word;
	ip_hdr->dst_addr = ip_hdr->src_addr;
	ip_hdr->src_addr.word = tmp;
#endif
	pktwr_send_out((struct qm_fd *)&dqrr->fd, (uint8_t)src_port, dqrr);

	return qman_cb_dqrr_defer;
}

#ifdef CGR_SUPPORT
static __PERCPU int congestion_state;

/**
 \brief Congestion Group Callback handler
 \param[inout] congested Provides state of congestion group
 */
static void pktwr_cgr_cb(struct qman_portal *qm __UNUSED, struct qman_cgr *cgr,
			 uint32_t congested)
{
	APP_INFO("%s: Congested Group %x is %s congested\n", __func__, cgr->cgrid, \
		congested ? "" : "not");
	congestion_state = congested;
}
#endif

/**
\brief		Switch IP and MAC address in frame and send it out

\param[in]	dqrr - Dequeue Receive Ring entry to be processed

\param[out]	None
*/
static enum qman_cb_dqrr_result pktwr_proc_rx_pcd(struct qman_portal *p,
						   struct qman_fq *fq,
						   const struct
						   qm_dqrr_entry *dqrr)
{
	uint8_t *p_pkt;
#ifdef ECHOMODE
	struct ip_header_t *ip_hdr = NULL;
	uint32_t tmp;
#endif
	uint32_t src_port;
	struct sg_entry_t *sgentry;
	phys_addr_t addr = 0;
	uint32_t offset;

	if (likely(dqrr->fd.format == 0)) {
		APP_DEBUG("%s: received packet is of contiguous format",
			 __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo;
		offset = dqrr->fd.offset;
	} else if (unlikely(dqrr->fd.format == QM_FD_FORMAT_SG)) {
		APP_DEBUG("%s: received packet is of sg format", __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo + dqrr->fd.offset;
		sgentry = (struct sg_entry_t *)ptov(addr);

		addr = sgentry->addr_hi;
		addr = (addr << 32) + sgentry->addr_lo;
		offset = sgentry->offset;
	} else {
		APP_ERROR("Unsupported Frame Format");
		return qman_cb_dqrr_consume;
	}

	pkt_annotations = ptov(addr);
	p_pkt = (uint8_t *)pkt_annotations + offset;
	src_port = pkt_annotations->parse.port_id;
	if (unlikely(INV_PORT == src_port)) {
		APP_ERROR("Frame from unknown port, Dropping it.");
		lwe_free_fd(buff_allocator, &dqrr->fd);
		return qman_cb_dqrr_consume;
	}

#ifdef ORP_ENABLED
	pkt_annotations->orp_fq = fq;
	pkt_annotations->seq = dqrr->seqnum;
#endif

#ifdef ECHOMODE
	ip_hdr = (struct ip_header_t *)(p_pkt + ETH_HDR_LEN);
	ether_header_swap((struct ethernet_header_t *)p_pkt);
	tmp = ip_hdr->dst_addr.word;
	ip_hdr->dst_addr = ip_hdr->src_addr;
	ip_hdr->src_addr.word = tmp;
#endif
	pktwr_send_out((struct qm_fd *)&dqrr->fd, (uint8_t)src_port, dqrr);

	return qman_cb_dqrr_defer;
}

/**
\Brief		DC portal Enqueue Rejection Handler

\param[out]	msg Message Ring Entry

\notes		As of now DC portal0 and DCportal 1 support ERN
		to MSG ring of a SW portal
*/
void pktwr_proc_dc_ern(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	APP_ERROR("%s: RC = %x, FQID = %x", __func__,\
			msg->dcern.rc, msg->dcern.fqid);
	lwe_free_fd(buff_allocator, &msg->dcern.fd);
	return;
}

/**
\Brief		Enqueue Rejection Handler

\param[out] 	msg Message Ring Entry
*/
void pktwr_proc_ern(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	APP_ERROR("%s: RC = %x, seqnum = %x", __func__,\
			msg->ern.rc, msg->ern.seqnum);
	lwe_free_fd(buff_allocator, &msg->ern.fd);
	return;
}

/**
 \brief 	PktWire Tx Error Frame Handler
 */
static enum qman_cb_dqrr_result pktwr_proc_tx_err(struct qman_portal *p,
					 struct qman_fq *fq,
					 const struct qm_dqrr_entry *dqrr)
{
	APP_INFO("%s: Frame queue ID = %d, status = %x",\
		 __func__, dqrr->fqid, dqrr->fd.status);
	lwe_free_fd(buff_allocator, &dqrr->fd);

	return qman_cb_dqrr_consume;
}

/**
\brief		PktWire Tx Confirmation Frame Handler
*/
static enum qman_cb_dqrr_result pktwr_proc_tx_confirm(struct qman_portal *p,
					 struct qman_fq *fq,
					 const struct qm_dqrr_entry *dqrr)
{
	APP_INFO("%s: Frame queue ID = %d, status = %x",\
		__func__, dqrr->fqid, dqrr->fd.status);
	lwe_free_fd(buff_allocator, &dqrr->fd);

	return qman_cb_dqrr_consume;
}

/**
\brief		PktWire Error Frame Handler
*/
static enum qman_cb_dqrr_result pktwr_proc_rx_err(struct qman_portal *p,
						  struct qman_fq *fq,
						  const struct
						  qm_dqrr_entry *dqrr)
{
	APP_DEBUG("%s: Frame queue ID = %d, status = %x",\
		 __func__, dqrr->fqid, dqrr->fd.status);

	rx_errors++;

	lwe_free_fd(buff_allocator, &dqrr->fd);

	return qman_cb_dqrr_consume;
}

static int cmd_print_stats(int argc, char *argv[])
{
	printf("rx_errors: %u\n", rx_errors);
	return 0;
}

/*
 * This function is used to move a FQ to OOS state, destroy it and free
 * the associated memory with FQ structure.
 */
static void close_fq(uint32_t swi_id, void *data)
{
	struct qman_fq *fq = data;
	uint32_t fqid;

	fqid = fq->fqid;
	if (0 > qman_oos_fq(fq)) {
		APP_ERROR("Moving FQID: %u to OOS failed", fq->fqid);
		free(fq);
		return;
	}

	qman_destroy_fq(fq, 0);
	free(fq);

	num_fq_close_pending--;

	APP_INFO("Closed FQID: %u", fqid);
}

/*This function waits till the FQ becomes empty and then closes it*/
static void wait_fq_empty(uint32_t swi_id, void *data)
{
	struct qman_fq *fq = data;
	uint32_t vdqcr;
	uint32_t flags;

	APP_INFO("Waiting for FQID: %u to become empty", fq->fqid);

	vdqcr = QM_VDQCR_PRECEDENCE_VDQCR | QM_VDQCR_NUMFRAMES_TILLEMPTY;

	/* NB: the use of QMAN_VOLATILE_FLAG_WAIT is prohibited in LWE
	 * applications. To avoid thrashing, we implement a throttled poll for
	 * completion. */
	if (qman_volatile_dequeue(fq, 0, vdqcr)) {
		APP_ERROR("qman_volatile_dequeue failure for FQID: %u",
			fq->fqid);
		free(fq);
		return;
	}
loop:
	qman_fq_state(fq, NULL, &flags);
	if (flags & QMAN_FQ_STATE_VDQCR) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	/*Now the FQ is empty, close it*/
	close_fq(SWI_ID_INV, data);
}

/* This function processes FQ state change notifications. This function is
 * called from ISR context. Thus this function must not make any blocking
 * and heavy weight APIs. When the FQ get retired and also gets empty, a
 * SWI is posted towards application's main loop for taking FQs OOS and
 * destroy them.
 */
static void pktwr_proc_fq_state_chg(struct qman_portal *qm,
				struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	uint8_t verb;
	uint8_t fqs;

	/*Extract the response verb*/
	verb = msg->verb & ~QM_MR_VERB_VBIT;

	if (QM_MR_VERB_FQRL == verb) {
		if (0 > swi_add(SWI_ID_INV, fq, close_fq, SWI_PRI_HIGH))
			APP_ERROR("Cannot add SWI for closing FQ");

		return;
	}

	fqs = msg->fq.fqs;

	if ((QM_MR_VERB_FQRN == verb) || (QM_MR_VERB_FQRNI == verb)) {
		if (fqs & QM_MR_FQS_NOTEMPTY) {
			APP_INFO("Retired FQID: %u not yet empty", fq->fqid);

			if (0 > swi_add(SWI_ID_INV, fq, wait_fq_empty,
					SWI_PRI_HIGH))
				APP_ERROR("Cannot add SWI for emptying FQ");
			return;
		}

		if (fqs & QM_MR_FQS_ORLPRESENT) {
			/*We need to wait till we get FQRL*/
			APP_INFO("Retired FQID: %u has ORL list", fq->fqid);
			return;
		}

		/*Else FQ is empty and can be destroyed*/
		if (0 > swi_add(SWI_ID_INV, fq,	close_fq, SWI_PRI_HIGH)) {
			APP_ERROR("Cannot add SWI for closing FQ");
			return;
		}

		return;
	}

	APP_ERROR("Unexpected FQ state change verb: %d", verb);
}

/*This function is called by all cores*/
static int pktwr_common_init(void)
{
	int ecode;


	buff_allocator = dpa_allocator_create(&ecode);
	if (NULL == buff_allocator) {
		APP_ERROR("dpa_buff_allocator_create failed");
		return -1;
	}

	APP_INFO("Buffer allocator created...");
	return 0;
}

/*BMAN pool initialisation function*/
static int pktwr_bpools_init(void)
{
	struct bman_pool_params params;
	uint32_t bpid;

	/*Create BMAN pools */
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		/*Else create the pool object */
		params.flags = BMAN_POOL_FLAG_STOCKPILE;
		params.cb = NULL;
		params.cb_ctx = NULL;
		params.bpid = bpid;

		bpool[bpid] = bman_new_pool(&params);
		if (NULL == bpool[bpid]) {
			APP_ERROR("bman_new_pool failed for BPID: %u", bpid);
			return -1;
		}

		if (0 != dpa_allocator_add_bpool(buff_allocator,
						dpa_bpools[bpid].size,
						bpid, bpool[bpid])) {
			APP_ERROR("dpa_allocator_add_bpool failed");
			return -1;
		}

		APP_INFO("Added BPID: %u, Size: %u", bpid,
			 dpa_bpools[bpid].size);
	}

	return 0;
}

/*BMAN pool destroy function*/
static void pktwr_bpools_destroy(void)
{
	uint32_t bpid;

	/*Free the BMAN buffer pools*/
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		bman_free_pool(bpool[bpid]);
		APP_INFO("BPID: %u destroyed", bpid);
	}

	/*Free the DPA allocator*/
	dpa_allocator_free(buff_allocator);
	APP_INFO("DPA allocator freed");
}

/*This function waits for shutdown process to complete and also cleans up
per core resources*/
static void wait_for_shutdown_complete(void)
{
	void *p_data;
	uint32_t swi_id;

	while (1) {
		if (SHUTDOWN_COMPLETE == pktwr_state)
			break;

		/*This would dequeue any remanining notifications from QMAN*/
		qman_poll();

		while (SWI_ID_INV != (swi_id = swi_activate(&p_data))) {
			switch (swi_id) {
			case SWI_ID_TIMER_EXPIRY:
				proc_timer_expiry((uint32_t) p_data);
				break;

			default:
				break;
			}
		}
	}

	if (0 != timer_framework_stop()) {
		APP_ERROR("timer_framework_stop failed");
		return;
	}

	pktwr_bpools_destroy();
	APP_INFO("Destroyed BMAN pools and buffer allocator");
}

static void pktwr_shutdown_complete(uint32_t timer_id, void *p_data)
{
	if (0 == num_fq_close_pending) {
		APP_INFO("Moving application state to SHUTDOWN_COMPLETE");
		pktwr_state = SHUTDOWN_COMPLETE;
		return;
	}

	APP_WARN("Num Tx FQ close pending: %d", num_fq_close_pending);
	APP_WARN("Waiting for some more time for FQs to retire");

	timer_id = reset_timer(timer_id, 0);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Resetting shutdown complete timer failed");
}

static void pktwr_shutdown_rx_complete(uint32_t timer_id, void *p_data)
{
	if (num_fq_close_pending) {
		APP_WARN("Num Rx FQ close pending: %d", num_fq_close_pending);
		APP_WARN("Waiting for some more time for FQs to retire");

		timer_id = reset_timer(timer_id, 0);
		if (INV_TIMER_ID == timer_id)
			APP_ERROR("Reset shutdown Rx complete timer failed");
		return;
	}

	APP_INFO("Moving application state to SHUTDOWN_RX_COMPLETE");
	pktwr_state = SHUTDOWN_RX_COMPLETE;

	/* At this stage, all the traffic which was received has been
	 * processed and we can now move forward to retire Tx FQs
	 */
	num_fq_close_pending = dpa_eth_tx_fq_retire();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring Tx FQs failed");
		return;
	}

	APP_INFO("Number of Tx FQs close pending: %d", num_fq_close_pending);

	/*
	 * If there are no Tx FQs which are pending to be closed, directly
	 * shutdown the application
	 */
	if (!num_fq_close_pending) {
		pktwr_shutdown_complete(INV_TIMER_ID, NULL);
		return;
	}

	/*Start a single shot timer to wait for all Tx FQs to get retired*/
	timer_id = start_timer(PKTWIRE_SHUTDOWN_TIMER_MS, false,
				NULL, SWI_PRI_HIGH,
				pktwr_shutdown_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown complete timer failed");
	else
		APP_INFO("Shutdown complete timer started...");
}


static void pktwr_process_shutdown(void)
{
	uint32_t timer_id;

	APP_INFO("Processing shutdown command ..");
	/*Bootcore is responsible for partition cleanup*/
	pktwr_state = SHUTDOWN_PENDING;

	/*Disable both Rx mac ports*/
	if (-1 == dpa_eth_disable_ports(true, false))
		APP_ERROR("Disabling MAC-Rx port failed");
	else
		APP_INFO("Disabled MAC-Rx ports");

	/*Put all Rx frame queues in retired state*/
	/*This would effectively disable the MAC-less Rx port*/
	num_fq_close_pending = dpa_eth_rx_fq_retire();
	if (0 > num_fq_close_pending) {
		APP_ERROR("Retiring Rx FQs failed");
		return;
	}

	APP_INFO("Number of Rx FQs close pending: %d", num_fq_close_pending);

	/*If there is no FQ close pending, directly move to the next state*/
	if (!num_fq_close_pending) {
		pktwr_shutdown_rx_complete(INV_TIMER_ID, 0);
		return;
	}

	/*Start a timer to complete Rx shutdown*/
	timer_id = start_timer(PKTWIRE_SHUTDOWN_TIMER_MS,
				false, NULL, SWI_PRI_HIGH,
				pktwr_shutdown_rx_complete, NULL);
	if (INV_TIMER_ID == timer_id)
		APP_ERROR("Starting shutdown RX complete timer failed");
	else
		APP_INFO("Started shutdown Rx complete timer");
}

/**
\brief		Packet Wire entry function

\param[in]	argc - count of input parameters

\param[in]	argv - Value of the parameters passed

\return		Return status of application
*/
int main(int argc, char *argv[])
{
#ifdef TAIL_DROP
	struct td_param pcd_td_param;
#endif
#ifdef ORP_ENABLED
	struct qman_orp_pcd pcd_orp;
#endif
	struct td_param *pcd_td_ptr = NULL;
	struct qman_orp_pcd *pcd_orp_ptr = NULL;

	struct qman_fq_cb rx_default_cb = { pktwr_proc_rx_default,
		NULL, pktwr_proc_dc_ern, pktwr_proc_fq_state_chg };

	struct qman_fq_cb rx_pcd_cb = { pktwr_proc_rx_pcd,
		NULL, pktwr_proc_dc_ern, pktwr_proc_fq_state_chg };

	struct qman_fq_cb rx_err_cb = { pktwr_proc_rx_err,
		NULL, pktwr_proc_dc_ern, pktwr_proc_fq_state_chg };

	struct qman_fq_cb tx_cb = { NULL,
		pktwr_proc_ern, NULL, pktwr_proc_fq_state_chg };

	struct qman_fq_cb tx_confirm_cb = { pktwr_proc_tx_confirm,
		NULL, NULL, pktwr_proc_fq_state_chg };

	struct qman_fq_cb tx_err_cb = { pktwr_proc_tx_err,
		NULL, pktwr_proc_dc_ern, pktwr_proc_fq_state_chg };

	uint32_t swi_id;
	void *p_data;

	if (0 == lwe_instance) {
		APP_INFO("Welcome to PktWire Application ...");
		add_command("print_stats", cmd_print_stats, "Print app stats");
		pktwr_state = INITIALIZING;
	} else {
		while (INITIALIZING != pktwr_state)
			;
	}

	/*Control core initialises global database*/
	if (0 == lwe_instance) {
		if (unlikely(0 >= dpa_eth_read(&recv_channel_map))) {
			APP_ERROR("dpa_eth_read failed");
			goto err_out;
		}

		if (unlikely(0 != dpa_eth_fq_create(&rx_default_cb,
						    &rx_pcd_cb,
						    &rx_err_cb,
						    &tx_cb,
						    &tx_confirm_cb, &tx_err_cb,
						    0))) {
			APP_ERROR("dpa_eth_fq_create failed");
			goto err_out;
		}

		APP_INFO("Num ethernet ports discovered: %u",
			 g_num_dpa_eth_ports);
#ifdef TAIL_DROP
		pcd_td_param.mant = 1; /**< Mantissa is 1 */
		pcd_td_param.exp = 12; /**< Exponent is 12 */
		pcd_td_ptr = &pcd_td_param;
#ifdef CGR_SUPPORT
		/* Allow 4k Frames in CG */
		pcd_td_param.flag = QM_FQCTRL_CGE;
		pcd_td_param.cgr_cb = (qman_cb_cgr)pktwr_cgr_cb;
		pcd_td_param.cgr_id = (uint8_t)get_lwe_id(); /* First CGR ID */
		if (!init_cgr(pcd_td_ptr)) {
			APP_ERROR("Unabled to create Congestion Group");
			return -EINVAL;
		}
#else
		pcd_td_param.flag = QM_FQCTRL_TDE;
#endif
#endif
#ifdef ORP_ENABLED
		pcd_orp.orprws = 0; /* Window Size is 32 */
		pcd_orp.oa = 1; /* NESN Autoadvance */
		pcd_orp.olws = 3; /* Late arrival always accepted */
		pcd_orp_ptr = &pcd_orp;
#endif
		if (unlikely(0 != dpa_eth_fq_init(64, 64, 64,
				pcd_orp_ptr, pcd_td_ptr))) {
			APP_ERROR("dpa_eth_fq_init failed");
			goto err_out;
		}
	}
	/*Wait till DPA is Initialized */
	while (0 == g_num_dpa_eth_ports) ;

	/*Common initialisation routine for all cores*/
	if (0 > pktwr_common_init()) {
		APP_ERROR("pktwr_common_init failed");
		goto err_out;
	}

	/*Create BMAN pools */
	if (0 > pktwr_bpools_init()) {
		APP_ERROR("pktwr_bpool_init failed");
		goto err_out;
	}

	if (0 == lwe_instance) {
		if (unlikely(0 != dpa_eth_enable_ports(false))) {
			APP_ERROR("dpa_eth_enable_ports failed");
			goto err_out;
		}
		pktwr_state = RUNNING;

	} else {
		while (RUNNING != pktwr_state)
			;
	}

	qman_static_dequeue_add(recv_channel_map);
	APP_INFO("Entering Main loop with channel map %d", recv_channel_map);

	while (likely(1)) {
		if (unlikely(SHUTDOWN_PENDING == pktwr_state))
			break;
		qman_poll();
		while (unlikely(SWI_ID_INV !=
			(swi_id = swi_activate(&p_data)))) {

			switch (swi_id) {
			case SWI_ID_TIMER_EXPIRY:
				proc_timer_expiry((uint32_t) p_data);
				break;

			/*Shutdown SWI is received only on bootcore*/
			case SWI_ID_SHUTDOWN: {
				/* Initialise the timer framework */
				if (0 != timer_framework_start()) {
					APP_ERROR("timer_framework_start fail");
					return -1;
				}

				pktwr_process_shutdown();
				break;
			}
			default:
				break;
			}
		}
	}
	/*The following code is executed only in SHUTDOWN_PENDING state*/
	/*Here, all cores wait for shutdown to complete*/
	wait_for_shutdown_complete();
	APP_INFO("Shutdown completed on core: %u", lwe_instance);
	return 0;

err_out:
	APP_ERROR("FATAL error, exiting");
	return 0;
}
