/**
\file	pktwire.cpp
\brief	Basic Packet wire C++ (Echo back mechanism)
*/
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <pktwire.h>
#include <app_common.h>
#include <dpa/dpa_eth.h>
#include <net/frame_desc.h>
#include <timebase.h>

/** Master LWE core does required initialization first */
volatile uint32_t g_ctrl_lwe = INV_LWE_ID;

__PERCPU uint32_t curr_lwe_id = 0;	/**< LWE ID for each core */

/** dpa buffer allocator object per core */
__PERCPU struct dpa_buff_allocator *buff_allocator;

__PERCPU struct pkt_annot pkt_annotations;
uint32_t pktwire_lock;

uint32_t recv_channel_map;

/**
\brief		Send packet back to same interface

\param[in]	qm_fd - Frame descriptor to be send out

\param[out]	port - One of the FM port to send packet out to

\return		0 on success, -ve value for failure
*/
static void pktwr_send_out(struct qm_fd *p_fd, uint8_t port_index)
{
	struct qman_fq *p_fq;
	uint32_t ret;

	p_fq = dpa_eth_ports[port_index].egress.fq[0];

loop:
	ret = qman_enqueue_orp(p_fq, p_fd, 0,
				pkt_annotations.orp_fq,
				pkt_annotations.seq);
	if (unlikely(ret)) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	APP_DEBUG("Sent packet to Fman");
}

/**
\brief		Calculates Port Information from Frame Queue ID

\param[in]	fq_id - Frame Queue Identifier

\return		Port ID of incoming frame
*/
static uint32_t fqid_to_port(uint32_t fq_id)
{
	uint32_t fq_start;
	uint32_t fq_end;
	uint32_t port_index;

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		fq_start = dpa_eth_ports[port_index].pcd_ingress.fq_start;
		fq_end = fq_start + dpa_eth_ports[port_index].pcd_ingress.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end)
			return port_index;
	}

	for (port_index = 0; port_index < g_num_dpa_eth_ports; port_index++) {
		fq_start = dpa_eth_ports[port_index].ingress.fq_start;
		fq_end = fq_start + dpa_eth_ports[port_index].ingress.fq_count;

		if (fq_id >= fq_start && fq_id < fq_end)
			return port_index;
	}

	return INV_PORT;
}

/**
\brief		Switch IP and MAC address in frame and send it out

\param[in]	dqrr - Dequeue Receive Ring entry to be processed

\param[out]	None
*/
static enum qman_cb_dqrr_result pktwr_process_work(struct qman_portal *p,
						   struct qman_fq *fq,
						   const struct
						   qm_dqrr_entry *dqrr)
{
	uint8_t *p_pkt;
	struct ethernet_header_t eth;
	struct ip_header_t ip, *ip_hdr = NULL;
	uint32_t src_port;
	struct sg_entry_t *sgentry;
	phys_addr_t addr;

	src_port = fqid_to_port(dqrr->fqid);
	if (unlikely(INV_PORT == src_port)) {
		APP_ERROR("Frame from unknown port, Dropping it.");
		lwe_free_fd(buff_allocator, &dqrr->fd);
		return qman_cb_dqrr_consume;
	}

	if (dqrr->fd.format == QM_FD_FORMAT_SG) {
		APP_DEBUG("%s: received packet is of sg format", __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo + dqrr->fd.offset;
		sgentry = (struct sg_entry_t *)ptov(addr);

		addr = sgentry->addr_hi;
		addr = (addr << 32) + sgentry->addr_lo + sgentry->offset;
	} else {
		APP_DEBUG("%s: received packet is of contiguous format",
			 __func__);
		addr = dqrr->fd.addr_hi;
		addr = (addr << 32) + dqrr->fd.addr_lo + dqrr->fd.offset;
	}
	p_pkt = (uint8_t *) ptov(addr);

	pkt_annotations.orp_fq = fq;
	pkt_annotations.seq = dqrr->seqnum;
	/* Update Ethernet Header */
	memcpy(eth.source.bytes, p_pkt, ETH_ADDR_LEN);
	memcpy(eth.destination.bytes, p_pkt + ETH_ADDR_LEN, ETH_ADDR_LEN);
	memcpy(&(eth.proto), p_pkt + 2 * ETH_ADDR_LEN, 2);

	/* Update IP Hdr */
	ip_hdr = (struct ip_header_t *)(p_pkt + ETH_HDR_LEN);
	memcpy(&ip, ip_hdr, IP_HEADER_LENGTH_NO_OPTIONS - 2 * IP_ADDRESS_BYTES);
	memcpy(&ip.dst_addr, &ip_hdr->src_addr, IP_ADDRESS_BYTES);
	memcpy(&ip.src_addr, &ip_hdr->dst_addr, IP_ADDRESS_BYTES);
	memcpy(p_pkt, &eth, ETH_HDR_LEN);
	memcpy(ip_hdr, &ip, IP_HEADER_LENGTH_NO_OPTIONS);

	pktwr_send_out((struct qm_fd *)&dqrr->fd, (uint8_t)src_port);

	return qman_cb_dqrr_consume;
}

/**
 \Brief Enqueue Rejection Handler
 \param[out] msg Message Ring Entry
 */
void pktwr_ern_handler(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	APP_ERROR("%s: RC = %x, seqnum = %x", __func__,\
			msg->ern.rc, msg->ern.seqnum);
	lwe_free_fd(buff_allocator, &msg->ern.fd);
	return;
}

/**
 \brief PktWire Error Frame Handler
 */
static enum qman_cb_dqrr_result pktwr_err_handler(struct qman_portal *p,
						  struct qman_fq *fq,
						  const struct
						  qm_dqrr_entry *dqrr)
{
	APP_INFO("%s: Frame queue ID = %d, status = %x",\
		 __func__, dqrr->fqid, dqrr->fd.status);
	lwe_free_fd(buff_allocator, &dqrr->fd);

	return qman_cb_dqrr_consume;
}

/**
\brief		Packet Wire entry function

\param[in]	argc - count of input parameters

\param[in]	argv - Value of the parameters passed

\return	Return status of application
*/
int main(int argc, char *argv[])
{
	struct qman_fq_cb rx_default_cb =
	    { pktwr_process_work, NULL, NULL, NULL };
	struct qman_fq_cb err_cb = { pktwr_err_handler, NULL, NULL, NULL };
	struct qman_fq_cb tx_default_cb = { pktwr_err_handler,
		pktwr_ern_handler, NULL, NULL };

	uint32_t bpid;
	struct bman_pool *bpool;
	struct bman_pool_params params;
	int ecode;

	curr_lwe_id = get_lwe_id();
	spin_lock(&pktwire_lock);
	if (INV_LWE_ID == g_ctrl_lwe) {
		APP_INFO("[Boot Core]: Welcome to PktWire Application!");
		g_ctrl_lwe = curr_lwe_id;
	} else
		APP_INFO("PktWire Application");
	spin_unlock(&pktwire_lock);

	buff_allocator = dpa_allocator_create(&ecode);
	if (unlikely(NULL == buff_allocator)) {
		APP_ERROR("dpa_buff_allocator_create failed");
		goto err_out;
	}
	APP_INFO("Buffer allocator created...");

	if (curr_lwe_id == g_ctrl_lwe) {
		if (unlikely(0 >= dpa_eth_read(&recv_channel_map))) {
			APP_ERROR("dpa_eth_read failed");
			goto err_out;
		}

		if (unlikely(0 != dpa_eth_fq_create(&rx_default_cb,
						    &rx_default_cb,
						    &err_cb,
						    &tx_default_cb,
						    &err_cb, &err_cb, 0))) {
			APP_ERROR("dpa_eth_fq_create failed");
			goto err_out;
		}

		APP_INFO("Num ethernet ports discovered: %u",
			 g_num_dpa_eth_ports);

		if (unlikely(0 != dpa_eth_fq_init(64, 0, 64, NULL, NULL))) {
			APP_ERROR("dpa_eth_fq_init failed");
			goto err_out;
		}

	}

	/*Wait till DPA is Initialized */
	while (0 == g_num_dpa_eth_ports) ;
	APP_INFO("Frames to be recv on channel map: 0x%x", recv_channel_map);

	/*Create BMAN pools */
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		/*Else create the pool object */
		params.flags = 0;
		params.cb = NULL;
		params.cb_ctx = NULL;
		params.bpid = bpid;

		bpool = bman_new_pool(&params);
		if (unlikely(NULL == bpool)) {
			APP_ERROR("bman_new_pool failed for BPID: %u", bpid);
			goto err_out;
		}

		if (unlikely(0 != dpa_allocator_add_bpool(buff_allocator,
							  dpa_bpools[bpid].size,
							  bpid, bpool))) {
			APP_ERROR("dpa_allocator_add_bpool failed");
			goto err_out;
		}

		APP_INFO("Added BPID: %u, Size: %u", bpid,
			 dpa_bpools[bpid].size);
	}

	qman_static_dequeue_add(recv_channel_map);

	APP_INFO("Entering Main loop with channel map %d", recv_channel_map);

	if (curr_lwe_id == g_ctrl_lwe) {
		if (unlikely(0 != dpa_eth_enable_ports(false))) {
			APP_ERROR("dpa_eth_enable_ports failed");
			goto err_out;
		}

	}

	while (likely(1))
		qman_poll();

err_out:
	APP_ERROR("FATAL error, exiting");
	return 0;
}
