/** 
\file	simple_crypto_cp.c
\brief	Basic SEC 4.0 Test Config Tool. It takes SEC 4.0 parameters from user
	as inputs and copy to LWE data plane application's WB area using mmap().
 */
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <argp.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <app_cpdp.h>

#include "../dp/simple_crypto_cpdp.h"

/** start address of LWE data plane WB area */
#define ADDR	LWE_CTRL_SHM_ADDR

/** size of LWE data plane WB area */
#define SIZE	LWE_CTRL_SHM_SIZE

uint32_t g_cmd_params;	/**< bit mask of all parameters provided by user */

/** A string describing how the non-option arguments should look */
static char args_doc[] = "WILL BE DOCUMENTED LATER";

/** A descriptive string about this program */
static char doc[] = "WILL BE DOCUMENTED LATER";

/**
\brief		The OPTIONS field contains a pointer to a vector of struct
		argp_option's 
     
\details	structure has the following fields
\n		name - The name of this option's long option (may be zero)

		key - The KEY to pass to the PARSER function when parsing this
		option,	and the name of this option's short option, if it is
		a printable ascii character

		ARG - The name of this option's argument, if any;
 
		FLAGS - Flags describing this option; some of them are:
\n			OPTION_ARG_OPTIONAL - The argument to this option is
					      optional
\n			OPTION_ALIAS	- This option is an alias for the
					      previous option
\n			OPTION_HIDDEN       - Don't show this option in --help output

     		DOC - A documentation string for this option, shown in --help output

\note		An options vector should be terminated by an option with all fields zero
*/
static struct argp_option options[] = {
	{"mode", 'm', "TEST MODE", 0, "test mode: provide following number\
		\n 1 for perf \
		\n 2 for cipher \
		\n Following two combinations are valid only \
		   and all options are mandatory: \
		\n\t -m 1 -s <buf_size> -n <buf_num_per_core> -o <algo> \
		   -l <itr_num> \
		\n\t -m 2 -t <test_set> -n <buf_num_per_core> -o <algo> \
		   -l <itr_num>\n", 0},
	{"testset", 't', "TEST SET", 0, "THIS OPTION IS VALID ONLY IN \
		   CIPHER MODE \
		\n\t provide following test set number:- \
		\n\t test set for AES_CBC is 1 to 4 \
		\n\t test set for TDES_CBC is 1 to 2 \
		\n\t test set for SNOW_F8 is 1 to 5 \
		\n\t test set for SNOW_F9 is 1 to 5 \
		\n\t test set for KASUMI_F8 is 1 to 5 \
		\n\t test set for KASUMI_F9 is 1 to 5 \
		\n\t test set for CRC is 1 to 5 \
		\n\t test set for HMAC_SHA1 is 1 to 2 \
		\n\t test set for SNOW_F8_F9 is 1 to 1\n",
	 0},
	{"bufsize", 's', "BUFFER SIZE", 0, "THIS OPTION IS VALID ONLY IN PERF MODE \
		\n\t Buffer size (64, 128 ...upto 6400) \
		\n\t For AES_CBC buffer size should be 16 byte aligned \
		\n\t For TDES_CBC buffer size should be 8 byte aligned\n",
	 0},
	{"bufnum", 'n', "NUMBER OF BUFFERS per core", 0,
	 "Number of buffers per core (1 to 800)\n", 0},
	{"itrnum", 'l', "NUMBER OF ITERATIONS", 0,
	 "Number of iteration to repeat\n", 0},
	{"algo", 'o', "SEC 4.0 ALGORITHM", 0,
	 "Cryptographic operation, provide following number \
		\n\t 1 for AES_CBC \
		\n\t 2 for TDES_CBC \
		\n\t 3 for SNOW_F8 \
		\n\t 4 for SNOW_F9 \
		\n\t 5 for KASUMI_F8 \
		\n\t 6 for KASUMI_F9 \
		\n\t 7 for CRC \
		\n\t 8 for HMAC_SHA1\
		\n\t 9 for SNOW_F8_F9", 0},
	{0, 0, 0, 0, 0, 0}
};

/**
\brief		Parse a single option

\param[in]	key - An integer specifying which option this is (taken	from
		the KEY field in each struct argp_option), or a special key
		specifying something else. We do not use any special key here

\param[in]	arg - For an option KEY, the string value of its argument, or
		NULL if it has none

\param[in]	state - A pointer to a struct argp_state, containing various
		useful information about the parsing state; used here are the
		INPUT field, which reflects the INPUT argument to argp_parse

\return		It should return either 0, meaning success, ARGP_ERR_UNKNOWN,
		meaning the given KEY wasn't recognized, or an errno value
		indicating some other error
*/
static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
	struct crypto_param *crypto_info = state->input;

	switch (key) {
	case 'm':
		crypto_info->mode = atoi(arg);
		g_cmd_params |= BMASK_SEC_TEST_MODE;
		printf("\nTest mode = %s", arg);
		break;

	case 't':
		crypto_info->test_set = atoi(arg);
		g_cmd_params |= BMASK_SEC_TEST_SET;
		printf("\nTest set = %d", crypto_info->test_set);
		break;

	case 's':
		crypto_info->buf_size = atoi(arg);
		g_cmd_params |= BMASK_SEC_BUFFER_SIZE;
		printf("\nBuffer size = %d", crypto_info->buf_size);
		break;

	case 'n':
		crypto_info->buf_num_per_core = atoi(arg);
		g_cmd_params |= BMASK_SEC_BUFFER_NUM;
		printf("\nNumber of Buffers per core = %d",
			crypto_info->buf_num_per_core);
		break;

	case 'o':
		crypto_info->algo = atoi(arg);
		g_cmd_params |= BMASK_SEC_ALG;
		printf("\nSEC4.0 cryptographic operation = %s", arg);
		break;

	case 'l':
		crypto_info->itr_num = atoi(arg);
		g_cmd_params |= BMASK_SEC_ITR_NUM;
		printf("\nNumber of iteration = %u", crypto_info->itr_num);
		break;

	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

static int validate_test_set(struct crypto_param *crypto_info)
{
	switch (crypto_info->algo) {
	case AES_CBC:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 5)
			return 0;
		else
			goto err;
	case TDES_CBC:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 3)
			return 0;
		else
			goto err;
	case SNOW_F8:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 6)
			return 0;
		else
			goto err;
	case SNOW_F9:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 6)
			return 0;
		else
			goto err;
	case KASUMI_F8:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 6)
			return 0;
		else
			goto err;
	case KASUMI_F9:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 6)
			return 0;
		else
			goto err;
	case CRC:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 6)
			return 0;
		else
			goto err;
	case HMAC_SHA1:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 3)
			return 0;
		else
			goto err;
	case SNOW_F8_F9:
		if (crypto_info->test_set > 0 && crypto_info->test_set < 2)
			return 0;
		else
			goto err;
	default:
		printf("\n%s: Invalid SEC algorithm", __func__);
		return -1;
	}
err:
	printf("\n%s: Test set number is invalid", __func__);
	return -1;
}

/**
\brief		Check SEC 4.0 parameters provided by user whether valid or not

\param[in]	crypto_info - Paramters provided by user

\param[in]	g_cmd_params - Bit mask of all parameters provided by user

\return		0 on success, otherwise -ve value
*/
static int validate_params(struct crypto_param *crypto_info,
			   uint32_t g_cmd_params)
{
	if ((PERF == crypto_info->mode)
	    && BMASK_SEC_PERF_MODE == g_cmd_params) {
		/* do nothing */
	} else if ((CIPHER == crypto_info->mode)
		   && g_cmd_params == BMASK_SEC_CIPHER_MODE) {
		if (0 > validate_test_set(crypto_info)) {
			printf("\n%s: validate_test_set failed", __func__);
			return -1;
		}
	} else {
		printf
		    ("\nERROR: provide a valid combination of mandatory arguments, ");
		printf("see --help option");
		printf("\nDEBUG Info: g_cmd_params = %0x", g_cmd_params);
		return -1;
	}

	if (crypto_info->buf_num_per_core > BUFF_NUM_PER_CORE) {
		printf("\n%s: Invalid number of buffers", __func__);
		printf("see --help option");
		return -1;
	}

	if (crypto_info->buf_size % 64 != 0
			|| crypto_info->buf_size > BUFF_SIZE) {
		printf("\n%s: Invalid number of buffers", __func__);
		printf("see --help option");
	}

	switch (crypto_info->algo) {
	case AES_CBC:
	case TDES_CBC:
	case SNOW_F8:
	case SNOW_F9:
	case KASUMI_F8:
	case KASUMI_F9:
	case CRC:
	case HMAC_SHA1:
	case SNOW_F8_F9:
		break;
	default:
		printf("\n%s: SEC algorithm not supported", __func__);
		return -1;
	}

	return 0;

}

/**
\brief		Copy sec parameters information from control plane application
		to LWE data plane application's WB area using mmap()

\param[in]	crypto_info - Paramters provided by user
*/
static int copy_to_shm(struct crypto_param *crypto_info)
{
	int fd;
	void *map_addr;

	if ((fd = open("/dev/mem", O_RDWR)) < 0) {
		printf("\nERROR: failed in opening /dev/mem");
		return -1;
	}

	map_addr = mmap(0, SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, ADDR);
	if (map_addr == MAP_FAILED) {
		close(fd);
		printf("\nERROR: mmapping the file");
		return -1;
	}

	memcpy(map_addr, (void *)crypto_info, sizeof(struct crypto_param));

	if (munmap(map_addr, SIZE) == -1) {
		close(fd);
		printf("\nERROR: un-mmapping the file");
		return -1;
	}

	close(fd);
	return 0;
}

/** argp structure itself of argp parser */
static struct argp argp =
    { options, parse_opt, args_doc, doc, NULL, NULL, NULL };

/**
\brief		Main function of SEC 4.0 control plane application

\param[in]	argc - Argument count

\param[in]	argv - Argument list pointer
*/
int main(int argc, char **argv)
{
	struct crypto_param crypto_info;

	/* set default value 0 to all */
	memset(&crypto_info, 0, sizeof(struct crypto_param));

	g_cmd_params = 0;

	/* Where the magic happens */
	argp_parse(&argp, argc, argv, 0, 0, &crypto_info);

	if (0 > validate_params(&crypto_info, g_cmd_params)) {
		printf("\nERROR: validate_params failed!\n");
		return -1;
	}

	crypto_info.valid_params = true;

	if (copy_to_shm(&crypto_info) < 0) {
		printf("\nERROR: copy_to_shm failed!\n");
		return -1;
	}

	printf("\nSEC4.0 test parameters are valid, start DP application\n");
	return 0;
}
