/**
\file	simple_crypto.c
\brief	Basic SEC 4.0 test application. It operates on user defined SEC
	parameters through CP application and reports throughput for
	various SEC 4.0 raw algorithm.
*/
/*
 * Copyright (C) 2008-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define WANT_MARKPOINTS

#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include "simple_crypto_cpdp.h"
#include "simple_crypto.h"
#include "algo_desc.h"
#include "test_vector.h"
#include <app_common.h>
#include <common/cpu_freq.h>
#include <net/frame_desc.h>
#include <dpa/dpa_bman.h>
#include <dpa/dpa_qman.h>
#include <app_cpdp.h>

extern uint32_t get_cpu_freq(void);

/** user defined parameters through CP */
volatile struct crypto_param *crypto_info;

/** refrence test vector consists of key, iv, plain text, cipher text etc*/
struct ref_vector_s ref_test_vector;

/** dpa buffer allocator object per core */
__PERCPU struct dpa_buff_allocator *buff_allocator;

__PERCPU uint32_t curr_lwe_id;	/**< LWE ID for each core */

/** Master LWE core does required initialization first */
uint32_t g_ctrl_lwe = INV_LWE_ID;

/** local lwe instance which starts from 0 to <num_cores> */
__PERCPU uint8_t my_lwe_instance;

/** output buffer size(varies for authentication algos) */
uint32_t output_buf_size;

/** input buffer capacity(varies for authentication algos) */
uint32_t input_buf_capacity;

/** The size required for a buffer that can hold encrypt or decrypt
 *  descriptors */
uint32_t job_desc_buf_size;

uint8_t authnct = 0;	/**< processing authentication algorithm */

/** total number of frame(s) sending to SEC4.0 */
volatile uint32_t *packet_to_sec;

/** total number of encrypted frmae(s) returned from SEC4.0 */
volatile uint32_t *enc_packet_from_sec;

/** total number of decrypted frame(s) returned from SEC4.0 */
volatile uint32_t *dec_packet_from_sec;

uint32_t index = 0;
struct qm_fd fd[BUFF_NUM_PER_CORE*7];	/**< storage for frame descriptor */

/** start FQ no. for encryption flows */
uint32_t fq_base_encrypt = SEC40_FQ_BASE;

/** start FQ no. for decryption flows */
uint32_t fq_base_decrpyt = SEC40_FQ_BASE + 2 * FQ_COUNT;

/** storage for encryption FQ's(from SEC4.0 to software portal) object*/
struct qman_fq *enc_fq_from_sec[FQ_COUNT];

/** storage for encryption FQ's(from software portal to SEC4.0) object*/
struct qman_fq *enc_fq_to_sec[FQ_COUNT];

/** storage for decryption FQ's(from SEC4.0 to software portal) object*/
struct qman_fq *dec_fq_from_sec[FQ_COUNT];

/** storage for decryption FQ's(from software portal to SEC4.0) object*/
struct qman_fq *dec_fq_to_sec[FQ_COUNT];

/* retire flag associated with each of the frame queues */
volatile static bool enc_fq_from_sec_retire[FQ_COUNT]
= {[0 ... FQ_COUNT - 1] = false };

volatile static bool enc_fq_to_sec_retire[FQ_COUNT]
= {[0 ... FQ_COUNT - 1] = false };

volatile static bool dec_fq_from_sec_retire[FQ_COUNT]
= {[0 ... FQ_COUNT - 1] = false };

volatile static bool dec_fq_to_sec_retire[FQ_COUNT]
= {[0 ... FQ_COUNT - 1] = false };

/** pool Channel Offset */
volatile int32_t pool_channel_offset;

uint32_t sec_lock;
struct lwe_barrier sec40_barrier;
struct frame_fragment_t in_frag;
struct frame_fragment_t jdesc_frag;

volatile static bool ctrl_error;
volatile static bool ctrl_init_fq_error;

volatile uint32_t total_buf_num;
uint8_t num_cores;


/**
\brief		Enqueue Rejection Notification Handler for SEC Tx FQ

\param[in]	msg - Message Ring entry to be processed

\param[out]	NULL
 */
static void cb_ern(struct qman_portal *qm, struct qman_fq *fq,
				const struct qm_mr_entry *msg)
{
	APP_ERROR("%s: RC = %x, seqnum = %x", __func__,\
			msg->ern.rc, msg->ern.seqnum);
	lwe_free_fd(buff_allocator, &msg->ern.fd);
	return;
}

/** callback handler for dequeued frames and fq's(from SEC40) state change */
const struct qman_fq_cb sec40_rx_cb = { cb_dqrr, NULL, NULL, cb_fqs };

/** callback handler for fq's(to SEC40) state change */
const struct qman_fq_cb sec40_tx_cb = { NULL, cb_ern, NULL, cb_fqs };

char algorithm[20];	/**< string corresponding to integral value */
char mode_type[20];	/**< string corresponding to integral value */

void print_frame_desc(struct qm_fd *frame);

/**
\brief		Stats initialisation

\details	Allocates region of memory for storing statistics
 
\return		None

*/
static void sec_stats_init(void)
{
	packet_to_sec = stats_memalign(CACHE_LINE_SIZE,
				       sizeof(struct stat32_pair_t));
	if (unlikely(NULL == packet_to_sec)) {
		LWE_ERROR("stats_memalign failed for packet_to_sec!");
		assert(0);
	}

	enc_packet_from_sec = stats_memalign(CACHE_LINE_SIZE,
					     sizeof(struct stat32_pair_t));
	if (unlikely(NULL == enc_packet_from_sec)) {
		LWE_ERROR("stats_memalign failed for enc_packet_from_sec!");
		assert(0);
	}

	dec_packet_from_sec = stats_memalign(CACHE_LINE_SIZE,
					     sizeof(struct stat32_pair_t));
	if (unlikely(NULL == dec_packet_from_sec)) {
		LWE_ERROR("stats_memalign failed for dec_packet_from_sec!");
		assert(0);
	}
}

static int init_ref_test_vector(void)
{
	switch (crypto_info->algo) {
	case AES_CBC:
		strcpy(algorithm, "AES_CBC");
		ref_test_vector.key =
		    aes_cbc_reference_key[crypto_info->test_set - 1];
		ref_test_vector.iv.init_vec =
		    aes_cbc_reference_iv[crypto_info->test_set - 1];
		ref_test_vector.length =
		    aes_cbc_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
		    aes_cbc_reference_plaintext[crypto_info->test_set - 1];
		ref_test_vector.ciphertext =
		    aes_cbc_reference_ciphertext[crypto_info->test_set - 1];

		break;

	case TDES_CBC:
		strcpy(algorithm, "TDES_CBC");
		ref_test_vector.key =
		    tdes_cbc_reference_key[crypto_info->test_set - 1];
		ref_test_vector.iv.init_vec =
		    tdes_cbc_reference_iv[crypto_info->test_set - 1];
		ref_test_vector.length =
		    tdes_cbc_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
		    tdes_cbc_reference_plaintext[crypto_info->test_set - 1];
		ref_test_vector.ciphertext =
		    tdes_cbc_reference_ciphertext[crypto_info->test_set - 1];

		break;

	case SNOW_F8:
		strcpy(algorithm, "SNOW_F8");
		ref_test_vector.key =
		    snow_f8_reference_key[crypto_info->test_set - 1];
		ref_test_vector.iv.f8.count =
		    snow_f8_reference_count[crypto_info->test_set - 1];
		ref_test_vector.iv.f8.bearer =
		    snow_f8_reference_bearer[crypto_info->test_set - 1];
		ref_test_vector.iv.f8.direction =
		    snow_f8_reference_dir[crypto_info->test_set - 1];
		ref_test_vector.length =
		    snow_f8_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
		    snow_f8_reference_plaintext[crypto_info->test_set - 1];
		ref_test_vector.ciphertext =
		    snow_f8_reference_ciphertext[crypto_info->test_set - 1];

		break;

	case SNOW_F9:
		strcpy(algorithm, "SNOW_F9");
		ref_test_vector.key =
		    snow_f9_reference_key[crypto_info->test_set - 1];
		ref_test_vector.iv.f9.count =
		    snow_f9_reference_count[crypto_info->test_set - 1];
		ref_test_vector.iv.f9.fresh =
		    snow_f9_reference_fresh[crypto_info->test_set - 1];
		ref_test_vector.iv.f9.direction =
		    snow_f9_reference_dir[crypto_info->test_set - 1];
		ref_test_vector.length =
		    snow_f9_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
		    snow_f9_reference_plaintext[crypto_info->test_set - 1];
		ref_test_vector.ciphertext = NULL;
		ref_test_vector.digest =
		    snow_f9_reference_digest[crypto_info->test_set - 1];

		authnct = 1;
		break;

	case KASUMI_F8:
		strcpy(algorithm, "KASUMI_F8");
		ref_test_vector.key =
		    kasumi_f8_reference_key[crypto_info->test_set - 1];
		ref_test_vector.iv.f8.count =
		    kasumi_f8_reference_count[crypto_info->test_set - 1];
		ref_test_vector.iv.f8.bearer =
		    kasumi_f8_reference_bearer[crypto_info->test_set - 1];
		ref_test_vector.iv.f8.direction =
		    kasumi_f8_reference_dir[crypto_info->test_set - 1];
		ref_test_vector.length =
		    kasumi_f8_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
		    kasumi_f8_reference_plaintext[crypto_info->test_set - 1];
		ref_test_vector.ciphertext =
		    kasumi_f8_reference_ciphertext[crypto_info->test_set - 1];

		break;

	case KASUMI_F9:
		strcpy(algorithm, "KASUMI_F9");
		ref_test_vector.key =
		    kasumi_f9_reference_key[crypto_info->test_set - 1];
		ref_test_vector.iv.f9.count =
		    kasumi_f9_reference_count[crypto_info->test_set - 1];
		ref_test_vector.iv.f9.fresh =
		    kasumi_f9_reference_fresh[crypto_info->test_set - 1];
		ref_test_vector.iv.f9.direction =
		    kasumi_f9_reference_dir[crypto_info->test_set - 1];
		ref_test_vector.length =
		    kasumi_f9_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
		    kasumi_f9_reference_plaintext[crypto_info->test_set - 1];
		ref_test_vector.ciphertext = NULL;
		ref_test_vector.digest =
		    kasumi_f9_reference_digest[crypto_info->test_set - 1];

		authnct = 1;
		break;

	case CRC:
		strcpy(algorithm, "CRC");
		ref_test_vector.length =
		    crc_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
		    crc_reference_plaintext[crypto_info->test_set - 1];
		ref_test_vector.ciphertext = NULL;
		ref_test_vector.digest =
		    crc_reference_digest[crypto_info->test_set - 1];

		authnct = 1;
		break;

	case HMAC_SHA1:
		strcpy(algorithm, "HMAC_SHA1");
		ref_test_vector.key =
		    hmac_sha1_reference_key[crypto_info->test_set - 1];
		ref_test_vector.length =
		    hamc_sha1_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
		    hmac_sha1_reference_plaintext[crypto_info->test_set - 1];
		ref_test_vector.ciphertext = NULL;
		ref_test_vector.digest =
		    hamc_sha1_reference_digest[crypto_info->test_set - 1];

		authnct = 1;
		break;

	case SNOW_F8_F9:
		strcpy(algorithm, "SNOW_F8_F9");
		ref_test_vector.length =
		    snow_enc_f8_f9_reference_length[crypto_info->test_set - 1];
		ref_test_vector.plaintext =
			 snow_enc_f8_f9_reference_plaintext
						[crypto_info->test_set - 1];
		ref_test_vector.ciphertext =
			 snow_enc_f8_f9_reference_ciphertext
						[crypto_info->test_set - 1];

		break;


	default:
		APP_ERROR("%s: algorithm not supported", __func__);
		return -1;
	}

	return 0;
}

/**
\brief		Create a compound frame descriptor understood by SEC 4.0

\return		0 on success, otherwise -ve value 
*/
static int create_compound_fd(void)
{
	uint8_t *in_buf, *out_buf;
	struct sg_entry_priv_t *sg_priv_and_data;
	struct sg_entry_t *sg;
	uint32_t input_buf_length = 0;
	uint32_t total_size;
	input_buf_capacity = crypto_info->buf_size;
	input_buf_length = crypto_info->buf_size;

	switch (crypto_info->algo) {
	case AES_CBC:
	case TDES_CBC:
	case SNOW_F8:
		output_buf_size = crypto_info->buf_size;
		break;
	case SNOW_F9:
		output_buf_size = SNOW_F9_DIGEST_SIZE;
		break;
	case KASUMI_F8:
		output_buf_size = crypto_info->buf_size;
		break;
	case KASUMI_F9:
		output_buf_size = KASUMI_F9_DIGEST_SIZE;
		break;
	case CRC:
		output_buf_size = CRC_DIGEST_SIZE;
		break;
	case HMAC_SHA1:
		output_buf_size = HMAC_SHA1_DIGEST_SIZE;
		break;
	case SNOW_F8_F9:
		output_buf_size = crypto_info->buf_size + SNOW_F9_DIGEST_SIZE;

		/* For this algorithm a  Job Descriptor will be added to the
		 head of the SEC frame. Increase the buffer capacity and
		 length. The same buffer will be used for holding the
		 plain-text data + encrypt job descriptor and later the
		 encrypted data + SNOW F9 digest + decrypt job descriptor.
		*/
		input_buf_capacity += job_desc_buf_size + SNOW_F9_DIGEST_SIZE;
		input_buf_length += job_desc_buf_size;

		break;
	default:
		APP_ERROR("%s: algorithm not supported", __func__);
		return -1;
	}

	for (index = 0; index < total_buf_num; index++) {

		total_size = sizeof(struct sg_entry_priv_t) + output_buf_size
				+ input_buf_capacity;

		sg_priv_and_data = (struct sg_entry_priv_t *)memalign
			(CACHE_LINE_SIZE, total_size);

		if (unlikely(sg_priv_and_data == NULL)) {
			APP_ERROR("Unable to allocate memory for buffer");
			return -1;
		}

		memset(sg_priv_and_data, 0, total_size);

		out_buf = (uint8_t *)(sg_priv_and_data)
				+ sizeof(struct sg_entry_priv_t);
		in_buf = (uint8_t *)sg_priv_and_data + (total_size
				- input_buf_capacity);

		sg = (struct sg_entry_t *)sg_priv_and_data;

		/* output buffer */
		sg->addr_lo = (uint32_t)out_buf;
		sg->length = output_buf_size;

		/* input buffer */
		sg++;
		sg->addr_lo = (uint32_t)in_buf;
		sg->length = input_buf_length;
		sg->final = 1;
		sg--;

		fd[index].addr_lo = (uint32_t)APP_VTOP(sg);
		fd[index]._format1 = qm_fd_compound;

		sg_priv_and_data->index = index;
	}

	return 0;
}

static void *setup_preheader(uint32_t shared_desc_len, uint8_t pool_id,
			 uint16_t pool_buf_size, uint8_t abs, uint8_t add_buf)
{
	struct preheader_s *prehdr = NULL;

	/* Request a buffer from the buffer allocator, with the preheader
	 *  size */
	prehdr = memalign(CACHE_LINE_SIZE, PREHDR_LEN * sizeof(uint32_t));

	if (unlikely(prehdr == NULL)) {
		APP_ERROR("\r\n%s: memalign failed for preheader"
			, __func__);
		return NULL;
	}

	prehdr->hi.field.rsvd63_48 = 0;
	prehdr->hi.field.rsvd47_39 = 0;
	/* the shared descriptor length is 0, meaning that no shared
	   descriptor follows the preheader in the context A */
	prehdr->hi.field.idlen = shared_desc_len;
	prehdr->lo.field.rsvd31_30 = 0;
	prehdr->lo.field.fsgt = 0;
	prehdr->lo.field.lng = 0;
	prehdr->lo.field.offset = 0;
	prehdr->lo.field.abs = abs;
	prehdr->lo.field.add_buf = add_buf;
	prehdr->lo.field.pool_id = pool_id;
	prehdr->lo.field.pool_buffer_size = pool_buf_size;

	return (void *)prehdr;

}

/**
\brief 		Create SEC 4.0 shared descriptor consists of sequence of commands
		to SEC 4.0 with	necessary key, iv etc initialisation

\param[in] 	mode -	To check whether descriptor is for encryption or decryption

\return 	Shared descriptor pointer on success, otherwise NULL
*/
static void *setup_init_descriptor(bool mode)
{
	uint32_t i;
	sec_descriptor_t *prehdr_desc;
	uint32_t *shared_desc = NULL;
	uint16_t shared_desc_len;
	phys_addr_t addr;

	prehdr_desc = memalign(CACHE_LINE_SIZE, sizeof(sec_descriptor_t));
	if (unlikely(prehdr_desc == NULL)) {
		APP_ERROR("\r\n%s: memalign failed for preheader"
			, __func__);
		return NULL;
	}

	memset(prehdr_desc, 0, sizeof(sec_descriptor_t));

	shared_desc =
	    (uint32_t *) ((uint32_t) prehdr_desc + sizeof(struct preheader_s));

	switch (crypto_info->algo) {
	case AES_CBC:
		if (0 > cnstr_shdsc_cbc_blkcipher(shared_desc, &shared_desc_len,
						  ref_test_vector.key,
						  AES_CBC_KEY_LEN *
						  BITS_PER_BYTE,
						  ref_test_vector.iv.init_vec,
						  AES_CBC_IV_LEN *
						  BITS_PER_BYTE,
						  mode ? DIR_ENCRYPT :
						  DIR_DECRYPT,
						  OP_ALG_ALGSEL_AES, 0))
			goto error;

		break;

	case TDES_CBC:
		if (0 > cnstr_shdsc_cbc_blkcipher(shared_desc, &shared_desc_len,
						  ref_test_vector.key,
						  TDES_CBC_KEY_LEN *
						  BITS_PER_BYTE,
						  ref_test_vector.iv.init_vec,
						  TDES_CBC_IV_LEN *
						  BITS_PER_BYTE,
						  mode ? DIR_ENCRYPT :
						  DIR_DECRYPT,
						  OP_ALG_ALGSEL_3DES, 0))
			goto error;

		break;

	case SNOW_F8:
		if (0 > cnstr_shdsc_snow_f8(shared_desc, &shared_desc_len,
					    ref_test_vector.key,
					    F8_KEY_LEN * BITS_PER_BYTE,
					    mode ? DIR_ENCRYPT : DIR_DECRYPT,
					    ref_test_vector.iv.f8.count,
					    ref_test_vector.iv.f8.bearer,
					    ref_test_vector.iv.f8.direction, 0))
			goto error;

		break;

	case SNOW_F9:
		if (mode == DECRYPT) {
			APP_ERROR("%s: enc bit not selected as protect",
				  __func__);
			free(prehdr_desc);
			return NULL;
		}

		/* assume here same data in each frame */
		addr = in_frag.addr_hi;
		addr = (addr << 32) | in_frag.addr_lo;

		if (0 > cnstr_shdsc_snow_f9(shared_desc, &shared_desc_len,
					    ref_test_vector.key,
					    F9_KEY_LEN * BITS_PER_BYTE,
					    DIR_ENCRYPT,
					    ref_test_vector.iv.f9.count,
					    ref_test_vector.iv.f9.fresh,
					    ref_test_vector.iv.f9.direction, 0,
					    addr, ref_test_vector.length))
			goto error;

		break;

	case KASUMI_F8:
		if (0 > cnstr_shdsc_kasumi_f8(shared_desc, &shared_desc_len,
					    ref_test_vector.key,
					    F8_KEY_LEN * BITS_PER_BYTE,
					    mode ? DIR_ENCRYPT : DIR_DECRYPT,
					    ref_test_vector.iv.f8.count,
					    ref_test_vector.iv.f8.bearer,
					    ref_test_vector.iv.f8.direction, 0))
			goto error;

		break;

	case KASUMI_F9:
		if (mode == DECRYPT) {
			APP_ERROR("%s: enc bit not selected as protect",
				  __func__);
			free(prehdr_desc);
			return NULL;
		}

		/* assume here same data in each frame */
		addr = in_frag.addr_hi;
		addr = (addr << 32) | in_frag.addr_lo;

		if (0 > cnstr_shdsc_kasumi_f9(shared_desc, &shared_desc_len,
					    ref_test_vector.key,
					    F9_KEY_LEN * BITS_PER_BYTE,
					    DIR_ENCRYPT,
					    ref_test_vector.iv.f9.count,
					    ref_test_vector.iv.f9.fresh,
					    ref_test_vector.iv.f9.direction, 0,
					    addr, ref_test_vector.length))
			goto error;

		break;

	case CRC:
		if (mode == DECRYPT) {
			APP_ERROR("%s: enc bit not selected as protect",
				  __func__);
			free(prehdr_desc);
			return NULL;
		}

		if (0 > cnstr_shdsc_crc(shared_desc, &shared_desc_len, 0))
			goto error;

		break;

	case HMAC_SHA1:
		if (mode == DECRYPT) {
			APP_ERROR("%s: enc bit not selected as protect",
				  __func__);
			free(prehdr_desc);
			return NULL;
		}

		if (0 > cnstr_shdsc_hmac(shared_desc, &shared_desc_len,
					 ref_test_vector.key,
					 OP_ALG_ALGSEL_SHA1, NULL, 0))
			goto error;

		break;

	default:
		APP_ERROR("%s: algorithm not supported", __func__);
		free(prehdr_desc);
		return NULL;
	}

	prehdr_desc->prehdr.hi.field.rsvd63_48 = 0;
	prehdr_desc->prehdr.hi.field.rsvd47_39 = 0;
	prehdr_desc->prehdr.hi.field.idlen = shared_desc_len;
	prehdr_desc->prehdr.lo.field.rsvd31_30 = 0;
	prehdr_desc->prehdr.lo.field.fsgt = 0;
	prehdr_desc->prehdr.lo.field.lng = 0;
	prehdr_desc->prehdr.lo.field.offset = 0;
	prehdr_desc->prehdr.lo.field.abs = 0;
	prehdr_desc->prehdr.lo.field.add_buf = 0;
	prehdr_desc->prehdr.lo.field.pool_id = 0;
	prehdr_desc->prehdr.lo.field.pool_buffer_size = 0;

	APP_DEBUG("SEC4.0 %s shared descriptor:", algorithm);
	for (i = 0; i < shared_desc_len; i++)
		APP_DEBUG("0x%x", *shared_desc++);

	return (void *)prehdr_desc;

error:
	APP_ERROR("%s: %s shared descriptor initilization failed",
		  __func__, algorithm);
	return NULL;
}

static void *setup_sec_descriptor(bool mode)
{
	void *descriptor = NULL;

	if (crypto_info->algo == SNOW_F8_F9) {
		descriptor = setup_preheader
				(0, /* shared descriptor length is 0, meaning
					 there is no shared desc in context A*/
				 0, /* pool buffer id*/
				 0, /* pool buffer size*/
				 0, /* abs = 0 and add_buf = 0 means that the
					 output buffer is provided inside
					 compound frame*/
				 0); /* add_buf = 0*/

	} else {
		descriptor = setup_init_descriptor(mode);
	}
	return descriptor;
}


struct qman_fq *create_sec_frame_queue(uint32_t fq_id, uint16_t channel,
				       uint16_t wq_id, uint32_t ctxAhi,
				       uint32_t ctxAlo, uint32_t ctxB)
{
	struct qm_mcc_initfq fq_opts;
	struct qman_fq *fq;
	uint32_t flags;

	fq = (struct qman_fq *)memalign(CACHE_LINE_SIZE,
			sizeof(struct qman_fq));
	if (unlikely(NULL == fq)) {
		APP_ERROR("malloc failed in create_fqs for FQ ID: %u", fq_id);
		return NULL;
	}
	if (ctxAlo) {
		flags = QMAN_FQ_FLAG_LOCKED | QMAN_FQ_FLAG_TO_DCPORTAL;
		fq->cb = sec40_tx_cb;
	} else {
		flags = QMAN_FQ_FLAG_LOCKED | QMAN_FQ_FLAG_NO_ENQUEUE;
		fq->cb = sec40_rx_cb;
	}

	if (unlikely(0 != qman_create_fq(fq_id, flags, fq))) {
		APP_ERROR("qman_create_fq failed for FQ ID: %u", fq_id);
		return NULL;
	}

	flags = QMAN_INITFQ_FLAG_SCHED;
	fq_opts.we_mask = QM_INITFQ_WE_DESTWQ | QM_INITFQ_WE_CONTEXTA;
	if (ctxAlo) {
		fq_opts.we_mask |= QM_INITFQ_WE_CONTEXTB;
		fq_opts.fqd.context_a.hi = ctxAhi;
		fq_opts.fqd.context_a.lo = ctxAlo;
		fq_opts.fqd.context_b = ctxB;
	} else {
		uint32_t ctx_a_excl, ctx_a_len;
		ctx_a_excl = (QM_STASHING_EXCL_DATA | QM_STASHING_EXCL_CTX);
		ctx_a_len = (1 << 2) | 1;
		fq_opts.fqd.context_a.hi = (ctx_a_excl << 24)
			| (ctx_a_len << 16);
	}

	fq_opts.fqd.dest.wq = wq_id;
	fq_opts.fqd.dest.channel = channel;

	if (unlikely(0 != qman_init_fq(fq, flags, &fq_opts))) {
		APP_ERROR("qm_init_fq failed for fq_id: %u", fq_id);
		return NULL;
	}

	return fq;
}

static int init_sec_frame_queues(enum SEC_MODE mode)
{
	uint32_t fq_base;
	uint32_t fq_from_sec, fq_to_sec;
	void *ctxtA;
	uint32_t ctxtA_hi, ctxtA_lo;
	phys_addr_t addr;
	uint16_t lwe_pool_channel =
	    (qm_channel_pool1 - 1) + pool_channel_offset;

	ctxtA = setup_sec_descriptor(mode);
	addr = vtop(ctxtA);
	ctxtA_hi = (uint32_t) (addr >> 32);
	ctxtA_lo = (uint32_t) (addr);

	if (NULL == ctxtA) {
		APP_ERROR("%s: Initializing shared descriptor failure!",
			  __func__);
		return -1;
	}

	if (mode == ENCRYPT) {
		fq_base = fq_base_encrypt;
		for (int i = 0; i < FQ_COUNT; i++) {
			fq_from_sec = fq_base++;
			fq_to_sec = fq_base++;
			enc_fq_from_sec[i] =
			    create_sec_frame_queue(fq_from_sec,
						   lwe_pool_channel, 0, 0, 0,
						   0);
			if (!enc_fq_from_sec[i]) {
				APP_ERROR
				    ("%s : Encrypt FQ(from SEC) %d couldn't be allocated, ID = %d",
				     __func__, i, fq_from_sec);
				return -1;
			}

			enc_fq_to_sec[i] =
			    create_sec_frame_queue(fq_to_sec, qm_channel_caam,
						   0, ctxtA_hi, ctxtA_lo,
						   fq_from_sec);
			if (!enc_fq_to_sec[i]) {
				APP_ERROR
				    ("%s : Encrypt FQ(to SEC) %d couldn't be allocated, ID = %d",
				     __func__, i, fq_to_sec);
				return -1;
			}
		}
	} else {
		fq_base = fq_base_decrpyt;
		for (int i = 0; i < FQ_COUNT; i++) {
			fq_from_sec = fq_base++;
			fq_to_sec = fq_base++;
			dec_fq_from_sec[i] =
			    create_sec_frame_queue(fq_from_sec,
						   lwe_pool_channel, 0, 0, 0,
						   0);
			if (!dec_fq_from_sec[i]) {
				APP_ERROR
				    ("%s : Decrypt FQ(from SEC) %d couldn't be allocated, ID = %d",
				     __func__, i, fq_from_sec);
				return -1;
			}

			dec_fq_to_sec[i] =
			    create_sec_frame_queue(fq_to_sec, qm_channel_caam,
						   0, ctxtA_hi, ctxtA_lo,
						   fq_from_sec);
			if (!dec_fq_to_sec[i]) {
				APP_ERROR
				    ("%s : Decrypt FQ(to SEC) %d couldn't be allocated, ID = %d",
				     __func__, i, fq_to_sec);
				return -1;
			}
		}
	}
	return 0;
}

/**
\brief		Initialize frame queues to enqueue and dequeue frames to SEC
		and from SEC respectively

\param[in] 	None

\return		0 on success, otherwise -ve value
*/
static int init_sec_fq(void)
{

	if (init_sec_frame_queues(ENCRYPT)) {
		APP_ERROR("%s: couldn't Initialize SEC 4.0 Encrypt Queues",
			  __func__);
		return -1;
	}

	if (!authnct) {
		if (init_sec_frame_queues(DECRYPT)) {
			APP_ERROR
			    ("%s: couldn't Initialize SEC 4.0 Decrypt Queues",
			     __func__);
			return -1;
		}
	}

	APP_INFO("Initialized FQs");
	return 0;
}

void cb_fqs(struct qman_portal *qm, struct qman_fq *fq,
	    const struct qm_mr_entry *msg)
{
	uint32_t fqid;
	uint32_t rem, offset;

	fqid = fq->fqid;

	if ((fqid >= fq_base_encrypt) && (fqid < fq_base_decrpyt)) {
		offset = fqid - fq_base_encrypt;
		rem = offset % 2;
		if (!rem)
			enc_fq_from_sec_retire[offset / 2] = true;
		else
			enc_fq_to_sec_retire[offset / 2] = true;
	} else if ((fqid >= fq_base_decrpyt)
		   && (fqid < (fq_base_decrpyt + 2 * FQ_COUNT))) {
		offset = fqid - fq_base_decrpyt;
		rem = offset % 2;
		if (!rem)
			dec_fq_from_sec_retire[offset / 2] = true;
		else
			dec_fq_to_sec_retire[offset / 2] = true;
	}

	return;
}

static int free_sec_frame_queues(struct qman_fq **fq,
				 volatile bool * fq_retire_flag)
{
	int i, res;
	uint32_t flags;

	for (i = 0; i < FQ_COUNT; i++) {
		if (0 > (res = qman_retire_fq(fq[i], &flags))) {
			APP_ERROR("qman_retire_fq failed for fq %d", i);
			return -1;
		}
		wait_event(NULL, *((unsigned char *)(fq_retire_flag) + i));

		if (flags & QMAN_FQ_STATE_BLOCKOOS) {
			APP_ERROR("leaking frames for fq %d", i);
			return -1;
		}
		if (qman_oos_fq(fq[i])) {
			APP_ERROR("qman_oos_fq failed for fq %d", i);
			return -1;
		}
		qman_destroy_fq(fq[i], 0);
	}

	return 0;
}

static int free_sec_fq(void)
{
	if (unlikely(0 > free_sec_frame_queues(enc_fq_from_sec,
					       enc_fq_from_sec_retire))) {
		APP_ERROR("free_sec_frame_queues failed for enc_fq_from_sec");
		return -1;
	}

	if (unlikely(0 > free_sec_frame_queues(enc_fq_to_sec,
					       enc_fq_to_sec_retire))) {
		APP_ERROR("free_sec_frame_queues failed for enc_fq_to_sec");
		return -1;
	}

	if (!authnct) {
		if (unlikely(0 > free_sec_frame_queues(dec_fq_from_sec,
						       dec_fq_from_sec_retire)))
		{
			APP_ERROR
			    ("free_sec_frame_queues failed for dec_fq_from_sec");
			return -1;
		}

		if (unlikely(0 > free_sec_frame_queues(dec_fq_to_sec,
						       dec_fq_to_sec_retire))) {
			APP_ERROR
			    ("free_sec_frame_queues failed for dec_fq_to_sec");
			return -1;
		}
	}

	return 0;
}

/**
\brief		Initialize input buffer plain text data	and set	output buffer
		as 0 in compound frame descriptor
*/
static void set_enc_buf(void)
{
	uint32_t i;
	struct sg_entry_t *sgentry;
	uint8_t *out_buf, *in_buf;
	uint8_t plain_data = 0;
	phys_addr_t addr;
	uint8_t *enc_job_descriptor = NULL;

	for (index = 0; index < total_buf_num; index++) {
		markpoint(1);

		addr = fd[index].addr_hi;
		addr = (addr << 32) | fd[index].addr_lo;
		sgentry = (struct sg_entry_t *)ptov(addr);

		addr = sgentry->addr_hi;
		addr = (addr << 32) | sgentry->addr_lo;
		out_buf = (uint8_t *) ptov(addr);
		memset(out_buf, 0, output_buf_size);

		sgentry++;
		addr = sgentry->addr_hi;
		addr = (addr << 32) | sgentry->addr_lo;
		in_buf = (uint8_t *) ptov(addr);
		memset(in_buf, 0, input_buf_capacity);


		/* In case of SNOW_F8_F9 algorithm, a Job Descriptor must be
		 * inlined at the head of the input buffer. Set the encrypt
		 *  job descriptor here.
		 */
		if (crypto_info->algo == SNOW_F8_F9) {
			/* Update the input frame length. If this is not the
			 *  first iteration, the input len will remain set
			 *  from the decryption phase. The input len for
			 *  encrypt is different than for decrypt.
			 */

			sgentry->length = input_buf_capacity -
						 SNOW_F9_DIGEST_SIZE;

			/* Update the out frame length. If this is not the
			 *  first iteration, the output len will remain set
			 *  from the decryption phase. The output len for
			 *  encrypt is different than for decrypt.
			*/
			sgentry--;
			sgentry->length = output_buf_size;

			/* Convert the descriptor to an array of uint8_t items*/
			enc_job_descriptor = (uint8_t *) snow_jdesc_enc_f8_f9;

			/* Copy the job descriptor */
			for (i = 0; i < job_desc_buf_size; i++)
				in_buf[i] = enc_job_descriptor[i];
			in_buf += i;
		}

		/* Copy the input plain-text data */
		for (i = 0; i < crypto_info->buf_size; i++) {
			if (CIPHER == crypto_info->mode)
				in_buf[i] = ref_test_vector.plaintext[i];
			else
				in_buf[i] = plain_data++;
		}
	}
}

/**
\brief		Initialize input buffer as cipher text data and	set output
		buffer as 0 in compound frame descriptor
*/
static void set_dec_buf(void)
{

	struct sg_entry_t *sg_out;
	struct sg_entry_t *sg_in;

	uint16_t addr_hi;
	uint32_t addr_lo;
	uint32_t length;
	uint32_t offset;
	uint8_t bpid;
	uint8_t *out_buf;
	phys_addr_t addr;

	for (index = 0; index < total_buf_num; index++) {
		markpoint(4);

		addr = fd[index].addr_hi;
		addr = (addr << 32) | fd[index].addr_lo;
		sg_out = (struct sg_entry_t *)ptov(addr);
		sg_in = sg_out + 1;

		addr_hi = sg_out->addr_hi;
		addr_lo = sg_out->addr_lo;
		length = sg_out->length;
		offset = sg_out->offset;
		bpid = sg_out->bpid;

		sg_out->addr_hi = sg_in->addr_hi;
		sg_out->addr_lo = sg_in->addr_lo;
		sg_out->length = sg_in->length;
		sg_out->offset = sg_in->offset;
		sg_out->bpid = sg_in->bpid;

		sg_in->addr_hi = addr_hi;
		sg_in->addr_lo = addr_lo;
		sg_in->length = length;
		sg_in->offset = offset;
		sg_in->bpid = bpid;

		addr = sg_out->addr_hi;
		addr = (addr << 32) | sg_out->addr_lo;
		out_buf = (uint8_t *) ptov(addr);
		memset(out_buf, 0, crypto_info->buf_size);
	}
}

static void set_dec_auth_buf(void)
{

	struct sg_entry_t *sg_out;
	struct sg_entry_t *sg_in;
	uint8_t *out_buf = NULL;
	uint8_t *in_buf = NULL;
	phys_addr_t addr;
	uint8_t *dec_job_descriptor = NULL;
	int i;

	if (crypto_info->algo != SNOW_F8_F9) {
		APP_ERROR("%s: algorithm not supported", __func__);
		return;
	}

	for (index = 0; index < total_buf_num; index++) {
		markpoint(4);

		addr = fd[index].addr_hi;
		addr = (addr << 32) | fd[index].addr_lo;
		sg_out = (struct sg_entry_t *)ptov(addr);


		addr = sg_out->addr_hi;
		addr = (addr << 32) | sg_out->addr_lo;
		out_buf = (uint8_t *) ptov(addr);

		sg_in = sg_out + 1;

		addr = sg_in->addr_hi;
		addr = (addr << 32) | sg_in->addr_lo;
		in_buf = (uint8_t *) ptov(addr);
		memset(in_buf, 0, input_buf_capacity);

		/* Convert the descriptor to an array of uint8_t items */
		dec_job_descriptor = (uint8_t *) snow_jdesc_dec_f8_f9;

		/* A Job Descriptor must be inlined at the head of the input
		 *buffer. Set the decrypt job descriptor here. */
		for (i = 0; i < job_desc_buf_size; i++)
			in_buf[i] = dec_job_descriptor[i];
		in_buf += i;

		/* Validate that the output buffer size is equal with the
		 *  size of the reference cyphertext for decryption
		*/
		assert(output_buf_size ==
			 (snow_dec_f8_f9_reference_length
			[crypto_info->test_set - 1]/8));

		/* Use the reference encrypted data as input for decryption  */
		for (i = 0; i < output_buf_size; i++)
			in_buf[i] =
				 snow_dec_f8_f9_reference_ciphertext
				[crypto_info->test_set - 1][i];

		sg_in->length = output_buf_size + job_desc_buf_size;

		/* The output buffer will contain only the decoded F8 data */
		sg_out->length = crypto_info->buf_size;

		/* Clear the output buffer  */
		memset(out_buf, 0, sg_out->length);
	}
}


/**
\brief		Bitwise comparison of two vectors

\param[in]	left - The lefthand-side vector to enter the comparison

\param[in]	right - The righthand-side vector to enter the comparison

\param[in]	bitlength - The length(in bits) on which the comparison
		will be made

\retval		0 if and only if the vectors are identical up to
 		(including) the given bitlength.
 
\pre Neither of the buffer pointers is allowed to be NULL.

\pre Both buffers must be at least of size ceil(bitlength / sizeof(32)).
*/
static int test_vector_match(uint32_t *left, uint32_t *right,
						 uint32_t bitlen)
{
	uint32_t i;
	uint8_t reminder_bitlen;
	uint32_t bitmasks[32];

	assert(left != NULL);
	assert(right != NULL);

	/* initialize bitmasks */
	bitmasks[0] = 0xFFFFFFFF;
	for (i = 1; i <= 31; i++) {
		bitmasks[i] = bitmasks[i - 1] >> 1;
	}
	assert(bitmasks[31] == 1);

	/* compare the full 32-bit quantities */
	for (i = 0; i < (bitlen >> 5); i++) {
		if (left[i] != right[i]) {
			APP_ERROR
				("%s(): Bytes at offset %d don't match "
				"(0x%x, 0x%x)", __func__, i, left[i],
				 right[i]);
			return -1;
		}
	}

	/* compare the reminder dword starting with its most significant
	 *  bits */
	reminder_bitlen = bitlen & 0x1F;
	if (reminder_bitlen) {
		/* compare left[bitlen >> 5] with right[bitlen >> 5]
		 *  on the remaining number of bits
		 */
		uint32_t left_last = left[bitlen >> 5];
		uint32_t right_last = right[bitlen >> 5];

		if ((left_last | bitmasks[reminder_bitlen])
		    != (right_last | bitmasks[reminder_bitlen])) {
			APP_ERROR("%s(): Last bytes (%d) don't match on full"
				" %d bitlength", __func__, bitlen >> 5,
				 reminder_bitlen);
			return -1;
		}
	}

	return 0;
}

/**
\brief		Compare encrypted data returned by SEC 4.0 with	standard
		cipher text

\param[in]	None

\return		0 on success, otherwise -ve value 
*/
static int test_enc_match(void)
{
	struct sg_entry_t *sgentry;
	uint8_t *enc_buf;
	phys_addr_t addr;

	for (index = 0; index < total_buf_num; index++) {
		addr = fd[index].addr_hi;
		addr = (addr << 32) | fd[index].addr_lo;
		sgentry = (struct sg_entry_t *)ptov(addr);

		addr = sgentry->addr_hi;
		addr = (addr << 32) | sgentry->addr_lo;
		enc_buf = (uint8_t *) ptov(addr);
		if (test_vector_match((uint32_t *) enc_buf,
				      authnct ? (uint32_t *)
				      ref_test_vector.digest : (uint32_t *)
				      ref_test_vector.ciphertext,
				      authnct ? output_buf_size *
				      BITS_PER_BYTE : ref_test_vector.length) !=
		    0) {
			if (!authnct)
				APP_ERROR("%s: Encrypted frame %d with"
					" CIPHERTEXT test vector doesn't match"
					, __func__, index + 1);
			else
				APP_ERROR("%s digest match failed", algorithm);
			print_frame_desc(&fd[index]);
			return -1;
		}
	}
	if (!authnct)
		APP_INFO
		    ("All %s encrypted frame match found with cipher text",
		     algorithm);
	else
		APP_INFO("All %s digest successfully matched", algorithm);

	return 0;
}

/**
\brief		Compare decrypted data returned by SEC 4.0 with plain text
		input data

\param[in]	None

\return		0 on success, otherwise -ve value 
*/
static int test_dec_match(void)
{
	uint32_t i;
	struct sg_entry_t *sgentry;
	uint8_t *dec_buf;
	uint8_t plain_data = 0;
	phys_addr_t addr;

	for (index = 0; index < total_buf_num; index++) {
		markpoint(5);

		addr = fd[index].addr_hi;
		addr = (addr << 32) | fd[index].addr_lo;
		sgentry = (struct sg_entry_t *)ptov(addr);

		addr = sgentry->addr_hi;
		addr = (addr << 32) | sgentry->addr_lo;
		dec_buf = (uint8_t *) ptov(addr);
		if (CIPHER == crypto_info->mode) {
			if (test_vector_match((uint32_t *) dec_buf, (uint32_t *)
					      ref_test_vector.plaintext,
					      ref_test_vector.length) != 0) {
				APP_ERROR("%s: Decrypted frame %d with"
					" PLAINTEXT test vector doesn't match",
					 __func__, index + 1);
				print_frame_desc(&fd[index]);
				return -1;
			}
		} else {
			for (i = 0; i < crypto_info->buf_size; i++) {
				if (dec_buf[i] != plain_data) {
					APP_ERROR("%s: %s decrypted frame %d"
						" doesn't match!", __func__,
						 algorithm, index + 1);
					print_frame_desc(&fd[index]);
					return -1;
				}
				plain_data++;
			}
		}
	}
	APP_INFO("All %s decrypted frame match found with plain text",
		 algorithm);
	return 0;
}

/**
\brief		Enqueue frames to SEC 4.0 on Encrypt/Decrypt FQ's

\param[in]	mode - Encrypt/Decrypt

\return		0 on success, otherwise -ve value 
*/
static void do_enqueues(enum SEC_MODE mode)
{
	struct qman_fq *fq_to_sec;
	int i = 0;
	uint32_t ret;

	do {
		if (i >= crypto_info->buf_num_per_core)
			return;

		decorated_load_inc_32(packet_to_sec);

		if (mode == ENCRYPT)
			fq_to_sec = enc_fq_to_sec[(i*num_cores +
					 my_lwe_instance) % FQ_COUNT];
		else
			fq_to_sec = dec_fq_to_sec[(i*num_cores +
					 my_lwe_instance) % FQ_COUNT];

		APP_DEBUG("%s mode: Enqueue packet ->%d",
			  mode ? "Encrypt" : "Decrypt", i + 1);

		markpoint(2);
loop:
		ret = qman_enqueue(fq_to_sec, (struct qm_fd *)&fd[i*num_cores
					+ my_lwe_instance], 0);

		if (unlikely(ret)) {
			uint64_t now, then = get_atb();
			do {
				now = get_atb();
			} while (now < (then + 1000));
			goto loop;
		}

		i++;
	} while (1);
}

__PERCPU uint32_t enc_underruns;
__PERCPU uint32_t dec_underruns;
bool busy;

/**
\brief		call back handler for dequeued frames; Counts number of dequeued
		packets returned by SEC 4.0 on Encrypt/Decrypt FQ's

\param[in] 	mode - Encrypt/Decrypt

\return		None
*/
enum qman_cb_dqrr_result cb_dqrr(struct qman_portal *qm, struct qman_fq *fq,
				 const struct qm_dqrr_entry *dqrr)
{
	struct sg_entry_priv_t *sgentry_priv;
	phys_addr_t addr;
	enum SEC_MODE mode;

	busy = 1;
	markpoint(3);
	if ((dqrr->fqid >= fq_base_encrypt)
	    && (dqrr->fqid < fq_base_decrpyt)) {
		decorated_notify_inc_32(enc_packet_from_sec);
		mode = ENCRYPT;
	} else if ((dqrr->fqid >= fq_base_decrpyt)
		   && (dqrr->fqid < (fq_base_decrpyt + 2 * FQ_COUNT))) {
		decorated_notify_inc_32(dec_packet_from_sec);
		mode = DECRYPT;
	} else {
		APP_ERROR
		    ("%s: Invalid Frame Queue ID Returned by SEC 4.0=%d",
		     __func__, dqrr->fqid);
		abort();
	}

	APP_DEBUG("%s mode: Packet dequeued ->%d",
		  mode ? "Encrypt" : "Decrypt",
		  mode ? *enc_packet_from_sec : *dec_packet_from_sec);

	addr = dqrr->fd.addr_hi;
	addr = (addr << 32) | dqrr->fd.addr_lo;
	sgentry_priv = (struct sg_entry_priv_t *)ptov(addr);
	fd[sgentry_priv->index].status = dqrr->fd.status;

	return qman_cb_dqrr_consume;
}

/** Poll qman DQCR for encrypted frames */
static void enc_qman_poll(void)
{
	while (*enc_packet_from_sec < total_buf_num) {
		qman_poll();
		if (busy == 0)
			enc_underruns++;
		busy = 0;
	}

	return;
}

/** Poll qman DQCR for decrypted frames */
static void dec_qman_poll(void)
{
	while (*dec_packet_from_sec < total_buf_num) {
		qman_poll();
		if (busy == 0)
			dec_underruns++;
		busy = 0;
	}

	return;
}


/**
\brief		Checks if the status received in FD from CAAM block is valid

\param[in]	Void

\param[in]	Void

\return		0 - if status is correct (i.e. 0)
		-1 - if CAAM returned an error status (i.e. non 0)
*/
static int check_fd_status()
{
	uint32_t index;

	for (index = 0; index < total_buf_num; index++) {
		if (fd[index].status) {
			APP_ERROR("Bad status return from SEC");
			print_frame_desc(&fd[index]);
			return -1;
		}
	}
	return 0;
}

/**
\brief		Main function of SEC 4.0 Test Application

\param[in]	argc - Argument count

\param[in]	argv - Argument list pointer
*/
int main(int argc, char *argv[])
{
	int ret;
	int j = 1;
	uint32_t loop;
	uint64_t cpu_freq = 0;
	uint64_t atb_start_enc = 0;
	uint64_t atb_start_dec = 0;
	uint64_t enc_delta = 0, dec_delta = 0;
	uint32_t enc_cycles_per_frame = 0;
	uint32_t dec_cycles_per_frame = 0;
	struct sg_entry_t *sgentry;

	uint32_t bpid;
	struct bman_pool *bpool;
	struct bman_pool_params params;
	int ecode;
	phys_addr_t addr;

	spin_lock(&sec_lock);
	curr_lwe_id = get_lwe_id();
	crypto_info = (struct crypto_param *)&_whiteboard_start;

	/* Store the number of bytes for the job descriptor.
	   The SNOW_JDESC_ENC_F8_F9_LEN and SNOW_JDESC_DEC_F8_F9_LEN macros
	   return the length in 32 bit words.
	   Calculate the maximum between the encrypt and decrypt descriptors
	   because the same compound FD will be used for both processes, so
	   will allocate the memory area that can hold both descriptors.
	 */
	job_desc_buf_size = MAX(SNOW_JDESC_ENC_F8_F9_LEN,
				 SNOW_JDESC_DEC_F8_F9_LEN);
	if (0 == lwe_instance) {
		APP_INFO("Welcome to FSL SEC 4.0 application!");
		g_ctrl_lwe = curr_lwe_id;
	}
	spin_unlock(&sec_lock);
	loop = crypto_info->itr_num;

	if (false == crypto_info->valid_params) {
		APP_ERROR("SEC parameters are not initialized through CP");
		goto err_out;
	}

	buff_allocator = dpa_allocator_create(&ecode);
	if (unlikely(NULL == buff_allocator)) {
		APP_ERROR("dpa_buff_allocator_create failed");
		goto err_out;
	}
	APP_INFO("Buffer allocator created...");

	if (curr_lwe_id == g_ctrl_lwe) {
		sec_stats_init();
		cpu_freq = get_cpu_freq()/1000000; /* cpu_freq in MHz */

		ret = barrier_init(&sec40_barrier, get_online_core_mask());
		if (unlikely(ret != 0)) {
			APP_ERROR("Barrier initialization failed");
			ctrl_error = 1;
			goto err_out;
		}
		/* Extract pool Channel Offset */
		if (unlikely(0 > (pool_channel_offset =
				  extract_qman_pool_channel(devtree_ptr)))) {
			APP_ERROR("extract_qman_channel failed");
			ctrl_error = 1;
			goto err_out;
		}
		if (unlikely(0 > dpa_bpool_init())) {
			APP_ERROR("dpa_bpool_init failed");
			ctrl_error = 1;
			goto err_out;
		}
	}

	/* Wait till initialization completed */
	while ((0 == pool_channel_offset) && (ctrl_error == 0))
		;
	APP_INFO("LWE pool channel offset = %d", pool_channel_offset);

	if (ctrl_error)
		goto err_out;

	/* Get the number of cores working */
	spin_lock(&sec_lock);
	my_lwe_instance = num_cores++;
	spin_unlock(&sec_lock);
	if (unlikely(0 > barrier_sync(&sec40_barrier)))
		LWE_PANIC("barrier sync failed before creating bman pools");

	/* Calculate total number of buffers */
	total_buf_num = crypto_info->buf_num_per_core * num_cores;

	/*Create BMAN pools */
	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		/*Else create the pool object */
		params.flags = 0;
		params.cb = NULL;
		params.cb_ctx = NULL;
		params.bpid = bpid;

		bpool = bman_new_pool(&params);
		if (unlikely(NULL == bpool)) {
			APP_ERROR("bman_new_pool failed for BPID: %u", bpid);
			goto err_out;
		}

		if (unlikely(0 != dpa_allocator_add_bpool(buff_allocator,
							  dpa_bpools[bpid].size,
							  bpid, bpool))) {
			APP_ERROR("dpa_allocator_add_bpool failed");
			goto err_out;
		}

		APP_INFO("Added BPID: %u, Size: %u", bpid,
			 dpa_bpools[bpid].size);
	}

	if (curr_lwe_id == g_ctrl_lwe) {
		if (PERF == crypto_info->mode) {
			strcpy(mode_type, "PERF");
			crypto_info->test_set = 1;
		}

		if (unlikely(0 > init_ref_test_vector())) {
			APP_ERROR("%s: init_ref_test_vector() failed!",
				  __func__);
			ctrl_error = 1;
			goto error1;
		}

		if (CIPHER == crypto_info->mode) {
			strcpy(mode_type, "CIPHER");
			crypto_info->buf_size =
			    NO_OF_BYTES(ref_test_vector.length);
		}

		if (unlikely(0 > create_compound_fd())) {
			APP_ERROR("%s: create_compound_fd() failed!",
				  __func__);
			ctrl_error = 1;
			goto error1;
		}

		if (unlikely(0 > init_sec_fq())) {
			APP_ERROR("%s: init_sec_fq() failure", __func__);
			ctrl_init_fq_error = 1;
			goto error1;
		}

		APP_INFO("Processing %s for %d Frames", algorithm,
			 total_buf_num);
		APP_INFO("%s mode, buffer length = %d", mode_type,
			 crypto_info->buf_size);
		APP_INFO("Number of iterations = %d", crypto_info->itr_num);
	}
error1:
	if (unlikely(0 > barrier_sync(&sec40_barrier)))
		LWE_PANIC("barrier sync failed before qman_static_dequeue_add");
	if (ctrl_error)
		goto err_out;
	else if (ctrl_init_fq_error)
		goto err_free_fd;

	qman_static_dequeue_add(QM_SDQCR_CHANNELS_POOL(pool_channel_offset));
	MAGIC_BREAKPOINT;

	while (loop) {
		/* Set encryption buffer */
		if (curr_lwe_id == g_ctrl_lwe) {
			APP_INFO("Iteration %d start here", j);
			set_enc_buf();
			*packet_to_sec = 0;
			*enc_packet_from_sec = 0;
		}

		if (unlikely(0 > barrier_sync(&sec40_barrier)))
			LWE_PANIC
			    ("Encrypt mode: barrier sync failed before enqueue");

		if (curr_lwe_id == g_ctrl_lwe)
			/* encrypt mode: start time */
			atb_start_enc = get_atb();

		/* Send data to SEC40 for encryption/authentication */
		do_enqueues(ENCRYPT);

		if (unlikely(0 > barrier_sync(&sec40_barrier)))
			LWE_PANIC("Encrypt mode: barrier sync failed \
				   before enqueue");

		APP_DEBUG("Encrypt mode: Total packet sent to SEC = %d",
			  *packet_to_sec);

		/* Recieve encrypted or MAC data from SEC40 */
		enc_qman_poll();
		APP_DEBUG("Encrypt mode: Total packet returned from SEC = %d",
			  *enc_packet_from_sec);

		if (curr_lwe_id == g_ctrl_lwe)
			/* encrypt mode: accumulated time taken */
			enc_delta += (get_atb() - atb_start_enc);

		/* Test ciphertext or MAC generated by SEC40 */
		if (curr_lwe_id == g_ctrl_lwe) {
			if (CIPHER == crypto_info->mode) {
				if (unlikely(0 > check_fd_status())) {
					ctrl_error = 1;
					goto error2;
				}
				if (unlikely(0 > test_enc_match())) {
					ctrl_error = 1;
					goto error2;
				}
			}

			/* Set decryption buffer */
			if (crypto_info->algo == SNOW_F8_F9)
				set_dec_auth_buf();
			else if (!authnct)
				set_dec_buf();

			*packet_to_sec = 0;
			*dec_packet_from_sec = 0;
		}
error2:
		if (unlikely(0 > barrier_sync(&sec40_barrier)))
			LWE_PANIC
			    ("Decrypt mode: barrier sync failed before enqueue");
		if (ctrl_error)
			goto err_free_fq;

		if (authnct)
			goto result;

		if (curr_lwe_id == g_ctrl_lwe)
			/* decrypt mode: start time */
			atb_start_dec = get_atb();

		/* Send data to SEC40 for decryption */
		do_enqueues(DECRYPT);

		if (unlikely(0 > barrier_sync(&sec40_barrier)))
			LWE_PANIC("Encrypt mode: barrier sync failed \
				   before enqueue");

		APP_DEBUG("Decrypt mode: Total packet sent to SEC = %d",
			  *packet_to_sec);

		/* Recieve decrypted data from SEC40 */
		dec_qman_poll();
		APP_DEBUG("Decrypt mode: Total packet returned from SEC = %d",
			  *dec_packet_from_sec);

		if (curr_lwe_id == g_ctrl_lwe)
			/* decrypt mode: accumulated time taken */
			dec_delta += (get_atb() - atb_start_dec);

		/* Test decrypted data with original plaintext */
		if (curr_lwe_id == g_ctrl_lwe) {
			if (CIPHER == crypto_info->mode) {
				if (unlikely(0 > check_fd_status())) {
					ctrl_error = 1;
					goto error2;
				}
			}
			if (unlikely(0 > test_dec_match())) {
				ctrl_error = 1;
				goto error3;
			}
		}
error3:
		if (unlikely(0 > barrier_sync(&sec40_barrier)))
			LWE_PANIC
			    ("barrier sync failed after test_dec_match");
		if (ctrl_error)
			goto err_free_fq;

result:
		if (curr_lwe_id == g_ctrl_lwe)
			APP_INFO("Iteration %d finished", j++);
		loop--;
	}

	if (unlikely(0 > barrier_sync(&sec40_barrier)))
		LWE_PANIC
		    ("barrier sync failed before result");

	APP_INFO("%s: Underruns = %d",
		authnct ? "Authenticate" : "Encrypt",
		enc_underruns);
	if (!authnct)
		APP_INFO("Decrypt: Underruns = %d",
			dec_underruns);
	if (curr_lwe_id == g_ctrl_lwe) {
		enc_cycles_per_frame = (enc_delta) /
			 (crypto_info->itr_num * total_buf_num);

		APP_INFO("%s: Throughput = %llu Mbps",
		authnct ? "Authenticate" : "Encrypt",
		(cpu_freq * BITS_PER_BYTE * crypto_info->buf_size) /
		enc_cycles_per_frame);
		if (!authnct) {
			dec_cycles_per_frame = dec_delta /
				 (crypto_info->itr_num * total_buf_num);

			APP_INFO("%s: Throughput = %llu Mbps",
			"Decrypt",
			(cpu_freq * BITS_PER_BYTE * crypto_info->buf_size) /
			dec_cycles_per_frame);
		}

		if (curr_lwe_id == g_ctrl_lwe) {
			for (index = 0; index < total_buf_num; index++) {
				addr = fd[index].addr_hi;
				addr = (addr << 32) | fd[index].addr_lo;
				sgentry = (struct sg_entry_t *)ptov(addr);
				free(sgentry);
			}
			APP_INFO("All FDs are free");
		}

		if (unlikely(0 > free_sec_fq())) {
			APP_ERROR("%s: free_sec_fq failed", __func__);
				APP_ERROR("partition restart would not work");
			goto err_out;
		}

		APP_INFO("SEC 4.0 test PASSED");
	}
	if (unlikely(0 > barrier_sync(&sec40_barrier)))
		LWE_PANIC("barrier sync failed before markpoint 31");
	markpoint(31);
	return 0;

err_free_fq:
	if (curr_lwe_id == g_ctrl_lwe) {
		if (unlikely(0 > free_sec_fq())) {
			APP_ERROR("%s: free_sec_fq failed", __func__);
			APP_ERROR("partition restart would not work");
			return -1;
		}
		APP_INFO("All FQs are free");
	}

err_free_fd:
	if (curr_lwe_id == g_ctrl_lwe) {
		for (index = 0; index < total_buf_num; index++) {
			addr = fd[index].addr_hi;
			addr = (addr << 32) | fd[index].addr_lo;
			sgentry = (struct sg_entry_t *)ptov(addr);
			free(sgentry);
		}
		APP_INFO("All FDs are free");
	}

err_out:
	APP_ERROR("FATAL error, exiting");
	markpoint(31);
	return -1;
}

/**
\brief		Prints the Frame Descriptor on Console

\param[in]	struct qm_fd * Pointer to the Frame Descriptor

\return		None
*/
void print_frame_desc(struct qm_fd *fd)
{
	uint8_t *v;
	phys_addr_t addr;

	APP_ERROR("Frame Description at address 0x%8p", (void *)fd);
	if (NULL == fd) {
		APP_ERROR(" - NULL pointer");
	} else {
		APP_ERROR(" - debug	: %d", fd->dd);
		APP_ERROR(" - bpid	: %d", fd->bpid);
		APP_ERROR(" - address	: 0x%04x%08x",
			  fd->addr_hi, fd->addr_lo);

		switch ((int)fd->format) {
		case 0:
			APP_ERROR
			    (" - format		: 0 - Short single buffer FD");
			APP_ERROR(" - offset	: %d", fd->offset);
			APP_ERROR(" - length	: %d", fd->length20);
			break;
		case 1:
			APP_ERROR(" - format	: 1 - Compound FD");
			APP_ERROR(" - congestion weight	: %d",
				  fd->cong_weight);
			break;
		case 2:
			APP_ERROR
			    (" - format		: 2 - Long single buffer FD");
			APP_ERROR(" - length	: %d", fd->length29);
			break;
		case 4:
			APP_ERROR
			    (" - format		: 4 - Short multi buffer FD");
			APP_ERROR(" - offset	: %d", fd->offset);
			APP_ERROR(" - length	: %d", fd->length29);
			break;
		case 6:
			APP_ERROR
			    (" - format		: 6 - Long multi buffer FD");
			APP_ERROR(" - length	: %d", fd->length29);
			break;
		default:
			APP_ERROR
			    (" - format		: INVALID format %d",
			     fd->format);
		}

		APP_ERROR(" - status/command	: 0x%08x", fd->cmd);

		if (fd->format == qm_fd_compound) {
			struct sg_entry_t *sgentry;

			addr = fd->addr_hi;
			addr = (addr << 32) | fd->addr_lo;
			sgentry = (struct sg_entry_t *)ptov(addr);

			APP_ERROR
			    (" - compound FD S/G list at 0x%04x%08x",
			     fd->addr_hi, fd->addr_lo);
			APP_ERROR("   - SG Entry");
			APP_ERROR
			    ("      - address       0x%04x%08x",
			     sgentry->addr_hi, sgentry->addr_lo);

			APP_ERROR("      - F	     %d", sgentry->final);
			APP_ERROR("      - E	     %d",
				  sgentry->extension);
			APP_ERROR("      - length	%d", sgentry->length);
			APP_ERROR("      - bpid	  %d", sgentry->bpid);
			APP_ERROR("      - offset	%d", sgentry->offset);

			APP_ERROR("      - Output buffer data at 0x%04x%08x",
				sgentry->addr_hi, sgentry->addr_lo);
			addr = sgentry->addr_hi;
			addr = (addr << 32) | sgentry->addr_lo;
			v = (uint8_t *) ptov(addr);
			for (int i = 0; i < output_buf_size; i++)
				APP_ERROR("	0x%x", *v++);

			sgentry++;
			APP_ERROR("   - Next SG Entry");
			APP_ERROR
			    ("      - address       0x%04x%08x",
			     sgentry->addr_hi, sgentry->addr_lo);

			APP_ERROR("      - F	     %d", sgentry->final);
			APP_ERROR("      - E	     %d",
				  sgentry->extension);
			APP_ERROR("      - length	%d", sgentry->length);
			APP_ERROR("      - bpid	  %d", sgentry->bpid);
			APP_ERROR("      - offset	%d", sgentry->offset);

			APP_ERROR("      - Input buffer data at 0x%04x%08x",
				sgentry->addr_hi, sgentry->addr_lo);
			addr = sgentry->addr_hi;
			addr = (addr << 32) | sgentry->addr_lo;
			v = (uint8_t *) ptov(addr);
			for (int i = 0; i < crypto_info->buf_size; i++)
				APP_ERROR("	0x%x", *v++);
		}
	}
}
