#!/bin/sh
#
#  Copyright (C) 2009 Freescale Semiconductor, Inc.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
# NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

usage_and_exit() {
	cat 1>&2 <<EOF
Usage:  $0 [-p <num_pkt_events>[,<num_pkt_events>]] file [file ...]

Process all (or last num_pkt_events) events in the specified file(s),
and output stats.  Note that only the rate stats honor the -p
flag.

If the more-complicated XX,YY value is given, it will only honor
the YY events starting after event number XX.  (tail -n +XX | head -n YY)
EOF
exit 1
}

# If we are given a -p flag for the last P packets, honor it.
TEMP=`getopt "p:" $*`
if [ $? != 0 ]; then usage_and_exit; fi
eval set -- "$TEMP"

for opt in $*; do
	case $opt in
		-p)
			if expr index $2 ',' >/dev/null; then
				# Got the complex XX,YY format.
				pkt_first_arg=`echo $2 | sed -e 's/,.*$//'`
				pkt_second_arg=`echo $2 | sed -e 's/.*,//'`
			else
				pkt_tail=$2;
			fi
			shift 2;;
		--) shift;;
	esac
done

for f in $*; do
	# If the file is not a regular file, then first cp it to a
	# temp file.
	if [ ! -f $f ]; then
		set -x
		cp $f /tmp/summ_pkts.$$.out
		
		f=/tmp/summ_pkts.$$.out
		remove_me="$remove_me $f"
		set +x
	fi
	# Check if there is a stats.after file.
	statfile=`echo $f | sed -e 's/.out$/.stats.after/'`
	# If no file, use the output file.
	if [ ! -f ${statfile} ]; then statfile=$f; fi
	printf "=== Packet summary for $f ===\n"
	# See if it is running at 600MHz or 800MHz etc.
	sys_mhz=`grep ^sim.system_clock_freq_mhz $f | awk '{print $(NF)}'`
	qman_mhz=`grep ^sim.clock_ratio $f | awk -v sys_mhz=$sys_mhz '
	{
		ratio=$(NF);
		# Split by colon.
		split(ratio, parts, ":");
		print sys_mhz * parts[3] / parts[2];
	}
	'`
	echo "*** Assuming Platform is at 800 MHz, QMan is at 400 MHz."
	
	# If the test did not complete, print a warning.
	if grep "test PASSED" $f >/dev/null; then
		: # Do nothing.
	else
		echo "WARNING:  No SUCCESS line detected -- simulation still in progress, or failed?"
		exit
	fi

	# Extract qman trace for faster processing
	qf=`echo $f | sed -e 's/.out$/.qmantrace/'`
	if [ "$qf" = "$f" ]; then
		  # If the given filename doesn't have .out in the name, then
		  # we better be careful, because $qf will be the same, and
		  # the command below will clobber our input!
		  qf=/tmp/summ_pkts.$$.qmantrace
		  remove_me="$remove_me $qf"
	fi
	grep "qman\]" $f > $qf
	if [ "$pkt_tail" != "" ]; then
		mv $qf $qf.full
		tail -n $pkt_tail $qf.full > $qf
	fi
	if [ "$pkt_first_arg" != "" ]; then
		mv $qf $qf.full
		tail -n +$pkt_first_arg $qf.full | head -n $pkt_second_arg > $qf
	fi

	awk -v qman_mhz=$qman_mhz '
	BEGIN {
		first_sw_dq=1;
		num_pkt = 500;
	}
	#/enqueue event on DC portal 2/ {
	($4=="enqueue" && $7=="DC" && $9=="2" && dc2_eq<num_pkt) {
		dc2_eq++;
		if(dc2_eq==1)
			cyc_first_dc2_eq = $3;
		if(dc2_eq==num_pkt)
			cyc_last_dc2_eq = $3;
   	}
	#/enqueue event on SW/ {
	($4=="enqueue" && $7=="SW" && sw_eq<num_pkt) {
		sw_eq++;
		if(sw_eq==1)
			cyc_first_sw_eq = $3;
		if(sw_eq==num_pkt)
			cyc_last_sw_eq = $3;
	}	
   	#/dequeue event on DC portal 2/ {
	($4=="dequeue" && $7=="DC" && $9=="2" && dc2_dq<num_pkt) {
		dc2_dq++;
		if(dc2_dq==1){
			cyc_first_dc2_dq = $3;
			bklog_pre_dc2_dq = sw_eq;
		}
		if(sw_eq==num_pkt)
			cyc_last_dc2_dq = $3;
   	}
	#/dequeue event on SW/ {
	($4=="dequeue" && $7=="SW" && sw_dq<num_pkt) {
		sw_dq++;
		if(sw_dq==1)
			cyc_first_sw_dq = $3;

	      if(sw_eq==num_pkt)
			cyc_last_sw_dq = $3;
	}
	# Always calculate current num packets in system.
	{
		pkts_in_pre_caam = sw_eq - dc2_dq;
		if (pkts_in_pre_caam > max_pkts_in_pre_caam)
			max_pkts_in_pre_caam = pkts_in_pre_caam;
	}
function print_mpps_gbps(prompt, num, delta_cycs) {
	# QMan runs at 400MHz, hence the 400 factor to get from packets/cycles into
	# packets/second.
	if (delta_cycs) {
		mpps = num/delta_cycs * 400;
	} else
		mpps = 0;
	printf "%-12s rate:  %6d in %8d QMan cycs = %5.4f Mpps\n", prompt, num, delta_cycs, mpps;
}

END {
	printf "===============\n"
	if(dc2_dq>0)
		printf "SEC \t Dequeues: %d\n", dc2_dq;
	if(dc2_eq>0)
		printf "SEC \t Enqueues: %d\n", dc2_eq;

	printf "===============\n"
	if(dc2_dq>0)
		printf "SW enqueues before first SEC dequeue:  %5d\n", bklog_pre_dc2_dq;
	if(dc2_dq>0)
		printf "SW enqueues  after first SEC dequeue:  %5d\n", sw_eq - bklog_pre_dc2_dq;

	printf "===============\n"
	if(dc2_dq>0)
		printf "Max packet backlog pre-SEC:  %d\n", max_pkts_in_pre_caam;

	printf "===============\n"
	if(sw_eq>0)
		print_mpps_gbps("SW  enqueue", num_pkt, cyc_last_sw_eq - cyc_first_sw_eq);
	if(dc2_dq>0)
		print_mpps_gbps("SEC  dequeue", num_pkt, cyc_last_dc2_dq - cyc_first_dc2_dq);
	if(dc2_eq>0)
		print_mpps_gbps("SEC  enqueue", num_pkt, cyc_last_dc2_eq - cyc_first_dc2_eq);
	if(sw_dq>0)
		print_mpps_gbps("SW  dequeue", num_pkt, cyc_last_sw_dq - cyc_first_sw_dq);
}
	' $qf

done

if [ "$remove_me" != "" ]; then
	rm $remove_me
fi
