/**
 \file simple_pme_cp.c
 \brief  Basic LWE PME Config Tool
 */
/*
 * Copyright (C) 2009 - 2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <argp.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include "simple_pme_cp.h"
#include "pme/simple_pme_cpdp.h"
#include "app_cpdp.h"
#include "wb/wb.h"

#ifdef DEBUG
#define lwe_pme_log_print		printf
#else
#define lwe_pme_log_print(a, ...)
#endif

unsigned int g_mndtr_param;
volatile struct lwe_ctrl_op_info *g_shm_pme_info;
error_t g_parse_error;

static error_t parse_opt(int key, char *arg, struct argp_state *state);
static error_t parse_pme_add_del_opt(int key, char *arg,
				     struct argp_state *state);

static char args_doc[] = "WILL BE DOCUMENTED LATER";

static char doc[] = "WILL BE DOCUMENTED LATER";

static struct argp pme_add_argp =
    { pme_add_options, parse_pme_add_del_opt, args_doc, doc, NULL, NULL, NULL };

static struct argp pme_del_argp =
    { pme_del_options, parse_pme_add_del_opt, args_doc, doc, NULL, NULL, NULL };

static struct argp pme_go_argp =
    { pme_go_options, parse_pme_add_del_opt, args_doc, doc, NULL, NULL, NULL };

static struct argp_option options[] = {
	{"pmeadd", 'J', "TYPE", 0, "adding a pme context entry", 0},
	{"pmedel", 'K', "TYPE", 0, "deleting a pme context entry", 0},
	{"Start/ Go", 'O', "TYPE", OPTION_ARG_OPTIONAL,
	 "Start the steady phase/ processing of packets", 0},
	{0, 0, 0, 0, 0, 0}
};

/*
   The ARGP structure itself.
*/

static struct argp argp =
    { options, parse_opt, args_doc, doc, NULL, NULL, NULL };

static struct wb_handle wb_handle;
static struct wb_msg *wb_msg;

/**
 \brief Defines actions for parsing the add/ delete command options; 
	it is called for each option parsed
 \param[in] key For each option that is parsed, parser is called with a value of 
		key from that option's key field in the option vector
 \param[in] arg If key is an option, arg is its given value.
 \param[in] state state points to a struct argp_state, containing pointer to pme_info structure
 \return
 */
static error_t parse_pme_add_del_opt(int key, char *arg,
				     struct argp_state *state)
{
	struct lwe_ctrl_op_info *op_info = state->input;
	struct lwe_ctrl_spa_cfg_info *pme_info = &(op_info->app_info.pme_info);
	uint32_t tmp = 0;
	int32_t val;

	lwe_pme_log_print("\r\nKey = %c, arg = %s", key, arg);

	switch (key) {

		/* flow_id */
	case 'f':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_FLOWID_MIN) ||
		    (val > LWE_CTRL_PME_FLOWID_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->flow_id = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_FLOWID;
		break;

		/* set */
	case 's':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_SET_MIN) ||
		    (val > LWE_CTRL_PME_SET_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->set = (uint8_t)val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_SET;
		break;

		/* subset */
	case 'u':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_SUBSET_MIN) ||
		    (val > LWE_CTRL_PME_SUBSET_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->subset = (uint16_t)atoi(arg);

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_SUBSET;
		break;

		/* scan mode type - direct/ flow */
	case 'm':
		tmp = atoi(arg);
		if (tmp == 1)
			pme_info->flag |= SET_SCAN_MODE_FLOW;
		else
			pme_info->flag &= SET_SCAN_MODE_DIRECT;
		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_FLAG;
		break;

		/* residue - enable/ disable */
	case 'r':
		tmp = atoi(arg);
		if (tmp == 1)
			pme_info->flag |= SET_RESIDUE_ENABLE;
		else
			pme_info->flag &= SET_RESIDUE_DISABLE;
		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_FLAG;
		break;

		/* report verbosity */
	case 'p':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_REPORT_VERBOSITY_MIN)
		    || (val > LWE_CTRL_PME_REPORT_VERBOSITY_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->report_verbosity = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_REPORT_VERBOSITY;
		break;

		/* session id */
	case 'i':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_SESSION_ID_MIN) ||
		    (val > LWE_CTRL_PME_SESSION_ID_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->session_id = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_SESSION_ID;
		break;

		/* output buffer size */
	case 'b':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_INPUT_BUFF_CHUNK_MIN)
		    || (val > LWE_CTRL_PME_INPUT_BUFF_CHUNK_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->input_buff_chunk = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_INPUT_BUFF_CHUNK;
		break;

		/* Qos In */
	case 'q':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_QOS_IN_MIN) ||
		    (val > LWE_CTRL_PME_QOS_IN_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->qos_in = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_QOS_IN;
		break;

		/* Qos out */
	case 't':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_QOS_OUT_MIN) ||
		    (val > LWE_CTRL_PME_QOS_OUT_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->qos_out = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_QOS_OUT;
		break;

		/* match limit */
	case 'l':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_MATCH_LIMIT_MIN) ||
		    (val > LWE_CTRL_PME_MATCH_LIMIT_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->match_limit = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_MATCH_LIMIT;
		break;

		/* compare limit */
	case 'o':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_COMPARE_LIMIT_MIN)
		    || (val > LWE_CTRL_PME_COMPARE_LIMIT_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->compare_limit = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_COMPARE_LIMIT;
		break;

		/* scan function type - block/ non-block */
	case 'n':
		tmp = atoi(arg);
		if (tmp == 1)
			pme_info->flag |= SET_SCAN_FN_TYPE_BLOCK;
		else
			pme_info->flag &= SET_SCAN_FN_TYPE_NON_BLOCK;
		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_FLAG;
		break;

		/* output buffer from App */
	case 'e':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_OUTPUT_BUFF_FRM_APP_MIN)
		    || (val > LWE_CTRL_PME_OUTPUT_BUFF_FRM_APP_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->output_buff_frm_app = val;

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_OUTPUT_BUFF_FRM_APP;
		break;

		/* input buffer size */
	case 'z':
		pme_info->input_buff_size = atoi(arg);
		if ((pme_info->input_buff_size <
		     LWE_CTRL_PME_INPUT_BUFF_SIZE_MIN)
		    || (pme_info->input_buff_size >
			LWE_CTRL_PME_INPUT_BUFF_SIZE_MAX)) {
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}

		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_INPUT_BUFF_SIZE;
		break;

		/* input file to be read */
	case 'd':
		{
			FILE *fd;
			uint32_t file_size = 1024 * 3;
			uint32_t i = 0;
			fd = fopen(arg, "r");
			if (fd == NULL) {
				printf("\nFile %s could not be opened", arg);
				return ARGP_ERR_UNKNOWN;
			}

			pme_info->size_of_file =
			    fread(pme_info->file, 1, 10, fd);
			while (1) {
				if (strncmp
				    ((char *)(pme_info->file), "0x12344321",
				     10) == 0)
					break;
				memcpy(pme_info->file, &pme_info->file[1], 9);

				pme_info->size_of_file =
				    fread(&(pme_info->file[9]), 1, 1, fd);

				i++;
				if (i > file_size) {
					fclose(fd);
					return ARGP_ERR_UNKNOWN;
				}
			}

			pme_info->size_of_file =
			    fread(pme_info->file, 1, PME_MAX_FILE_SIZE, fd);
			fclose(fd);

			if (pme_info->size_of_file)
				g_mndtr_param |=
				    LWE_CTRL_PARAM_BMASK_PME_INPUT_FILE;
			else
				return ARGP_ERR_UNKNOWN;
			break;

		}

	case 'v':
		val = atoi(arg);
		if ((val < LWE_CTRL_PME_REPORT_PARSE_ENABLE_MIN) ||
			(val > LWE_CTRL_PME_REPORT_PARSE_ENABLE_MAX))
		{
			printf("Invalid Value \"%s\" for '%c'\n", arg, key);
			g_parse_error = ERANGE;
			return ERANGE;
		}
		pme_info->report_parse_enable = (uint8_t)val;
		g_mndtr_param |= LWE_CTRL_PARAM_BMASK_PME_REPORT_PARSE_ENABLE;
		break;

	default:
		return ARGP_ERR_UNKNOWN;
	}

	return 0;
}

/**
 \brief Copies data from shared memory area into pme_info structure
 \param[in] pme_info Pointer to the structure buffer for copying data
 \return
 */
void lwe_pme_copy_from_shared_memory(struct lwe_ctrl_op_info *pme_info)
{
	lwe_pme_log_print("lwe_pme_copy_from_shared_memory: Enter\n");

	memcpy(pme_info, (void *)g_shm_pme_info,
	       sizeof(struct lwe_ctrl_op_info));

	lwe_pme_log_print("lwe_pme_copy_from_shared_memory: Exit\n");
	return;
}

/**
 \brief Copies data from the pme_info structure to the shared memory area.
 \param[in] pme_info Pointer to the structure that needs to be copies onto shared memory area
 \return
 */
void lwe_pme_copy_to_shared_memory(struct lwe_ctrl_op_info *pme_info)
{
	unsigned int result = LWE_CTRL_RSLT_SUCCESSFULL;
	unsigned int part_id = 0;
	unsigned int part_mask = 0;
	char label[10] ;
#if 0
	char error_str[][2] = {
		/* Addition */
		{"PME Entry Addition failed", "PME Entry Added successfully"},

		/* Deletion */
		{"PME Entry Deletion failed", "PME Entry Deleted successfully"},
	};
#endif
	lwe_pme_log_print("lwe_pme_copy_to_shared_memory: Enter\n");
	/*
	 ** Copy the structure onto the white board shared memory 
	 */
	memcpy((void *)g_shm_pme_info, pme_info,
	       sizeof(struct lwe_ctrl_op_info));
	strcpy(label, "p3-lwe");
	part_id = wb_get_id_from_part_name(label);
	part_mask = 1 << part_id;
	result = wb_send(wb_msg, part_mask, 0);
	if (result == WB_TIMEOUT) {
		result = LWE_CTRL_RSLT_FAILURE;
	} else {
		/*
		** Check for result of Request
		*/
		result = WB_GET_RSLT(wb_msg, part_id);
	}

	if (WB_GET_MSG_TYPE(wb_msg) == LWE_CTRL_CMD_TYPE_PME_ADD) {
		if (result == LWE_CTRL_RSLT_SUCCESSFULL) {
			printf("PME Entry Added successfully\n");

		} else {
			printf("PME Entry Addition failed\n");
		}
	} else if (WB_GET_MSG_TYPE(wb_msg) == LWE_CTRL_CMD_TYPE_PME_DEL) {
		if (result == LWE_CTRL_RSLT_SUCCESSFULL) {
			printf("PME Entry Deleted successfully\n");
		} else {
			printf("PME Entry Deletion failed\n");
		}
	} else if (WB_GET_MSG_TYPE(wb_msg) == LWE_CTRL_CMD_TYPE_PME_GO) {
		if (result == LWE_CTRL_RSLT_SUCCESSFULL) {
			printf("PME Application Started successfully\n");
		} else {
			printf("PME Application failed\n");
		}

	} else {
		lwe_pme_copy_from_shared_memory(pme_info);
	}

	lwe_pme_log_print("lwe_pme_copy_to_shared_memory: Exit\n");
	WB_RESET_RSLT(wb_msg, part_id);
	return;
}

/**
 \brief 
 \param[in] argc
 \param[in] argv 
 \param[in] type Message Type for the CP request - Add or delete
 \return
 */
void lwe_pme_add_del_command(int argc, char **argv, char *type)
{
	unsigned int i = 0;
	struct lwe_ctrl_op_info pme_info;
	unsigned int mndtr_param_map[] = { LWE_CTRL_PME_ADD_MDTR_PARAM_MAP,
		LWE_CTRL_PME_DEL_MDTR_PARAM_MAP,
		LWE_CTRL_PME_GO_MDTR_PARAM_MAP
	};
	struct argp *pme_argp[] = { &pme_add_argp, &pme_del_argp, &pme_go_argp};
	struct lwe_ctrl_spa_cfg_info *tmp_pme_info =
		    &pme_info.app_info.pme_info;


	lwe_pme_log_print("\r\nlwe_pme_add_del_command: Enter");

	/*
	 ** Initializing the route info structure
	 */
	memset(&pme_info, 0, sizeof(pme_info));
	pme_info.state = LWE_CTRL_CMD_STATE_IDLE;
	WB_SET_MSG_TYPE(wb_msg, (unsigned int)type);
	pme_info.result = LWE_CTRL_RSLT_FAILURE;
	pme_info.free_offset = g_shm_pme_info->free_offset;

	if (WB_GET_MSG_TYPE(wb_msg) == LWE_CTRL_CMD_TYPE_PME_ADD) {
		/* 
		 ** Set the default values before 
		 ** extracting the values coming  
		 ** from the CLI command
		 */
		tmp_pme_info->set = LWE_CTRL_PME_DEF_SET;
		tmp_pme_info->subset = LWE_CTRL_PME_DEF_SUBSET;
		tmp_pme_info->flag = LWE_CTRL_PME_DEF_FLAG;
		tmp_pme_info->report_verbosity =
		    LWE_CTRL_PME_DEF_REPORT_VERBOSITY;
		tmp_pme_info->session_id = LWE_CTRL_PME_DEF_SESSION_ID;
		tmp_pme_info->input_buff_chunk =
		    LWE_CTRL_PME_DEF_INPUT_BUFF_CHUNK;
		tmp_pme_info->qos_in = LWE_CTRL_PME_DEF_QOS_IN;
		tmp_pme_info->qos_out = LWE_CTRL_PME_DEF_QOS_OUT;
		tmp_pme_info->match_limit = LWE_CTRL_PME_DEF_MATCH_LIMIT;
		tmp_pme_info->compare_limit = LWE_CTRL_PME_DEF_COMPARE_LIMIT;
		tmp_pme_info->output_buff_frm_app =
		    LWE_CTRL_PME_DEF_OUTPUT_BUFF_FRM_APP;
		tmp_pme_info->input_buff_size =
		    LWE_CTRL_PME_DEF_INPUT_BUFF_SIZE;
	} else if (WB_GET_MSG_TYPE(wb_msg) == LWE_CTRL_CMD_TYPE_PME_GO) {
		tmp_pme_info->report_parse_enable =
		    LWE_CTRL_PME_DEF_REPORT_PARSE_DISABLE;
	}

	g_mndtr_param = 0;
	g_parse_error = 0;

	/* Where the magic happens */
	argp_parse(
		pme_argp[WB_GET_MSG_TYPE(wb_msg) - LWE_CTRL_CMD_TYPE_PME_ADD],
		argc, argv, 0, 0, &pme_info);

	if (pme_info.app_info.pme_info.input_buff_size >
	    pme_info.app_info.pme_info.size_of_file)
		pme_info.app_info.pme_info.input_buff_size =
		    pme_info.app_info.pme_info.size_of_file;

	if (g_parse_error != 0)
		return;

	/*
	 ** If all the mandatory parameters for the operation are present,
	 ** Copy the data onto the shared memory area
	 */
	if ((g_mndtr_param &
	mndtr_param_map[WB_GET_MSG_TYPE(wb_msg) - LWE_CTRL_CMD_TYPE_PME_ADD]) ==
	mndtr_param_map[WB_GET_MSG_TYPE(wb_msg) - LWE_CTRL_CMD_TYPE_PME_ADD])
		goto copy;

	/*
	 ** Check for the mandatory parameter which is misisng
	 */
	for (i = 0; i < LWE_CTRL_PARAM_PME_MAX_BIT_NO; i++) {
		if (((mndtr_param_map
		      [WB_GET_MSG_TYPE(wb_msg) -
		       LWE_CTRL_CMD_TYPE_PME_ADD] & (1 << i)) != 0)
		    && ((g_mndtr_param & (1 << i)) == 0)) {
			printf
			    ("Route Entry Operation failed as mandatory parameters missing; --%s or -%c\n",
			     pme_add_options[i].name, pme_add_options[i].key);
			return;
		}
	}			/* end of for (i = 0; i < LWE_CTRL_PARAM_PME_MAX_B .... */

copy:
	lwe_pme_copy_to_shared_memory(&pme_info);

	lwe_pme_log_print("\r\nlwe_pme_add_del_command: Exit");
	return;
}

/**
 \brief Defines actions for parsing the lwe_ip command options - add/ delete/ sastats/ ifstats/ salist/ iflist; 
	it is called for each option parsed
 \param[in] key For each option that is parsed, parser is called with a value of 
		key from that option's key field in the option vector
 \param[in] arg If key is an option, arg is its given value.
 \param[in] state state points to a struct argp_state, containing pointer to pme_info structure
 \return
 */
static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
	switch (key) {

	case 'J':
		WB_SET_MSG_TYPE(wb_msg, LWE_CTRL_CMD_TYPE_PME_ADD);
		lwe_pme_log_print
		    ("\nFILE: %s : LINE: %d :IN PARSE_OPT::PME ADD OPTION SELECTED",
		     __FILE__, __LINE__);
		break;

	case 'K':
		WB_SET_MSG_TYPE(wb_msg, LWE_CTRL_CMD_TYPE_PME_DEL);
		lwe_pme_log_print
		    ("\nFILE: %s : LINE: %d :IN PARSE_OPT::PME DEL OPTION SELECTED",
		     __FILE__, __LINE__);
		break;

	case 'O':
		WB_SET_MSG_TYPE(wb_msg, LWE_CTRL_CMD_TYPE_PME_GO);
		lwe_pme_log_print
		    ("\nFILE: %s : LINE: %d :IN PARSE_OPT::PME GO OPTION SELECTED",
		     __FILE__, __LINE__);
		break;

	default:
		return ARGP_ERR_UNKNOWN;
	}

	return ARGP_KEY_END;
}

/**
 \brief The main function. The only function call needed to process
   	all command-line options and arguments nicely is argp_parse.
  \param[in] argc
  \param[in] argv
*/
int main(int argc, char **argv)
{
	struct lwe_ctrl_op_info pme_info;
	unsigned int part_id = 0;
	char label[10];
	FILE *fd;
	size_t count;

	fd = fopen("/proc/device-tree/label", "r");
	if (fd == NULL) {
		printf("\nFile /proc/device-tree/label could not be opened");
		goto _close;
	}
	count = fread(label, sizeof(char), sizeof(label), fd);
	/* count < sizeof(label) is a "correct" result so we can't use it to check for a possible
	 * error. Just use ferror() anyway
	 */
	if (ferror(fd)) {
		perror("fread");
		goto _close;
	}
	/* In case the label was longer then 9 characters, make sure we have a null terminated
	 * string
	 */
	label[count] = 0;
	fclose(fd);
	part_id = wb_get_id_from_part_name(label);

	wb_register(part_id, NULL, &wb_handle, (void *)(LWE_CTRL_SHM_ADDR));
	wb_acquire_lock(&wb_handle, &wb_msg);
	g_shm_pme_info = (struct lwe_ctrl_op_info *)wb_msg->data;

	memset(&pme_info, 0, sizeof(struct lwe_ctrl_op_info));

	/* State of Command */
	pme_info.state = LWE_CTRL_CMD_STATE_IDLE;

	/** Result - Successful, Failure */
	pme_info.result = LWE_CTRL_RSLT_FAILURE;
	pme_info.free_offset = g_shm_pme_info->free_offset;

	if (argc == 1) {
		printf("Mandatory Parameter missing\n");
		printf("Try `simple_pme --help' for more information\n");
		goto _close;
	}

	/* Where the magic happens */
	argp_parse(&argp, argc, argv, 0, 0, &pme_info);

	switch (WB_GET_MSG_TYPE(wb_msg)) {
	case LWE_CTRL_CMD_TYPE_PME_ADD:
		{
			lwe_pme_log_print
			    ("\nFILE: %s : LINE %d : IN MAIN : THE TYPE PROVIDED FOR ADD OPTION IS : %d",
			     __FILE__, __LINE__, WB_GET_MSG_TYPE(wb_msg));
			lwe_pme_add_del_command(argc - 1, &argv[1],
					(char *)WB_GET_MSG_TYPE(wb_msg));
		}
		break;

	case LWE_CTRL_CMD_TYPE_PME_DEL:
		{
			lwe_pme_log_print
			    ("\nFILE: %s : LINE %d : IN MAIN : THE TYPE PROVIDED FOR DELETE OPTION IS : %d",
			     __FILE__, __LINE__, WB_GET_MSG_TYPE(wb_msg));
			lwe_pme_add_del_command(argc - 1, &argv[1],
					(char *)WB_GET_MSG_TYPE(wb_msg));
		}
		break;

	case LWE_CTRL_CMD_TYPE_PME_GO:
		{
			lwe_pme_log_print
			    ("\nFILE: %s : LINE %d : IN MAIN : THE TYPE PROVIDED FOR GO OPTION IS : %d",
			     __FILE__, __LINE__, WB_GET_MSG_TYPE(wb_msg));
			lwe_pme_add_del_command(argc - 1, &argv[1],
					(char *)WB_GET_MSG_TYPE(wb_msg));

		}
		break;
	default:
		printf("Invalid Option\n");
	}

_close:
	wb_release_lock(&wb_handle, &wb_msg);
	wb_unregister(&wb_handle);

	return 0;
}
