/**
 \file simple_pme.c
 \brief Simple PME Application DP code
 */
/*
 * Copyright (C) 2009-2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <assert.h>
#include <compat.h>
#include <stdlib.h>
#include "init.h"
#include <app_common.h>
#include <app_cpdp.h>
#include "dpa_alloc.h"
#include <fsl_bman.h>
#include "dpa/dpa_bman.h"
#include "dpa/dpa_qman.h"
#include "net/frame_desc.h"
#include "pme/simple_pme_cpdp.h"
#include "pme/common_pme.h"
#include <common/cpu_freq.h>
#include <statistics.h>
#include "cache.h"
#include "barrier.h"
#include "sg_api.h"
#include "timebase.h"
#include "swi.h"
#include "wb/wb.h"

#define INV_LWE_ID        255
#define LWE_BOOT_CORE	0

#define MAX_BMAN_POOLS  8

#define MAX_SPA_PME_PKTS (4096 * 4)
#define ERR_DEL_UNINIT_FLOW 		1
#define ERR_FLOW_ALREADY_INITIALIZED 	2
#define ERR_BLOCKING_FLOW_ADDED 	3

uint32_t pme_test_lock;
uint8_t core_count = 0;
uint32_t pme_lock;
uint32_t g_total_flow = 0;
uint32_t g_first_flow = 1;
uint32_t g_block_scan_flow_id = 0;
uint32_t g_total_input_frames = 0;
uint8_t g_all_context_initialized = 0;
struct lwe_barrier pme_barrier;
bool pme_generate_report;
bool parse_report_enable;

struct spa_context_info_t spa_context_info[MAX_SPA_PME_CNTX];
struct scan_cb_func_t scan_cb_func = { NULL };

uint64_t total_buffer_size = 0;
static uint64_t atb_start;
static uint64_t atb_stop;
uint32_t timer_lwe_id;
bool start_timer_flag = false;

uint32_t block_scan = 0;

/* variable for block scan */
volatile uint32_t next;
static volatile uint64_t min_cycles = 0xffffffffffffffff;
static volatile uint64_t max_cycles;
static volatile uint64_t total_cycles;
static volatile uint64_t avg_cycles;

volatile uint32_t *g_context_init_done = NULL;
volatile uint32_t *g_total_scanned_response = NULL;

uint32_t g_boot_lwe_core = INV_LWE_ID;
volatile struct lwe_ctrl_op_info *lwe_ctrl_info;

__PERCPU volatile uint32_t qman_poll_success = 0;
__PERCPU volatile uint32_t underrun = 0;

__PERCPU uint32_t curr_lwe_id;
__PERCPU struct dpa_buff_allocator *buff_allocator;
uint32_t pool_channel_id;
int32_t pool_channel_num;

volatile bool shutting_down;
volatile bool done;
struct wb_handle wb_handle;
struct wb_msg *wb_msg;

int spa_scan_simple(const struct qm_fd *fd1, uint16_t seq);
int spa_scan_compound(const struct qm_fd *fd1, uint16_t seq);

void ctx_cb_ern(struct pme_ctx *ctx, const struct qm_mr_entry *msg,
			struct pme_ctx_token *token);
void ctx_cb_bpid_ern(struct pme_ctx *ctx, const struct qm_mr_entry *msg,
			struct pme_ctx_token *token);
int get_flow_id_frm_fqid(uint32_t fqid)
{
	uint32_t flow_id;

	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if ((spa_context_info[flow_id].valid) &&
			(spa_context_info[flow_id].fq_id == fqid))
			return flow_id;
	}
	return -1;
}

void release_loopback_fqs(void)
{
	uint32_t flow_id;
	uint32_t flags;
	int ret;
	enum qman_fq_state state;

	APP_DEBUG("%s : release loopback FQ resources", __func__);

	/* Send RETIRE Command to all loopback FQs */
	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if ((!spa_context_info[flow_id].valid) ||
			(unlikely(spa_context_info[flow_id].qman_fq == NULL)))
			continue;

		qman_fq_state(spa_context_info[flow_id].qman_fq,
							&state, &flags);
		spa_context_info[flow_id].fq_state = state;

		if ((state == qman_fq_state_parked) ||
				(state == qman_fq_state_sched)) {
			ret = qman_retire_fq(spa_context_info[flow_id].qman_fq,
							&flags);
			if (unlikely(ret < 0))
				APP_ERROR("Retire fails: fqid = %x error = %x",
					spa_context_info[flow_id].qman_fq->fqid,
					ret);

			if (ret == 1)
				APP_DEBUG("qman_retire_fq will happen "
							"asynchronously");

		}

	}

	/* Wait for All Loopback FQs to get retired and
	 * volatile deque all frames from FQ if not empty when retired */
	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if ((!spa_context_info[flow_id].valid) ||
			(unlikely(spa_context_info[flow_id].qman_fq == NULL)))
			continue;

		while (spa_context_info[flow_id].fq_state !=
						qman_fq_state_retired)
			qman_poll();

		if (spa_context_info[flow_id].retired_non_empty) {
			/* NB: the use of QMAN_VOLATILE_FLAG_WAIT is prohibited
			 * in LWE applications. To avoid thrashing, we implement
			 * a throttled poll for completion. */
			if (qman_volatile_dequeue(
					spa_context_info[flow_id].qman_fq, 0,
					(QM_VDQCR_PRECEDENCE_VDQCR |
					QM_VDQCR_NUMFRAMES_TILLEMPTY))) {
				/* TODO: the original code had no error handling
				 * here, so I won't pretend to know what's
				 * appropriate here. GT */
			}
loop:
			qman_fq_state(spa_context_info[flow_id].qman_fq, NULL,
					&flags);
			if (flags & QMAN_FQ_STATE_VDQCR) {
				uint64_t now, then = get_atb();
				do {
					now = get_atb();
				} while (now < (then + 1000));
				goto loop;
			}

			spa_context_info[flow_id].retired_non_empty =
								false;
		}
	}

	/* Let the DQRR get flushed (Flush the frames which are already
	* in DQRR when a FQ was retired).
	* Here the loop count = 100 is random but looks very much
	* sufficient as there can be atmost 16 entries in DQRR */
	for (int i = 0;  i < 100; i++)
		qman_poll();

	/* Now Turn the FQs OOS and then destroy */
	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if ((!spa_context_info[flow_id].valid) ||
			(unlikely(spa_context_info[flow_id].qman_fq == NULL)))
			continue;

		if (0 > qman_oos_fq(spa_context_info[flow_id].qman_fq))
			APP_ERROR("qman_oos_fq() fails");
	}

	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if ((!spa_context_info[flow_id].valid) ||
			(unlikely(spa_context_info[flow_id].qman_fq == NULL)))
			continue;

		qman_destroy_fq(spa_context_info[flow_id].qman_fq, 0);
		free(spa_context_info[flow_id].qman_fq);
	}

	APP_DEBUG("%s : All Loopback FQs released", __func__);
}

struct disable_token {
	struct pme_ctx_ctrl_token driver_token;
	int done;
};

static void disable_cb(struct pme_ctx *ctx, const struct qm_fd *fd,
			struct pme_ctx_ctrl_token *__token)
{
	struct disable_token *token = (void *)__token;
	token->done = 1;
}

static int blocking_pme_disable(struct pme_ctx *ctx)
{
	int ret;
	struct disable_token *token = malloc(sizeof(*token));
	if (!token)
		return -ENOMEM;
	memset(token, 0, sizeof(*token));
	token->driver_token.cb = disable_cb;
	/* NB: the use of PME_CTX_OP_WAIT is prohibited in LWE applications. */
	ret = pme_ctx_disable(ctx, 0, &token->driver_token);
	if (ret <= 0)
		return ret;
	while (!token->done)
		qman_poll();
	free(token);
	return 0;
}

void release_pme_resources(void)
{
	uint32_t flow_id;

	shutting_down = true;
	if (curr_lwe_id != g_boot_lwe_core) {
		while (done != true)
			qman_poll();

		return;
	}

	APP_DEBUG("%s : release PME resources", __func__);
	release_loopback_fqs();

	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if ((!spa_context_info[flow_id].valid) ||
			(spa_context_info[flow_id].spa_pme_ctx == NULL))
			continue;

		/* Disable and then destroy the PME context */
		if (blocking_pme_disable(spa_context_info[flow_id].spa_pme_ctx))
			APP_ERROR("pme_ctx_disable() fails ");

		pme_ctx_finish(spa_context_info[flow_id].spa_pme_ctx);
	}
	done = true;
	APP_DEBUG("%s : All SPA resources released", __func__);
}

static void update_flow_cb(struct pme_ctx *ctx, const struct qm_fd *fd,
				struct pme_ctx_ctrl_token *token)
{
	free(token);
}

/**
 \brief Creates, and Initializes PME Context in PME Driver
 \param[in] flow_id Flow Id associated with the Pme Context
 \param[in] spa_cfg_info Pointer to PME Configuration Info structure
 \return 0 if successfull, else -1
 */
int spa_pme_ctxt_init(uint32_t flow_id,
		      volatile struct lwe_ctrl_spa_cfg_info *spa_cfg_info)
{
	int ret;
	uint32_t flags = 0;
	struct pme_ctx *pme_ctx;
	struct spa_pme_ctxt_stats_t *pme_ctx_stats;
	struct pme_ctx_ctrl_token *token;
	struct pme_flow params;

	APP_DEBUG("spa_pme_ctxt_init() flow_id = %x", flow_id);
	token = malloc(sizeof(*token));
	pme_ctx = (struct pme_ctx *)malloc(sizeof(struct pme_ctx));
	if (unlikely(!pme_ctx || !token)) {
		APP_ERROR("malloc fails");
		free(token);
		free(pme_ctx);
		return -1;
	}
	token->cb = update_flow_cb;

	pme_ctx_stats =
	    (struct spa_pme_ctxt_stats_t *)stats_memalign(CACHE_LINE_SIZE,
					 sizeof(struct spa_pme_ctxt_stats_t));
	if (unlikely(pme_ctx_stats == NULL)) {
		APP_ERROR("stats_memalign fails");
		free(token);
		free(pme_ctx);
		return -1;
	}

	memset(pme_ctx_stats, 0, sizeof(struct spa_pme_ctxt_stats_t));
	memset(pme_ctx, 0, sizeof(struct pme_ctx));

	spa_context_info[flow_id].spa_pme_ctx = pme_ctx;
	spa_context_info[flow_id].pme_ctx_stats = pme_ctx_stats;

	/* Initialize the callback handler */
	if (spa_cfg_info->report_verbosity == REPORT_DISABLED) {
		pme_ctx->cb = scan_cb_func.scan_cb_simple;
		pme_ctx->ern_cb = ctx_cb_ern;
	} else if (spa_cfg_info->output_buff_frm_app) {
		pme_ctx->cb = scan_cb_func.scan_cb_cmpd_buff;
		pme_ctx->ern_cb = ctx_cb_ern;
	} else {
		pme_ctx->cb = scan_cb_func.scan_cb_cmpd_bpid;
		pme_ctx->ern_cb = ctx_cb_bpid_ern;
	}

	/*  Configure the PME mode */
	if (!(spa_cfg_info->flag & SET_SCAN_MODE_FLOW))
		flags |= PME_CTX_FLAG_DIRECT;

	/* If block scan than use Dedicated portal */
	if (block_scan)
		flags |= PME_CTX_FLAG_LOCAL;

	if (unlikely(pme_ctx_init(pme_ctx, flags,
				  spa_context_info[flow_id].bpid,
				  (uint8_t)spa_cfg_info->qos_in,
				  (uint8_t)spa_cfg_info->qos_out,
				  pool_channel_id, NULL))) {
		APP_ERROR("Error in pme_ctx_init");
		stats_free(pme_ctx_stats);
		free(token);
		free(pme_ctx);
		return -1;
	}

	/* enable the context */
	if (unlikely(0 != pme_ctx_enable(pme_ctx))) {
		APP_ERROR("Error in enabling pme context");
		stats_free(pme_ctx_stats);
		free(token);
		free(pme_ctx);
		return -1;
	}

	/* If direct mode than no need to update below params */
	if (!(spa_cfg_info->flag & SET_SCAN_MODE_FLOW)) {
		free(token);
		return 0;
	}

	/* initialize all params to ZERO */
	pme_sw_flow_init(&params);

	flags = 0;

	params.sos = 1;
	params.sessionid = spa_cfg_info->session_id;
	flags |= PME_CMD_FCW_SEQ;

	params.clim = (uint16_t)spa_cfg_info->compare_limit;
	params.mlim = (uint16_t)spa_cfg_info->match_limit;
	flags |= PME_CMD_FCW_DXE;
	if (spa_cfg_info->flag & SET_RESIDUE_ENABLE) {
		flags |= PME_CMD_FCW_RES;
		params.ren = 1;
		params.rlen = 126;
	}

	if (spa_cfg_info->report_verbosity == REPORT_DISABLED) {
		params.esee = 0;
		flags |= PME_CMD_FCW_SRE;
	} else {
		params.srvm = (uint8_t)spa_cfg_info->report_verbosity;
		params.esee = 1;
		flags |= PME_CMD_FCW_SRE;
	}

	ret = pme_ctx_ctrl_update_flow(spa_context_info[flow_id].spa_pme_ctx,
				       flags, &params, token);

	if (unlikely(ret != 0))
		APP_ERROR("Error in PME flow updation ");

	return ret;
}

/**
 \brief Print the statistics information for block scan request.

 \param[in] - None

 \return -  None
 */
void display_block_scan_stats(void)
{
	struct spa_pme_ctxt_stats_t *pme_ctx_stats;

	pme_ctx_stats = spa_context_info[g_block_scan_flow_id].pme_ctx_stats;
	APP_INFO("Total packets scanned = %d ",
		 pme_ctx_stats->simple_report.frames_scanned);
	APP_INFO("Total match found = %d ", pme_ctx_stats->
				simple_report.frames_with_atleast_one_pattern);
	APP_INFO("Total fail response = %d ", pme_ctx_stats->
				simple_report.frames_with_fail_response);
	APP_INFO("Minimum Cycles in scanning a packet = %lld", min_cycles);
	APP_INFO("Maximum Cycles in scanning a packet = %lld", max_cycles);
	APP_INFO("Total cycles taken in all packet = %lld", total_cycles);
	APP_INFO("Number of cycles/pkts = %lld", avg_cycles);
	APP_INFO("Test Passed");
}

void display_report_parse_scan_stats(void)
{
	struct spa_pme_ctxt_stats_t *stats;
	uint32_t frames_scanned = 0;
	uint32_t total_match_count = 0;
	uint32_t frames_with_atlest_one_match = 0;
	uint32_t fail_responses = 0;
	uint32_t truncated_reports = 0;
	uint32_t sre_rules_hit = 0;
	uint32_t flow_id;
	uint32_t atb_freq;
	uint64_t pps, bps;
	uint64_t atb_count = atb_stop - atb_start;
	uint8_t i, j;

	APP_INFO("**************** PER FLOW REPORT ***********************");
	APP_INFO("Flowid | Total   | Frames  | Trunca- | SRE    | Total  ");
	APP_INFO("       | Frames  | With    | ted     | RULE   | Matches");
	APP_INFO("       | Scanned | Pattern | Reports | Hit    | Found  ");

	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if (!spa_context_info[flow_id].valid)
			continue;

		stats = spa_context_info[flow_id].pme_ctx_stats;
		APP_INFO("%d %11d %8d %8d %22d",
			flow_id,
			stats->simple_report.frames_scanned,
			stats->simple_report.frames_with_atleast_one_pattern,
			stats->detail_report.truncated_report_count,
			stats->detail_report.patterns_count_in_all_frames);

		for (i = 0, j = 0; i < MAX_RULE_ID ; i++) {
			if (!(stats->detail_report.sre_rule_hit_count[i]))
				continue;
			if (!j++)
				APP_INFO("                                     "
								" ID - COUNT");
			APP_INFO(" %38d - %d", i,
				stats->detail_report.sre_rule_hit_count[i]);
			sre_rules_hit +=
				stats->detail_report.sre_rule_hit_count[i];
		}

		frames_scanned += stats->simple_report.frames_scanned;
		frames_with_atlest_one_match +=
			stats->simple_report.frames_with_atleast_one_pattern;
		total_match_count +=
			stats->detail_report.patterns_count_in_all_frames;
		truncated_reports +=
			stats->detail_report.truncated_report_count;
		fail_responses +=
				stats->simple_report.frames_with_fail_response;
	}

	APP_INFO("**************** SYSTEM REPORT ***************");
	APP_INFO("Total frames scanned = %d", frames_scanned);
	APP_INFO("Total Frames with at least one match found = %d",
				frames_with_atlest_one_match);
	APP_INFO("Total Match Count in all Frames = %d", total_match_count);
	APP_INFO("Total Truncated Reports = %d", truncated_reports);
	APP_INFO("Total Frames With Reports = %d",
			frames_with_atlest_one_match - truncated_reports);
	APP_INFO("Total SRE Rules hit = %d", sre_rules_hit);
	APP_INFO("Total fail response = %d", fail_responses);
	APP_INFO("Total  time taken  = %lld", atb_count);
	APP_INFO("Number of cycles/frame = %lld",
			 atb_count / frames_scanned);

	atb_freq  = get_cpu_freq();
	APP_INFO("ATB FREQ = %d", atb_freq);

	pps = (uint64_t)(atb_freq * frames_scanned) / atb_count;
	APP_INFO("Performance in PPS = %lld", pps);
	APP_INFO("total_buffer_size = %lld ", total_buffer_size);

	bps = (uint64_t)(atb_freq * total_buffer_size *
				BITS_PER_BYTE) / atb_count;
	APP_INFO("Performance in bps = %lld", bps);
	APP_INFO("Test Passed");
}

void display_noraml_scan_stats(void)
{
	struct spa_pme_ctxt_stats_t *stats;
	uint32_t frames_scanned = 0;
	uint32_t frames_with_atlest_one_match = 0;
	uint32_t fail_responses = 0;
	uint32_t truncated_reports = 0;
	uint32_t flow_id;
	uint32_t atb_freq;
	uint64_t pps, bps;
	uint64_t atb_count = atb_stop - atb_start;

	APP_INFO("**************** PER FLOW REPORT ******************");
	APP_INFO("Flowid | Totals  | Frames  | Truncated | Total     ");
	APP_INFO("       | Frames  | With    | Report    | Fail      ");
	APP_INFO("       | Scanned | Pattern |           | Responses ");

	for (flow_id = 0; flow_id < MAX_SPA_PME_CNTX; flow_id++) {
		if (!spa_context_info[flow_id].valid)
			continue;

		stats = spa_context_info[flow_id].pme_ctx_stats;
		APP_INFO("%d %11d %8d %8d %10d ", flow_id,
			stats->simple_report.frames_scanned,
			stats->simple_report.frames_with_atleast_one_pattern,
			stats->detail_report.truncated_report_count,
			stats->simple_report.frames_with_fail_response);

		frames_scanned += stats->simple_report.frames_scanned;
		frames_with_atlest_one_match +=
			stats->simple_report.frames_with_atleast_one_pattern;
		truncated_reports +=
			stats->detail_report.truncated_report_count;
		fail_responses +=
			stats->simple_report.frames_with_fail_response;
	}

	APP_INFO("**************** SYSTEM REPORT ***************");
	APP_INFO("Total frames scanned = %d", frames_scanned);
	APP_INFO("Total Frames with at least one match found = %d",
					frames_with_atlest_one_match);
	APP_INFO("Total Truncated Reports = %d", truncated_reports);
	APP_INFO("Total Frames With Reports = %d",
			frames_with_atlest_one_match - truncated_reports);
	APP_INFO("Total fail response = %d", fail_responses);
	APP_INFO("Total  time taken  = %lld", atb_count);
	APP_INFO("Number of cycles/frame = %lld", atb_count / frames_scanned);

	atb_freq  = get_cpu_freq();
	APP_INFO("ATB FREQ = %d", atb_freq);

	pps = (uint64_t)(atb_freq * frames_scanned) / atb_count;
	APP_INFO("Performance in PPS = %lld", pps);
	APP_INFO("total_buffer_size = %lld ", total_buffer_size);

	bps = (uint64_t)(atb_freq * total_buffer_size *
				BITS_PER_BYTE) / atb_count;
	APP_INFO("Performance in bps = %lld", bps);
	APP_INFO("Test Passed");
}

/**
 \brief This function used to update the minimum , maximum , total and average
 	cycles stats for block scan request type.

 \param[in] - cycles taken in last packet scan.

 \return -  None
 */
void block_scan_stats(uint64_t cycles)
{
	static uint32_t count = 0;

	count++;
	if (cycles < min_cycles)
		min_cycles = cycles;

	if (cycles > max_cycles)
		max_cycles = cycles;

	total_cycles += cycles;
	avg_cycles = total_cycles / count;
}

/**
 \brief Dump the context information.

 \param[in] None

 \return None
*/
void dump_context_info(void)
{
	int i = 0;
	for (i = 0; i < MAX_SPA_PME_CNTX; i++) {
		if (spa_context_info[i].valid) {
			APP_DEBUG(" [flowid = %x].number_of_frames = %x",
				  i, spa_context_info[i].number_of_frames);
			APP_DEBUG(" [flowid = %x].bpid = %x",
				  i, spa_context_info[i].bpid);
			APP_DEBUG(" [flowid = %x].fq_id  = %x",
				  i, spa_context_info[i].fq_id);
			APP_DEBUG(" [flowid = %x].qman_fq = %p",
				  i, spa_context_info[i].qman_fq);
			APP_DEBUG(" [flowid = %x].spa_pme_ctx = %p",
				  i, spa_context_info[i].spa_pme_ctx);
		}
	}
}

/**
 \brief Loopback FQ callback function for simple FD.
	If this fuction called after GO command from CP then it send
		the frames from loopback FQ to PME otherwise this function
		frees FD and frees SG buffer.

 \param[in] - struct qman_portal *qm : Qman portal pointer

 \param[in] - struct qman_fq *fq: Frame Queue pointer

 \param[in] - struct qm_dqrr_entry *dqrr: dequeue entry pointer

 \return - enum qman_cb_dqrr_result
 */
enum qman_cb_dqrr_result lb_fq_cb_dqrr_simple(struct qman_portal *qm,
					      struct qman_fq *fq,
					      const struct qm_dqrr_entry *dqrr)
{
	APP_DEBUG("lb_fq_cb_dqrr_simple");
	markpoint(10);

	qman_poll_success = 1;

	if (likely(g_all_context_initialized) &&
				likely(shutting_down == false)) {
again:
		if (0 > spa_scan_simple(&(dqrr->fd), dqrr->seqnum)) {
			APP_ERROR("spa_scan_simple fails");
			goto again;
		}
		return qman_cb_dqrr_consume;
	}

	/* Control reaches here only if the all contexts is to be deleted */
	/* FIX-ME free fd */
	return qman_cb_dqrr_consume;
}

/**
 \brief Loopback FQ callback function for compound FD.
	If this fuction called after GO command from CP then it send
		the frames from loopback FQ to PME otherwise this function
		frees FD and frees SG buffer.

 \param[in] - struct qman_portal *qm : Qman portal pointer

 \param[in] - struct qman_fq *fq: Frame Queue pointer

 \param[in] - struct qm_dqrr_entry *dqrr: dequeue entry pointer

 \return - enum qman_cb_dqrr_result
 */
enum qman_cb_dqrr_result lb_fq_cb_dqrr_compound(struct qman_portal *qm,
						struct qman_fq *fq,
						const struct qm_dqrr_entry
						*dqrr)
{
	APP_DEBUG("lb_fq_cb_dqrr_compound");
	markpoint(20);

	qman_poll_success = 1;

	if (likely(g_all_context_initialized) &&
				likely(shutting_down == false)) {
again:
		if (0 > spa_scan_compound(&(dqrr->fd), dqrr->seqnum)) {
			APP_ERROR("spa_scan_compound fails");
			goto again;
		}
		return qman_cb_dqrr_consume;
	}

	/* Control reaches here only if the context/flow are to be deletd */
	/* FIX-ME Free fd buffer */
	APP_DEBUG("Free resources as deletion of flow requested ");

	return qman_cb_dqrr_consume;
}

void ctx_cb_ern(struct pme_ctx *ctx, const struct qm_mr_entry *msg,
			struct pme_ctx_token *token)
{
	/* All rejected enque frames are freed. */
	/* FIX-ME free fd */
	APP_INFO("Rejected enqueued frames");
}

void ctx_cb_bpid_ern(struct pme_ctx *ctx, const struct qm_mr_entry *msg,
			struct pme_ctx_token *token)
{
	struct bm_buffer buf;
	struct sg_entry_t *sgentry = NULL;
	struct sg_entry_t *sgentry1 = NULL;
	phys_addr_t addr;
	struct qm_fd fd;

	memcpy(&fd, &(msg->ern.fd), sizeof(struct qm_fd));
	addr = fd.addr_hi;
	addr = (addr << 32) | fd.addr_lo;
	sgentry = (struct sg_entry_t *)ptov(addr);
	if (unlikely(sgentry->extension)) {
		addr = sgentry->addr_hi;
		addr = (addr << 32) | sgentry->addr_lo;
		sgentry1 = (struct sg_entry_t *)ptov(addr);
		if (likely(sgentry1)) {
			buf.bpid = sgentry1->bpid;
			buf.hi = sgentry1->addr_hi;
			buf.lo = sgentry1->addr_lo;
		}
	} else {
		buf.bpid = sgentry->bpid;
		buf.hi = sgentry->addr_hi;
		buf.lo = sgentry->addr_lo;
	}

	/* Free output buffer */
	dpa_allocator_free_buff(buff_allocator, &buf);
}
/**
 \brief ERN callback handler for Loopback FQs.

 \param[in] - struct qman_portal *qm : Qman portal pointer

 \param[in] - struct qman_fq *fq: Frame Queue pointer

 \param[in] - struct qm_mr_entry *dqrr: messare entry pointer

 \return - None
 */
void lb_fq_cb_mr_ern(struct qman_portal *qm, struct qman_fq *fq,
		 const struct qm_mr_entry *msg)
{
	/* All rejected enque frames are freed. */
	/* FIX-ME free fd */
	APP_INFO("Error in loopback FQs");
}

/**
 \brief ERN callback handler for Loopback FQs when output buffer
	is allocated from PME.

 \param[in] - struct qman_portal *qm : Qman portal pointer

 \param[in] - struct qman_fq *fq: Frame Queue pointer

 \param[in] - struct qm_mr_entry *dqrr: messare entry pointer

 \return - None
 */
void lb_fq_cb_mr_bpid_ern(struct qman_portal *qm, struct qman_fq *fq,
		 const struct qm_mr_entry *msg)
{
	struct bm_buffer buf;
	struct sg_entry_t *sgentry = NULL;
	struct sg_entry_t *sgentry1 = NULL;
	phys_addr_t addr;
	struct qm_fd fd;

	memcpy(&fd, &(msg->ern.fd), sizeof(struct qm_fd));
	addr = fd.addr_hi;
	addr = (addr << 32) | fd.addr_lo;
	sgentry = (struct sg_entry_t *)ptov(addr);
	if (unlikely(sgentry->extension)) {
		addr = sgentry->addr_hi;
		addr = (addr << 32) | sgentry->addr_lo;
		sgentry1 = (struct sg_entry_t *)ptov(addr);
		if (likely(sgentry1)) {
			buf.bpid = sgentry1->bpid;
			buf.hi = sgentry1->addr_hi;
			buf.lo = sgentry1->addr_lo;
		}
	} else {
		buf.bpid = sgentry->bpid;
		buf.hi = sgentry->addr_hi;
		buf.lo = sgentry->addr_lo;
	}

	/* Free output buffer */
	dpa_allocator_free_buff(buff_allocator, &buf);
}
/**
 \brief State change callback handler for Loopback FQs.

 \param[in] - struct qman_portal *qm : Qman portal pointer

 \param[in] - struct qman_fq *fq: Frame Queue pointer

 \param[in] - struct qm_mr_entry *dqrr: messare entry pointer

 \return - None
 */
void lb_fq_cb_mr_state_change_handler(struct qman_portal *qm,
		struct qman_fq *fq, const struct qm_mr_entry *msg)
{
	int flow_id;

	flow_id = get_flow_id_frm_fqid(fq->fqid);

	if (unlikely(0 > flow_id)) {
		APP_ERROR("Flow id not found for FQ ID = %d", fq->fqid);
		return;
	}

	uint8_t verb = msg->verb;

	switch (verb & QM_DQRR_VERB_MASK) {
	case QM_MR_VERB_FQRN:
	case QM_MR_VERB_FQRNI:
		if (msg->fq.fqs & QM_MR_FQS_ORLPRESENT) {
			/* There are pending rejected enque on this FQ.
			* Also Handle ERN messages. So wait for FQRL. */
			return;
		}
		if (msg->fq.fqs & QM_MR_FQS_NOTEMPTY) {
			/* Need to put in non-empty list */
			spa_context_info[flow_id].retired_non_empty = true;
		}
		break;

	case QM_MR_VERB_FQRL:
		break;

	case QM_MR_VERB_FQPN:
		APP_INFO("Frame Queue Parked Notification message");
	default:
		APP_ERROR("Unknown state change notification verb = %x", verb);
		return;
	}

	spa_context_info[flow_id].fq_state = qman_fq_state_retired;
	return;
}

/**
 \brief This function handles the context delete command from control core.
	    Dequeue all frames from the loopback FQ and destroy the loopback FQ
	    Disable and free the pme context.

 \param[in] - None

 \return  0 if successful, otherwise -ve
 */
int spa_del_context(void)
{
	volatile struct lwe_ctrl_spa_cfg_info *spa_cfg_info =
	    &lwe_ctrl_info->app_info.pme_info;
	uint32_t flow_id, flags;

	flow_id = spa_cfg_info->flow_id;

	if (unlikely(spa_context_info[flow_id].valid == false)) {
		APP_INFO("Trying to delete Uninitialized PME Context");
		WB_SET_RSLT(wb_handle, ERR_DEL_UNINIT_FLOW);
		return 0;
	}

	if (unlikely(spa_context_info[flow_id].qman_fq == NULL)) {
		APP_INFO("Trying to delete Uninitialized FQ");
		WB_SET_RSLT(wb_handle, ERR_DEL_UNINIT_FLOW);
		return 0;
	}

	APP_DEBUG("calling qman_volatile_dequeue ... ");
	/* NB: the use of QMAN_VOLATILE_FLAG_WAIT is prohibited in LWE
	 * applications. To avoid thrashing, we implement a throttled poll for
	 * completion. */
	if (qman_volatile_dequeue(spa_context_info[flow_id].qman_fq, 0,
				(QM_VDQCR_PRECEDENCE_VDQCR |
				QM_VDQCR_NUMFRAMES_TILLEMPTY))) {
		/* TODO: the original code had no error handling here, so I
		 * won't pretend to know what's appropriate here. GT */
	}
loop:
	qman_fq_state(spa_context_info[flow_id].qman_fq, NULL, &flags);
	if (flags & QMAN_FQ_STATE_VDQCR) {
		uint64_t now, then = get_atb();
		do {
			now = get_atb();
		} while (now < (then + 1000));
		goto loop;
	}

	APP_DEBUG("calling qman_destroy_fq ... ");
	qman_destroy_fq(spa_context_info[flow_id].qman_fq,
			QMAN_FQ_DESTROY_PARKED);

	free(spa_context_info[flow_id].qman_fq);

	if (unlikely(spa_context_info[flow_id].spa_pme_ctx == NULL)) {
		APP_ERROR("Trying to delete Uninitialized PME Context");
		WB_SET_RSLT(wb_handle, ERR_DEL_UNINIT_FLOW);
		return -1;
	}

#if (LOG_LEVEL > 3)
	struct pme_ctx *pme_ctx;

	/* Disable and then destroy the PME context */
	pme_ctx = spa_context_info[flow_id].spa_pme_ctx;

	APP_DEBUG("calling pme_ctx_disable flow_id = %x, spa_pme_ctx = %p",
		  flow_id, spa_context_info[flow_id].spa_pme_ctx);
	APP_DEBUG("pme_ctx->flags = %x ", pme_ctx->flags);
	APP_DEBUG("pme_ctx->fq.fqid = %x ", pme_ctx->fq.fqid);
#endif

	if (unlikely(blocking_pme_disable(spa_context_info[flow_id].spa_pme_ctx))) {
		APP_ERROR("ERROR:  PME flow disable failing ");
		return -1;
	}

	APP_DEBUG(" calling pme_ctx_finish ... ");
	pme_ctx_finish(spa_context_info[flow_id].spa_pme_ctx);

	spa_context_info[flow_id].valid = false;
	g_total_flow--;
	g_total_input_frames -= spa_context_info[flow_id].number_of_frames;

	if (!g_total_flow) {
		g_first_flow = 1;
		block_scan = 0;
	}
	WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_SUCCESSFULL);
	return 0;
}

/**
 \brief Schedules the Loopback FQ of all the context added.

 \param[in]  None

 \return  0 if successful, otherwise -ve
 */
int simple_pme_schd_lp_fq(void)
{
	int i = 0;
	for (i = 0; i < MAX_SPA_PME_CNTX; i++) {
		if (likely(spa_context_info[i].valid)) {
			if (unlikely
			    (0 >
			     qman_schedule_fq(spa_context_info[i].qman_fq))) {
				APP_ERROR(" fman_schedule_fq return error");
				return -1;
			}
		}
	}
	return 0;
}

/**
 \brief Allocate and initialize scatter-gather buffer.

 \param[in] - *buff :  Source buffer pointer.
 		If null then it means no input buffer to initialize the
 		frame but still it will create the buffer frame.

 \param[in] - csize :  buffer chunk size.

 \param[in] - bsize :  buffer frame size. This bsize is divided into multiple
			csize for creating the compund buffer.

 \param[in] - offset in frame where buffer will fit in.

 \param[out] - struct bm_buffer *bman_buf_sgt :  This function return the
			SG buffer information in this structure.

 \return - 0 on success, -ve on Failure.
 */
int create_sg_frame(uint8_t * buff, uint32_t csize, uint32_t bsize,
		    uint32_t offset, uint32_t **frame_add)
{
	struct sg_entry_t *sgt = NULL;
	uint8_t *src_addr;
	uint8_t *dest_addr;
	uint32_t num_sg_entry;
	phys_addr_t addr;
	uint32_t sg_entry, j, num;
	uint32_t size;
	uint32_t last_sgt = 0;
	uint32_t *buff_add, *sgt_add;
	phys_addr_t addr1;

	APP_DEBUG("create_sg_frame(&buff= %p, csize = %x, bsize = %x "
		 " offset = %x", buff, csize, bsize, offset);
	num_sg_entry = bsize / csize;
	if (bsize % csize)
		num_sg_entry++;

	APP_DEBUG("num_sg_entry = %x", num_sg_entry);

	/* input_buffer_array is char */
	src_addr = (uint8_t *) buff;

	for (sg_entry = 0; sg_entry < num_sg_entry;) {
		if ((num_sg_entry - sg_entry) <= 16) {
			num = num_sg_entry - sg_entry;
			last_sgt = 1;
		} else {
			num = 16;
		}

		APP_DEBUG("sg_entry = %x", sg_entry);
		APP_DEBUG("num = %x", num);
		APP_DEBUG("last_sgt = %x", last_sgt);

		sgt_add = malloc((num * sizeof(struct sg_entry_t)) + offset);
		if (unlikely(sgt_add == NULL)) {
			APP_ERROR("Create sgt: No memory");
			WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
			return -1;
		}
		memset(sgt_add, 0,
			(num * sizeof(struct sg_entry_t)) + offset);
		addr = vtop(sgt_add);
		if (sgt) {
			APP_DEBUG("Init SGT entry");
			sgt->bpid = 0;
			sgt->addr_hi = (uint16_t) (addr >> 32);
			sgt->addr_lo = (uint32_t) addr;
			sgt->length = sizeof(struct sg_entry_t);
			sgt->offset = 0;
			sgt->reserved_offset = 0;
			sgt->reserved_zero = 0;
			sgt->reserved_zero2 = 0;
			sgt->extension = 1;
			sgt->final = 0;
		}
		if (sg_entry == 0)
			*frame_add = sgt_add;

		sgt = (struct sg_entry_t *)((char *)sgt_add + offset);
		APP_DEBUG("sgt = %p ", sgt);

		for (j = 0; j < num; j++) {
			if (!last_sgt && j == (num - 1)) {
				APP_DEBUG
				    (" breaking as this will be extention SGT entry ");
				break;
			}

			if (unlikely((sg_entry + j) == (num_sg_entry - 1))) {
				size = bsize - csize * (sg_entry + j);
			} else {
				size = csize;
			}

			APP_DEBUG("size = %x", size);
			buff_add = malloc(size);
			if (unlikely(buff_add == NULL)) {
				APP_ERROR("Create simple buff: No memory");
				WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
				return -1;
			}
			memset(buff_add, 0, size);

			if (unlikely(buff != NULL)) {
				dest_addr = (uint8_t *)buff_add;
				APP_DEBUG(" src_addr = %p", src_addr);
				APP_DEBUG(" dest_addr = %p", dest_addr);
				APP_DEBUG(" size = %x", size);
				memcpy(dest_addr, src_addr, size);
				src_addr += size;
			}

			addr1 = vtop(buff_add);
			sgt->bpid = 0;
			sgt->addr_hi = (uint16_t) (addr1 >> 32);;
			sgt->addr_lo = (uint32_t) addr1;
			sgt->length = size;
			sgt->offset = 0;
			sgt->reserved_offset = 0;
			sgt->reserved_zero = 0;
			sgt->reserved_zero2 = 0;
			sgt->extension = 0;
			sgt->final = 0;
			sgt++;
		}

		if (!last_sgt)
			sg_entry += (num - 1);
		else
			sg_entry += num;

		offset = 0;
	}

	sgt--;
	sgt->extension = 0;
	sgt->final = 1;
	APP_DEBUG("SGT init Done");
	return 0;
}

/**
 \brief Allocate and initialize simple/scatter-gather buffer.

 \param[in] - *buff :  Source buffer pointer.
 		If null then it means no input buffer to initialize the
 		frame but still it will create the buffer frame.

 \param[in] - csize :  buffer chunk size.

 \param[in] - bsize :  buffer frame size. This bsize is divided into multiple
			csize for creating the compund buffer.

 \param[in] - offset in frame where buffer will fit in.

 \param[out] - struct frame_info_t *frame:  This function return the buffer
			frame information in this structure.

 \return - 0 on success, -ve on Failure.
 */
int create_buffer_frame(uint8_t * buff, uint32_t csize, uint32_t bsize,
			uint32_t offset, struct frame_info_t *frame)
{
	uint32_t bpid;
	uint32_t bpsize;
	uint32_t size;
	uint8_t *dest_addr;
	uint32_t *frame_addr = NULL;

	APP_DEBUG("create_buffer_frame(&buff= %p, csize = %x, bsize = %x "
		 " offset = %x", buff, csize, bsize, offset);
	if (!bsize) {
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
		return -1;
	}

	if (csize > bsize) {
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAIL_CSIZE_GT_BUFFSIZE);
		return -1;
	}

	if (!csize)
		csize = bsize;

	if (0 > get_bpid_bpsize(csize, &bpid, &bpsize)) {
		APP_ERROR(" Bpools not initialised ");
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
		return -1;
	}

	if (bpsize < csize)
		csize = bpsize;

	size = bsize + offset;
	if (0 > get_bpid_bpsize(size, &bpid, &bpsize)) {
		APP_ERROR(" Bpools not initialised ");
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
		return -1;
	}

	if (csize != bsize || bpsize < size) {
		APP_DEBUG(" Compound buffer frame ");
		/* Compound buffer frame */
		if (0 > create_sg_frame(buff, csize, bsize,
					offset, &frame_addr)) {
			APP_ERROR("SG buffer initialization failed");
			return -1;
		}
		frame->buff = frame_addr;
		frame->format = 1;
		frame->bpid = 0;
	} else {
		APP_DEBUG(" Simple buffer frame ");
		/* Simple buffer frame */
		frame->buff = malloc(bsize + offset);
		if (unlikely(frame->buff == NULL)) {
			APP_ERROR("Create simple buff: No memory");
			WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
			return -1;
		}
		memset(frame->buff, 0, bsize + offset);
		if (buff != NULL) {
			dest_addr = (uint8_t *)frame->buff + offset;
			APP_DEBUG(" src_addr = %p", buff);
			APP_DEBUG(" dest_addr = %p", dest_addr);
			APP_DEBUG(" size = %x", bsize);
			memcpy(dest_addr, buff, bsize);
		}

		frame->format = 0;
		frame->bpid = 0;
	}
	return 0;
}

/**
 \brief This function handles the Context add request from Control plane.
	    Initialize the loopback FQs and enqueue the buffer in loopback FQs.
	    Create and initialize the PME context.

 \param[in] - None

 \return -	0  on success and more context can be added,
		-1 on failure
		1  on sucess and no more context can be added as the upper limit
		   of number of context or number of frames reached
 */
int spa_add_context(void)
{
	volatile struct lwe_ctrl_spa_cfg_info *spa_cfg_info =
	    &lwe_ctrl_info->app_info.pme_info;
	struct sg_entry_t *sg = NULL;
	uint32_t flow_id;
	uint32_t number_of_frames;
	struct qm_fd fd;
	struct spa_pme_ctx_token_t *spa_pme_frm_token = NULL;
	int i = 0;
	uint32_t offset;
	uint32_t args = 0;
	struct qman_fq *fq = NULL;
	uint32_t bpid, bpsize;
	uint32_t in_bsize;
	uint32_t in_csize;
	uint32_t report_size = 0;
	uint8_t *src_addr;
	struct frame_info_t inframe;
	struct frame_info_t outframe;
	uint32_t flags = 0;
	phys_addr_t addr;

	/* Copy the context information from whiteboard */
	flow_id = spa_cfg_info->flow_id;

	/* If spa_cfg_info->input_buff_size is zero, then make it equal to file
	 * size or max size of all availavle bpools (whichever in minimum)
	 * */
	if (!(spa_cfg_info->input_buff_size))
		spa_cfg_info->input_buff_size = spa_cfg_info->size_of_file;

	APP_DEBUG("File size = %x ", spa_cfg_info->size_of_file);
	APP_DEBUG("Input_buff_size = %x ", spa_cfg_info->input_buff_size);

	total_buffer_size += spa_cfg_info->size_of_file;

	number_of_frames = spa_cfg_info->size_of_file /
	    spa_cfg_info->input_buff_size;
	if (spa_cfg_info->size_of_file % spa_cfg_info->input_buff_size)
		number_of_frames++;

	spa_context_info[flow_id].number_of_frames = number_of_frames;
	APP_DEBUG("Number_of_frames = %x ", number_of_frames);

	/* There can be only one block scan context in system */
	if (unlikely(g_first_flow)) {
		APP_INFO("Configuration of PME flows started ...");
		g_first_flow = 0;
		if (spa_cfg_info->flag & SET_SCAN_FN_TYPE_BLOCK) {
			APP_INFO("Blocking flow ");
			block_scan = 1;
			g_block_scan_flow_id = flow_id;
		}
	}

	if (spa_cfg_info->report_verbosity != REPORT_DISABLED) {
		APP_DEBUG("Report enabled");
		if (!(spa_cfg_info->flag & SET_SCAN_MODE_FLOW))
			spa_cfg_info->match_limit = 65535;

		/* Get output buffer size from verbosity and mlimit */
		report_size = get_report_size(spa_cfg_info->report_verbosity,
					      spa_cfg_info->match_limit);

		/* Get Buffer Pool ID for output buffer */
		if (0 > get_bpid_bpsize(report_size, &bpid, &bpsize)) {
			APP_ERROR("Bpools not initialised ");
			WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
			return -1;
		}

		spa_context_info[flow_id].bpid = bpid;

		if (bpsize < report_size)
			report_size = bpsize;

		pme_generate_report = true;
		APP_DEBUG("Report_size = %x", report_size);
	}

	/* Initialize FQ */
	fq = (struct qman_fq *)malloc(sizeof(struct qman_fq));
	if (!fq) {
		APP_ERROR(" malloc fails ");
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
		return -1;
	}

	if (spa_cfg_info->report_verbosity == REPORT_DISABLED) {
		fq->cb.dqrr = lb_fq_cb_dqrr_simple;
		fq->cb.ern = lb_fq_cb_mr_ern;
	} else if (spa_cfg_info->output_buff_frm_app) {
		fq->cb.dqrr = lb_fq_cb_dqrr_compound;
		fq->cb.ern = lb_fq_cb_mr_ern;
	} else {
		fq->cb.dqrr = lb_fq_cb_dqrr_compound;
		fq->cb.ern = lb_fq_cb_mr_bpid_ern;
	}

	fq->cb.dc_ern = NULL; /* not needed as we do not have any DC
				in Simple PME application. */
	fq->cb.fqs = lb_fq_cb_mr_state_change_handler;

	if (qman_create_fq(0, QMAN_FQ_FLAG_LOCKED |
			   QMAN_FQ_FLAG_DYNAMIC_FQID, fq) < 0) {
		APP_ERROR(" qman_create_fq fails");
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
		return -1;
	}

	struct qm_mcc_initfq opts;
	memset(&opts, 0, sizeof(struct qm_mcc_initfq));

	opts.fqd.dest.channel = (uint16_t)pool_channel_id;
	opts.we_mask = QM_INITFQ_WE_ORPC |
			QM_INITFQ_WE_FQCTRL | QM_INITFQ_WE_DESTWQ;
	opts.fqd.dest.wq = 1;
	opts.fqd.orprws = 1;
	opts.fqd.oa = 0;
	opts.fqd.olws = 0;
	opts.fqd.fq_ctrl = QM_FQCTRL_ORP;

	/* If block scan than use Dedicated portal */
	if (block_scan)
		flags = QMAN_INITFQ_FLAG_LOCAL;

	if (qman_init_fq(fq, flags, &opts) < 0) {
		APP_ERROR(" qman_init_fq fails");
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
		return -1;
	}

	spa_context_info[flow_id].fq_id = fq->fqid;
	spa_context_info[flow_id].qman_fq = fq;
	/* Copy data from whiteboard to bpool buffer and enqueue
	 * the buffer FQs
	 */
	for (i = 0; i < number_of_frames; i++) {
		/*  Allocate space in bpool for input buffer. This is the buffer
		 *  in which the data coming from whiteboard will be copied for
		 *  sent for scan */
		memset(&inframe, 0, sizeof(struct frame_info_t));
		memset(&outframe, 0, sizeof(struct frame_info_t));

		src_addr = (uint8_t *) spa_cfg_info->file +
		    spa_cfg_info->input_buff_size * i;

		if (unlikely(i == number_of_frames - 1)) {
			in_bsize =
			    spa_cfg_info->size_of_file -
			    (spa_cfg_info->input_buff_size * i);
			if (in_bsize < spa_cfg_info->input_buff_chunk)
				in_csize = in_bsize;
			else
				in_csize = spa_cfg_info->input_buff_chunk;
		} else {
			in_bsize = spa_cfg_info->input_buff_size;
			in_csize = spa_cfg_info->input_buff_chunk;
		}

		offset = sizeof(struct spa_pme_ctx_token_t);
		if (unlikely(create_buffer_frame(src_addr, in_csize, in_bsize,
						 offset, &inframe) < 0)) {
			APP_ERROR(" Frame creation failed ");
			WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
			return -1;
		}

		if (unlikely(inframe.buff == NULL)) {
			APP_ERROR(" Frame creation failed ");
			WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
			return -1;
		}

		spa_pme_frm_token = (struct spa_pme_ctx_token_t *)inframe.buff;
		spa_pme_frm_token->flow_id = flow_id;
		spa_pme_frm_token->ip_ctx = NULL;
		spa_pme_frm_token->notes = NULL;

		APP_DEBUG(" IN BUFF spa_pme_frm_token = %p ",
			  spa_pme_frm_token);
		APP_DEBUG(" IN BUFF spa_pme_frm_token->flow_id = %x ",
			  spa_pme_frm_token->flow_id);
		APP_DEBUG(" spa_cfg_info->set = %x ", spa_cfg_info->set);
		APP_DEBUG(" spa_cfg_info->subset = %x ", spa_cfg_info->subset);
		flags = 0;
		if ((spa_cfg_info->flag & SET_SCAN_MODE_FLOW)) {
			if (unlikely(i == 0)) {
				if (unlikely(i == (number_of_frames - 1))) {
					flags |=
					    PME_CMD_SCAN_SR | PME_CMD_SCAN_E;
				} else {
					flags |= PME_CMD_SCAN_SR;
				}
			} else if (unlikely(i == (number_of_frames - 1))) {
				flags |= PME_CMD_SCAN_E;
			}
		} else {
			flags = PME_CMD_SCAN_SR | PME_CMD_SCAN_E;
		}

		if (spa_cfg_info->report_verbosity != REPORT_DISABLED)
			flags |=
			    PME_CMD_SCAN_SRVM(spa_cfg_info->report_verbosity);

		args =
		    PME_SCAN_ARGS(flags, spa_cfg_info->set,
				  spa_cfg_info->subset);
		spa_pme_frm_token->args = args;
		APP_DEBUG(" spa_pme_frm_token->args = %x", args);

		/* Initialize FD to zero */
		memset(&fd, 0, sizeof(struct qm_fd));

		/* Simple FD, When no report needed */
		if (spa_cfg_info->report_verbosity == REPORT_DISABLED) {
			addr = vtop(inframe.buff);
			fd.addr_hi = (uint8_t) (addr >> 32);
			fd.addr_lo = (uint32_t) addr;
			if (inframe.format)
				fd.format = qm_fd_sg;
			else
				fd.format = qm_fd_contig;

			fd.offset = (uint8_t)offset;
			fd.length20 = in_bsize;
			fd.bpid = (uint8_t)inframe.bpid;
		} else {
			/* Acquire buffer for compound FD */
			sg = (struct sg_entry_t *)malloc(2 *
				 sizeof(struct sg_entry_t));
			if (unlikely(sg == NULL)) {
				APP_ERROR("Create sg: No memory");
				WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
				free(inframe.buff);
				return -1;
			}

			/* Init SG to zero */
			memset(sg, 0, 2 *
			       sizeof(struct sg_entry_t));

			if (spa_cfg_info->output_buff_frm_app) {
				if (create_buffer_frame(NULL, 0, report_size, 0,
							&outframe) < 0) {
					APP_ERROR(" Frame creation failed ");
					WB_SET_RSLT(wb_handle,
					    LWE_CTRL_RSLT_FAILURE);
					free(inframe.buff);
					free(sg);
					return -1;
				}

				if (outframe.buff == NULL) {
					APP_ERROR(" Frame creation failed ");
					WB_SET_RSLT(wb_handle,
					    LWE_CTRL_RSLT_FAILURE);
					free(inframe.buff);
					free(sg);
					return -1;
				}

				/* Initialized O/P buffer info */
				sg->bpid = (uint8_t)outframe.bpid;
				spa_context_info[flow_id].bpid = outframe.bpid;

				addr = vtop(outframe.buff);
				sg->addr_hi = (uint8_t) (addr >> 32);
				sg->addr_lo = (uint32_t) addr;
				sg->length = report_size;
			} else {
				sg->length = 0;
				sg->addr_lo = NULL;
				sg->bpid = (uint8_t)bpid;
			}

			/* input buffer */

			sg++;
			addr = vtop(inframe.buff);
			sg->addr_hi = (uint8_t) (addr >> 32);
			sg->addr_lo = (uint32_t) addr;
			sg->length = in_bsize;
			sg->offset = offset;
			sg->bpid = (uint8_t)inframe.bpid;
			sg->final = 1;
			if (inframe.format)
				sg->extension = 1;
			sg--;

			addr = vtop(sg);
			fd.addr_hi = (uint8_t) (addr >> 32);
			fd.addr_lo = (uint32_t) addr;
			fd._format2 = qm_fd_compound;
			fd.cong_weight = 0;
			fd.bpid = 0;
			fd.cmd = 0;
			fd.status = 0;
		}

		if (qman_enqueue(spa_context_info[flow_id].qman_fq, &fd, 0)) {
			APP_ERROR("Unable to Enqueue Initial Frame");
			/* FIX-ME free fd */
			WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
			return -1;
		}
	}
	if (0 > spa_pme_ctxt_init(flow_id, spa_cfg_info)) {
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAILURE);
		return -1;
	}

	spa_context_info[flow_id].valid = true;

	g_total_flow++;
	g_total_input_frames += spa_context_info[flow_id].number_of_frames;

	APP_DEBUG("g_total_input_frames = %x", g_total_input_frames);
	APP_DEBUG("g_total_flow = %x", g_total_flow);

	WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_SUCCESSFULL);
	return 0;
}

void parse_report(uint8_t *report, uint32_t length,
				struct spa_pme_ctxt_stats_t *stats)
{
	uint32_t offset = 0;
	enum pme_report_type type;
	struct pme_report_rule_record_t *rule_record_p = NULL;

	APP_DEBUG("length = %d ", length);
	while (offset <  length) {
		type = report[offset];

		switch (type) {
		case pme_end_of_report_type:
			APP_DEBUG("END of report");
			return;

		case pme_rule_report_type:
			APP_DEBUG("type = %d (pme_rule_report_type)", type);
			rule_record_p = (struct pme_report_rule_record_t *)
							(report + offset);
			APP_DEBUG("Rule id %x", rule_record_p->rule_id);
			decorated_notify_inc_32(&(stats->detail_report.
				sre_rule_hit_count[rule_record_p->rule_id &
				SRE_RULE_ID_MASK]));
			offset += rule_record_p->length;

			APP_DEBUG("report length = %d", rule_record_p->length);
			break;

		case pme_simple_match_type:
			APP_DEBUG("type = %d (pme_simple_match_type)", type);
			APP_DEBUG("report length = %d", sizeof(
				struct pme_report_simple_match_record_t));
			offset += sizeof(
				struct pme_report_simple_match_record_t);
			break;

		case pme_verbose_match_type:
			APP_DEBUG("type = %d (pme_verbose_match_type)", type);
			APP_DEBUG("report length = %d", sizeof(
				struct pme_report_verbose_match_record_t));
			offset += sizeof(
				struct pme_report_verbose_match_record_t);
			break;

		case pme_stateless_rule:
			APP_DEBUG("type = %d (pme_stateless_rule)", type);
			APP_DEBUG("report length = %d", sizeof(
				struct pme_report_stateless_match_record_t));
			offset += sizeof(
				struct pme_report_stateless_match_record_t);
			break;

		case pme_dual_state_rule_no_context:
		case pme_dual_state_rule_with_context:
		case pme_multi_state_rule:
			APP_DEBUG("type = %d (report_verbose2_match_record)",
									type);
			APP_DEBUG("Report length = %d", sizeof(
				struct pme_report_verbose2_match_record_t));
			offset += sizeof(
				struct pme_report_verbose2_match_record_t);
			break;

		case pme_dual_state_rule_with_context_verbose:
		case pme_multi_state_rule_verbose:
			APP_DEBUG("type = %d (report_verbose3_match_record)",
									type);
			APP_DEBUG("Report length = %d", sizeof(
				struct pme_report_verbose3_match_record_t));
			offset += sizeof(
				struct pme_report_verbose3_match_record_t);
			break;

		default:
			APP_ERROR("Unsupported type of report");
			APP_INFO("type = %d", type);
			APP_INFO("length = %d", length);
			hexdump(&report[offset], 4);
			while (1)
			;
		}
	}
}

/**
 \brief PME context control command parser. This function receives the context
	creation, deletion and run command from control plane.

 \param[in] - None

 \return - None
 */

static int check_to_add_context(void)
{
	volatile struct lwe_ctrl_spa_cfg_info *spa_cfg_info =
	    &lwe_ctrl_info->app_info.pme_info;
	uint32_t flow_id;

	/* Only one block_scan context can be added */
	if (block_scan) {
		APP_INFO(" 1st Flow is block_scan, "
					 " so no more flow added");
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAIL_BLOCKING_FLOW_ADDED);
		return -1;
	}

	if (g_total_flow == MAX_SPA_PME_CNTX) {
		APP_ERROR("MAX Number of Context Limit in System is Already"
			" Reached. so No more context can be added");
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAIL_MAX_FLOW_LIMIT);
		return -1;
	}

	if (g_total_input_frames >= MAX_SPA_PME_PKTS) {
		APP_ERROR("MAX Packets Limit Already Reached. "
				"So no more context can be added");
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_FAIL_MAX_FRAME_LIMIT);
		return -1;
	}

	flow_id = spa_cfg_info->flow_id;

	if (unlikely(spa_context_info[flow_id].valid)) {
		APP_ERROR("Flow id %d is already initialized", flow_id);
		APP_ERROR("To re-initialize , first delete the Flow");
		WB_SET_RSLT(wb_handle, ERR_FLOW_ALREADY_INITIALIZED);
		return -1;
	}
	return 0;
}

uint32_t spa_context_setup(struct wb_handle *handle, struct wb_msg *msg)
{
	void *p_data;

	lwe_ctrl_info->free_offset =
	    sizeof(*lwe_ctrl_info) + (uint8_t *) lwe_ctrl_info;

	volatile struct lwe_ctrl_spa_cfg_info *spa_cfg_info =
	    &lwe_ctrl_info->app_info.pme_info;

	if (unlikely(curr_lwe_id == LWE_BOOT_CORE)) {
		if (unlikely(SWI_ID_SHUTDOWN ==
				swi_activate(&p_data))) {
			APP_INFO("Shutdown request received");
			shutting_down = true;
		}
	}
	if (unlikely(shutting_down == true)) {
		if (!block_scan) {
			if (simple_pme_schd_lp_fq() < 0) {
				APP_ERROR("FQ not scheduled");
				assert(0);
			}
		}
		return 0;
	}

	switch (WB_GET_MSG_TYPE(wb_msg)) {
	case LWE_CTRL_CMD_TYPE_PME_ADD:
		APP_DEBUG(" LWE_CTRL_CMD_TYPE_PME_ADD  ");
		if (check_to_add_context() < 0)
			break;

		if (unlikely(spa_add_context() < 0))
			APP_ERROR("Error In Context Addition");
		break;

	case LWE_CTRL_CMD_TYPE_PME_DEL:
		APP_DEBUG(" LWE_CTRL_CMD_TYPE_PME_DEL  ");
		/* Delete a Simple PME context */
		if (unlikely(spa_del_context() < 0))
			assert(0);
		break;

	case LWE_CTRL_CMD_TYPE_PME_GO:
		APP_DEBUG(" LWE_CTRL_CMD_TYPE_PME_GO  ");
		APP_DEBUG("spa_cfg_info-> Report Parse _enable = %d",
			spa_cfg_info->report_parse_enable);
		if (spa_cfg_info->report_parse_enable) {
			if (pme_generate_report == false) {
				APP_ERROR("Report Parsing can not be"
				"enabled as pme is not configured to"
				"generate report for any flow/context");

			WB_SET_RSLT(wb_handle,
				LWE_CTRL_RSLT_FAIL_REPORT_PARSE_NOT_ENABLED);
			break;
			}
			parse_report_enable = true;
		}
		/* Start to run all initialised PME context */
		g_all_context_initialized = 1;
		while (*g_context_init_done != (core_count - 1))
			;

		dump_context_info();
		if (!block_scan)
			if (simple_pme_schd_lp_fq() < 0) {
				APP_ERROR("Error: Loopback FQ not"
						" Scheduled");
				assert(0);
			}
		WB_SET_RSLT(wb_handle, LWE_CTRL_RSLT_SUCCESSFULL);
		return 0;

	default:
		APP_INFO(" Invalid command from CP ");
		break;
	}

	return 0;
}

/**
 \brief This is callback function of the scan request when the
	input buffer is simple fd (No output buffer).
	This updates the PME scan stats.

 \param[in] - struct pme context.

 \param[in] - Frame descriptor.

 \param[in] - PME context token.

 \return - None
 */
void scan_cb_simple(struct pme_ctx *ctx, const struct qm_fd *fd,
		    struct pme_ctx_token *ctx_token)
{
	uint32_t flow_id;
	struct spa_pme_ctxt_stats_t *pme_ctx_stats;
	uint8_t res_flags;

	markpoint(11);

	qman_poll_success = 1;

	if (block_scan) {
		APP_DEBUG(" block_scan ");
		atb_stop = get_atb();
		block_scan_stats(atb_stop - atb_start);
		next = 1;
	}

	flow_id = ((struct spa_pme_ctx_token_t *)(ctx_token))->flow_id;
	APP_DEBUG(" flow id = %x ", flow_id);

	pme_ctx_stats = spa_context_info[flow_id].pme_ctx_stats;

	decorated_notify_inc_32(&(pme_ctx_stats->simple_report.frames_scanned));
	decorated_notify_inc_32(g_total_scanned_response);

	if (likely(pme_fd_res_status(fd) == pme_status_ok)) {
		res_flags = pme_fd_res_flags(fd);

		if (res_flags & PME_STATUS_TRUNCATED) {
			decorated_notify_inc_32(&(pme_ctx_stats->
				simple_report.frames_with_atleast_one_pattern));
		}
	} else {
		APP_ERROR("Error in scan response, fd.status = %x",
			  pme_fd_res_status(fd));
		decorated_notify_inc_32(&(pme_ctx_stats->
				simple_report.frames_with_fail_response));
		assert(0);
	}

	/* FIXE-ME Free fd */
	/* This is used only for block_scan context */
	APP_DEBUG("scan_cb_simple() invoked\n");
	markpoint(12);
}

/**
 \brief This is callback function of the scan request when the output
	buffer is allocated by application and input buffer is compound fd.
	 This updates the PME scan stats.

 \param[in] - struct pme context.

 \param[in] - Frame descriptor.

 \param[in] - PME context token.

 \return - None
 */
void scan_cb_cmpd_buff(struct pme_ctx *ctx, const struct qm_fd *fd,
		       struct pme_ctx_token *ctx_token)
{
	uint32_t flow_id;
	struct spa_pme_ctxt_stats_t *pme_ctx_stats;
	struct sg_entry_t *sgentry;
	struct sg_entry_t *sgentry1;
	uint32_t *out_buf = NULL;
	phys_addr_t addr;
	uint32_t length = 0;
	uint8_t res_flags;

	markpoint(21);

	qman_poll_success = 1;

	if (block_scan) {
		APP_DEBUG(" block_scan ");
		atb_stop = get_atb();
		block_scan_stats(atb_stop - atb_start);
		next = 1;
	}

	flow_id = ((struct spa_pme_ctx_token_t *)(ctx_token))->flow_id;
	pme_ctx_stats = spa_context_info[flow_id].pme_ctx_stats;

	decorated_notify_inc_32(&(pme_ctx_stats->simple_report.frames_scanned));
	decorated_notify_inc_32(g_total_scanned_response);

	if (unlikely(pme_fd_res_status(fd) != pme_status_ok)) {
		APP_ERROR("Error in scan response, fd.status = %x",
			  pme_fd_res_status(fd));
		decorated_notify_inc_32(&(pme_ctx_stats->
				simple_report.frames_with_fail_response));
		assert(0);
		/* FIXE-ME Free fd */
		return;
	}

	res_flags = pme_fd_res_flags(fd);

	if (res_flags & PME_STATUS_TRUNCATED) {
		APP_DEBUG(" TRUNCATED bit set");
		decorated_notify_inc_32(&(pme_ctx_stats->
			simple_report.frames_with_atleast_one_pattern));
		/* FIXE-ME Free fd */
		return;
	}

	addr = fd->addr_hi;
	addr = (addr << 32) | fd->addr_lo;
	sgentry = (struct sg_entry_t *)ptov(addr);
	if (likely(sgentry->length)) {
		if (unlikely(parse_report_enable == true)) {
			if (unlikely(sgentry->extension)) {
				APP_DEBUG(" SG buffer in compound fd");
				addr = sgentry->addr_hi;
				addr = (addr << 32) | sgentry->addr_lo;
				sgentry1 = (struct sg_entry_t *)ptov(addr);
				addr = sgentry1->addr_hi;
				addr = (addr << 32) | sgentry1->addr_lo;
				out_buf = (uint32_t *) ptov(addr);
				length = sgentry1->length;
			} else {
				APP_DEBUG(" Simple buffer in compound fd");
				addr = sgentry->addr_hi;
				addr = (addr << 32) | sgentry->addr_lo;
				out_buf = (uint32_t *) ptov(addr);
				length = sgentry->length;
			}

			parse_report((uint8_t *)out_buf, length, pme_ctx_stats);
		}

		decorated_notify_inc_32(&(pme_ctx_stats->
			simple_report.frames_with_atleast_one_pattern));
	}

	/* FIXE-ME Free fd */
	markpoint(22);
}

/**
 \brief This is callback function of the scan request when the output
	buffer is allocated by PME. This updates the PME scan stats.

 \param[in] - struct pme context.

 \param[in] - Frame descriptor.

 \param[in] - PME context token.

 \return - None
 */
void scan_cb_cmpd_bpid(struct pme_ctx *ctx, const struct qm_fd *fd,
		       struct pme_ctx_token *ctx_token)
{
	uint32_t flow_id;
	struct spa_pme_ctxt_stats_t *pme_ctx_stats;
	struct sg_entry_t *sgentry = NULL;
	struct sg_entry_t *sgentry1 = NULL;
	uint32_t *out_buf = NULL;
	uint32_t length = 0;
	phys_addr_t addr;
	struct bm_buffer buf;

	if (block_scan) {
		atb_stop = get_atb();
		block_scan_stats(atb_stop - atb_start);
		/* This is used only for block_scan context */
		next = 1;
	}

	flow_id = ((struct spa_pme_ctx_token_t *)(ctx_token))->flow_id;
	APP_DEBUG(" flow id = %x ", flow_id);

	pme_ctx_stats = spa_context_info[flow_id].pme_ctx_stats;

	decorated_notify_inc_32(&(pme_ctx_stats->simple_report.frames_scanned));
	decorated_notify_inc_32(g_total_scanned_response);

	addr = fd->addr_hi;
	addr = (addr << 32) | fd->addr_lo;
	sgentry = (struct sg_entry_t *)ptov(addr);

	if (unlikely(pme_fd_res_status(fd) != pme_status_ok)) {
		APP_ERROR("Error in scan response, fd.status = %x",
			  pme_fd_res_status(fd));
		decorated_notify_inc_32(&(pme_ctx_stats->
				simple_report.frames_with_fail_response));
		assert(0);
		/* FIXE-ME Free fd */
		return;
	}

	uint8_t res_flags = pme_fd_res_flags(fd);
	if (unlikely(res_flags & PME_STATUS_TRUNCATED)) {
		APP_DEBUG("PME_STATUS_TRUNCATED set\n");
		decorated_notify_inc_32(&(pme_ctx_stats->
				simple_report.frames_with_atleast_one_pattern));
		/* FIXE-ME Free fd */
		return;
	}

	if (unlikely(sgentry->extension)) {
		addr = sgentry->addr_hi;
		addr = (addr << 32) | sgentry->addr_lo;
		sgentry1 = (struct sg_entry_t *)ptov(addr);
		if (likely(sgentry1)) {
			addr = sgentry1->addr_hi;
			addr = (addr << 32) | sgentry1->addr_lo;
			out_buf = (uint32_t *) ptov(addr);
			length = sgentry1->length;
			buf.bpid = sgentry1->bpid;
			buf.hi = sgentry1->addr_hi;
			buf.lo = sgentry1->addr_lo;
		}
	} else {
		addr = sgentry->addr_hi;
		addr = (addr << 32) | sgentry->addr_lo;
		out_buf = (uint32_t *) ptov(addr);
		length = sgentry->length;
		buf.bpid = sgentry->bpid;
		buf.hi = sgentry->addr_hi;
		buf.lo = sgentry->addr_lo;
	}

	if (unlikely(out_buf)) {
		if (unlikely(parse_report_enable == true))
			parse_report((uint8_t *)out_buf, length, pme_ctx_stats);
		decorated_notify_inc_32(&(pme_ctx_stats->simple_report.
					frames_with_atleast_one_pattern));
	}

	/* FIXE-ME Free fd */
	/* Free output buffer */
	dpa_allocator_free_buff(buff_allocator, &buf);
}

/**
 \brief Send the simple single frames to PME for scanning.

 \param[in] - Frame descriptor.

 \return - 0 on success, -ve on Failure.
 */
int spa_scan_simple(const struct qm_fd *fd1, uint16_t seq)
{
	struct spa_pme_ctx_token_t *token;
	struct pme_ctx *pme_ctx;
	uint32_t flow_id;
	uint32_t scan_flags = 0;	/* NO-WAIT */
	struct qm_fd fd;
	phys_addr_t addr;

	memcpy(&fd, fd1, sizeof(struct qm_fd));

	APP_DEBUG(" In spa_scan_simple");
	/* Get pme context token from single buffer descriptor pointer */
	addr = fd.addr_hi;
	addr = (addr << 32) | fd.addr_lo;
	token = (struct spa_pme_ctx_token_t *)ptov(addr);

	/* Get pme context from global data structure */
	flow_id = token->flow_id;

	pme_ctx = spa_context_info[flow_id].spa_pme_ctx;

	/* Send scan request */
	APP_DEBUG(" token->args = %x", token->args);
	if (block_scan)
		atb_start = get_atb();

	APP_DEBUG(" send scan request to pme ");
	return pme_ctx_scan_orp(pme_ctx, scan_flags, &fd, token->args,
			(struct pme_ctx_token *)token,
			spa_context_info[flow_id].qman_fq, seq);
}

/**
 \brief Send the compound frames to PME for scanning.

 \param[in] - Frame descriptor.

 \return - 0 on success, otherwise -ve value
 */
int spa_scan_compound(const struct qm_fd *fd1, uint16_t seq)
{
	struct spa_pme_ctx_token_t *token;
	struct pme_ctx *pme_ctx;
	struct sg_entry_t *sg;
	uint32_t flow_id;
	uint32_t scan_flags = 0;	/* NO-WAIT */
	struct qm_fd fd;
	int ret;
	phys_addr_t addr;

	memcpy(&fd, fd1, sizeof(struct qm_fd));

	APP_DEBUG(" In spa_scan_compound");
	/* Get pme context token from compound buffer descriptor pointer */
	addr = fd.addr_hi;
	addr = (addr << 32) | fd.addr_lo;
	sg = (struct sg_entry_t *)ptov(addr);

	sg++;

	addr = sg->addr_hi;
	addr = (addr << 32) | sg->addr_lo;
	token = (struct spa_pme_ctx_token_t *)ptov(addr);

	/* Get pme context from global data structure */
	flow_id = token->flow_id;
	APP_DEBUG(" token->flow_id = %x", token->flow_id);

	pme_ctx = spa_context_info[flow_id].spa_pme_ctx;

	/* Send scan request */
	APP_DEBUG(" token->args = %x", token->args);
	if (block_scan)
		atb_start = get_atb();

	APP_DEBUG(" send scan request to pme ");

	ret = pme_ctx_scan_orp(pme_ctx, scan_flags, &fd, token->args,
			(struct pme_ctx_token *)token,
			spa_context_info[flow_id].qman_fq, seq);

	if (ret < 0)
		APP_ERROR("ret  = %x , pme_ctx_is_disabled()  %d ",
			  ret, pme_ctx_is_disabled(pme_ctx));
	return ret;
}

/**
 \brief This function dequeues the frames from loopback frame queue and send then
	to PME for scan 

 \param[in]	None

 \return	None
 */
void process_block_scan_req(void)
{
	void *p_data;

	if (curr_lwe_id != g_boot_lwe_core) {
		while (1) {
			if (unlikely(SWI_ID_SHUTDOWN ==
						swi_activate(&p_data))) {
				APP_INFO("Shutdown request received");
				shutting_down = true;
			}
			if (shutting_down == true)
				return;
		}
	}

	while (1) {
		if (unlikely(curr_lwe_id == LWE_BOOT_CORE)) {
			if (unlikely(SWI_ID_SHUTDOWN ==
						swi_activate(&p_data))) {
				APP_INFO("Shutdown request received");
				shutting_down = true;
			}
		}

		/* Exit if packet scanned exceed the limit */
		if (*g_total_scanned_response >= g_total_input_frames ||
				shutting_down == true) {
			markpoint(25);
			display_block_scan_stats();
			markpoint(26);
			shutting_down = true;
			return;
		}

		next = 0;
		qman_volatile_dequeue(spa_context_info
				      [g_block_scan_flow_id].qman_fq, 0,
				      (QM_VDQCR_PRECEDENCE_VDQCR |
				       QM_VDQCR_NUMFRAMES_SET(1) |
				       QM_VDQCR_EXACT));

		markpoint(5);
		/* Scan callback handler will set _next_ to 1; */
		while (!next) {
			qman_poll();
			markpoint(4);
		}

		markpoint(6);
	}
}

/**
 \brief Main function of Simple PME Application

 \param[in]	argc - Argument count

 \param[in]	argv - Argument list pointer

 \return	0 on successfull completion of application
		, otherwise -ve value
 */
int main(int argc, char *argv[])
{
	struct bman_pool_params params;
	uint64_t timer_start, timer_stop;
	uint64_t time1 = 0, time2 = 0;
	uint32_t count1 = 0;
	int ecode = 0;
	uint32_t bpid;
	struct bman_pool *bpool;
	void *p_data;
	char label[10];
	uint32_t part_id = 0;

	curr_lwe_id = get_lwe_id();
	spin_lock(&pme_test_lock);
	core_count++;

	if (INV_LWE_ID == g_boot_lwe_core) {
		parse_report_enable = false;
		APP_INFO("[Boot Core  %d]: Welcome to PME Test"
			 " Application!\r", curr_lwe_id);
		g_boot_lwe_core = curr_lwe_id;
		memcpy(label, get_part_name(&ecode), 10);
		part_id = wb_get_id_from_part_name(label);
		wb_register(part_id, spa_context_setup, &wb_handle,
				(void *)(&_whiteboard_start));
		wb_acquire_lock(&wb_handle, &wb_msg);
		lwe_ctrl_info = (struct lwe_ctrl_op_info *)wb_msg->data;
		memset((void *)wb_msg, 0, sizeof(struct wb_msg));
		memset((void *)lwe_ctrl_info, 0, sizeof(*lwe_ctrl_info));

		APP_INFO("extract_qman_pool_channel");

		if (0 >
		    (pool_channel_num =
		     extract_qman_pool_channel(devtree_ptr))) {
			APP_ERROR("local_extract_qman_channel_id failed");
			abort();
			return -EINVAL;
		}
		pool_channel_id = pool_channel_num + qm_channel_pool1 - 1;
	} else {
		APP_INFO("PME Test Got Core ID %x\n", curr_lwe_id);
	}

	spin_unlock(&pme_test_lock);

	APP_INFO("Pool Channel for lwe %d", pool_channel_id);

	buff_allocator = dpa_allocator_create(&ecode);
	if (unlikely(NULL == buff_allocator)) {
		APP_ERROR("dpa_buff_allocator_create failed");
		abort();
		return -ENOMEM;
	}
	APP_INFO("Buffer allocator created...");

	barrier_init(&pme_barrier, get_online_core_mask());
	if (curr_lwe_id == g_boot_lwe_core) {
		if (unlikely(dpa_bpool_init() != 0)) {
			APP_ERROR("dpa_buff_allocator_create failed");
			return -EINVAL;
		}
	}
	/*Wait for Control LWE Apps core init bpools */
	if (0 > barrier_sync(&pme_barrier)) {
		APP_ERROR(" barrier sync failed before dpa_allocator init");
		return -EINVAL;
	}

	for (bpid = 0; bpid < MAX_NUM_BMAN_POOLS; bpid++) {
		if (0 == dpa_bpools[bpid].size)
			continue;

		params.bpid = bpid;
		params.flags = BMAN_POOL_FLAG_STOCKPILE;
		params.cb = NULL;
		params.cb_ctx = NULL;
		bpool = bman_new_pool(&params);

		if (unlikely(NULL == bpool)) {
			APP_ERROR("bman_new_pool failed");
			abort();
			return -EINVAL;
		}
		if (unlikely(0 != dpa_allocator_add_bpool(buff_allocator,
							  dpa_bpools[bpid].size,
							  bpid, bpool))) {
			APP_ERROR("dpa_allocator_add_bpool failed");
			return -EINVAL;
		}
	}

	qman_static_dequeue_add(QM_SDQCR_CHANNELS_POOL(pool_channel_num));
	APP_INFO("Initialization Complete");
	if (curr_lwe_id == g_boot_lwe_core) {
		g_context_init_done =
		    (uint32_t *) stats_memalign(CACHE_LINE_SIZE,
						sizeof(uint32_t));
		g_total_scanned_response =
		    (uint32_t *) stats_memalign(CACHE_LINE_SIZE,
						sizeof(uint32_t));

		*g_context_init_done = 0;
		*g_total_scanned_response = 0;

		scan_cb_func.scan_cb_simple = &scan_cb_simple;
		scan_cb_func.scan_cb_cmpd_buff = &scan_cb_cmpd_buff;
		scan_cb_func.scan_cb_cmpd_bpid = &scan_cb_cmpd_bpid;
		APP_INFO("Waiting for configuration from CP ... ");
		while (g_all_context_initialized != 1)
			wb_poll(&wb_handle);

	} else {
		do {
			if (unlikely(curr_lwe_id == LWE_BOOT_CORE)) {
				if (unlikely(SWI_ID_SHUTDOWN ==
						swi_activate(&p_data))) {
					APP_INFO("Shutdown request received");
					shutting_down = true;
				}
			}

			if (g_all_context_initialized ||
					shutting_down == true) {
				break;
			}

			qman_poll();

		} while (1);
		decorated_notify_inc_32(g_context_init_done);
	}

	if (unlikely(shutting_down == true))
		goto exit;

	APP_DEBUG("g_context_init_done = %d", *g_context_init_done);
	APP_INFO("Simple PME Test");

	if (0 > barrier_sync(&pme_barrier)) {
		APP_ERROR(" barrier sync failed before dequeue");
		return -1;
	}
	MAGIC_BREAKPOINT;

	markpoint(1);

	if (block_scan) {
		APP_DEBUG(" block_scan ... ");
		markpoint(2);
		if (likely(curr_lwe_id != g_boot_lwe_core)) {
			APP_INFO(" This is not lwe boot core");
			qman_stop_dequeues();
		}
		if (0 > barrier_sync(&pme_barrier)) {
			APP_ERROR(" barrier sync failed before dequeue");
			return -1;
		}
		markpoint(3);

		if (likely(curr_lwe_id == g_boot_lwe_core) ||
				unlikely(curr_lwe_id == LWE_BOOT_CORE)) {
			process_block_scan_req();
		}
		goto exit;
	}

	if (unlikely(curr_lwe_id == LWE_BOOT_CORE)) {
		while (1) {
			if (unlikely(SWI_ID_SHUTDOWN ==
						swi_activate(&p_data))) {
				APP_INFO("Shutdown request received");
				shutting_down = true;
				break;
			}

			/* start timer after First response prom PME received */
			if (unlikely(start_timer_flag == false) &&
				unlikely(*g_total_scanned_response > 0)) {
				spin_lock(&pme_lock);
				start_timer_flag = true;
				if (!atb_start) {
					timer_lwe_id = get_lwe_id();
					atb_start = get_atb();
				}
				spin_unlock(&pme_lock);
			}

			/* Exit if packet scanned exceed the limit */
			if (*g_total_scanned_response >= g_total_input_frames)
				break;

			timer_start = get_atb();
			qman_poll();
			timer_stop = get_atb();
			markpoint(4);

			if (likely(qman_poll_success)) {
				time1 += timer_stop - timer_start;
				count1++;
				qman_poll_success = 0;
			} else {
				underrun++;
				time2 += timer_stop - timer_start;
			}
		}

	} else {
		while (1) {
			if (unlikely(shutting_down == true)) {
				APP_INFO("Shutdown Request Received ");
				break;
			}

			/* start timer after First response prom PME received */
			if (unlikely(start_timer_flag == false) &&
				unlikely(*g_total_scanned_response > 0)) {
				spin_lock(&pme_lock);
				start_timer_flag = true;
				if (!atb_start) {
					timer_lwe_id = get_lwe_id();
					atb_start = get_atb();
				}
				spin_unlock(&pme_lock);
			}

			/* Exit if packet scanned exceed the limit */
			if (*g_total_scanned_response >= g_total_input_frames)
				break;

			timer_start = get_atb();
			qman_poll();
			timer_stop = get_atb();
			markpoint(4);

			if (likely(qman_poll_success)) {
				time1 += timer_stop - timer_start;
				count1++;
				qman_poll_success = 0;
			} else {
				underrun++;
				time2 += timer_stop - timer_start;
			}
		}
	}

	markpoint(30);
	/* Stop the Timer */
	if (get_lwe_id() == timer_lwe_id)
		atb_stop = get_atb();


	if (count1) {
		APP_INFO("Successfull qman poll callback time = %lld", time1);
		APP_INFO("Successfull qman poll callback count = %d", count1);
		time1 = time1/count1;
		APP_INFO("Average qman_poll() callback time = %lld", time1);
	}

	APP_INFO("Underrun = %d", underrun);
	if (underrun) {
		APP_INFO("UN - Successfull callback time = %lld",
							time2 / underrun);
		APP_INFO("Add more core if underrun more than = %lld",
					((time1 * 16) / (time2 / underrun)));
	} else
		APP_INFO("Add more cores as under-run is ZERO");

	if (0 > barrier_sync(&pme_barrier)) {
		APP_ERROR(" barrier sync failed before dequeue");
		return -1;
	}

	if (curr_lwe_id == g_boot_lwe_core) {
		if (parse_report_enable)
			display_report_parse_scan_stats();
		else
			display_noraml_scan_stats();
	}

exit:
	if (0 > barrier_sync(&pme_barrier)) {
		APP_ERROR(" barrier sync failed before dequeue");
		return -1;
	}

	release_pme_resources();

	markpoint(31);

	APP_INFO("ALL DONE");
	return 0;
}
