#!/bin/bash
# Copyright (C) 2009 Freescale Semiconductor, Inc.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
# NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Global variables used in this script
expct=""
host=""
user=""
passwd=""
port=0
board=""
core_mask=""
test_script=""
num_xterms=3 # one for uboot and one for Linux + 1 (bug: TODO)
mux_port_string=""

# Prints the usage information and exits with code 2
print_usage_and_exit()
{
	echo -e "\nTestbench for execution of 'simple_pme' on P4080 Board.\n"
	echo "This script is the interface to launch the simple_pme LWE application on"
	echo "the P4080 boards. Currently, it only supports the 2p configuration of the "
	echo "hypervisor. Before executing the script, make sure that the following"
	echo "conditions are satisfied:"
	echo "  - The 'expect' utility must be installed on the machine from where"
	echo "    this script is being launched."
	echo "  - you have a user_id and password for the gleep server (10.82.176.8)"
	echo "  - you have claimed one of the boards (ex. onefish, horton, etc.)"
	echo "  - the board you have claimed is correctly flashed"
	echo -e "  - the board is in POWERON condition (the script will reset the board)\n\n"
	echo -e "The following are the arguments that this script expects:\n"
	echo "-H"
	echo "   Display this help"
	echo "-u <username>"
	echo "   The user name to be used to connect to the gleep server"
	echo "-x <password>"
	echo "   The password to connect the user to the gleep server"
	echo "-b <board_name | board_ip_address>"
	echo "   This option specifies the name of the board on the gleep server on which"
	echo "   the test will run."
	echo "-e <path_to_expect>"
	echo "   This option specifies the path to the 'expect' utility."
	echo "-h <hostname | host_ip_address>"
	echo "   This option specifies the hostname or host IP address for the gleep server,"
	echo "   to which the board is connected, and on which the mux_server will launch."
	echo "-p <starting port number for the mux_server>"
	echo "   This option specifies the starting port number for the mux_server command."
	echo "   The script will launch the mux_server with the port number starting from"
	echo "   what is specified here, and increment the port number for each byte channel"
	echo "   opened after the first one."
	echo "-c <core_mask>"
	echo "   This option specifies the coremask value to be passed onto partman. The"
	echo "   number must be an 8-bit hex number (prefixed with a 0x), the maximum"
	echo "   allowed value being 0x7f. Each ON bit in the value of the coremask enables"
	echo "   a core in the partition. This script automatically launches the mux_server"
	echo "   with the required number of ports and opens up as many xterms, based on "
	echo "   the number of cores specified through this option."
	echo "-t <test_script>"
	echo "   This option specifies the test shell script to use."
	exit 2
}

# Checks whether the coremask value specified is valid, extracts
# the bits which are ON, and based on this, sets the value of the
# 'num_xterms' global variable. '$1' is the coremask value to check.
check_coremask()
{
	err=0
	# $1 is in 0xZZZZ form ... check whether it is a valid hex number
	# extract the 0x part
	hex_spec=${1:0:2}
	hex_num=${1:2}

	if [ "$hex_spec" != "0x" -a "$hex_spec" != "0X" ]
	then
		echo "ERROR: incorrect hex specifier for value of coremask ... use 0x or 0X"
		err=$((err+1))
	fi

	dummy=`echo $hex_num | egrep "^[0-9a-fA-F]+$"`
	if [ "$dummy" = "" ]
	then
		echo "ERROR: incorrect characters in the hex number of the value of coremask"
		err=$((err+1))
	fi

	# now, check whether it is in range
	num=`printf "%d" $1`
	if [ "$num" -gt "127" ]
	then
		echo "ERROR: maximum value of coremask is 0x7f"
		err=$((err+1))
	fi

	result=""

	# now, for each hex digit, generate 4 binary digits
	num_chars=${#hex_num}
	i=0
	while [ "$i" -lt "$num_chars" ]
	do
		hex_char=${hex_num:$i:1}
		case "$hex_char" in
		0) result=`echo -n "$result""0000"`
			;;
		1) result=`echo -n "$result""0001"`
			;;
		2) result=`echo -n "$result""0010"`
			;;
		3) result=`echo -n "$result""0011"`
			;;
		4) result=`echo -n "$result""0100"`
			;;
		5) result=`echo -n "$result""0101"`
			;;
		6) result=`echo -n "$result""0110"`
			;;
		7) result=`echo -n "$result""0111"`
			;;
		8) result=`echo -n "$result""1000"`
			;;
		9) result=`echo -n "$result""1001"`
			;;
		a|A) result=`echo -n "$result""1010"`
			;;
		b|B) result=`echo -n "$result""1011"`
			;;
		c|C) result=`echo -n "$result""1100"`
			;;
		d|D) result=`echo -n "$result""1101"`
			;;
		e|E) result=`echo -n "$result""1110"`
			;;
		f|F) result=`echo -n "$result""1111"`
			;;
		esac
		i=$((i+1))
	done

	for (( j=1; j<=8; j++ ))
	do
		enabled=${result:(-$j)}
		enabled=${enabled:0:1}
		if [ "$enabled" = "1" ]
		then
			num_xterms=$((num_xterms+1))
		fi
	done

	if [ "$err" -gt "0" ]
	then
		echo -e "\nsimple_pme_test_2p.sh: Please use the -H option to get detailed help"
		exit 3
	fi

}

# Parses the command line and fills the following global variables:
# host, port, core_mask, mode, buffer_size, test_set, sec_algo, frame_count, iter_count
parse_command_line()
{
	errors=0; u_flag=0; x_flag=0; b_flag=0
	e_flag=0; h_flag=0; p_flag=0; c_flag=0; t_flag=0
	
	while getopts "Hu:x:b:e:h:p:c:t:" OPTION
	do
		case $OPTION in
		H)
		print_usage_and_exit
		;;
		u)
		user="$OPTARG"
		u_flag=1
		;;
		x)
		passwd="$OPTARG"
		x_flag=1
		;;
		b)
		if [ "$b_flag" -eq "1" ]
		then
			if [ "$board" != "$OPTARG" ]
			then
				echo -e "\nsimple_pme_test_2p.sh: ERROR: -b specified multiple times with different arguments"
				errors=$((errors+1))
			else
				echo -e "\nsimple_pme_test_2p.sh: -b specified multiple times ... ignored"
			fi
		fi
		board=$OPTARG
		b_flag=1
		;;
		e)
		if [ "$e_flag" -eq "1" ]
		then
			if [ "$expct" != "$OPTARG" ]
			then
				echo -e "\nsimple_pme_test_2p.sh: ERROR: -e specified multiple times with different arguments"
				errors=$((errors+1))
			else
				echo -e "\nsimple_pme_test_2p.sh: -e specified multiple times ... ignored"
			fi
		fi
		$OPTARG -v 1>/dev/null 2>/dev/null
		if [ "$?" -ne "0" ]
		then
			echo -e "\nsimple_pme_test_2p.sh: ERROR: cannot find expect in specified path $OPTARG"
			errors=$((errors+1))
		fi
		expct=$OPTARG
		e_flag=1
		;;
		h)
		if [ "$h_flag" -eq "1" ]
		then
			if [ "$host" != "$OPTARG" ]
			then
				echo -e "\nsimple_pme_test_2p.sh: ERROR: -h specified multiple times with different arguments"
				errors=$((errors+1))
			else
				echo -e "\nsimple_pme_test_2p.sh: -h specified multiple times ... ignored"
			fi
		fi
		ping -c 1 $OPTARG 1>/dev/null 2>/dev/null
		if [ "$?" -ne "0" ]
		then
			echo -e "\nsimple_pme_test_2p.sh: ERROR: host $OPTARG is unreachable!"
			errors=$((errors+1))
		fi
		host=$OPTARG
		h_flag=1
		;;
		p)
		if [ "$p_flag" -eq "1" ]
		then
			if [ "$port" != "$OPTARG" ]
			then
				echo -e "\nsimple_pme_test_2p.sh: ERROR: -p specified multiple times with different arguments"
				errors=$((errors+1))
			else
				echo -e "\nsimple_pme_test_2p.sh: -p specified multiple times ... ignored"
			fi
		fi
		dummy=`echo "$OPTARG" | egrep "^[0-9]+$"`
		if [ "$dummy" = "" ]
		then
			echo -e "\nsimple_pme_test_2p.sh: ERROR: unrecognized characters in the port number specified with -p"
			errors=$((errors+1))
		fi
		port=$OPTARG
		p_flag=1
		;;
		c)
		if [ "$c_flag" -eq "1" ]
		then
			if [ "$core_mask" != "$OPTARG" ]
			then
				echo -e "\nsimple_pme_test_2p.sh: ERROR: -c specified multiple times with different arguments"
				errors=$((errors+1))
			else
				echo -e "\nsimple_pme_test_2p.sh: -c specified multiple times ... ignored"
			fi
		fi
		check_coremask $OPTARG
		core_mask=$OPTARG
		c_flag=1
		;;
		t)
		if [ "$t_flag" -eq "1" ]
		then
			if [ "$test_script" != "$OPTARG" ]
			then
				echo -e "\nsimple_pme_test_2p.sh: ERROR: -t specified multiple times with different arguments"
				errors=$((errors+1))
			else
				echo -e "\nsimple_pme_test_2p.sh: -t specified multiple times ... ignored"
			fi
		fi
		test_script=$OPTARG
		t_flag=1
		;;
		?)
		print_usage_and_exit
		;;
		esac
	done

	if [ "$b_flag" -eq "0" -o "$e_flag" -eq "0" -o "$h_flag" -eq "0" -o "$p_flag" -eq "0" -o "$c_flag" -eq "0" -o "$t_flag" -eq "0" ]
	then
		echo -e "\nsimple_pme_test_2p.sh: ERROR: required arguments missing"
		errors=$((errors+1))
	fi

	if [ "$errors" -gt "0" ]
	then
		echo "\nsimple_pme_test_2p.sh: Please use the -H option to get detailed help"
		exit 1
	fi
}

# Constructs the mux_server port strings based on the
# first port number specified and the number of xterms to open.
# Fills up the 'mux_port_string' global variable.
construct_mux_port_string()
{
	mux_port_string=`echo "$port"`
	temp="$port"
	for (( i=1; i<"$num_xterms"; i++ ))
	do
		temp=$((temp+1))
		mux_port_string=`echo "$mux_port_string $temp"`
	done
	echo "Using mux_port_string = $mux_port_string"
}

# main
parse_command_line $*
rm -f "simple_pme.$host.*.log"
construct_mux_port_string
spawned_procs=""

# reset the board first
"$expct" `pwd`/board_start.exp "$host" "$user" "$passwd" "$board" "$mux_port_string" &
spawned_procs=`echo "$!"`

sleep 5
# start the uboot console
"$expct" `pwd`/uboot.exp "$host" "$port" &
spawned_procs=`echo "$spawned_procs $!"`

# start the Linux console
port=$((port+1))
"$expct" `pwd`/simple_pme.exp "$host" "$port" "$core_mask" "$test_script" &
spawned_procs=`echo "$spawned_procs $!"`

# start the application consoles
port=$((port+1))
xterm -exec "$expct" `pwd`/exit.exp "$host" "$port" &
app_console_pids=`echo "$!"`

for (( i=3; i<"$num_xterms"; i++ ))
do
	port=$((port+1))
	xterm -exec "$expct" `pwd`/exit.exp "$host" "$port" &
	app_console_pids=`echo "$app_console_pids $!"`
done

num_app_consoles=$((num_xterms-2))

echo "Going into monitoring mode ..."
sleep 25

# monitor for the app consoles to exit
# take short naps and keep monitoring
while [ true ]
do
	sleep 5
	can_exit=0
	for (( i=1; i<=$num_app_consoles; i++ ))
	do
		pid=`echo "$app_console_pids" | cut -d' ' -f"$i"`
		kill -0 "$pid" 1>/dev/null 2>/dev/null
		# as soon as one ends, get out of here
		if [ "$?" -eq "1" ]
		then
			echo "process detected to have ended ... exiting"
			can_exit=1
			for j in `echo "$app_console_pids $spawned_procs"`
			do
				kill "$j" 1>/dev/null 2>/dev/null
			done
		fi
	done
	if [ "$can_exit" -eq "1" ]
	then
		break
	fi
done

# parting remarks
# scan the log files for test passed and report accordingly

for i in `ls simple_pme.$host.*.log`
do
	grep -i pass "$i" 1>/dev/null 2>/dev/null
	if [ "$?" -eq "0" ]
	then
		echo "Test PASSED"
		exit 0
	fi
done

echo "Test FAILED"
exit 3
	
