/**
 \file torture_test.h
 \brief Common datatypes, hash-defines and function prototypes of Qman Torture application
 */
/*
 * Copyright (C) 2008, 2010 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _TORTURE_TEST_H
#define _TORTURE_TEST_H

#include "compat.h"
#include <io.h>
#include <bitops.h>
#include <init.h>
#include <misc.h>
#include <statistics.h>
#include <mmap.h>
#include <app_common.h>
#include <app_cpdp.h>
#include <dpa_alloc.h>
#include <dpa/dpa_eth.h>
#include "errno.h"
#include "cache.h"
#include "barrier.h"

#define BUFFER_SIZE 		128
#define STATS_PRINT_INTERVAL_MS 1000

/**
 \brief Structure to keep Configuartion parameters.
        This structure is kept in shared no heap area for CP-DP communication.
        Parameters entered by user through CP are written in this structure,
        which are used by the corresponding application running on 
        data plane for initializations.
 */
struct torture_config_s {
	uint32_t fq_start;		/**< Start Frame Number */
	uint32_t num_fq;		/**< Number of frame queues need to be initialized */
	uint32_t channel;		/**< Channel number in which frame queues will be created */
	uint32_t wq_id;			/**< Work queue id of the frame queues */
	uint32_t num_buff_per_fq;	/**< Number of buffers in each frame queue */
	uint32_t no_of_iterations;	/**< Number of times qman_poll should iterate in hybrid mode */
	uint32_t barrier;		/**< Barrier variable for CP-DP synchronization */
};

/**
 \brief Configuration structure initialization routine.
        Initialize fields of configuration structure.
	this contains every parameter for creating and initializing 
	frame queues. The value of each field is picked from 
	the one filled by the user through control plane part of the application.
 \param[in] Void
 \return NULL
 \note Called only once from control partition.
 */
void init_config_structure(void);

/**
 \brief Dequeue Handler for Qman Torture Application
        Enqueues the currently dequeued frame into 
	next initialized Frame Queue. It is called as a handler 
        routine whenever a frame is dequeued by the core.
 \param[in] struct qman_portal * Qman Portal Pointer
 \param[in] struct qman_fq * Pointer to a Qman frame Queue descriptor
 \param[in] struct qm_dqrr_entry * Dequeue Receive Ring entry to be processed
 \param[out] NULL
 */
static enum qman_cb_dqrr_result dqrr_entry_handler(struct qman_portal *qm,
						   struct qman_fq *fq,
						   const struct qm_dqrr_entry
						   *dqrr);

/**
 \brief Frame queues creation routine.
        Create specified number of frame queues
	and store their index in a structure in shared heap area
	(to make it accessed by each partition.)
 \param[in] Void
 \return Negative - Failure
	Non-Negative - Success
 \note Called once only from control partition.
 */
int create_frame_queues(void);

/**
 \brief Frame queues initialization routine.
        Initializes the already created frame queues 
	whose indexes are there in a structure in shared heap area.
        Associates these frame queues to work queue and
	to LWE partition's pool channel
	Set the queue flag for scheduling 
 \param[in] Void
 \return Negative - Failure
	Non-Negative - Success
 \note Called once only from control partition.
 */
int initialize_fqs(void);

/**
 \brief Buffers enqueuing routine.
        Preparing specified number of Frame Descriptors
	for each frame queue and enqueue them in frame queues.
 \param[in] Void
 \return NULL
 \note Called only once from control partition.
 */
void test_buffers_enq(void);

/**
 \brief Stats printing routine
        Prints the statistics like:
	frames enqueued/dequeued by the current core and rate of enqueuing/dequeuing.
	Timer framework has been initialized through which this function is called
	after specified periodicity.
 \param[in] Timer identifier
 \param[in] void pointer
 \return NULL
 \note Called by the timer, as a interrupt handling routine 
*/
void print_stats(uint32_t timer_id, void *p_data);

/**
 \brief Dummy function to take checkpoint 
        for running the application on hybrid simics model
 \param[in] part_id Partition id
 \return NULL
*/
void qman_test(uint8_t part_id);

#endif /* _TORTURE_TEST_H */
